URL_LIST = {
	'srtm1': {
		'N00W177.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N00W177.hgt.zip',
		'N05W163.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N05W163.hgt.zip',
		'N06W163.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N06W163.hgt.zip',
		'N16W170.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N16W170.hgt.zip',
		'N17W065.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N17W065.hgt.zip',
		'N17W066.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N17W066.hgt.zip',
		'N17W067.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N17W067.hgt.zip',
		'N17W068.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N17W068.hgt.zip',
		'N18W065.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N18W065.hgt.zip',
		'N18W066.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N18W066.hgt.zip',
		'N18W067.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N18W067.hgt.zip',
		'N18W068.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N18W068.hgt.zip',
		'N18W076.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N18W076.hgt.zip',
		'N18W156.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N18W156.hgt.zip',
		'N19W155.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N19W155.hgt.zip',
		'N19W156.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N19W156.hgt.zip',
		'N19W157.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N19W157.hgt.zip',
		'N20W156.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N20W156.hgt.zip',
		'N20W157.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N20W157.hgt.zip',
		'N20W158.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N20W158.hgt.zip',
		'N21W157.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N21W157.hgt.zip',
		'N21W158.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N21W158.hgt.zip',
		'N21W159.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N21W159.hgt.zip',
		'N21W160.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N21W160.hgt.zip',
		'N21W161.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N21W161.hgt.zip',
		'N22W160.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N22W160.hgt.zip',
		'N22W161.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N22W161.hgt.zip',
		'N23W162.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N23W162.hgt.zip',
		'N23W165.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N23W165.hgt.zip',
		'N23W167.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N23W167.hgt.zip',
		'N24W081.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N24W081.hgt.zip',
		'N24W082.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N24W082.hgt.zip',
		'N24W083.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N24W083.hgt.zip',
		'N25W081.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N25W081.hgt.zip',
		'N25W082.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N25W082.hgt.zip',
		'N25W098.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N25W098.hgt.zip',
		'N26W081.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N26W081.hgt.zip',
		'N26W082.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N26W082.hgt.zip',
		'N26W083.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N26W083.hgt.zip',
		'N26W098.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N26W098.hgt.zip',
		'N26W099.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N26W099.hgt.zip',
		'N26W100.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N26W100.hgt.zip',
		'N26W174.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N26W174.hgt.zip',
		'N27W081.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N27W081.hgt.zip',
		'N27W082.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N27W082.hgt.zip',
		'N27W083.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N27W083.hgt.zip',
		'N27W097.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N27W097.hgt.zip',
		'N27W098.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N27W098.hgt.zip',
		'N27W099.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N27W099.hgt.zip',
		'N27W100.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N27W100.hgt.zip',
		'N27W176.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N27W176.hgt.zip',
		'N28W081.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N28W081.hgt.zip',
		'N28W082.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N28W082.hgt.zip',
		'N28W083.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N28W083.hgt.zip',
		'N28W090.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N28W090.hgt.zip',
		'N28W096.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N28W096.hgt.zip',
		'N28W097.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N28W097.hgt.zip',
		'N28W098.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N28W098.hgt.zip',
		'N28W099.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N28W099.hgt.zip',
		'N28W100.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N28W100.hgt.zip',
		'N28W101.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N28W101.hgt.zip',
		'N28W104.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N28W104.hgt.zip',
		'N28W178.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N28W178.hgt.zip',
		'N28W179.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N28W179.hgt.zip',
		'N29W081.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N29W081.hgt.zip',
		'N29W082.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N29W082.hgt.zip',
		'N29W083.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N29W083.hgt.zip',
		'N29W084.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N29W084.hgt.zip',
		'N29W085.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N29W085.hgt.zip',
		'N29W086.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N29W086.hgt.zip',
		'N29W089.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N29W089.hgt.zip',
		'N29W090.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N29W090.hgt.zip',
		'N29W091.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N29W091.hgt.zip',
		'N29W092.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N29W092.hgt.zip',
		'N29W093.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N29W093.hgt.zip',
		'N29W094.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N29W094.hgt.zip',
		'N29W095.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N29W095.hgt.zip',
		'N29W096.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N29W096.hgt.zip',
		'N29W097.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N29W097.hgt.zip',
		'N29W098.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N29W098.hgt.zip',
		'N29W099.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N29W099.hgt.zip',
		'N29W100.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N29W100.hgt.zip',
		'N29W101.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N29W101.hgt.zip',
		'N29W102.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N29W102.hgt.zip',
		'N29W103.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N29W103.hgt.zip',
		'N29W104.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N29W104.hgt.zip',
		'N29W105.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N29W105.hgt.zip',
		'N30W082.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N30W082.hgt.zip',
		'N30W083.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N30W083.hgt.zip',
		'N30W084.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N30W084.hgt.zip',
		'N30W085.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N30W085.hgt.zip',
		'N30W086.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N30W086.hgt.zip',
		'N30W087.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N30W087.hgt.zip',
		'N30W088.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N30W088.hgt.zip',
		'N30W089.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N30W089.hgt.zip',
		'N30W090.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N30W090.hgt.zip',
		'N30W091.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N30W091.hgt.zip',
		'N30W092.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N30W092.hgt.zip',
		'N30W093.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N30W093.hgt.zip',
		'N30W094.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N30W094.hgt.zip',
		'N30W095.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N30W095.hgt.zip',
		'N30W096.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N30W096.hgt.zip',
		'N30W097.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N30W097.hgt.zip',
		'N30W098.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N30W098.hgt.zip',
		'N30W099.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N30W099.hgt.zip',
		'N30W100.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N30W100.hgt.zip',
		'N30W101.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N30W101.hgt.zip',
		'N30W102.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N30W102.hgt.zip',
		'N30W103.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N30W103.hgt.zip',
		'N30W104.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N30W104.hgt.zip',
		'N30W105.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N30W105.hgt.zip',
		'N30W106.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N30W106.hgt.zip',
		'N31W081.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N31W081.hgt.zip',
		'N31W082.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N31W082.hgt.zip',
		'N31W083.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N31W083.hgt.zip',
		'N31W084.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N31W084.hgt.zip',
		'N31W085.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N31W085.hgt.zip',
		'N31W086.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N31W086.hgt.zip',
		'N31W087.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N31W087.hgt.zip',
		'N31W088.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N31W088.hgt.zip',
		'N31W089.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N31W089.hgt.zip',
		'N31W090.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N31W090.hgt.zip',
		'N31W091.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N31W091.hgt.zip',
		'N31W092.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N31W092.hgt.zip',
		'N31W093.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N31W093.hgt.zip',
		'N31W094.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N31W094.hgt.zip',
		'N31W095.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N31W095.hgt.zip',
		'N31W096.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N31W096.hgt.zip',
		'N31W097.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N31W097.hgt.zip',
		'N31W098.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N31W098.hgt.zip',
		'N31W099.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N31W099.hgt.zip',
		'N31W100.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N31W100.hgt.zip',
		'N31W101.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N31W101.hgt.zip',
		'N31W102.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N31W102.hgt.zip',
		'N31W103.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N31W103.hgt.zip',
		'N31W104.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N31W104.hgt.zip',
		'N31W105.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N31W105.hgt.zip',
		'N31W106.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N31W106.hgt.zip',
		'N31W107.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N31W107.hgt.zip',
		'N31W108.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N31W108.hgt.zip',
		'N31W109.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N31W109.hgt.zip',
		'N31W110.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N31W110.hgt.zip',
		'N31W111.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N31W111.hgt.zip',
		'N31W112.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N31W112.hgt.zip',
		'N31W113.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N31W113.hgt.zip',
		'N31W114.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N31W114.hgt.zip',
		'N32W080.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N32W080.hgt.zip',
		'N32W081.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N32W081.hgt.zip',
		'N32W082.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N32W082.hgt.zip',
		'N32W083.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N32W083.hgt.zip',
		'N32W084.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N32W084.hgt.zip',
		'N32W085.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N32W085.hgt.zip',
		'N32W086.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N32W086.hgt.zip',
		'N32W087.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N32W087.hgt.zip',
		'N32W088.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N32W088.hgt.zip',
		'N32W089.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N32W089.hgt.zip',
		'N32W090.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N32W090.hgt.zip',
		'N32W091.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N32W091.hgt.zip',
		'N32W092.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N32W092.hgt.zip',
		'N32W093.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N32W093.hgt.zip',
		'N32W094.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N32W094.hgt.zip',
		'N32W095.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N32W095.hgt.zip',
		'N32W096.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N32W096.hgt.zip',
		'N32W097.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N32W097.hgt.zip',
		'N32W098.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N32W098.hgt.zip',
		'N32W099.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N32W099.hgt.zip',
		'N32W100.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N32W100.hgt.zip',
		'N32W101.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N32W101.hgt.zip',
		'N32W102.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N32W102.hgt.zip',
		'N32W103.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N32W103.hgt.zip',
		'N32W104.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N32W104.hgt.zip',
		'N32W105.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N32W105.hgt.zip',
		'N32W106.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N32W106.hgt.zip',
		'N32W107.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N32W107.hgt.zip',
		'N32W108.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N32W108.hgt.zip',
		'N32W109.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N32W109.hgt.zip',
		'N32W110.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N32W110.hgt.zip',
		'N32W111.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N32W111.hgt.zip',
		'N32W112.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N32W112.hgt.zip',
		'N32W113.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N32W113.hgt.zip',
		'N32W114.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N32W114.hgt.zip',
		'N32W115.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N32W115.hgt.zip',
		'N32W116.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N32W116.hgt.zip',
		'N32W117.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N32W117.hgt.zip',
		'N32W118.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N32W118.hgt.zip',
		'N32W119.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N32W119.hgt.zip',
		'N33W078.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N33W078.hgt.zip',
		'N33W079.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N33W079.hgt.zip',
		'N33W080.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N33W080.hgt.zip',
		'N33W081.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N33W081.hgt.zip',
		'N33W082.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N33W082.hgt.zip',
		'N33W083.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N33W083.hgt.zip',
		'N33W084.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N33W084.hgt.zip',
		'N33W085.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N33W085.hgt.zip',
		'N33W086.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N33W086.hgt.zip',
		'N33W087.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N33W087.hgt.zip',
		'N33W088.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N33W088.hgt.zip',
		'N33W089.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N33W089.hgt.zip',
		'N33W090.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N33W090.hgt.zip',
		'N33W091.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N33W091.hgt.zip',
		'N33W092.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N33W092.hgt.zip',
		'N33W093.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N33W093.hgt.zip',
		'N33W094.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N33W094.hgt.zip',
		'N33W095.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N33W095.hgt.zip',
		'N33W096.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N33W096.hgt.zip',
		'N33W097.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N33W097.hgt.zip',
		'N33W098.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N33W098.hgt.zip',
		'N33W099.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N33W099.hgt.zip',
		'N33W100.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N33W100.hgt.zip',
		'N33W101.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N33W101.hgt.zip',
		'N33W102.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N33W102.hgt.zip',
		'N33W103.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N33W103.hgt.zip',
		'N33W104.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N33W104.hgt.zip',
		'N33W105.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N33W105.hgt.zip',
		'N33W106.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N33W106.hgt.zip',
		'N33W107.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N33W107.hgt.zip',
		'N33W108.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N33W108.hgt.zip',
		'N33W109.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N33W109.hgt.zip',
		'N33W110.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N33W110.hgt.zip',
		'N33W111.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N33W111.hgt.zip',
		'N33W112.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N33W112.hgt.zip',
		'N33W113.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N33W113.hgt.zip',
		'N33W114.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N33W114.hgt.zip',
		'N33W115.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N33W115.hgt.zip',
		'N33W116.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N33W116.hgt.zip',
		'N33W117.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N33W117.hgt.zip',
		'N33W118.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N33W118.hgt.zip',
		'N33W119.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N33W119.hgt.zip',
		'N33W120.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N33W120.hgt.zip',
		'N33W121.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N33W121.hgt.zip',
		'N34W077.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N34W077.hgt.zip',
		'N34W078.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N34W078.hgt.zip',
		'N34W079.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N34W079.hgt.zip',
		'N34W080.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N34W080.hgt.zip',
		'N34W081.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N34W081.hgt.zip',
		'N34W082.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N34W082.hgt.zip',
		'N34W083.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N34W083.hgt.zip',
		'N34W084.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N34W084.hgt.zip',
		'N34W085.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N34W085.hgt.zip',
		'N34W086.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N34W086.hgt.zip',
		'N34W087.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N34W087.hgt.zip',
		'N34W088.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N34W088.hgt.zip',
		'N34W089.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N34W089.hgt.zip',
		'N34W090.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N34W090.hgt.zip',
		'N34W091.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N34W091.hgt.zip',
		'N34W092.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N34W092.hgt.zip',
		'N34W093.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N34W093.hgt.zip',
		'N34W094.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N34W094.hgt.zip',
		'N34W095.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N34W095.hgt.zip',
		'N34W096.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N34W096.hgt.zip',
		'N34W097.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N34W097.hgt.zip',
		'N34W098.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N34W098.hgt.zip',
		'N34W099.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N34W099.hgt.zip',
		'N34W100.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N34W100.hgt.zip',
		'N34W101.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N34W101.hgt.zip',
		'N34W102.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N34W102.hgt.zip',
		'N34W103.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N34W103.hgt.zip',
		'N34W104.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N34W104.hgt.zip',
		'N34W105.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N34W105.hgt.zip',
		'N34W106.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N34W106.hgt.zip',
		'N34W107.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N34W107.hgt.zip',
		'N34W108.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N34W108.hgt.zip',
		'N34W109.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N34W109.hgt.zip',
		'N34W110.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N34W110.hgt.zip',
		'N34W111.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N34W111.hgt.zip',
		'N34W112.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N34W112.hgt.zip',
		'N34W113.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N34W113.hgt.zip',
		'N34W114.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N34W114.hgt.zip',
		'N34W115.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N34W115.hgt.zip',
		'N34W116.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N34W116.hgt.zip',
		'N34W117.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N34W117.hgt.zip',
		'N34W118.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N34W118.hgt.zip',
		'N34W119.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N34W119.hgt.zip',
		'N34W120.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N34W120.hgt.zip',
		'N34W121.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N34W121.hgt.zip',
		'N35W076.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N35W076.hgt.zip',
		'N35W077.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N35W077.hgt.zip',
		'N35W078.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N35W078.hgt.zip',
		'N35W079.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N35W079.hgt.zip',
		'N35W080.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N35W080.hgt.zip',
		'N35W081.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N35W081.hgt.zip',
		'N35W082.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N35W082.hgt.zip',
		'N35W083.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N35W083.hgt.zip',
		'N35W084.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N35W084.hgt.zip',
		'N35W085.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N35W085.hgt.zip',
		'N35W086.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N35W086.hgt.zip',
		'N35W087.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N35W087.hgt.zip',
		'N35W088.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N35W088.hgt.zip',
		'N35W089.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N35W089.hgt.zip',
		'N35W090.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N35W090.hgt.zip',
		'N35W091.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N35W091.hgt.zip',
		'N35W092.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N35W092.hgt.zip',
		'N35W093.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N35W093.hgt.zip',
		'N35W094.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N35W094.hgt.zip',
		'N35W095.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N35W095.hgt.zip',
		'N35W096.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N35W096.hgt.zip',
		'N35W097.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N35W097.hgt.zip',
		'N35W098.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N35W098.hgt.zip',
		'N35W099.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N35W099.hgt.zip',
		'N35W100.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N35W100.hgt.zip',
		'N35W101.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N35W101.hgt.zip',
		'N35W102.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N35W102.hgt.zip',
		'N35W103.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N35W103.hgt.zip',
		'N35W104.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N35W104.hgt.zip',
		'N35W105.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N35W105.hgt.zip',
		'N35W106.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N35W106.hgt.zip',
		'N35W107.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N35W107.hgt.zip',
		'N35W108.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N35W108.hgt.zip',
		'N35W109.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N35W109.hgt.zip',
		'N35W110.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N35W110.hgt.zip',
		'N35W111.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N35W111.hgt.zip',
		'N35W112.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N35W112.hgt.zip',
		'N35W113.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N35W113.hgt.zip',
		'N35W114.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N35W114.hgt.zip',
		'N35W115.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N35W115.hgt.zip',
		'N35W116.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N35W116.hgt.zip',
		'N35W117.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N35W117.hgt.zip',
		'N35W118.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N35W118.hgt.zip',
		'N35W119.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N35W119.hgt.zip',
		'N35W120.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N35W120.hgt.zip',
		'N35W121.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N35W121.hgt.zip',
		'N35W122.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N35W122.hgt.zip',
		'N36W076.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N36W076.hgt.zip',
		'N36W077.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N36W077.hgt.zip',
		'N36W078.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N36W078.hgt.zip',
		'N36W079.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N36W079.hgt.zip',
		'N36W080.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N36W080.hgt.zip',
		'N36W081.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N36W081.hgt.zip',
		'N36W082.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N36W082.hgt.zip',
		'N36W083.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N36W083.hgt.zip',
		'N36W084.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N36W084.hgt.zip',
		'N36W085.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N36W085.hgt.zip',
		'N36W086.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N36W086.hgt.zip',
		'N36W087.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N36W087.hgt.zip',
		'N36W088.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N36W088.hgt.zip',
		'N36W089.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N36W089.hgt.zip',
		'N36W090.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N36W090.hgt.zip',
		'N36W091.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N36W091.hgt.zip',
		'N36W092.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N36W092.hgt.zip',
		'N36W093.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N36W093.hgt.zip',
		'N36W094.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N36W094.hgt.zip',
		'N36W095.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N36W095.hgt.zip',
		'N36W096.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N36W096.hgt.zip',
		'N36W097.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N36W097.hgt.zip',
		'N36W098.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N36W098.hgt.zip',
		'N36W099.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N36W099.hgt.zip',
		'N36W100.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N36W100.hgt.zip',
		'N36W101.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N36W101.hgt.zip',
		'N36W102.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N36W102.hgt.zip',
		'N36W103.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N36W103.hgt.zip',
		'N36W104.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N36W104.hgt.zip',
		'N36W105.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N36W105.hgt.zip',
		'N36W106.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N36W106.hgt.zip',
		'N36W107.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N36W107.hgt.zip',
		'N36W108.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N36W108.hgt.zip',
		'N36W109.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N36W109.hgt.zip',
		'N36W110.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N36W110.hgt.zip',
		'N36W111.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N36W111.hgt.zip',
		'N36W112.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N36W112.hgt.zip',
		'N36W113.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N36W113.hgt.zip',
		'N36W114.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N36W114.hgt.zip',
		'N36W115.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N36W115.hgt.zip',
		'N36W116.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N36W116.hgt.zip',
		'N36W117.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N36W117.hgt.zip',
		'N36W118.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N36W118.hgt.zip',
		'N36W119.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N36W119.hgt.zip',
		'N36W120.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N36W120.hgt.zip',
		'N36W121.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N36W121.hgt.zip',
		'N36W122.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N36W122.hgt.zip',
		'N36W123.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N36W123.hgt.zip',
		'N37W076.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N37W076.hgt.zip',
		'N37W077.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N37W077.hgt.zip',
		'N37W078.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N37W078.hgt.zip',
		'N37W079.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N37W079.hgt.zip',
		'N37W080.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N37W080.hgt.zip',
		'N37W081.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N37W081.hgt.zip',
		'N37W082.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N37W082.hgt.zip',
		'N37W083.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N37W083.hgt.zip',
		'N37W084.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N37W084.hgt.zip',
		'N37W085.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N37W085.hgt.zip',
		'N37W086.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N37W086.hgt.zip',
		'N37W087.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N37W087.hgt.zip',
		'N37W088.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N37W088.hgt.zip',
		'N37W089.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N37W089.hgt.zip',
		'N37W090.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N37W090.hgt.zip',
		'N37W091.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N37W091.hgt.zip',
		'N37W092.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N37W092.hgt.zip',
		'N37W093.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N37W093.hgt.zip',
		'N37W094.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N37W094.hgt.zip',
		'N37W095.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N37W095.hgt.zip',
		'N37W096.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N37W096.hgt.zip',
		'N37W097.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N37W097.hgt.zip',
		'N37W098.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N37W098.hgt.zip',
		'N37W099.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N37W099.hgt.zip',
		'N37W100.hgt': 'https://srtm.kurviger.de/SRTM1//Region_05//N37W100.hgt.zip',
		'N37W101.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N37W101.hgt.zip',
		'N37W102.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N37W102.hgt.zip',
		'N37W103.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N37W103.hgt.zip',
		'N37W104.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N37W104.hgt.zip',
		'N37W105.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N37W105.hgt.zip',
		'N37W106.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N37W106.hgt.zip',
		'N37W107.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N37W107.hgt.zip',
		'N37W108.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N37W108.hgt.zip',
		'N37W109.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N37W109.hgt.zip',
		'N37W110.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N37W110.hgt.zip',
		'N37W111.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N37W111.hgt.zip',
		'N37W112.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N37W112.hgt.zip',
		'N37W113.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N37W113.hgt.zip',
		'N37W114.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N37W114.hgt.zip',
		'N37W115.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N37W115.hgt.zip',
		'N37W116.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N37W116.hgt.zip',
		'N37W117.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N37W117.hgt.zip',
		'N37W118.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N37W118.hgt.zip',
		'N37W119.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N37W119.hgt.zip',
		'N37W120.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N37W120.hgt.zip',
		'N37W121.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N37W121.hgt.zip',
		'N37W122.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N37W122.hgt.zip',
		'N37W123.hgt': 'https://srtm.kurviger.de/SRTM1//Region_04//N37W123.hgt.zip',
		'N38W075.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N38W075.hgt.zip',
		'N38W076.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N38W076.hgt.zip',
		'N38W077.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N38W077.hgt.zip',
		'N38W078.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N38W078.hgt.zip',
		'N38W079.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N38W079.hgt.zip',
		'N38W080.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N38W080.hgt.zip',
		'N38W081.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N38W081.hgt.zip',
		'N38W082.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N38W082.hgt.zip',
		'N38W083.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N38W083.hgt.zip',
		'N38W084.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N38W084.hgt.zip',
		'N38W085.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N38W085.hgt.zip',
		'N38W086.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N38W086.hgt.zip',
		'N38W087.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N38W087.hgt.zip',
		'N38W088.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N38W088.hgt.zip',
		'N38W089.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N38W089.hgt.zip',
		'N38W090.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N38W090.hgt.zip',
		'N38W091.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N38W091.hgt.zip',
		'N38W092.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N38W092.hgt.zip',
		'N38W093.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N38W093.hgt.zip',
		'N38W094.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N38W094.hgt.zip',
		'N38W095.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N38W095.hgt.zip',
		'N38W096.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N38W096.hgt.zip',
		'N38W097.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N38W097.hgt.zip',
		'N38W098.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N38W098.hgt.zip',
		'N38W099.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N38W099.hgt.zip',
		'N38W100.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N38W100.hgt.zip',
		'N38W101.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N38W101.hgt.zip',
		'N38W102.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N38W102.hgt.zip',
		'N38W103.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N38W103.hgt.zip',
		'N38W104.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N38W104.hgt.zip',
		'N38W105.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N38W105.hgt.zip',
		'N38W106.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N38W106.hgt.zip',
		'N38W107.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N38W107.hgt.zip',
		'N38W108.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N38W108.hgt.zip',
		'N38W109.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N38W109.hgt.zip',
		'N38W110.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N38W110.hgt.zip',
		'N38W111.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N38W111.hgt.zip',
		'N38W112.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N38W112.hgt.zip',
		'N38W113.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N38W113.hgt.zip',
		'N38W114.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N38W114.hgt.zip',
		'N38W115.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N38W115.hgt.zip',
		'N38W116.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N38W116.hgt.zip',
		'N38W117.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N38W117.hgt.zip',
		'N38W118.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N38W118.hgt.zip',
		'N38W119.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N38W119.hgt.zip',
		'N38W120.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N38W120.hgt.zip',
		'N38W121.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N38W121.hgt.zip',
		'N38W122.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N38W122.hgt.zip',
		'N38W123.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N38W123.hgt.zip',
		'N38W124.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N38W124.hgt.zip',
		'N39W075.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N39W075.hgt.zip',
		'N39W076.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N39W076.hgt.zip',
		'N39W077.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N39W077.hgt.zip',
		'N39W078.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N39W078.hgt.zip',
		'N39W079.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N39W079.hgt.zip',
		'N39W080.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N39W080.hgt.zip',
		'N39W081.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N39W081.hgt.zip',
		'N39W082.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N39W082.hgt.zip',
		'N39W083.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N39W083.hgt.zip',
		'N39W084.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N39W084.hgt.zip',
		'N39W085.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N39W085.hgt.zip',
		'N39W086.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N39W086.hgt.zip',
		'N39W087.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N39W087.hgt.zip',
		'N39W088.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N39W088.hgt.zip',
		'N39W089.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N39W089.hgt.zip',
		'N39W090.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N39W090.hgt.zip',
		'N39W091.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N39W091.hgt.zip',
		'N39W092.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N39W092.hgt.zip',
		'N39W093.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N39W093.hgt.zip',
		'N39W094.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N39W094.hgt.zip',
		'N39W095.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N39W095.hgt.zip',
		'N39W096.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N39W096.hgt.zip',
		'N39W097.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N39W097.hgt.zip',
		'N39W098.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N39W098.hgt.zip',
		'N39W099.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N39W099.hgt.zip',
		'N39W100.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N39W100.hgt.zip',
		'N39W101.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N39W101.hgt.zip',
		'N39W102.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N39W102.hgt.zip',
		'N39W103.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N39W103.hgt.zip',
		'N39W104.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N39W104.hgt.zip',
		'N39W105.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N39W105.hgt.zip',
		'N39W106.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N39W106.hgt.zip',
		'N39W107.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N39W107.hgt.zip',
		'N39W108.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N39W108.hgt.zip',
		'N39W109.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N39W109.hgt.zip',
		'N39W110.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N39W110.hgt.zip',
		'N39W111.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N39W111.hgt.zip',
		'N39W112.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N39W112.hgt.zip',
		'N39W113.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N39W113.hgt.zip',
		'N39W114.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N39W114.hgt.zip',
		'N39W115.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N39W115.hgt.zip',
		'N39W116.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N39W116.hgt.zip',
		'N39W117.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N39W117.hgt.zip',
		'N39W118.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N39W118.hgt.zip',
		'N39W119.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N39W119.hgt.zip',
		'N39W120.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N39W120.hgt.zip',
		'N39W121.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N39W121.hgt.zip',
		'N39W122.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N39W122.hgt.zip',
		'N39W123.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N39W123.hgt.zip',
		'N39W124.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N39W124.hgt.zip',
		'N39W125.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N39W125.hgt.zip',
		'N40W073.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N40W073.hgt.zip',
		'N40W074.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N40W074.hgt.zip',
		'N40W075.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N40W075.hgt.zip',
		'N40W076.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N40W076.hgt.zip',
		'N40W077.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N40W077.hgt.zip',
		'N40W078.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N40W078.hgt.zip',
		'N40W079.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N40W079.hgt.zip',
		'N40W080.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N40W080.hgt.zip',
		'N40W081.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N40W081.hgt.zip',
		'N40W082.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N40W082.hgt.zip',
		'N40W083.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N40W083.hgt.zip',
		'N40W084.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N40W084.hgt.zip',
		'N40W085.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N40W085.hgt.zip',
		'N40W086.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N40W086.hgt.zip',
		'N40W087.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N40W087.hgt.zip',
		'N40W088.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N40W088.hgt.zip',
		'N40W089.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N40W089.hgt.zip',
		'N40W090.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N40W090.hgt.zip',
		'N40W091.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N40W091.hgt.zip',
		'N40W092.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N40W092.hgt.zip',
		'N40W093.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N40W093.hgt.zip',
		'N40W094.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N40W094.hgt.zip',
		'N40W095.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N40W095.hgt.zip',
		'N40W096.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N40W096.hgt.zip',
		'N40W097.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N40W097.hgt.zip',
		'N40W098.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N40W098.hgt.zip',
		'N40W099.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N40W099.hgt.zip',
		'N40W100.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N40W100.hgt.zip',
		'N40W101.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N40W101.hgt.zip',
		'N40W102.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N40W102.hgt.zip',
		'N40W103.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N40W103.hgt.zip',
		'N40W104.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N40W104.hgt.zip',
		'N40W105.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N40W105.hgt.zip',
		'N40W106.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N40W106.hgt.zip',
		'N40W107.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N40W107.hgt.zip',
		'N40W108.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N40W108.hgt.zip',
		'N40W109.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N40W109.hgt.zip',
		'N40W110.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N40W110.hgt.zip',
		'N40W111.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N40W111.hgt.zip',
		'N40W112.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N40W112.hgt.zip',
		'N40W113.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N40W113.hgt.zip',
		'N40W114.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N40W114.hgt.zip',
		'N40W115.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N40W115.hgt.zip',
		'N40W116.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N40W116.hgt.zip',
		'N40W117.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N40W117.hgt.zip',
		'N40W118.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N40W118.hgt.zip',
		'N40W119.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N40W119.hgt.zip',
		'N40W120.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N40W120.hgt.zip',
		'N40W121.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N40W121.hgt.zip',
		'N40W122.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N40W122.hgt.zip',
		'N40W123.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N40W123.hgt.zip',
		'N40W124.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N40W124.hgt.zip',
		'N40W125.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N40W125.hgt.zip',
		'N41W070.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N41W070.hgt.zip',
		'N41W071.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N41W071.hgt.zip',
		'N41W072.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N41W072.hgt.zip',
		'N41W073.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N41W073.hgt.zip',
		'N41W074.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N41W074.hgt.zip',
		'N41W075.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N41W075.hgt.zip',
		'N41W076.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N41W076.hgt.zip',
		'N41W077.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N41W077.hgt.zip',
		'N41W078.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N41W078.hgt.zip',
		'N41W079.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N41W079.hgt.zip',
		'N41W080.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N41W080.hgt.zip',
		'N41W081.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N41W081.hgt.zip',
		'N41W082.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N41W082.hgt.zip',
		'N41W083.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N41W083.hgt.zip',
		'N41W084.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N41W084.hgt.zip',
		'N41W085.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N41W085.hgt.zip',
		'N41W086.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N41W086.hgt.zip',
		'N41W087.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N41W087.hgt.zip',
		'N41W088.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N41W088.hgt.zip',
		'N41W089.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N41W089.hgt.zip',
		'N41W090.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N41W090.hgt.zip',
		'N41W091.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N41W091.hgt.zip',
		'N41W092.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N41W092.hgt.zip',
		'N41W093.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N41W093.hgt.zip',
		'N41W094.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N41W094.hgt.zip',
		'N41W095.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N41W095.hgt.zip',
		'N41W096.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N41W096.hgt.zip',
		'N41W097.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N41W097.hgt.zip',
		'N41W098.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N41W098.hgt.zip',
		'N41W099.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N41W099.hgt.zip',
		'N41W100.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N41W100.hgt.zip',
		'N41W101.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N41W101.hgt.zip',
		'N41W102.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N41W102.hgt.zip',
		'N41W103.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N41W103.hgt.zip',
		'N41W104.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N41W104.hgt.zip',
		'N41W105.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N41W105.hgt.zip',
		'N41W106.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N41W106.hgt.zip',
		'N41W107.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N41W107.hgt.zip',
		'N41W108.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N41W108.hgt.zip',
		'N41W109.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N41W109.hgt.zip',
		'N41W110.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N41W110.hgt.zip',
		'N41W111.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N41W111.hgt.zip',
		'N41W112.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N41W112.hgt.zip',
		'N41W113.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N41W113.hgt.zip',
		'N41W114.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N41W114.hgt.zip',
		'N41W115.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N41W115.hgt.zip',
		'N41W116.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N41W116.hgt.zip',
		'N41W117.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N41W117.hgt.zip',
		'N41W118.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N41W118.hgt.zip',
		'N41W119.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N41W119.hgt.zip',
		'N41W120.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N41W120.hgt.zip',
		'N41W121.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N41W121.hgt.zip',
		'N41W122.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N41W122.hgt.zip',
		'N41W123.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N41W123.hgt.zip',
		'N41W124.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N41W124.hgt.zip',
		'N41W125.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N41W125.hgt.zip',
		'N42W071.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N42W071.hgt.zip',
		'N42W072.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N42W072.hgt.zip',
		'N42W073.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N42W073.hgt.zip',
		'N42W074.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N42W074.hgt.zip',
		'N42W075.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N42W075.hgt.zip',
		'N42W076.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N42W076.hgt.zip',
		'N42W077.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N42W077.hgt.zip',
		'N42W078.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N42W078.hgt.zip',
		'N42W079.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N42W079.hgt.zip',
		'N42W080.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N42W080.hgt.zip',
		'N42W081.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N42W081.hgt.zip',
		'N42W082.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N42W082.hgt.zip',
		'N42W083.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N42W083.hgt.zip',
		'N42W084.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N42W084.hgt.zip',
		'N42W085.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N42W085.hgt.zip',
		'N42W086.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N42W086.hgt.zip',
		'N42W087.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N42W087.hgt.zip',
		'N42W088.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N42W088.hgt.zip',
		'N42W089.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N42W089.hgt.zip',
		'N42W090.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N42W090.hgt.zip',
		'N42W091.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N42W091.hgt.zip',
		'N42W092.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N42W092.hgt.zip',
		'N42W093.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N42W093.hgt.zip',
		'N42W094.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N42W094.hgt.zip',
		'N42W095.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N42W095.hgt.zip',
		'N42W096.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N42W096.hgt.zip',
		'N42W097.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N42W097.hgt.zip',
		'N42W098.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N42W098.hgt.zip',
		'N42W099.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N42W099.hgt.zip',
		'N42W100.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N42W100.hgt.zip',
		'N42W101.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N42W101.hgt.zip',
		'N42W102.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N42W102.hgt.zip',
		'N42W103.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N42W103.hgt.zip',
		'N42W104.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N42W104.hgt.zip',
		'N42W105.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N42W105.hgt.zip',
		'N42W106.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N42W106.hgt.zip',
		'N42W107.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N42W107.hgt.zip',
		'N42W108.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N42W108.hgt.zip',
		'N42W109.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N42W109.hgt.zip',
		'N42W110.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N42W110.hgt.zip',
		'N42W111.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N42W111.hgt.zip',
		'N42W112.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N42W112.hgt.zip',
		'N42W113.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N42W113.hgt.zip',
		'N42W114.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N42W114.hgt.zip',
		'N42W115.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N42W115.hgt.zip',
		'N42W116.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N42W116.hgt.zip',
		'N42W117.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N42W117.hgt.zip',
		'N42W118.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N42W118.hgt.zip',
		'N42W119.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N42W119.hgt.zip',
		'N42W120.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N42W120.hgt.zip',
		'N42W121.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N42W121.hgt.zip',
		'N42W122.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N42W122.hgt.zip',
		'N42W123.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N42W123.hgt.zip',
		'N42W124.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N42W124.hgt.zip',
		'N42W125.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N42W125.hgt.zip',
		'N43W069.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N43W069.hgt.zip',
		'N43W070.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N43W070.hgt.zip',
		'N43W071.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N43W071.hgt.zip',
		'N43W072.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N43W072.hgt.zip',
		'N43W073.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N43W073.hgt.zip',
		'N43W074.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N43W074.hgt.zip',
		'N43W075.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N43W075.hgt.zip',
		'N43W076.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N43W076.hgt.zip',
		'N43W077.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N43W077.hgt.zip',
		'N43W078.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N43W078.hgt.zip',
		'N43W079.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N43W079.hgt.zip',
		'N43W080.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N43W080.hgt.zip',
		'N43W083.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N43W083.hgt.zip',
		'N43W084.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N43W084.hgt.zip',
		'N43W085.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N43W085.hgt.zip',
		'N43W086.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N43W086.hgt.zip',
		'N43W087.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N43W087.hgt.zip',
		'N43W088.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N43W088.hgt.zip',
		'N43W089.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N43W089.hgt.zip',
		'N43W090.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N43W090.hgt.zip',
		'N43W091.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N43W091.hgt.zip',
		'N43W092.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N43W092.hgt.zip',
		'N43W093.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N43W093.hgt.zip',
		'N43W094.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N43W094.hgt.zip',
		'N43W095.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N43W095.hgt.zip',
		'N43W096.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N43W096.hgt.zip',
		'N43W097.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N43W097.hgt.zip',
		'N43W098.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N43W098.hgt.zip',
		'N43W099.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N43W099.hgt.zip',
		'N43W100.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N43W100.hgt.zip',
		'N43W101.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N43W101.hgt.zip',
		'N43W102.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N43W102.hgt.zip',
		'N43W103.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N43W103.hgt.zip',
		'N43W104.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N43W104.hgt.zip',
		'N43W105.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N43W105.hgt.zip',
		'N43W106.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N43W106.hgt.zip',
		'N43W107.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N43W107.hgt.zip',
		'N43W108.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N43W108.hgt.zip',
		'N43W109.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N43W109.hgt.zip',
		'N43W110.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N43W110.hgt.zip',
		'N43W111.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N43W111.hgt.zip',
		'N43W112.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N43W112.hgt.zip',
		'N43W113.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N43W113.hgt.zip',
		'N43W114.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N43W114.hgt.zip',
		'N43W115.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N43W115.hgt.zip',
		'N43W116.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N43W116.hgt.zip',
		'N43W117.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N43W117.hgt.zip',
		'N43W118.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N43W118.hgt.zip',
		'N43W119.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N43W119.hgt.zip',
		'N43W120.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N43W120.hgt.zip',
		'N43W121.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N43W121.hgt.zip',
		'N43W122.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N43W122.hgt.zip',
		'N43W123.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N43W123.hgt.zip',
		'N43W124.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N43W124.hgt.zip',
		'N43W125.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N43W125.hgt.zip',
		'N44W067.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N44W067.hgt.zip',
		'N44W068.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N44W068.hgt.zip',
		'N44W069.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N44W069.hgt.zip',
		'N44W070.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N44W070.hgt.zip',
		'N44W071.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N44W071.hgt.zip',
		'N44W072.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N44W072.hgt.zip',
		'N44W073.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N44W073.hgt.zip',
		'N44W074.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N44W074.hgt.zip',
		'N44W075.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N44W075.hgt.zip',
		'N44W076.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N44W076.hgt.zip',
		'N44W077.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N44W077.hgt.zip',
		'N44W083.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N44W083.hgt.zip',
		'N44W084.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N44W084.hgt.zip',
		'N44W085.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N44W085.hgt.zip',
		'N44W086.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N44W086.hgt.zip',
		'N44W087.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N44W087.hgt.zip',
		'N44W088.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N44W088.hgt.zip',
		'N44W089.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N44W089.hgt.zip',
		'N44W090.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N44W090.hgt.zip',
		'N44W091.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N44W091.hgt.zip',
		'N44W092.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N44W092.hgt.zip',
		'N44W093.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N44W093.hgt.zip',
		'N44W094.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N44W094.hgt.zip',
		'N44W095.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N44W095.hgt.zip',
		'N44W096.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N44W096.hgt.zip',
		'N44W097.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N44W097.hgt.zip',
		'N44W098.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N44W098.hgt.zip',
		'N44W099.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N44W099.hgt.zip',
		'N44W100.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N44W100.hgt.zip',
		'N44W101.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N44W101.hgt.zip',
		'N44W102.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N44W102.hgt.zip',
		'N44W103.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N44W103.hgt.zip',
		'N44W104.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N44W104.hgt.zip',
		'N44W105.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N44W105.hgt.zip',
		'N44W106.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N44W106.hgt.zip',
		'N44W107.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N44W107.hgt.zip',
		'N44W108.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N44W108.hgt.zip',
		'N44W109.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N44W109.hgt.zip',
		'N44W110.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N44W110.hgt.zip',
		'N44W111.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N44W111.hgt.zip',
		'N44W112.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N44W112.hgt.zip',
		'N44W113.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N44W113.hgt.zip',
		'N44W114.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N44W114.hgt.zip',
		'N44W115.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N44W115.hgt.zip',
		'N44W116.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N44W116.hgt.zip',
		'N44W117.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N44W117.hgt.zip',
		'N44W118.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N44W118.hgt.zip',
		'N44W119.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N44W119.hgt.zip',
		'N44W120.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N44W120.hgt.zip',
		'N44W121.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N44W121.hgt.zip',
		'N44W122.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N44W122.hgt.zip',
		'N44W123.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N44W123.hgt.zip',
		'N44W124.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N44W124.hgt.zip',
		'N44W125.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N44W125.hgt.zip',
		'N45W068.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N45W068.hgt.zip',
		'N45W069.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N45W069.hgt.zip',
		'N45W070.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N45W070.hgt.zip',
		'N45W071.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N45W071.hgt.zip',
		'N45W072.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N45W072.hgt.zip',
		'N45W073.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N45W073.hgt.zip',
		'N45W074.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N45W074.hgt.zip',
		'N45W075.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N45W075.hgt.zip',
		'N45W083.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N45W083.hgt.zip',
		'N45W084.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N45W084.hgt.zip',
		'N45W085.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N45W085.hgt.zip',
		'N45W086.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N45W086.hgt.zip',
		'N45W087.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N45W087.hgt.zip',
		'N45W088.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N45W088.hgt.zip',
		'N45W089.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N45W089.hgt.zip',
		'N45W090.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N45W090.hgt.zip',
		'N45W091.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N45W091.hgt.zip',
		'N45W092.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N45W092.hgt.zip',
		'N45W093.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N45W093.hgt.zip',
		'N45W094.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N45W094.hgt.zip',
		'N45W095.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N45W095.hgt.zip',
		'N45W096.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N45W096.hgt.zip',
		'N45W097.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N45W097.hgt.zip',
		'N45W098.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N45W098.hgt.zip',
		'N45W099.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N45W099.hgt.zip',
		'N45W100.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N45W100.hgt.zip',
		'N45W101.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N45W101.hgt.zip',
		'N45W102.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N45W102.hgt.zip',
		'N45W103.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N45W103.hgt.zip',
		'N45W104.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N45W104.hgt.zip',
		'N45W105.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N45W105.hgt.zip',
		'N45W106.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N45W106.hgt.zip',
		'N45W107.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N45W107.hgt.zip',
		'N45W108.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N45W108.hgt.zip',
		'N45W109.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N45W109.hgt.zip',
		'N45W110.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N45W110.hgt.zip',
		'N45W111.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N45W111.hgt.zip',
		'N45W112.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N45W112.hgt.zip',
		'N45W113.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N45W113.hgt.zip',
		'N45W114.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N45W114.hgt.zip',
		'N45W115.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N45W115.hgt.zip',
		'N45W116.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N45W116.hgt.zip',
		'N45W117.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N45W117.hgt.zip',
		'N45W118.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N45W118.hgt.zip',
		'N45W119.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N45W119.hgt.zip',
		'N45W120.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N45W120.hgt.zip',
		'N45W121.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N45W121.hgt.zip',
		'N45W122.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N45W122.hgt.zip',
		'N45W123.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N45W123.hgt.zip',
		'N45W124.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N45W124.hgt.zip',
		'N45W125.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N45W125.hgt.zip',
		'N46W068.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N46W068.hgt.zip',
		'N46W069.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N46W069.hgt.zip',
		'N46W070.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N46W070.hgt.zip',
		'N46W071.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N46W071.hgt.zip',
		'N46W084.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N46W084.hgt.zip',
		'N46W085.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N46W085.hgt.zip',
		'N46W086.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N46W086.hgt.zip',
		'N46W087.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N46W087.hgt.zip',
		'N46W088.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N46W088.hgt.zip',
		'N46W089.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N46W089.hgt.zip',
		'N46W090.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N46W090.hgt.zip',
		'N46W091.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N46W091.hgt.zip',
		'N46W092.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N46W092.hgt.zip',
		'N46W093.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N46W093.hgt.zip',
		'N46W094.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N46W094.hgt.zip',
		'N46W095.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N46W095.hgt.zip',
		'N46W096.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N46W096.hgt.zip',
		'N46W097.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N46W097.hgt.zip',
		'N46W098.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N46W098.hgt.zip',
		'N46W099.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N46W099.hgt.zip',
		'N46W100.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N46W100.hgt.zip',
		'N46W101.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N46W101.hgt.zip',
		'N46W102.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N46W102.hgt.zip',
		'N46W103.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N46W103.hgt.zip',
		'N46W104.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N46W104.hgt.zip',
		'N46W105.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N46W105.hgt.zip',
		'N46W106.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N46W106.hgt.zip',
		'N46W107.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N46W107.hgt.zip',
		'N46W108.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N46W108.hgt.zip',
		'N46W109.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N46W109.hgt.zip',
		'N46W110.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N46W110.hgt.zip',
		'N46W111.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N46W111.hgt.zip',
		'N46W112.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N46W112.hgt.zip',
		'N46W113.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N46W113.hgt.zip',
		'N46W114.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N46W114.hgt.zip',
		'N46W115.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N46W115.hgt.zip',
		'N46W116.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N46W116.hgt.zip',
		'N46W117.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N46W117.hgt.zip',
		'N46W118.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N46W118.hgt.zip',
		'N46W119.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N46W119.hgt.zip',
		'N46W120.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N46W120.hgt.zip',
		'N46W121.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N46W121.hgt.zip',
		'N46W122.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N46W122.hgt.zip',
		'N46W123.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N46W123.hgt.zip',
		'N46W124.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N46W124.hgt.zip',
		'N46W125.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N46W125.hgt.zip',
		'N47W068.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N47W068.hgt.zip',
		'N47W069.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N47W069.hgt.zip',
		'N47W070.hgt': 'https://srtm.kurviger.de/SRTM1//Region_06//N47W070.hgt.zip',
		'N47W086.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N47W086.hgt.zip',
		'N47W088.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N47W088.hgt.zip',
		'N47W089.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N47W089.hgt.zip',
		'N47W090.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N47W090.hgt.zip',
		'N47W091.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N47W091.hgt.zip',
		'N47W092.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N47W092.hgt.zip',
		'N47W093.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N47W093.hgt.zip',
		'N47W094.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N47W094.hgt.zip',
		'N47W095.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N47W095.hgt.zip',
		'N47W096.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N47W096.hgt.zip',
		'N47W097.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N47W097.hgt.zip',
		'N47W098.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N47W098.hgt.zip',
		'N47W099.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N47W099.hgt.zip',
		'N47W100.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N47W100.hgt.zip',
		'N47W101.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N47W101.hgt.zip',
		'N47W102.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N47W102.hgt.zip',
		'N47W103.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N47W103.hgt.zip',
		'N47W104.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N47W104.hgt.zip',
		'N47W105.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N47W105.hgt.zip',
		'N47W106.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N47W106.hgt.zip',
		'N47W107.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N47W107.hgt.zip',
		'N47W108.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N47W108.hgt.zip',
		'N47W109.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N47W109.hgt.zip',
		'N47W110.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N47W110.hgt.zip',
		'N47W111.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N47W111.hgt.zip',
		'N47W112.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N47W112.hgt.zip',
		'N47W113.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N47W113.hgt.zip',
		'N47W114.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N47W114.hgt.zip',
		'N47W115.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N47W115.hgt.zip',
		'N47W116.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N47W116.hgt.zip',
		'N47W117.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N47W117.hgt.zip',
		'N47W118.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N47W118.hgt.zip',
		'N47W119.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N47W119.hgt.zip',
		'N47W120.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N47W120.hgt.zip',
		'N47W121.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N47W121.hgt.zip',
		'N47W122.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N47W122.hgt.zip',
		'N47W123.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N47W123.hgt.zip',
		'N47W124.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N47W124.hgt.zip',
		'N47W125.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N47W125.hgt.zip',
		'N48W088.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N48W088.hgt.zip',
		'N48W089.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N48W089.hgt.zip',
		'N48W090.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N48W090.hgt.zip',
		'N48W091.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N48W091.hgt.zip',
		'N48W092.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N48W092.hgt.zip',
		'N48W093.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N48W093.hgt.zip',
		'N48W094.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N48W094.hgt.zip',
		'N48W095.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N48W095.hgt.zip',
		'N48W096.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N48W096.hgt.zip',
		'N48W097.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N48W097.hgt.zip',
		'N48W098.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N48W098.hgt.zip',
		'N48W099.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N48W099.hgt.zip',
		'N48W100.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N48W100.hgt.zip',
		'N48W101.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N48W101.hgt.zip',
		'N48W102.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N48W102.hgt.zip',
		'N48W103.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N48W103.hgt.zip',
		'N48W104.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N48W104.hgt.zip',
		'N48W105.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N48W105.hgt.zip',
		'N48W106.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N48W106.hgt.zip',
		'N48W107.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N48W107.hgt.zip',
		'N48W108.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N48W108.hgt.zip',
		'N48W109.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N48W109.hgt.zip',
		'N48W110.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N48W110.hgt.zip',
		'N48W111.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N48W111.hgt.zip',
		'N48W112.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N48W112.hgt.zip',
		'N48W113.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N48W113.hgt.zip',
		'N48W114.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N48W114.hgt.zip',
		'N48W115.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N48W115.hgt.zip',
		'N48W116.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N48W116.hgt.zip',
		'N48W117.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N48W117.hgt.zip',
		'N48W118.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N48W118.hgt.zip',
		'N48W119.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N48W119.hgt.zip',
		'N48W120.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N48W120.hgt.zip',
		'N48W121.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N48W121.hgt.zip',
		'N48W122.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N48W122.hgt.zip',
		'N48W123.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N48W123.hgt.zip',
		'N48W124.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N48W124.hgt.zip',
		'N48W125.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N48W125.hgt.zip',
		'N48W126.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N48W126.hgt.zip',
		'N49W095.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N49W095.hgt.zip',
		'N49W096.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N49W096.hgt.zip',
		'N49W097.hgt': 'https://srtm.kurviger.de/SRTM1//Region_03//N49W097.hgt.zip',
		'N49W098.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N49W098.hgt.zip',
		'N49W099.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N49W099.hgt.zip',
		'N49W100.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N49W100.hgt.zip',
		'N49W105.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N49W105.hgt.zip',
		'N49W107.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N49W107.hgt.zip',
		'N49W108.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N49W108.hgt.zip',
		'N49W110.hgt': 'https://srtm.kurviger.de/SRTM1//Region_02//N49W110.hgt.zip',
		'N49W115.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N49W115.hgt.zip',
		'N49W116.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N49W116.hgt.zip',
		'N49W117.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N49W117.hgt.zip',
		'N49W118.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N49W118.hgt.zip',
		'N49W119.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N49W119.hgt.zip',
		'N49W120.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N49W120.hgt.zip',
		'N49W121.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N49W121.hgt.zip',
		'N49W122.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N49W122.hgt.zip',
		'N49W123.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N49W123.hgt.zip',
		'N49W124.hgt': 'https://srtm.kurviger.de/SRTM1//Region_01//N49W124.hgt.zip',
		'N51E178.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N51E178.hgt.zip',
		'N51E179.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N51E179.hgt.zip',
		'N51W176.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N51W176.hgt.zip',
		'N51W177.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N51W177.hgt.zip',
		'N51W178.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N51W178.hgt.zip',
		'N51W179.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N51W179.hgt.zip',
		'N51W180.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N51W180.hgt.zip',
		'N52E178.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N52E178.hgt.zip',
		'N52E179.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N52E179.hgt.zip',
		'N52W169.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N52W169.hgt.zip',
		'N52W170.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N52W170.hgt.zip',
		'N52W171.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N52W171.hgt.zip',
		'N52W172.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N52W172.hgt.zip',
		'N52W173.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N52W173.hgt.zip',
		'N52W174.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N52W174.hgt.zip',
		'N52W175.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N52W175.hgt.zip',
		'N52W176.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N52W176.hgt.zip',
		'N52W177.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N52W177.hgt.zip',
		'N53W167.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N53W167.hgt.zip',
		'N53W168.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N53W168.hgt.zip',
		'N53W169.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N53W169.hgt.zip',
		'N53W170.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N53W170.hgt.zip',
		'N54W131.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N54W131.hgt.zip',
		'N54W132.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N54W132.hgt.zip',
		'N54W133.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N54W133.hgt.zip',
		'N54W134.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N54W134.hgt.zip',
		'N54W160.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N54W160.hgt.zip',
		'N54W161.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N54W161.hgt.zip',
		'N54W162.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N54W162.hgt.zip',
		'N54W163.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N54W163.hgt.zip',
		'N54W164.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N54W164.hgt.zip',
		'N54W165.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N54W165.hgt.zip',
		'N54W166.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N54W166.hgt.zip',
		'N54W167.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N54W167.hgt.zip',
		'N55W130.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N55W130.hgt.zip',
		'N55W131.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N55W131.hgt.zip',
		'N55W132.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N55W132.hgt.zip',
		'N55W133.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N55W133.hgt.zip',
		'N55W134.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N55W134.hgt.zip',
		'N55W156.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N55W156.hgt.zip',
		'N55W157.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N55W157.hgt.zip',
		'N55W159.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N55W159.hgt.zip',
		'N55W160.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N55W160.hgt.zip',
		'N55W161.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N55W161.hgt.zip',
		'N55W162.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N55W162.hgt.zip',
		'N55W163.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N55W163.hgt.zip',
		'N55W164.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N55W164.hgt.zip',
		'N56W131.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N56W131.hgt.zip',
		'N56W132.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N56W132.hgt.zip',
		'N56W133.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N56W133.hgt.zip',
		'N56W134.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N56W134.hgt.zip',
		'N56W135.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N56W135.hgt.zip',
		'N56W136.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N56W136.hgt.zip',
		'N56W154.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N56W154.hgt.zip',
		'N56W155.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N56W155.hgt.zip',
		'N56W157.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N56W157.hgt.zip',
		'N56W158.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N56W158.hgt.zip',
		'N56W159.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N56W159.hgt.zip',
		'N56W160.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N56W160.hgt.zip',
		'N56W161.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N56W161.hgt.zip',
		'N56W162.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N56W162.hgt.zip',
		'N56W170.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N56W170.hgt.zip',
		'N57W133.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N57W133.hgt.zip',
		'N57W134.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N57W134.hgt.zip',
		'N57W135.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N57W135.hgt.zip',
		'N57W136.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N57W136.hgt.zip',
		'N57W137.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N57W137.hgt.zip',
		'N57W153.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N57W153.hgt.zip',
		'N57W154.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N57W154.hgt.zip',
		'N57W155.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N57W155.hgt.zip',
		'N57W156.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N57W156.hgt.zip',
		'N57W157.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N57W157.hgt.zip',
		'N57W158.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N57W158.hgt.zip',
		'N57W159.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N57W159.hgt.zip',
		'N57W170.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N57W170.hgt.zip',
		'N57W171.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N57W171.hgt.zip',
		'N58W134.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N58W134.hgt.zip',
		'N58W135.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N58W135.hgt.zip',
		'N58W136.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N58W136.hgt.zip',
		'N58W137.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N58W137.hgt.zip',
		'N58W138.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N58W138.hgt.zip',
		'N58W139.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N58W139.hgt.zip',
		'N58W152.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N58W152.hgt.zip',
		'N58W153.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N58W153.hgt.zip',
		'N58W154.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N58W154.hgt.zip',
		'N58W155.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N58W155.hgt.zip',
		'N58W156.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N58W156.hgt.zip',
		'N58W157.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N58W157.hgt.zip',
		'N58W158.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N58W158.hgt.zip',
		'N58W159.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N58W159.hgt.zip',
		'N58W160.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N58W160.hgt.zip',
		'N58W161.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N58W161.hgt.zip',
		'N58W162.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N58W162.hgt.zip',
		'N58W163.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N58W163.hgt.zip',
		'N59W135.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N59W135.hgt.zip',
		'N59W136.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N59W136.hgt.zip',
		'N59W137.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N59W137.hgt.zip',
		'N59W138.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N59W138.hgt.zip',
		'N59W139.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N59W139.hgt.zip',
		'N59W140.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N59W140.hgt.zip',
		'N59W141.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N59W141.hgt.zip',
		'N59W142.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N59W142.hgt.zip',
		'N59W144.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N59W144.hgt.zip',
		'N59W145.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N59W145.hgt.zip',
		'N59W147.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N59W147.hgt.zip',
		'N59W148.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N59W148.hgt.zip',
		'N59W149.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N59W149.hgt.zip',
		'N59W150.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N59W150.hgt.zip',
		'N59W151.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N59W151.hgt.zip',
		'N59W152.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N59W152.hgt.zip',
		'N59W153.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N59W153.hgt.zip',
		'N59W154.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N59W154.hgt.zip',
		'N59W155.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N59W155.hgt.zip',
		'N59W156.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N59W156.hgt.zip',
		'N59W157.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N59W157.hgt.zip',
		'N59W158.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N59W158.hgt.zip',
		'N59W159.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N59W159.hgt.zip',
		'N59W160.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N59W160.hgt.zip',
		'N59W161.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N59W161.hgt.zip',
		'N59W162.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N59W162.hgt.zip',
		'N59W163.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N59W163.hgt.zip',
		'N59W164.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N59W164.hgt.zip',
		'N59W165.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//N59W165.hgt.zip',
		'S12W172.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//S12W172.hgt.zip',
		'S15W169.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//S15W169.hgt.zip',
		'S15W170.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//S15W170.hgt.zip',
		'S15W171.hgt': 'https://srtm.kurviger.de/SRTM1//Region_07//S15W171.hgt.zip',
	},
	'srtm3': {
		'N00E006.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N00E006.hgt.zip',
		'N00E009.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N00E009.hgt.zip',
		'N00E010.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N00E010.hgt.zip',
		'N00E011.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N00E011.hgt.zip',
		'N00E012.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N00E012.hgt.zip',
		'N00E013.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N00E013.hgt.zip',
		'N00E014.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N00E014.hgt.zip',
		'N00E015.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N00E015.hgt.zip',
		'N00E016.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N00E016.hgt.zip',
		'N00E017.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N00E017.hgt.zip',
		'N00E018.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N00E018.hgt.zip',
		'N00E019.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N00E019.hgt.zip',
		'N00E020.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N00E020.hgt.zip',
		'N00E021.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N00E021.hgt.zip',
		'N00E022.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N00E022.hgt.zip',
		'N00E023.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N00E023.hgt.zip',
		'N00E024.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N00E024.hgt.zip',
		'N00E025.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N00E025.hgt.zip',
		'N00E026.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N00E026.hgt.zip',
		'N00E027.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N00E027.hgt.zip',
		'N00E028.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N00E028.hgt.zip',
		'N00E029.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N00E029.hgt.zip',
		'N00E030.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N00E030.hgt.zip',
		'N00E031.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N00E031.hgt.zip',
		'N00E032.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N00E032.hgt.zip',
		'N00E033.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N00E033.hgt.zip',
		'N00E034.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N00E034.hgt.zip',
		'N00E035.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N00E035.hgt.zip',
		'N00E036.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N00E036.hgt.zip',
		'N00E037.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N00E037.hgt.zip',
		'N00E038.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N00E038.hgt.zip',
		'N00E039.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N00E039.hgt.zip',
		'N00E040.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N00E040.hgt.zip',
		'N00E041.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N00E041.hgt.zip',
		'N00E042.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N00E042.hgt.zip',
		'N00E043.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N00E043.hgt.zip',
		'N00E072.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N00E072.hgt.zip',
		'N00E073.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N00E073.hgt.zip',
		'N00E097.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N00E097.hgt.zip',
		'N00E098.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N00E098.hgt.zip',
		'N00E099.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N00E099.hgt.zip',
		'N00E100.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N00E100.hgt.zip',
		'N00E101.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N00E101.hgt.zip',
		'N00E102.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N00E102.hgt.zip',
		'N00E103.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N00E103.hgt.zip',
		'N00E104.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N00E104.hgt.zip',
		'N00E106.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N00E106.hgt.zip',
		'N00E107.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N00E107.hgt.zip',
		'N00E108.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N00E108.hgt.zip',
		'N00E109.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N00E109.hgt.zip',
		'N00E110.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N00E110.hgt.zip',
		'N00E111.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N00E111.hgt.zip',
		'N00E112.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N00E112.hgt.zip',
		'N00E113.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N00E113.hgt.zip',
		'N00E114.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N00E114.hgt.zip',
		'N00E115.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N00E115.hgt.zip',
		'N00E116.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N00E116.hgt.zip',
		'N00E117.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N00E117.hgt.zip',
		'N00E118.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N00E118.hgt.zip',
		'N00E119.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N00E119.hgt.zip',
		'N00E120.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N00E120.hgt.zip',
		'N00E121.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N00E121.hgt.zip',
		'N00E122.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N00E122.hgt.zip',
		'N00E123.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N00E123.hgt.zip',
		'N00E124.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N00E124.hgt.zip',
		'N00E126.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N00E126.hgt.zip',
		'N00E127.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N00E127.hgt.zip',
		'N00E128.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N00E128.hgt.zip',
		'N00E129.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N00E129.hgt.zip',
		'N00E130.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N00E130.hgt.zip',
		'N00E131.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N00E131.hgt.zip',
		'N00E134.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N00E134.hgt.zip',
		'N00E172.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N00E172.hgt.zip',
		'N00E173.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N00E173.hgt.zip',
		'N00W050.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N00W050.hgt.zip',
		'N00W051.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N00W051.hgt.zip',
		'N00W052.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N00W052.hgt.zip',
		'N00W053.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N00W053.hgt.zip',
		'N00W054.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N00W054.hgt.zip',
		'N00W055.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N00W055.hgt.zip',
		'N00W056.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N00W056.hgt.zip',
		'N00W057.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N00W057.hgt.zip',
		'N00W058.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N00W058.hgt.zip',
		'N00W059.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N00W059.hgt.zip',
		'N00W060.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N00W060.hgt.zip',
		'N00W061.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N00W061.hgt.zip',
		'N00W062.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N00W062.hgt.zip',
		'N00W063.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N00W063.hgt.zip',
		'N00W064.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N00W064.hgt.zip',
		'N00W065.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N00W065.hgt.zip',
		'N00W066.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N00W066.hgt.zip',
		'N00W067.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N00W067.hgt.zip',
		'N00W068.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N00W068.hgt.zip',
		'N00W069.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N00W069.hgt.zip',
		'N00W070.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N00W070.hgt.zip',
		'N00W071.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N00W071.hgt.zip',
		'N00W072.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N00W072.hgt.zip',
		'N00W073.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N00W073.hgt.zip',
		'N00W074.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N00W074.hgt.zip',
		'N00W075.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N00W075.hgt.zip',
		'N00W076.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N00W076.hgt.zip',
		'N00W077.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N00W077.hgt.zip',
		'N00W078.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N00W078.hgt.zip',
		'N00W079.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N00W079.hgt.zip',
		'N00W080.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N00W080.hgt.zip',
		'N00W081.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N00W081.hgt.zip',
		'N00W090.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N00W090.hgt.zip',
		'N00W091.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N00W091.hgt.zip',
		'N00W092.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N00W092.hgt.zip',
		'N00W177.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N00W177.hgt.zip',
		'N01E007.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N01E007.hgt.zip',
		'N01E009.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N01E009.hgt.zip',
		'N01E010.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N01E010.hgt.zip',
		'N01E011.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N01E011.hgt.zip',
		'N01E012.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N01E012.hgt.zip',
		'N01E013.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N01E013.hgt.zip',
		'N01E014.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N01E014.hgt.zip',
		'N01E015.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N01E015.hgt.zip',
		'N01E016.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N01E016.hgt.zip',
		'N01E017.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N01E017.hgt.zip',
		'N01E018.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N01E018.hgt.zip',
		'N01E019.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N01E019.hgt.zip',
		'N01E020.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N01E020.hgt.zip',
		'N01E021.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N01E021.hgt.zip',
		'N01E022.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N01E022.hgt.zip',
		'N01E023.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N01E023.hgt.zip',
		'N01E024.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N01E024.hgt.zip',
		'N01E025.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N01E025.hgt.zip',
		'N01E026.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N01E026.hgt.zip',
		'N01E027.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N01E027.hgt.zip',
		'N01E028.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N01E028.hgt.zip',
		'N01E029.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N01E029.hgt.zip',
		'N01E030.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N01E030.hgt.zip',
		'N01E031.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N01E031.hgt.zip',
		'N01E032.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N01E032.hgt.zip',
		'N01E033.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N01E033.hgt.zip',
		'N01E034.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N01E034.hgt.zip',
		'N01E035.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N01E035.hgt.zip',
		'N01E036.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N01E036.hgt.zip',
		'N01E037.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N01E037.hgt.zip',
		'N01E038.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N01E038.hgt.zip',
		'N01E039.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N01E039.hgt.zip',
		'N01E040.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N01E040.hgt.zip',
		'N01E041.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N01E041.hgt.zip',
		'N01E042.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N01E042.hgt.zip',
		'N01E043.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N01E043.hgt.zip',
		'N01E044.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N01E044.hgt.zip',
		'N01E045.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N01E045.hgt.zip',
		'N01E073.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N01E073.hgt.zip',
		'N01E097.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N01E097.hgt.zip',
		'N01E098.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N01E098.hgt.zip',
		'N01E099.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N01E099.hgt.zip',
		'N01E100.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N01E100.hgt.zip',
		'N01E101.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N01E101.hgt.zip',
		'N01E102.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N01E102.hgt.zip',
		'N01E103.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N01E103.hgt.zip',
		'N01E104.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N01E104.hgt.zip',
		'N01E106.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N01E106.hgt.zip',
		'N01E107.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N01E107.hgt.zip',
		'N01E108.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N01E108.hgt.zip',
		'N01E109.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N01E109.hgt.zip',
		'N01E110.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N01E110.hgt.zip',
		'N01E111.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N01E111.hgt.zip',
		'N01E112.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N01E112.hgt.zip',
		'N01E113.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N01E113.hgt.zip',
		'N01E114.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N01E114.hgt.zip',
		'N01E115.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N01E115.hgt.zip',
		'N01E116.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N01E116.hgt.zip',
		'N01E117.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N01E117.hgt.zip',
		'N01E118.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N01E118.hgt.zip',
		'N01E119.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N01E119.hgt.zip',
		'N01E120.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N01E120.hgt.zip',
		'N01E121.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N01E121.hgt.zip',
		'N01E122.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N01E122.hgt.zip',
		'N01E124.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N01E124.hgt.zip',
		'N01E125.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N01E125.hgt.zip',
		'N01E126.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N01E126.hgt.zip',
		'N01E127.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N01E127.hgt.zip',
		'N01E128.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N01E128.hgt.zip',
		'N01E131.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N01E131.hgt.zip',
		'N01E154.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N01E154.hgt.zip',
		'N01E172.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N01E172.hgt.zip',
		'N01E173.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N01E173.hgt.zip',
		'N01W050.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N01W050.hgt.zip',
		'N01W051.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N01W051.hgt.zip',
		'N01W052.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N01W052.hgt.zip',
		'N01W053.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N01W053.hgt.zip',
		'N01W054.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N01W054.hgt.zip',
		'N01W055.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N01W055.hgt.zip',
		'N01W056.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N01W056.hgt.zip',
		'N01W057.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N01W057.hgt.zip',
		'N01W058.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N01W058.hgt.zip',
		'N01W059.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N01W059.hgt.zip',
		'N01W060.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N01W060.hgt.zip',
		'N01W061.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N01W061.hgt.zip',
		'N01W062.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N01W062.hgt.zip',
		'N01W063.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N01W063.hgt.zip',
		'N01W064.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N01W064.hgt.zip',
		'N01W065.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N01W065.hgt.zip',
		'N01W066.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N01W066.hgt.zip',
		'N01W067.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N01W067.hgt.zip',
		'N01W068.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N01W068.hgt.zip',
		'N01W069.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N01W069.hgt.zip',
		'N01W070.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N01W070.hgt.zip',
		'N01W071.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N01W071.hgt.zip',
		'N01W072.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N01W072.hgt.zip',
		'N01W073.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N01W073.hgt.zip',
		'N01W074.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N01W074.hgt.zip',
		'N01W075.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N01W075.hgt.zip',
		'N01W076.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N01W076.hgt.zip',
		'N01W077.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N01W077.hgt.zip',
		'N01W078.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N01W078.hgt.zip',
		'N01W079.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N01W079.hgt.zip',
		'N01W080.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N01W080.hgt.zip',
		'N01W092.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N01W092.hgt.zip',
		'N01W158.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N01W158.hgt.zip',
		'N02E009.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N02E009.hgt.zip',
		'N02E010.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N02E010.hgt.zip',
		'N02E011.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N02E011.hgt.zip',
		'N02E012.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N02E012.hgt.zip',
		'N02E013.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N02E013.hgt.zip',
		'N02E014.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N02E014.hgt.zip',
		'N02E015.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N02E015.hgt.zip',
		'N02E016.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N02E016.hgt.zip',
		'N02E017.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N02E017.hgt.zip',
		'N02E018.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N02E018.hgt.zip',
		'N02E019.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N02E019.hgt.zip',
		'N02E020.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N02E020.hgt.zip',
		'N02E021.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N02E021.hgt.zip',
		'N02E022.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N02E022.hgt.zip',
		'N02E023.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N02E023.hgt.zip',
		'N02E024.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N02E024.hgt.zip',
		'N02E025.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N02E025.hgt.zip',
		'N02E026.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N02E026.hgt.zip',
		'N02E027.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N02E027.hgt.zip',
		'N02E028.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N02E028.hgt.zip',
		'N02E029.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N02E029.hgt.zip',
		'N02E030.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N02E030.hgt.zip',
		'N02E031.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N02E031.hgt.zip',
		'N02E032.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N02E032.hgt.zip',
		'N02E033.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N02E033.hgt.zip',
		'N02E034.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N02E034.hgt.zip',
		'N02E035.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N02E035.hgt.zip',
		'N02E036.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N02E036.hgt.zip',
		'N02E037.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N02E037.hgt.zip',
		'N02E038.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N02E038.hgt.zip',
		'N02E039.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N02E039.hgt.zip',
		'N02E040.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N02E040.hgt.zip',
		'N02E041.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N02E041.hgt.zip',
		'N02E042.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N02E042.hgt.zip',
		'N02E043.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N02E043.hgt.zip',
		'N02E044.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N02E044.hgt.zip',
		'N02E045.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N02E045.hgt.zip',
		'N02E046.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N02E046.hgt.zip',
		'N02E072.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N02E072.hgt.zip',
		'N02E073.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N02E073.hgt.zip',
		'N02E095.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N02E095.hgt.zip',
		'N02E096.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N02E096.hgt.zip',
		'N02E097.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N02E097.hgt.zip',
		'N02E098.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N02E098.hgt.zip',
		'N02E099.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N02E099.hgt.zip',
		'N02E100.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N02E100.hgt.zip',
		'N02E101.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N02E101.hgt.zip',
		'N02E102.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N02E102.hgt.zip',
		'N02E103.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N02E103.hgt.zip',
		'N02E104.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N02E104.hgt.zip',
		'N02E105.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N02E105.hgt.zip',
		'N02E106.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N02E106.hgt.zip',
		'N02E107.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N02E107.hgt.zip',
		'N02E108.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N02E108.hgt.zip',
		'N02E109.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N02E109.hgt.zip',
		'N02E111.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N02E111.hgt.zip',
		'N02E112.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N02E112.hgt.zip',
		'N02E113.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N02E113.hgt.zip',
		'N02E114.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N02E114.hgt.zip',
		'N02E115.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N02E115.hgt.zip',
		'N02E116.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N02E116.hgt.zip',
		'N02E117.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N02E117.hgt.zip',
		'N02E118.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N02E118.hgt.zip',
		'N02E125.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N02E125.hgt.zip',
		'N02E127.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N02E127.hgt.zip',
		'N02E128.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N02E128.hgt.zip',
		'N02E131.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N02E131.hgt.zip',
		'N02E173.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N02E173.hgt.zip',
		'N02W051.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N02W051.hgt.zip',
		'N02W052.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N02W052.hgt.zip',
		'N02W053.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N02W053.hgt.zip',
		'N02W054.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N02W054.hgt.zip',
		'N02W055.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N02W055.hgt.zip',
		'N02W056.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N02W056.hgt.zip',
		'N02W057.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N02W057.hgt.zip',
		'N02W058.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N02W058.hgt.zip',
		'N02W059.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N02W059.hgt.zip',
		'N02W060.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N02W060.hgt.zip',
		'N02W061.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N02W061.hgt.zip',
		'N02W062.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N02W062.hgt.zip',
		'N02W063.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N02W063.hgt.zip',
		'N02W064.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N02W064.hgt.zip',
		'N02W065.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N02W065.hgt.zip',
		'N02W066.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N02W066.hgt.zip',
		'N02W067.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N02W067.hgt.zip',
		'N02W068.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N02W068.hgt.zip',
		'N02W069.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N02W069.hgt.zip',
		'N02W070.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N02W070.hgt.zip',
		'N02W071.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N02W071.hgt.zip',
		'N02W072.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N02W072.hgt.zip',
		'N02W073.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N02W073.hgt.zip',
		'N02W074.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N02W074.hgt.zip',
		'N02W075.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N02W075.hgt.zip',
		'N02W076.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N02W076.hgt.zip',
		'N02W077.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N02W077.hgt.zip',
		'N02W078.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N02W078.hgt.zip',
		'N02W079.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N02W079.hgt.zip',
		'N02W158.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N02W158.hgt.zip',
		'N03E008.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N03E008.hgt.zip',
		'N03E009.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N03E009.hgt.zip',
		'N03E010.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N03E010.hgt.zip',
		'N03E011.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N03E011.hgt.zip',
		'N03E012.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N03E012.hgt.zip',
		'N03E013.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N03E013.hgt.zip',
		'N03E014.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N03E014.hgt.zip',
		'N03E015.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N03E015.hgt.zip',
		'N03E016.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N03E016.hgt.zip',
		'N03E017.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N03E017.hgt.zip',
		'N03E018.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N03E018.hgt.zip',
		'N03E019.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N03E019.hgt.zip',
		'N03E020.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N03E020.hgt.zip',
		'N03E021.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N03E021.hgt.zip',
		'N03E022.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N03E022.hgt.zip',
		'N03E023.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N03E023.hgt.zip',
		'N03E024.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N03E024.hgt.zip',
		'N03E025.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N03E025.hgt.zip',
		'N03E026.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N03E026.hgt.zip',
		'N03E027.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N03E027.hgt.zip',
		'N03E028.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N03E028.hgt.zip',
		'N03E029.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N03E029.hgt.zip',
		'N03E030.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N03E030.hgt.zip',
		'N03E031.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N03E031.hgt.zip',
		'N03E032.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N03E032.hgt.zip',
		'N03E033.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N03E033.hgt.zip',
		'N03E034.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N03E034.hgt.zip',
		'N03E035.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N03E035.hgt.zip',
		'N03E036.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N03E036.hgt.zip',
		'N03E037.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N03E037.hgt.zip',
		'N03E038.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N03E038.hgt.zip',
		'N03E039.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N03E039.hgt.zip',
		'N03E040.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N03E040.hgt.zip',
		'N03E041.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N03E041.hgt.zip',
		'N03E042.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N03E042.hgt.zip',
		'N03E043.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N03E043.hgt.zip',
		'N03E044.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N03E044.hgt.zip',
		'N03E045.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N03E045.hgt.zip',
		'N03E046.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N03E046.hgt.zip',
		'N03E047.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N03E047.hgt.zip',
		'N03E072.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N03E072.hgt.zip',
		'N03E073.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N03E073.hgt.zip',
		'N03E095.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N03E095.hgt.zip',
		'N03E096.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N03E096.hgt.zip',
		'N03E097.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N03E097.hgt.zip',
		'N03E098.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N03E098.hgt.zip',
		'N03E099.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N03E099.hgt.zip',
		'N03E100.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N03E100.hgt.zip',
		'N03E101.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N03E101.hgt.zip',
		'N03E102.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N03E102.hgt.zip',
		'N03E103.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N03E103.hgt.zip',
		'N03E105.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N03E105.hgt.zip',
		'N03E106.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N03E106.hgt.zip',
		'N03E107.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N03E107.hgt.zip',
		'N03E108.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N03E108.hgt.zip',
		'N03E112.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N03E112.hgt.zip',
		'N03E113.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N03E113.hgt.zip',
		'N03E114.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N03E114.hgt.zip',
		'N03E115.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N03E115.hgt.zip',
		'N03E116.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N03E116.hgt.zip',
		'N03E117.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N03E117.hgt.zip',
		'N03E125.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N03E125.hgt.zip',
		'N03E126.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N03E126.hgt.zip',
		'N03E131.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N03E131.hgt.zip',
		'N03E154.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N03E154.hgt.zip',
		'N03E172.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N03E172.hgt.zip',
		'N03E173.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N03E173.hgt.zip',
		'N03W051.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N03W051.hgt.zip',
		'N03W052.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N03W052.hgt.zip',
		'N03W053.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N03W053.hgt.zip',
		'N03W054.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N03W054.hgt.zip',
		'N03W055.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N03W055.hgt.zip',
		'N03W056.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N03W056.hgt.zip',
		'N03W057.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N03W057.hgt.zip',
		'N03W058.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N03W058.hgt.zip',
		'N03W059.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N03W059.hgt.zip',
		'N03W060.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N03W060.hgt.zip',
		'N03W061.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N03W061.hgt.zip',
		'N03W062.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N03W062.hgt.zip',
		'N03W063.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N03W063.hgt.zip',
		'N03W064.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N03W064.hgt.zip',
		'N03W065.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N03W065.hgt.zip',
		'N03W066.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N03W066.hgt.zip',
		'N03W067.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N03W067.hgt.zip',
		'N03W068.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N03W068.hgt.zip',
		'N03W069.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N03W069.hgt.zip',
		'N03W070.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N03W070.hgt.zip',
		'N03W071.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N03W071.hgt.zip',
		'N03W072.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N03W072.hgt.zip',
		'N03W073.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N03W073.hgt.zip',
		'N03W074.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N03W074.hgt.zip',
		'N03W075.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N03W075.hgt.zip',
		'N03W076.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N03W076.hgt.zip',
		'N03W077.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N03W077.hgt.zip',
		'N03W078.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N03W078.hgt.zip',
		'N03W079.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N03W079.hgt.zip',
		'N03W082.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N03W082.hgt.zip',
		'N03W160.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N03W160.hgt.zip',
		'N04E005.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N04E005.hgt.zip',
		'N04E006.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N04E006.hgt.zip',
		'N04E007.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N04E007.hgt.zip',
		'N04E008.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N04E008.hgt.zip',
		'N04E009.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N04E009.hgt.zip',
		'N04E010.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N04E010.hgt.zip',
		'N04E011.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N04E011.hgt.zip',
		'N04E012.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N04E012.hgt.zip',
		'N04E013.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N04E013.hgt.zip',
		'N04E014.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N04E014.hgt.zip',
		'N04E015.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N04E015.hgt.zip',
		'N04E016.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N04E016.hgt.zip',
		'N04E017.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N04E017.hgt.zip',
		'N04E018.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N04E018.hgt.zip',
		'N04E019.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N04E019.hgt.zip',
		'N04E020.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N04E020.hgt.zip',
		'N04E021.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N04E021.hgt.zip',
		'N04E022.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N04E022.hgt.zip',
		'N04E023.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N04E023.hgt.zip',
		'N04E024.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N04E024.hgt.zip',
		'N04E025.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N04E025.hgt.zip',
		'N04E026.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N04E026.hgt.zip',
		'N04E027.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N04E027.hgt.zip',
		'N04E028.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N04E028.hgt.zip',
		'N04E029.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N04E029.hgt.zip',
		'N04E030.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N04E030.hgt.zip',
		'N04E031.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N04E031.hgt.zip',
		'N04E032.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N04E032.hgt.zip',
		'N04E033.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N04E033.hgt.zip',
		'N04E034.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N04E034.hgt.zip',
		'N04E035.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N04E035.hgt.zip',
		'N04E036.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N04E036.hgt.zip',
		'N04E037.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N04E037.hgt.zip',
		'N04E038.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N04E038.hgt.zip',
		'N04E039.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N04E039.hgt.zip',
		'N04E040.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N04E040.hgt.zip',
		'N04E041.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N04E041.hgt.zip',
		'N04E042.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N04E042.hgt.zip',
		'N04E043.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N04E043.hgt.zip',
		'N04E044.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N04E044.hgt.zip',
		'N04E045.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N04E045.hgt.zip',
		'N04E046.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N04E046.hgt.zip',
		'N04E047.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N04E047.hgt.zip',
		'N04E048.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N04E048.hgt.zip',
		'N04E072.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N04E072.hgt.zip',
		'N04E073.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N04E073.hgt.zip',
		'N04E095.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N04E095.hgt.zip',
		'N04E096.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N04E096.hgt.zip',
		'N04E097.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N04E097.hgt.zip',
		'N04E098.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N04E098.hgt.zip',
		'N04E100.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N04E100.hgt.zip',
		'N04E101.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N04E101.hgt.zip',
		'N04E102.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N04E102.hgt.zip',
		'N04E103.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N04E103.hgt.zip',
		'N04E107.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N04E107.hgt.zip',
		'N04E108.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N04E108.hgt.zip',
		'N04E113.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N04E113.hgt.zip',
		'N04E114.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N04E114.hgt.zip',
		'N04E115.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N04E115.hgt.zip',
		'N04E116.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N04E116.hgt.zip',
		'N04E117.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N04E117.hgt.zip',
		'N04E118.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N04E118.hgt.zip',
		'N04E119.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N04E119.hgt.zip',
		'N04E125.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N04E125.hgt.zip',
		'N04E126.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N04E126.hgt.zip',
		'N04E127.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N04E127.hgt.zip',
		'N04E131.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N04E131.hgt.zip',
		'N04E132.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N04E132.hgt.zip',
		'N04E168.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N04E168.hgt.zip',
		'N04W002.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N04W002.hgt.zip',
		'N04W003.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N04W003.hgt.zip',
		'N04W006.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N04W006.hgt.zip',
		'N04W007.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N04W007.hgt.zip',
		'N04W008.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N04W008.hgt.zip',
		'N04W009.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N04W009.hgt.zip',
		'N04W010.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N04W010.hgt.zip',
		'N04W052.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N04W052.hgt.zip',
		'N04W053.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N04W053.hgt.zip',
		'N04W054.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N04W054.hgt.zip',
		'N04W055.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N04W055.hgt.zip',
		'N04W056.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N04W056.hgt.zip',
		'N04W057.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N04W057.hgt.zip',
		'N04W058.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N04W058.hgt.zip',
		'N04W059.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N04W059.hgt.zip',
		'N04W060.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N04W060.hgt.zip',
		'N04W061.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N04W061.hgt.zip',
		'N04W062.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N04W062.hgt.zip',
		'N04W063.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N04W063.hgt.zip',
		'N04W064.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N04W064.hgt.zip',
		'N04W065.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N04W065.hgt.zip',
		'N04W066.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N04W066.hgt.zip',
		'N04W067.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N04W067.hgt.zip',
		'N04W068.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N04W068.hgt.zip',
		'N04W069.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N04W069.hgt.zip',
		'N04W070.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N04W070.hgt.zip',
		'N04W071.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N04W071.hgt.zip',
		'N04W072.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N04W072.hgt.zip',
		'N04W073.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N04W073.hgt.zip',
		'N04W074.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N04W074.hgt.zip',
		'N04W075.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N04W075.hgt.zip',
		'N04W076.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N04W076.hgt.zip',
		'N04W077.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N04W077.hgt.zip',
		'N04W078.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N04W078.hgt.zip',
		'N04W082.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N04W082.hgt.zip',
		'N04W161.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N04W161.hgt.zip',
		'N05E000.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N05E000.hgt.zip',
		'N05E001.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N05E001.hgt.zip',
		'N05E004.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N05E004.hgt.zip',
		'N05E005.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N05E005.hgt.zip',
		'N05E006.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N05E006.hgt.zip',
		'N05E007.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N05E007.hgt.zip',
		'N05E008.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N05E008.hgt.zip',
		'N05E009.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N05E009.hgt.zip',
		'N05E010.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N05E010.hgt.zip',
		'N05E011.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N05E011.hgt.zip',
		'N05E012.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N05E012.hgt.zip',
		'N05E013.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N05E013.hgt.zip',
		'N05E014.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N05E014.hgt.zip',
		'N05E015.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N05E015.hgt.zip',
		'N05E016.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N05E016.hgt.zip',
		'N05E017.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N05E017.hgt.zip',
		'N05E018.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N05E018.hgt.zip',
		'N05E019.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N05E019.hgt.zip',
		'N05E020.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N05E020.hgt.zip',
		'N05E021.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N05E021.hgt.zip',
		'N05E022.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N05E022.hgt.zip',
		'N05E023.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N05E023.hgt.zip',
		'N05E024.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N05E024.hgt.zip',
		'N05E025.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N05E025.hgt.zip',
		'N05E026.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N05E026.hgt.zip',
		'N05E027.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N05E027.hgt.zip',
		'N05E028.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N05E028.hgt.zip',
		'N05E029.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N05E029.hgt.zip',
		'N05E030.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N05E030.hgt.zip',
		'N05E031.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N05E031.hgt.zip',
		'N05E032.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N05E032.hgt.zip',
		'N05E033.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N05E033.hgt.zip',
		'N05E034.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N05E034.hgt.zip',
		'N05E035.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N05E035.hgt.zip',
		'N05E036.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N05E036.hgt.zip',
		'N05E037.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N05E037.hgt.zip',
		'N05E038.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N05E038.hgt.zip',
		'N05E039.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N05E039.hgt.zip',
		'N05E040.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N05E040.hgt.zip',
		'N05E041.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N05E041.hgt.zip',
		'N05E042.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N05E042.hgt.zip',
		'N05E043.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N05E043.hgt.zip',
		'N05E044.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N05E044.hgt.zip',
		'N05E045.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N05E045.hgt.zip',
		'N05E046.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N05E046.hgt.zip',
		'N05E047.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N05E047.hgt.zip',
		'N05E048.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N05E048.hgt.zip',
		'N05E072.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N05E072.hgt.zip',
		'N05E073.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N05E073.hgt.zip',
		'N05E080.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N05E080.hgt.zip',
		'N05E094.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N05E094.hgt.zip',
		'N05E095.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N05E095.hgt.zip',
		'N05E096.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N05E096.hgt.zip',
		'N05E097.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N05E097.hgt.zip',
		'N05E100.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N05E100.hgt.zip',
		'N05E101.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N05E101.hgt.zip',
		'N05E102.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N05E102.hgt.zip',
		'N05E103.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N05E103.hgt.zip',
		'N05E114.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N05E114.hgt.zip',
		'N05E115.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N05E115.hgt.zip',
		'N05E116.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N05E116.hgt.zip',
		'N05E117.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N05E117.hgt.zip',
		'N05E118.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N05E118.hgt.zip',
		'N05E119.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N05E119.hgt.zip',
		'N05E120.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N05E120.hgt.zip',
		'N05E121.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N05E121.hgt.zip',
		'N05E124.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N05E124.hgt.zip',
		'N05E125.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N05E125.hgt.zip',
		'N05E126.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N05E126.hgt.zip',
		'N05E132.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N05E132.hgt.zip',
		'N05E153.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N05E153.hgt.zip',
		'N05E157.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N05E157.hgt.zip',
		'N05E162.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N05E162.hgt.zip',
		'N05E163.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N05E163.hgt.zip',
		'N05E168.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N05E168.hgt.zip',
		'N05E169.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N05E169.hgt.zip',
		'N05E172.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N05E172.hgt.zip',
		'N05W001.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N05W001.hgt.zip',
		'N05W002.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N05W002.hgt.zip',
		'N05W003.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N05W003.hgt.zip',
		'N05W004.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N05W004.hgt.zip',
		'N05W005.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N05W005.hgt.zip',
		'N05W006.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N05W006.hgt.zip',
		'N05W007.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N05W007.hgt.zip',
		'N05W008.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N05W008.hgt.zip',
		'N05W009.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N05W009.hgt.zip',
		'N05W010.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N05W010.hgt.zip',
		'N05W011.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N05W011.hgt.zip',
		'N05W053.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N05W053.hgt.zip',
		'N05W054.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N05W054.hgt.zip',
		'N05W055.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N05W055.hgt.zip',
		'N05W056.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N05W056.hgt.zip',
		'N05W057.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N05W057.hgt.zip',
		'N05W058.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N05W058.hgt.zip',
		'N05W059.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N05W059.hgt.zip',
		'N05W060.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N05W060.hgt.zip',
		'N05W061.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N05W061.hgt.zip',
		'N05W062.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N05W062.hgt.zip',
		'N05W063.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N05W063.hgt.zip',
		'N05W064.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N05W064.hgt.zip',
		'N05W065.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N05W065.hgt.zip',
		'N05W066.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N05W066.hgt.zip',
		'N05W067.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N05W067.hgt.zip',
		'N05W068.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N05W068.hgt.zip',
		'N05W069.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N05W069.hgt.zip',
		'N05W070.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N05W070.hgt.zip',
		'N05W071.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N05W071.hgt.zip',
		'N05W072.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N05W072.hgt.zip',
		'N05W073.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N05W073.hgt.zip',
		'N05W074.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N05W074.hgt.zip',
		'N05W075.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N05W075.hgt.zip',
		'N05W076.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N05W076.hgt.zip',
		'N05W077.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N05W077.hgt.zip',
		'N05W078.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N05W078.hgt.zip',
		'N05W088.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N05W088.hgt.zip',
		'N05W163.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N05W163.hgt.zip',
		'N06E000.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N06E000.hgt.zip',
		'N06E001.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N06E001.hgt.zip',
		'N06E002.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N06E002.hgt.zip',
		'N06E003.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N06E003.hgt.zip',
		'N06E004.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N06E004.hgt.zip',
		'N06E005.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N06E005.hgt.zip',
		'N06E006.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N06E006.hgt.zip',
		'N06E007.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N06E007.hgt.zip',
		'N06E008.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N06E008.hgt.zip',
		'N06E009.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N06E009.hgt.zip',
		'N06E010.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N06E010.hgt.zip',
		'N06E011.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N06E011.hgt.zip',
		'N06E012.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N06E012.hgt.zip',
		'N06E013.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N06E013.hgt.zip',
		'N06E014.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N06E014.hgt.zip',
		'N06E015.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N06E015.hgt.zip',
		'N06E016.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N06E016.hgt.zip',
		'N06E017.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N06E017.hgt.zip',
		'N06E018.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N06E018.hgt.zip',
		'N06E019.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N06E019.hgt.zip',
		'N06E020.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N06E020.hgt.zip',
		'N06E021.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N06E021.hgt.zip',
		'N06E022.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N06E022.hgt.zip',
		'N06E023.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N06E023.hgt.zip',
		'N06E024.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N06E024.hgt.zip',
		'N06E025.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N06E025.hgt.zip',
		'N06E026.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N06E026.hgt.zip',
		'N06E027.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N06E027.hgt.zip',
		'N06E028.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N06E028.hgt.zip',
		'N06E029.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N06E029.hgt.zip',
		'N06E030.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N06E030.hgt.zip',
		'N06E031.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N06E031.hgt.zip',
		'N06E032.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N06E032.hgt.zip',
		'N06E033.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N06E033.hgt.zip',
		'N06E034.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N06E034.hgt.zip',
		'N06E035.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N06E035.hgt.zip',
		'N06E036.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N06E036.hgt.zip',
		'N06E037.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N06E037.hgt.zip',
		'N06E038.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N06E038.hgt.zip',
		'N06E039.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N06E039.hgt.zip',
		'N06E040.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N06E040.hgt.zip',
		'N06E041.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N06E041.hgt.zip',
		'N06E042.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N06E042.hgt.zip',
		'N06E043.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N06E043.hgt.zip',
		'N06E044.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N06E044.hgt.zip',
		'N06E045.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N06E045.hgt.zip',
		'N06E046.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N06E046.hgt.zip',
		'N06E047.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N06E047.hgt.zip',
		'N06E048.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N06E048.hgt.zip',
		'N06E049.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N06E049.hgt.zip',
		'N06E072.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N06E072.hgt.zip',
		'N06E073.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N06E073.hgt.zip',
		'N06E079.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N06E079.hgt.zip',
		'N06E080.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N06E080.hgt.zip',
		'N06E081.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N06E081.hgt.zip',
		'N06E093.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N06E093.hgt.zip',
		'N06E095.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N06E095.hgt.zip',
		'N06E099.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N06E099.hgt.zip',
		'N06E100.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N06E100.hgt.zip',
		'N06E101.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N06E101.hgt.zip',
		'N06E102.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N06E102.hgt.zip',
		'N06E115.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N06E115.hgt.zip',
		'N06E116.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N06E116.hgt.zip',
		'N06E117.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N06E117.hgt.zip',
		'N06E118.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N06E118.hgt.zip',
		'N06E120.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N06E120.hgt.zip',
		'N06E121.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N06E121.hgt.zip',
		'N06E122.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N06E122.hgt.zip',
		'N06E123.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N06E123.hgt.zip',
		'N06E124.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N06E124.hgt.zip',
		'N06E125.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N06E125.hgt.zip',
		'N06E126.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N06E126.hgt.zip',
		'N06E134.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N06E134.hgt.zip',
		'N06E143.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N06E143.hgt.zip',
		'N06E149.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N06E149.hgt.zip',
		'N06E151.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N06E151.hgt.zip',
		'N06E152.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N06E152.hgt.zip',
		'N06E157.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N06E157.hgt.zip',
		'N06E158.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N06E158.hgt.zip',
		'N06E159.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N06E159.hgt.zip',
		'N06E160.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N06E160.hgt.zip',
		'N06E169.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N06E169.hgt.zip',
		'N06E171.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N06E171.hgt.zip',
		'N06E172.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N06E172.hgt.zip',
		'N06W001.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N06W001.hgt.zip',
		'N06W002.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N06W002.hgt.zip',
		'N06W003.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N06W003.hgt.zip',
		'N06W004.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N06W004.hgt.zip',
		'N06W005.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N06W005.hgt.zip',
		'N06W006.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N06W006.hgt.zip',
		'N06W007.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N06W007.hgt.zip',
		'N06W008.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N06W008.hgt.zip',
		'N06W009.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N06W009.hgt.zip',
		'N06W010.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N06W010.hgt.zip',
		'N06W011.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N06W011.hgt.zip',
		'N06W012.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N06W012.hgt.zip',
		'N06W056.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N06W056.hgt.zip',
		'N06W057.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N06W057.hgt.zip',
		'N06W058.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N06W058.hgt.zip',
		'N06W059.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N06W059.hgt.zip',
		'N06W060.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N06W060.hgt.zip',
		'N06W061.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N06W061.hgt.zip',
		'N06W062.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N06W062.hgt.zip',
		'N06W063.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N06W063.hgt.zip',
		'N06W064.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N06W064.hgt.zip',
		'N06W065.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N06W065.hgt.zip',
		'N06W066.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N06W066.hgt.zip',
		'N06W067.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N06W067.hgt.zip',
		'N06W068.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N06W068.hgt.zip',
		'N06W069.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N06W069.hgt.zip',
		'N06W070.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N06W070.hgt.zip',
		'N06W071.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N06W071.hgt.zip',
		'N06W072.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N06W072.hgt.zip',
		'N06W073.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N06W073.hgt.zip',
		'N06W074.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N06W074.hgt.zip',
		'N06W075.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N06W075.hgt.zip',
		'N06W076.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N06W076.hgt.zip',
		'N06W077.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N06W077.hgt.zip',
		'N06W078.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N06W078.hgt.zip',
		'N06W163.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N06W163.hgt.zip',
		'N07E000.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N07E000.hgt.zip',
		'N07E001.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N07E001.hgt.zip',
		'N07E002.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N07E002.hgt.zip',
		'N07E003.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N07E003.hgt.zip',
		'N07E004.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N07E004.hgt.zip',
		'N07E005.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N07E005.hgt.zip',
		'N07E006.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N07E006.hgt.zip',
		'N07E007.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N07E007.hgt.zip',
		'N07E008.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N07E008.hgt.zip',
		'N07E009.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N07E009.hgt.zip',
		'N07E010.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N07E010.hgt.zip',
		'N07E011.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N07E011.hgt.zip',
		'N07E012.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N07E012.hgt.zip',
		'N07E013.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N07E013.hgt.zip',
		'N07E014.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N07E014.hgt.zip',
		'N07E015.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N07E015.hgt.zip',
		'N07E016.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N07E016.hgt.zip',
		'N07E017.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N07E017.hgt.zip',
		'N07E018.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N07E018.hgt.zip',
		'N07E019.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N07E019.hgt.zip',
		'N07E020.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N07E020.hgt.zip',
		'N07E021.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N07E021.hgt.zip',
		'N07E022.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N07E022.hgt.zip',
		'N07E023.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N07E023.hgt.zip',
		'N07E024.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N07E024.hgt.zip',
		'N07E025.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N07E025.hgt.zip',
		'N07E026.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N07E026.hgt.zip',
		'N07E027.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N07E027.hgt.zip',
		'N07E028.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N07E028.hgt.zip',
		'N07E029.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N07E029.hgt.zip',
		'N07E030.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N07E030.hgt.zip',
		'N07E031.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N07E031.hgt.zip',
		'N07E032.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N07E032.hgt.zip',
		'N07E033.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N07E033.hgt.zip',
		'N07E034.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N07E034.hgt.zip',
		'N07E035.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N07E035.hgt.zip',
		'N07E036.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N07E036.hgt.zip',
		'N07E037.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N07E037.hgt.zip',
		'N07E038.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N07E038.hgt.zip',
		'N07E039.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N07E039.hgt.zip',
		'N07E040.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N07E040.hgt.zip',
		'N07E041.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N07E041.hgt.zip',
		'N07E042.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N07E042.hgt.zip',
		'N07E043.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N07E043.hgt.zip',
		'N07E044.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N07E044.hgt.zip',
		'N07E045.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N07E045.hgt.zip',
		'N07E046.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N07E046.hgt.zip',
		'N07E047.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N07E047.hgt.zip',
		'N07E048.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N07E048.hgt.zip',
		'N07E049.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N07E049.hgt.zip',
		'N07E072.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N07E072.hgt.zip',
		'N07E073.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N07E073.hgt.zip',
		'N07E079.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N07E079.hgt.zip',
		'N07E080.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N07E080.hgt.zip',
		'N07E081.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N07E081.hgt.zip',
		'N07E093.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N07E093.hgt.zip',
		'N07E098.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N07E098.hgt.zip',
		'N07E099.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N07E099.hgt.zip',
		'N07E100.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N07E100.hgt.zip',
		'N07E113.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N07E113.hgt.zip',
		'N07E116.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N07E116.hgt.zip',
		'N07E117.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N07E117.hgt.zip',
		'N07E118.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N07E118.hgt.zip',
		'N07E121.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N07E121.hgt.zip',
		'N07E122.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N07E122.hgt.zip',
		'N07E123.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N07E123.hgt.zip',
		'N07E124.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N07E124.hgt.zip',
		'N07E125.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N07E125.hgt.zip',
		'N07E126.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N07E126.hgt.zip',
		'N07E134.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N07E134.hgt.zip',
		'N07E143.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N07E143.hgt.zip',
		'N07E144.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N07E144.hgt.zip',
		'N07E145.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N07E145.hgt.zip',
		'N07E146.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N07E146.hgt.zip',
		'N07E147.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N07E147.hgt.zip',
		'N07E149.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N07E149.hgt.zip',
		'N07E151.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N07E151.hgt.zip',
		'N07E152.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N07E152.hgt.zip',
		'N07E155.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N07E155.hgt.zip',
		'N07E157.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N07E157.hgt.zip',
		'N07E158.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N07E158.hgt.zip',
		'N07E168.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N07E168.hgt.zip',
		'N07E171.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N07E171.hgt.zip',
		'N07W001.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N07W001.hgt.zip',
		'N07W002.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N07W002.hgt.zip',
		'N07W003.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N07W003.hgt.zip',
		'N07W004.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N07W004.hgt.zip',
		'N07W005.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N07W005.hgt.zip',
		'N07W006.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N07W006.hgt.zip',
		'N07W007.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N07W007.hgt.zip',
		'N07W008.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N07W008.hgt.zip',
		'N07W009.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N07W009.hgt.zip',
		'N07W010.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N07W010.hgt.zip',
		'N07W011.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N07W011.hgt.zip',
		'N07W012.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N07W012.hgt.zip',
		'N07W013.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N07W013.hgt.zip',
		'N07W014.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N07W014.hgt.zip',
		'N07W059.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N07W059.hgt.zip',
		'N07W060.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N07W060.hgt.zip',
		'N07W061.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N07W061.hgt.zip',
		'N07W062.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N07W062.hgt.zip',
		'N07W063.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N07W063.hgt.zip',
		'N07W064.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N07W064.hgt.zip',
		'N07W065.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N07W065.hgt.zip',
		'N07W066.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N07W066.hgt.zip',
		'N07W067.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N07W067.hgt.zip',
		'N07W068.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N07W068.hgt.zip',
		'N07W069.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N07W069.hgt.zip',
		'N07W070.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N07W070.hgt.zip',
		'N07W071.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N07W071.hgt.zip',
		'N07W072.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N07W072.hgt.zip',
		'N07W073.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N07W073.hgt.zip',
		'N07W074.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N07W074.hgt.zip',
		'N07W075.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N07W075.hgt.zip',
		'N07W076.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N07W076.hgt.zip',
		'N07W077.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N07W077.hgt.zip',
		'N07W078.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N07W078.hgt.zip',
		'N07W079.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N07W079.hgt.zip',
		'N07W080.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N07W080.hgt.zip',
		'N07W081.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N07W081.hgt.zip',
		'N07W082.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N07W082.hgt.zip',
		'N07W083.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N07W083.hgt.zip',
		'N08E000.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N08E000.hgt.zip',
		'N08E001.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N08E001.hgt.zip',
		'N08E002.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N08E002.hgt.zip',
		'N08E003.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N08E003.hgt.zip',
		'N08E004.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N08E004.hgt.zip',
		'N08E005.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N08E005.hgt.zip',
		'N08E006.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N08E006.hgt.zip',
		'N08E007.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N08E007.hgt.zip',
		'N08E008.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N08E008.hgt.zip',
		'N08E009.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N08E009.hgt.zip',
		'N08E010.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N08E010.hgt.zip',
		'N08E011.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N08E011.hgt.zip',
		'N08E012.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N08E012.hgt.zip',
		'N08E013.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N08E013.hgt.zip',
		'N08E014.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N08E014.hgt.zip',
		'N08E015.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N08E015.hgt.zip',
		'N08E016.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N08E016.hgt.zip',
		'N08E017.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N08E017.hgt.zip',
		'N08E018.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N08E018.hgt.zip',
		'N08E019.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N08E019.hgt.zip',
		'N08E020.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N08E020.hgt.zip',
		'N08E021.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N08E021.hgt.zip',
		'N08E022.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N08E022.hgt.zip',
		'N08E023.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N08E023.hgt.zip',
		'N08E024.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N08E024.hgt.zip',
		'N08E025.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N08E025.hgt.zip',
		'N08E026.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N08E026.hgt.zip',
		'N08E027.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N08E027.hgt.zip',
		'N08E028.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N08E028.hgt.zip',
		'N08E029.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N08E029.hgt.zip',
		'N08E030.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N08E030.hgt.zip',
		'N08E031.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N08E031.hgt.zip',
		'N08E032.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N08E032.hgt.zip',
		'N08E033.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N08E033.hgt.zip',
		'N08E034.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N08E034.hgt.zip',
		'N08E035.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N08E035.hgt.zip',
		'N08E036.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N08E036.hgt.zip',
		'N08E037.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N08E037.hgt.zip',
		'N08E038.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N08E038.hgt.zip',
		'N08E039.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N08E039.hgt.zip',
		'N08E040.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N08E040.hgt.zip',
		'N08E041.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N08E041.hgt.zip',
		'N08E042.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N08E042.hgt.zip',
		'N08E043.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N08E043.hgt.zip',
		'N08E044.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N08E044.hgt.zip',
		'N08E045.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N08E045.hgt.zip',
		'N08E046.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N08E046.hgt.zip',
		'N08E047.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N08E047.hgt.zip',
		'N08E048.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N08E048.hgt.zip',
		'N08E049.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N08E049.hgt.zip',
		'N08E050.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N08E050.hgt.zip',
		'N08E073.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N08E073.hgt.zip',
		'N08E076.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N08E076.hgt.zip',
		'N08E077.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N08E077.hgt.zip',
		'N08E078.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N08E078.hgt.zip',
		'N08E079.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N08E079.hgt.zip',
		'N08E080.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N08E080.hgt.zip',
		'N08E081.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N08E081.hgt.zip',
		'N08E092.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N08E092.hgt.zip',
		'N08E093.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N08E093.hgt.zip',
		'N08E097.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N08E097.hgt.zip',
		'N08E098.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N08E098.hgt.zip',
		'N08E099.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N08E099.hgt.zip',
		'N08E100.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N08E100.hgt.zip',
		'N08E104.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N08E104.hgt.zip',
		'N08E105.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N08E105.hgt.zip',
		'N08E106.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N08E106.hgt.zip',
		'N08E111.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N08E111.hgt.zip',
		'N08E116.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N08E116.hgt.zip',
		'N08E117.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N08E117.hgt.zip',
		'N08E118.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N08E118.hgt.zip',
		'N08E122.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N08E122.hgt.zip',
		'N08E123.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N08E123.hgt.zip',
		'N08E124.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N08E124.hgt.zip',
		'N08E125.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N08E125.hgt.zip',
		'N08E126.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N08E126.hgt.zip',
		'N08E134.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N08E134.hgt.zip',
		'N08E137.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N08E137.hgt.zip',
		'N08E140.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N08E140.hgt.zip',
		'N08E144.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N08E144.hgt.zip',
		'N08E146.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N08E146.hgt.zip',
		'N08E147.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N08E147.hgt.zip',
		'N08E149.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N08E149.hgt.zip',
		'N08E150.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N08E150.hgt.zip',
		'N08E151.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N08E151.hgt.zip',
		'N08E152.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N08E152.hgt.zip',
		'N08E154.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N08E154.hgt.zip',
		'N08E165.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N08E165.hgt.zip',
		'N08E166.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N08E166.hgt.zip',
		'N08E167.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N08E167.hgt.zip',
		'N08E168.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N08E168.hgt.zip',
		'N08E170.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N08E170.hgt.zip',
		'N08E171.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N08E171.hgt.zip',
		'N08W001.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N08W001.hgt.zip',
		'N08W002.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N08W002.hgt.zip',
		'N08W003.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N08W003.hgt.zip',
		'N08W004.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N08W004.hgt.zip',
		'N08W005.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N08W005.hgt.zip',
		'N08W006.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N08W006.hgt.zip',
		'N08W007.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N08W007.hgt.zip',
		'N08W008.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N08W008.hgt.zip',
		'N08W009.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N08W009.hgt.zip',
		'N08W010.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N08W010.hgt.zip',
		'N08W011.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N08W011.hgt.zip',
		'N08W012.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N08W012.hgt.zip',
		'N08W013.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N08W013.hgt.zip',
		'N08W014.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N08W014.hgt.zip',
		'N08W060.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N08W060.hgt.zip',
		'N08W061.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N08W061.hgt.zip',
		'N08W062.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N08W062.hgt.zip',
		'N08W063.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N08W063.hgt.zip',
		'N08W064.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N08W064.hgt.zip',
		'N08W065.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N08W065.hgt.zip',
		'N08W066.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N08W066.hgt.zip',
		'N08W067.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N08W067.hgt.zip',
		'N08W068.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N08W068.hgt.zip',
		'N08W069.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N08W069.hgt.zip',
		'N08W070.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N08W070.hgt.zip',
		'N08W071.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N08W071.hgt.zip',
		'N08W072.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N08W072.hgt.zip',
		'N08W073.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N08W073.hgt.zip',
		'N08W074.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N08W074.hgt.zip',
		'N08W075.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N08W075.hgt.zip',
		'N08W076.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N08W076.hgt.zip',
		'N08W077.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N08W077.hgt.zip',
		'N08W078.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N08W078.hgt.zip',
		'N08W079.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N08W079.hgt.zip',
		'N08W080.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N08W080.hgt.zip',
		'N08W081.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N08W081.hgt.zip',
		'N08W082.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N08W082.hgt.zip',
		'N08W083.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N08W083.hgt.zip',
		'N08W084.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N08W084.hgt.zip',
		'N09E000.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N09E000.hgt.zip',
		'N09E001.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N09E001.hgt.zip',
		'N09E002.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N09E002.hgt.zip',
		'N09E003.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N09E003.hgt.zip',
		'N09E004.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N09E004.hgt.zip',
		'N09E005.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N09E005.hgt.zip',
		'N09E006.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N09E006.hgt.zip',
		'N09E007.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N09E007.hgt.zip',
		'N09E008.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N09E008.hgt.zip',
		'N09E009.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N09E009.hgt.zip',
		'N09E010.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N09E010.hgt.zip',
		'N09E011.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N09E011.hgt.zip',
		'N09E012.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N09E012.hgt.zip',
		'N09E013.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N09E013.hgt.zip',
		'N09E014.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N09E014.hgt.zip',
		'N09E015.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N09E015.hgt.zip',
		'N09E016.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N09E016.hgt.zip',
		'N09E017.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N09E017.hgt.zip',
		'N09E018.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N09E018.hgt.zip',
		'N09E019.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N09E019.hgt.zip',
		'N09E020.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N09E020.hgt.zip',
		'N09E021.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N09E021.hgt.zip',
		'N09E022.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N09E022.hgt.zip',
		'N09E023.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N09E023.hgt.zip',
		'N09E024.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N09E024.hgt.zip',
		'N09E025.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N09E025.hgt.zip',
		'N09E026.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N09E026.hgt.zip',
		'N09E027.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N09E027.hgt.zip',
		'N09E028.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N09E028.hgt.zip',
		'N09E029.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N09E029.hgt.zip',
		'N09E030.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N09E030.hgt.zip',
		'N09E031.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N09E031.hgt.zip',
		'N09E032.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N09E032.hgt.zip',
		'N09E033.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N09E033.hgt.zip',
		'N09E034.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N09E034.hgt.zip',
		'N09E035.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N09E035.hgt.zip',
		'N09E036.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N09E036.hgt.zip',
		'N09E037.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N09E037.hgt.zip',
		'N09E038.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N09E038.hgt.zip',
		'N09E039.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N09E039.hgt.zip',
		'N09E040.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N09E040.hgt.zip',
		'N09E041.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N09E041.hgt.zip',
		'N09E042.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N09E042.hgt.zip',
		'N09E043.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N09E043.hgt.zip',
		'N09E044.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N09E044.hgt.zip',
		'N09E045.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N09E045.hgt.zip',
		'N09E046.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N09E046.hgt.zip',
		'N09E047.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N09E047.hgt.zip',
		'N09E048.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N09E048.hgt.zip',
		'N09E049.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N09E049.hgt.zip',
		'N09E050.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N09E050.hgt.zip',
		'N09E076.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N09E076.hgt.zip',
		'N09E077.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N09E077.hgt.zip',
		'N09E078.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N09E078.hgt.zip',
		'N09E079.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N09E079.hgt.zip',
		'N09E080.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N09E080.hgt.zip',
		'N09E092.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N09E092.hgt.zip',
		'N09E097.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N09E097.hgt.zip',
		'N09E098.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N09E098.hgt.zip',
		'N09E099.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N09E099.hgt.zip',
		'N09E100.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N09E100.hgt.zip',
		'N09E102.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N09E102.hgt.zip',
		'N09E103.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N09E103.hgt.zip',
		'N09E104.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N09E104.hgt.zip',
		'N09E105.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N09E105.hgt.zip',
		'N09E106.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N09E106.hgt.zip',
		'N09E109.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N09E109.hgt.zip',
		'N09E117.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N09E117.hgt.zip',
		'N09E118.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N09E118.hgt.zip',
		'N09E119.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N09E119.hgt.zip',
		'N09E120.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N09E120.hgt.zip',
		'N09E121.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N09E121.hgt.zip',
		'N09E122.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N09E122.hgt.zip',
		'N09E123.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N09E123.hgt.zip',
		'N09E124.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N09E124.hgt.zip',
		'N09E125.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N09E125.hgt.zip',
		'N09E126.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N09E126.hgt.zip',
		'N09E138.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N09E138.hgt.zip',
		'N09E139.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N09E139.hgt.zip',
		'N09E140.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N09E140.hgt.zip',
		'N09E145.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N09E145.hgt.zip',
		'N09E160.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N09E160.hgt.zip',
		'N09E165.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N09E165.hgt.zip',
		'N09E166.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N09E166.hgt.zip',
		'N09E167.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N09E167.hgt.zip',
		'N09E169.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N09E169.hgt.zip',
		'N09E170.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N09E170.hgt.zip',
		'N09W001.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N09W001.hgt.zip',
		'N09W002.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N09W002.hgt.zip',
		'N09W003.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N09W003.hgt.zip',
		'N09W004.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N09W004.hgt.zip',
		'N09W005.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N09W005.hgt.zip',
		'N09W006.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N09W006.hgt.zip',
		'N09W007.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N09W007.hgt.zip',
		'N09W008.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N09W008.hgt.zip',
		'N09W009.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N09W009.hgt.zip',
		'N09W010.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N09W010.hgt.zip',
		'N09W011.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N09W011.hgt.zip',
		'N09W012.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N09W012.hgt.zip',
		'N09W013.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N09W013.hgt.zip',
		'N09W014.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N09W014.hgt.zip',
		'N09W015.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N09W015.hgt.zip',
		'N09W061.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N09W061.hgt.zip',
		'N09W062.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N09W062.hgt.zip',
		'N09W063.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N09W063.hgt.zip',
		'N09W064.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N09W064.hgt.zip',
		'N09W065.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N09W065.hgt.zip',
		'N09W066.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N09W066.hgt.zip',
		'N09W067.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N09W067.hgt.zip',
		'N09W068.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N09W068.hgt.zip',
		'N09W069.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N09W069.hgt.zip',
		'N09W070.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N09W070.hgt.zip',
		'N09W071.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N09W071.hgt.zip',
		'N09W072.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N09W072.hgt.zip',
		'N09W073.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N09W073.hgt.zip',
		'N09W074.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N09W074.hgt.zip',
		'N09W075.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N09W075.hgt.zip',
		'N09W076.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N09W076.hgt.zip',
		'N09W077.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N09W077.hgt.zip',
		'N09W078.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N09W078.hgt.zip',
		'N09W079.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N09W079.hgt.zip',
		'N09W080.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N09W080.hgt.zip',
		'N09W081.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N09W081.hgt.zip',
		'N09W082.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N09W082.hgt.zip',
		'N09W083.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N09W083.hgt.zip',
		'N09W084.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N09W084.hgt.zip',
		'N09W085.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N09W085.hgt.zip',
		'N09W086.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N09W086.hgt.zip',
		'N10E000.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N10E000.hgt.zip',
		'N10E001.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N10E001.hgt.zip',
		'N10E002.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N10E002.hgt.zip',
		'N10E003.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N10E003.hgt.zip',
		'N10E004.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N10E004.hgt.zip',
		'N10E005.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N10E005.hgt.zip',
		'N10E006.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N10E006.hgt.zip',
		'N10E007.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N10E007.hgt.zip',
		'N10E008.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N10E008.hgt.zip',
		'N10E009.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N10E009.hgt.zip',
		'N10E010.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N10E010.hgt.zip',
		'N10E011.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N10E011.hgt.zip',
		'N10E012.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N10E012.hgt.zip',
		'N10E013.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N10E013.hgt.zip',
		'N10E014.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N10E014.hgt.zip',
		'N10E015.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N10E015.hgt.zip',
		'N10E016.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N10E016.hgt.zip',
		'N10E017.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N10E017.hgt.zip',
		'N10E018.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N10E018.hgt.zip',
		'N10E019.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N10E019.hgt.zip',
		'N10E020.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N10E020.hgt.zip',
		'N10E021.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N10E021.hgt.zip',
		'N10E022.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N10E022.hgt.zip',
		'N10E023.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N10E023.hgt.zip',
		'N10E024.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N10E024.hgt.zip',
		'N10E025.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N10E025.hgt.zip',
		'N10E026.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N10E026.hgt.zip',
		'N10E027.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N10E027.hgt.zip',
		'N10E028.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N10E028.hgt.zip',
		'N10E029.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N10E029.hgt.zip',
		'N10E030.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N10E030.hgt.zip',
		'N10E031.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N10E031.hgt.zip',
		'N10E032.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N10E032.hgt.zip',
		'N10E033.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N10E033.hgt.zip',
		'N10E034.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N10E034.hgt.zip',
		'N10E035.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N10E035.hgt.zip',
		'N10E036.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N10E036.hgt.zip',
		'N10E037.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N10E037.hgt.zip',
		'N10E038.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N10E038.hgt.zip',
		'N10E039.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N10E039.hgt.zip',
		'N10E040.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N10E040.hgt.zip',
		'N10E041.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N10E041.hgt.zip',
		'N10E042.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N10E042.hgt.zip',
		'N10E043.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N10E043.hgt.zip',
		'N10E044.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N10E044.hgt.zip',
		'N10E045.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N10E045.hgt.zip',
		'N10E046.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N10E046.hgt.zip',
		'N10E047.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N10E047.hgt.zip',
		'N10E048.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N10E048.hgt.zip',
		'N10E049.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N10E049.hgt.zip',
		'N10E050.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N10E050.hgt.zip',
		'N10E051.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N10E051.hgt.zip',
		'N10E072.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N10E072.hgt.zip',
		'N10E073.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N10E073.hgt.zip',
		'N10E075.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N10E075.hgt.zip',
		'N10E076.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N10E076.hgt.zip',
		'N10E077.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N10E077.hgt.zip',
		'N10E078.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N10E078.hgt.zip',
		'N10E079.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N10E079.hgt.zip',
		'N10E092.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N10E092.hgt.zip',
		'N10E097.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N10E097.hgt.zip',
		'N10E098.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N10E098.hgt.zip',
		'N10E099.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N10E099.hgt.zip',
		'N10E102.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N10E102.hgt.zip',
		'N10E103.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N10E103.hgt.zip',
		'N10E104.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N10E104.hgt.zip',
		'N10E105.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N10E105.hgt.zip',
		'N10E106.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N10E106.hgt.zip',
		'N10E107.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N10E107.hgt.zip',
		'N10E108.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N10E108.hgt.zip',
		'N10E114.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N10E114.hgt.zip',
		'N10E115.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N10E115.hgt.zip',
		'N10E118.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N10E118.hgt.zip',
		'N10E119.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N10E119.hgt.zip',
		'N10E120.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N10E120.hgt.zip',
		'N10E121.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N10E121.hgt.zip',
		'N10E122.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N10E122.hgt.zip',
		'N10E123.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N10E123.hgt.zip',
		'N10E124.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N10E124.hgt.zip',
		'N10E125.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N10E125.hgt.zip',
		'N10E126.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N10E126.hgt.zip',
		'N10E139.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N10E139.hgt.zip',
		'N10E165.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N10E165.hgt.zip',
		'N10E166.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N10E166.hgt.zip',
		'N10E168.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N10E168.hgt.zip',
		'N10E169.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N10E169.hgt.zip',
		'N10E170.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N10E170.hgt.zip',
		'N10W001.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N10W001.hgt.zip',
		'N10W002.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N10W002.hgt.zip',
		'N10W003.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N10W003.hgt.zip',
		'N10W004.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N10W004.hgt.zip',
		'N10W005.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N10W005.hgt.zip',
		'N10W006.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N10W006.hgt.zip',
		'N10W007.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N10W007.hgt.zip',
		'N10W008.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N10W008.hgt.zip',
		'N10W009.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N10W009.hgt.zip',
		'N10W010.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N10W010.hgt.zip',
		'N10W011.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N10W011.hgt.zip',
		'N10W012.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N10W012.hgt.zip',
		'N10W013.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N10W013.hgt.zip',
		'N10W014.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N10W014.hgt.zip',
		'N10W015.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N10W015.hgt.zip',
		'N10W016.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N10W016.hgt.zip',
		'N10W061.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N10W061.hgt.zip',
		'N10W062.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N10W062.hgt.zip',
		'N10W063.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N10W063.hgt.zip',
		'N10W064.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N10W064.hgt.zip',
		'N10W065.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N10W065.hgt.zip',
		'N10W066.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N10W066.hgt.zip',
		'N10W067.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N10W067.hgt.zip',
		'N10W068.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N10W068.hgt.zip',
		'N10W069.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N10W069.hgt.zip',
		'N10W070.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N10W070.hgt.zip',
		'N10W071.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N10W071.hgt.zip',
		'N10W072.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N10W072.hgt.zip',
		'N10W073.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N10W073.hgt.zip',
		'N10W074.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N10W074.hgt.zip',
		'N10W075.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N10W075.hgt.zip',
		'N10W076.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N10W076.hgt.zip',
		'N10W084.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N10W084.hgt.zip',
		'N10W085.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N10W085.hgt.zip',
		'N10W086.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N10W086.hgt.zip',
		'N10W110.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N10W110.hgt.zip',
		'N11E000.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N11E000.hgt.zip',
		'N11E001.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N11E001.hgt.zip',
		'N11E002.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N11E002.hgt.zip',
		'N11E003.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N11E003.hgt.zip',
		'N11E004.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N11E004.hgt.zip',
		'N11E005.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N11E005.hgt.zip',
		'N11E006.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N11E006.hgt.zip',
		'N11E007.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N11E007.hgt.zip',
		'N11E008.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N11E008.hgt.zip',
		'N11E009.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N11E009.hgt.zip',
		'N11E010.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N11E010.hgt.zip',
		'N11E011.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N11E011.hgt.zip',
		'N11E012.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N11E012.hgt.zip',
		'N11E013.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N11E013.hgt.zip',
		'N11E014.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N11E014.hgt.zip',
		'N11E015.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N11E015.hgt.zip',
		'N11E016.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N11E016.hgt.zip',
		'N11E017.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N11E017.hgt.zip',
		'N11E018.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N11E018.hgt.zip',
		'N11E019.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N11E019.hgt.zip',
		'N11E020.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N11E020.hgt.zip',
		'N11E021.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N11E021.hgt.zip',
		'N11E022.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N11E022.hgt.zip',
		'N11E023.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N11E023.hgt.zip',
		'N11E024.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N11E024.hgt.zip',
		'N11E025.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N11E025.hgt.zip',
		'N11E026.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N11E026.hgt.zip',
		'N11E027.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N11E027.hgt.zip',
		'N11E028.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N11E028.hgt.zip',
		'N11E029.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N11E029.hgt.zip',
		'N11E030.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N11E030.hgt.zip',
		'N11E031.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N11E031.hgt.zip',
		'N11E032.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N11E032.hgt.zip',
		'N11E033.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N11E033.hgt.zip',
		'N11E034.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N11E034.hgt.zip',
		'N11E035.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N11E035.hgt.zip',
		'N11E036.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N11E036.hgt.zip',
		'N11E037.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N11E037.hgt.zip',
		'N11E038.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N11E038.hgt.zip',
		'N11E039.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N11E039.hgt.zip',
		'N11E040.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N11E040.hgt.zip',
		'N11E041.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N11E041.hgt.zip',
		'N11E042.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N11E042.hgt.zip',
		'N11E043.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N11E043.hgt.zip',
		'N11E047.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N11E047.hgt.zip',
		'N11E048.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N11E048.hgt.zip',
		'N11E049.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N11E049.hgt.zip',
		'N11E050.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N11E050.hgt.zip',
		'N11E051.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N11E051.hgt.zip',
		'N11E072.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N11E072.hgt.zip',
		'N11E073.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N11E073.hgt.zip',
		'N11E075.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N11E075.hgt.zip',
		'N11E076.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N11E076.hgt.zip',
		'N11E077.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N11E077.hgt.zip',
		'N11E078.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N11E078.hgt.zip',
		'N11E079.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N11E079.hgt.zip',
		'N11E092.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N11E092.hgt.zip',
		'N11E093.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N11E093.hgt.zip',
		'N11E097.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N11E097.hgt.zip',
		'N11E098.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N11E098.hgt.zip',
		'N11E099.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N11E099.hgt.zip',
		'N11E102.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N11E102.hgt.zip',
		'N11E103.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N11E103.hgt.zip',
		'N11E104.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N11E104.hgt.zip',
		'N11E105.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N11E105.hgt.zip',
		'N11E106.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N11E106.hgt.zip',
		'N11E107.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N11E107.hgt.zip',
		'N11E108.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N11E108.hgt.zip',
		'N11E109.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N11E109.hgt.zip',
		'N11E114.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N11E114.hgt.zip',
		'N11E115.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N11E115.hgt.zip',
		'N11E119.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N11E119.hgt.zip',
		'N11E120.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N11E120.hgt.zip',
		'N11E121.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N11E121.hgt.zip',
		'N11E122.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N11E122.hgt.zip',
		'N11E123.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N11E123.hgt.zip',
		'N11E124.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N11E124.hgt.zip',
		'N11E125.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N11E125.hgt.zip',
		'N11E162.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N11E162.hgt.zip',
		'N11E165.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N11E165.hgt.zip',
		'N11E166.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N11E166.hgt.zip',
		'N11E167.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N11E167.hgt.zip',
		'N11E169.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N11E169.hgt.zip',
		'N11W001.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N11W001.hgt.zip',
		'N11W002.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N11W002.hgt.zip',
		'N11W003.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N11W003.hgt.zip',
		'N11W004.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N11W004.hgt.zip',
		'N11W005.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N11W005.hgt.zip',
		'N11W006.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N11W006.hgt.zip',
		'N11W007.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N11W007.hgt.zip',
		'N11W008.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N11W008.hgt.zip',
		'N11W009.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N11W009.hgt.zip',
		'N11W010.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N11W010.hgt.zip',
		'N11W011.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N11W011.hgt.zip',
		'N11W012.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N11W012.hgt.zip',
		'N11W013.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N11W013.hgt.zip',
		'N11W014.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N11W014.hgt.zip',
		'N11W015.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N11W015.hgt.zip',
		'N11W016.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N11W016.hgt.zip',
		'N11W017.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N11W017.hgt.zip',
		'N11W061.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N11W061.hgt.zip',
		'N11W062.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N11W062.hgt.zip',
		'N11W064.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N11W064.hgt.zip',
		'N11W065.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N11W065.hgt.zip',
		'N11W067.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N11W067.hgt.zip',
		'N11W068.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N11W068.hgt.zip',
		'N11W069.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N11W069.hgt.zip',
		'N11W070.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N11W070.hgt.zip',
		'N11W071.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N11W071.hgt.zip',
		'N11W072.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N11W072.hgt.zip',
		'N11W073.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N11W073.hgt.zip',
		'N11W074.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N11W074.hgt.zip',
		'N11W075.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N11W075.hgt.zip',
		'N11W084.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N11W084.hgt.zip',
		'N11W085.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N11W085.hgt.zip',
		'N11W086.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N11W086.hgt.zip',
		'N11W087.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N11W087.hgt.zip',
		'N12E000.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N12E000.hgt.zip',
		'N12E001.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N12E001.hgt.zip',
		'N12E002.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N12E002.hgt.zip',
		'N12E003.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N12E003.hgt.zip',
		'N12E004.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N12E004.hgt.zip',
		'N12E005.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N12E005.hgt.zip',
		'N12E006.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N12E006.hgt.zip',
		'N12E007.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N12E007.hgt.zip',
		'N12E008.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N12E008.hgt.zip',
		'N12E009.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N12E009.hgt.zip',
		'N12E010.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N12E010.hgt.zip',
		'N12E011.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N12E011.hgt.zip',
		'N12E012.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N12E012.hgt.zip',
		'N12E013.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N12E013.hgt.zip',
		'N12E014.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N12E014.hgt.zip',
		'N12E015.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N12E015.hgt.zip',
		'N12E016.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N12E016.hgt.zip',
		'N12E017.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N12E017.hgt.zip',
		'N12E018.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N12E018.hgt.zip',
		'N12E019.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N12E019.hgt.zip',
		'N12E020.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N12E020.hgt.zip',
		'N12E021.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N12E021.hgt.zip',
		'N12E022.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N12E022.hgt.zip',
		'N12E023.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N12E023.hgt.zip',
		'N12E024.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N12E024.hgt.zip',
		'N12E025.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N12E025.hgt.zip',
		'N12E026.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N12E026.hgt.zip',
		'N12E027.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N12E027.hgt.zip',
		'N12E028.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N12E028.hgt.zip',
		'N12E029.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N12E029.hgt.zip',
		'N12E030.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N12E030.hgt.zip',
		'N12E031.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N12E031.hgt.zip',
		'N12E032.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N12E032.hgt.zip',
		'N12E033.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N12E033.hgt.zip',
		'N12E034.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N12E034.hgt.zip',
		'N12E035.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N12E035.hgt.zip',
		'N12E036.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N12E036.hgt.zip',
		'N12E037.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N12E037.hgt.zip',
		'N12E038.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N12E038.hgt.zip',
		'N12E039.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N12E039.hgt.zip',
		'N12E040.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N12E040.hgt.zip',
		'N12E041.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N12E041.hgt.zip',
		'N12E042.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N12E042.hgt.zip',
		'N12E043.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N12E043.hgt.zip',
		'N12E044.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N12E044.hgt.zip',
		'N12E045.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N12E045.hgt.zip',
		'N12E052.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N12E052.hgt.zip',
		'N12E053.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N12E053.hgt.zip',
		'N12E054.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N12E054.hgt.zip',
		'N12E074.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N12E074.hgt.zip',
		'N12E075.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N12E075.hgt.zip',
		'N12E076.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N12E076.hgt.zip',
		'N12E077.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N12E077.hgt.zip',
		'N12E078.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N12E078.hgt.zip',
		'N12E079.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N12E079.hgt.zip',
		'N12E080.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N12E080.hgt.zip',
		'N12E092.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N12E092.hgt.zip',
		'N12E093.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N12E093.hgt.zip',
		'N12E097.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N12E097.hgt.zip',
		'N12E098.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N12E098.hgt.zip',
		'N12E099.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N12E099.hgt.zip',
		'N12E100.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N12E100.hgt.zip',
		'N12E101.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N12E101.hgt.zip',
		'N12E102.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N12E102.hgt.zip',
		'N12E103.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N12E103.hgt.zip',
		'N12E104.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N12E104.hgt.zip',
		'N12E105.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N12E105.hgt.zip',
		'N12E106.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N12E106.hgt.zip',
		'N12E107.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N12E107.hgt.zip',
		'N12E108.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N12E108.hgt.zip',
		'N12E109.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N12E109.hgt.zip',
		'N12E119.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N12E119.hgt.zip',
		'N12E120.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N12E120.hgt.zip',
		'N12E121.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N12E121.hgt.zip',
		'N12E122.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N12E122.hgt.zip',
		'N12E123.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N12E123.hgt.zip',
		'N12E124.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N12E124.hgt.zip',
		'N12E125.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N12E125.hgt.zip',
		'N12E170.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N12E170.hgt.zip',
		'N12W001.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N12W001.hgt.zip',
		'N12W002.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N12W002.hgt.zip',
		'N12W003.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N12W003.hgt.zip',
		'N12W004.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N12W004.hgt.zip',
		'N12W005.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N12W005.hgt.zip',
		'N12W006.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N12W006.hgt.zip',
		'N12W007.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N12W007.hgt.zip',
		'N12W008.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N12W008.hgt.zip',
		'N12W009.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N12W009.hgt.zip',
		'N12W010.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N12W010.hgt.zip',
		'N12W011.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N12W011.hgt.zip',
		'N12W012.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N12W012.hgt.zip',
		'N12W013.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N12W013.hgt.zip',
		'N12W014.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N12W014.hgt.zip',
		'N12W015.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N12W015.hgt.zip',
		'N12W016.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N12W016.hgt.zip',
		'N12W017.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N12W017.hgt.zip',
		'N12W062.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N12W062.hgt.zip',
		'N12W069.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N12W069.hgt.zip',
		'N12W070.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N12W070.hgt.zip',
		'N12W071.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N12W071.hgt.zip',
		'N12W072.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N12W072.hgt.zip',
		'N12W073.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N12W073.hgt.zip',
		'N12W082.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N12W082.hgt.zip',
		'N12W083.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N12W083.hgt.zip',
		'N12W084.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N12W084.hgt.zip',
		'N12W085.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N12W085.hgt.zip',
		'N12W086.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N12W086.hgt.zip',
		'N12W087.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N12W087.hgt.zip',
		'N12W088.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N12W088.hgt.zip',
		'N13E000.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N13E000.hgt.zip',
		'N13E001.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N13E001.hgt.zip',
		'N13E002.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N13E002.hgt.zip',
		'N13E003.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N13E003.hgt.zip',
		'N13E004.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N13E004.hgt.zip',
		'N13E005.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N13E005.hgt.zip',
		'N13E006.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N13E006.hgt.zip',
		'N13E007.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N13E007.hgt.zip',
		'N13E008.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N13E008.hgt.zip',
		'N13E009.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N13E009.hgt.zip',
		'N13E010.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N13E010.hgt.zip',
		'N13E011.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N13E011.hgt.zip',
		'N13E012.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N13E012.hgt.zip',
		'N13E013.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N13E013.hgt.zip',
		'N13E014.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N13E014.hgt.zip',
		'N13E015.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N13E015.hgt.zip',
		'N13E016.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N13E016.hgt.zip',
		'N13E017.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N13E017.hgt.zip',
		'N13E018.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N13E018.hgt.zip',
		'N13E019.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N13E019.hgt.zip',
		'N13E020.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N13E020.hgt.zip',
		'N13E021.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N13E021.hgt.zip',
		'N13E022.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N13E022.hgt.zip',
		'N13E023.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N13E023.hgt.zip',
		'N13E024.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N13E024.hgt.zip',
		'N13E025.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N13E025.hgt.zip',
		'N13E026.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N13E026.hgt.zip',
		'N13E027.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N13E027.hgt.zip',
		'N13E028.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N13E028.hgt.zip',
		'N13E029.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N13E029.hgt.zip',
		'N13E030.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N13E030.hgt.zip',
		'N13E031.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N13E031.hgt.zip',
		'N13E032.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N13E032.hgt.zip',
		'N13E033.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N13E033.hgt.zip',
		'N13E034.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N13E034.hgt.zip',
		'N13E035.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N13E035.hgt.zip',
		'N13E036.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N13E036.hgt.zip',
		'N13E037.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N13E037.hgt.zip',
		'N13E038.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N13E038.hgt.zip',
		'N13E039.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N13E039.hgt.zip',
		'N13E040.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N13E040.hgt.zip',
		'N13E041.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N13E041.hgt.zip',
		'N13E042.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N13E042.hgt.zip',
		'N13E043.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N13E043.hgt.zip',
		'N13E044.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N13E044.hgt.zip',
		'N13E045.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N13E045.hgt.zip',
		'N13E046.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N13E046.hgt.zip',
		'N13E047.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N13E047.hgt.zip',
		'N13E048.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N13E048.hgt.zip',
		'N13E074.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N13E074.hgt.zip',
		'N13E075.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N13E075.hgt.zip',
		'N13E076.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N13E076.hgt.zip',
		'N13E077.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N13E077.hgt.zip',
		'N13E078.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N13E078.hgt.zip',
		'N13E079.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N13E079.hgt.zip',
		'N13E080.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N13E080.hgt.zip',
		'N13E092.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N13E092.hgt.zip',
		'N13E093.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N13E093.hgt.zip',
		'N13E094.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N13E094.hgt.zip',
		'N13E097.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N13E097.hgt.zip',
		'N13E098.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N13E098.hgt.zip',
		'N13E099.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N13E099.hgt.zip',
		'N13E100.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N13E100.hgt.zip',
		'N13E101.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N13E101.hgt.zip',
		'N13E102.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N13E102.hgt.zip',
		'N13E103.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N13E103.hgt.zip',
		'N13E104.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N13E104.hgt.zip',
		'N13E105.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N13E105.hgt.zip',
		'N13E106.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N13E106.hgt.zip',
		'N13E107.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N13E107.hgt.zip',
		'N13E108.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N13E108.hgt.zip',
		'N13E109.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N13E109.hgt.zip',
		'N13E120.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N13E120.hgt.zip',
		'N13E121.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N13E121.hgt.zip',
		'N13E122.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N13E122.hgt.zip',
		'N13E123.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N13E123.hgt.zip',
		'N13E124.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N13E124.hgt.zip',
		'N13E144.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N13E144.hgt.zip',
		'N13W001.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N13W001.hgt.zip',
		'N13W002.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N13W002.hgt.zip',
		'N13W003.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N13W003.hgt.zip',
		'N13W004.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N13W004.hgt.zip',
		'N13W005.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N13W005.hgt.zip',
		'N13W006.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N13W006.hgt.zip',
		'N13W007.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N13W007.hgt.zip',
		'N13W008.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N13W008.hgt.zip',
		'N13W009.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N13W009.hgt.zip',
		'N13W010.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N13W010.hgt.zip',
		'N13W011.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N13W011.hgt.zip',
		'N13W012.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N13W012.hgt.zip',
		'N13W013.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N13W013.hgt.zip',
		'N13W014.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N13W014.hgt.zip',
		'N13W015.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N13W015.hgt.zip',
		'N13W016.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N13W016.hgt.zip',
		'N13W017.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N13W017.hgt.zip',
		'N13W060.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N13W060.hgt.zip',
		'N13W061.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N13W061.hgt.zip',
		'N13W062.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N13W062.hgt.zip',
		'N13W081.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N13W081.hgt.zip',
		'N13W082.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N13W082.hgt.zip',
		'N13W084.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N13W084.hgt.zip',
		'N13W085.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N13W085.hgt.zip',
		'N13W086.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N13W086.hgt.zip',
		'N13W087.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N13W087.hgt.zip',
		'N13W088.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N13W088.hgt.zip',
		'N13W089.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N13W089.hgt.zip',
		'N13W090.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N13W090.hgt.zip',
		'N13W091.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N13W091.hgt.zip',
		'N13W092.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N13W092.hgt.zip',
		'N14E000.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N14E000.hgt.zip',
		'N14E001.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N14E001.hgt.zip',
		'N14E002.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N14E002.hgt.zip',
		'N14E003.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N14E003.hgt.zip',
		'N14E004.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N14E004.hgt.zip',
		'N14E005.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N14E005.hgt.zip',
		'N14E006.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N14E006.hgt.zip',
		'N14E007.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N14E007.hgt.zip',
		'N14E008.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N14E008.hgt.zip',
		'N14E009.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N14E009.hgt.zip',
		'N14E010.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N14E010.hgt.zip',
		'N14E011.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N14E011.hgt.zip',
		'N14E012.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N14E012.hgt.zip',
		'N14E013.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N14E013.hgt.zip',
		'N14E014.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N14E014.hgt.zip',
		'N14E015.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N14E015.hgt.zip',
		'N14E016.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N14E016.hgt.zip',
		'N14E017.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N14E017.hgt.zip',
		'N14E018.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N14E018.hgt.zip',
		'N14E019.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N14E019.hgt.zip',
		'N14E020.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N14E020.hgt.zip',
		'N14E021.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N14E021.hgt.zip',
		'N14E022.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N14E022.hgt.zip',
		'N14E023.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N14E023.hgt.zip',
		'N14E024.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N14E024.hgt.zip',
		'N14E025.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N14E025.hgt.zip',
		'N14E026.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N14E026.hgt.zip',
		'N14E027.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N14E027.hgt.zip',
		'N14E028.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N14E028.hgt.zip',
		'N14E029.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N14E029.hgt.zip',
		'N14E030.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N14E030.hgt.zip',
		'N14E031.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N14E031.hgt.zip',
		'N14E032.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N14E032.hgt.zip',
		'N14E033.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N14E033.hgt.zip',
		'N14E034.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N14E034.hgt.zip',
		'N14E035.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N14E035.hgt.zip',
		'N14E036.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N14E036.hgt.zip',
		'N14E037.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N14E037.hgt.zip',
		'N14E038.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N14E038.hgt.zip',
		'N14E039.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N14E039.hgt.zip',
		'N14E040.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N14E040.hgt.zip',
		'N14E041.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N14E041.hgt.zip',
		'N14E042.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N14E042.hgt.zip',
		'N14E043.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N14E043.hgt.zip',
		'N14E044.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N14E044.hgt.zip',
		'N14E045.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N14E045.hgt.zip',
		'N14E046.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N14E046.hgt.zip',
		'N14E047.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N14E047.hgt.zip',
		'N14E048.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N14E048.hgt.zip',
		'N14E049.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N14E049.hgt.zip',
		'N14E050.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N14E050.hgt.zip',
		'N14E074.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N14E074.hgt.zip',
		'N14E075.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N14E075.hgt.zip',
		'N14E076.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N14E076.hgt.zip',
		'N14E077.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N14E077.hgt.zip',
		'N14E078.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N14E078.hgt.zip',
		'N14E079.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N14E079.hgt.zip',
		'N14E080.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N14E080.hgt.zip',
		'N14E093.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N14E093.hgt.zip',
		'N14E097.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N14E097.hgt.zip',
		'N14E098.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N14E098.hgt.zip',
		'N14E099.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N14E099.hgt.zip',
		'N14E100.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N14E100.hgt.zip',
		'N14E101.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N14E101.hgt.zip',
		'N14E102.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N14E102.hgt.zip',
		'N14E103.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N14E103.hgt.zip',
		'N14E104.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N14E104.hgt.zip',
		'N14E105.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N14E105.hgt.zip',
		'N14E106.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N14E106.hgt.zip',
		'N14E107.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N14E107.hgt.zip',
		'N14E108.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N14E108.hgt.zip',
		'N14E109.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N14E109.hgt.zip',
		'N14E120.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N14E120.hgt.zip',
		'N14E121.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N14E121.hgt.zip',
		'N14E122.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N14E122.hgt.zip',
		'N14E123.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N14E123.hgt.zip',
		'N14E124.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N14E124.hgt.zip',
		'N14E145.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N14E145.hgt.zip',
		'N14E168.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N14E168.hgt.zip',
		'N14E169.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N14E169.hgt.zip',
		'N14W001.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N14W001.hgt.zip',
		'N14W002.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N14W002.hgt.zip',
		'N14W003.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N14W003.hgt.zip',
		'N14W004.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N14W004.hgt.zip',
		'N14W005.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N14W005.hgt.zip',
		'N14W006.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N14W006.hgt.zip',
		'N14W007.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N14W007.hgt.zip',
		'N14W008.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N14W008.hgt.zip',
		'N14W009.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N14W009.hgt.zip',
		'N14W010.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N14W010.hgt.zip',
		'N14W011.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N14W011.hgt.zip',
		'N14W012.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N14W012.hgt.zip',
		'N14W013.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N14W013.hgt.zip',
		'N14W014.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N14W014.hgt.zip',
		'N14W015.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N14W015.hgt.zip',
		'N14W016.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N14W016.hgt.zip',
		'N14W017.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N14W017.hgt.zip',
		'N14W018.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N14W018.hgt.zip',
		'N14W024.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N14W024.hgt.zip',
		'N14W025.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N14W025.hgt.zip',
		'N14W061.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N14W061.hgt.zip',
		'N14W062.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N14W062.hgt.zip',
		'N14W081.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N14W081.hgt.zip',
		'N14W083.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N14W083.hgt.zip',
		'N14W084.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N14W084.hgt.zip',
		'N14W085.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N14W085.hgt.zip',
		'N14W086.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N14W086.hgt.zip',
		'N14W087.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N14W087.hgt.zip',
		'N14W088.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N14W088.hgt.zip',
		'N14W089.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N14W089.hgt.zip',
		'N14W090.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N14W090.hgt.zip',
		'N14W091.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N14W091.hgt.zip',
		'N14W092.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N14W092.hgt.zip',
		'N14W093.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//N14W093.hgt.zip',
		'N15E000.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N15E000.hgt.zip',
		'N15E001.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N15E001.hgt.zip',
		'N15E002.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N15E002.hgt.zip',
		'N15E003.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N15E003.hgt.zip',
		'N15E004.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N15E004.hgt.zip',
		'N15E005.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N15E005.hgt.zip',
		'N15E006.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N15E006.hgt.zip',
		'N15E007.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N15E007.hgt.zip',
		'N15E008.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N15E008.hgt.zip',
		'N15E009.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N15E009.hgt.zip',
		'N15E010.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N15E010.hgt.zip',
		'N15E011.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N15E011.hgt.zip',
		'N15E012.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N15E012.hgt.zip',
		'N15E013.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N15E013.hgt.zip',
		'N15E014.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N15E014.hgt.zip',
		'N15E015.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N15E015.hgt.zip',
		'N15E016.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N15E016.hgt.zip',
		'N15E017.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N15E017.hgt.zip',
		'N15E018.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N15E018.hgt.zip',
		'N15E019.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N15E019.hgt.zip',
		'N15E020.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N15E020.hgt.zip',
		'N15E021.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N15E021.hgt.zip',
		'N15E022.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N15E022.hgt.zip',
		'N15E023.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N15E023.hgt.zip',
		'N15E024.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N15E024.hgt.zip',
		'N15E025.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N15E025.hgt.zip',
		'N15E026.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N15E026.hgt.zip',
		'N15E027.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N15E027.hgt.zip',
		'N15E028.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N15E028.hgt.zip',
		'N15E029.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N15E029.hgt.zip',
		'N15E030.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N15E030.hgt.zip',
		'N15E031.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N15E031.hgt.zip',
		'N15E032.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N15E032.hgt.zip',
		'N15E033.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N15E033.hgt.zip',
		'N15E034.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N15E034.hgt.zip',
		'N15E035.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N15E035.hgt.zip',
		'N15E036.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N15E036.hgt.zip',
		'N15E037.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N15E037.hgt.zip',
		'N15E038.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N15E038.hgt.zip',
		'N15E039.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N15E039.hgt.zip',
		'N15E040.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N15E040.hgt.zip',
		'N15E041.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N15E041.hgt.zip',
		'N15E042.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N15E042.hgt.zip',
		'N15E043.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N15E043.hgt.zip',
		'N15E044.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N15E044.hgt.zip',
		'N15E045.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N15E045.hgt.zip',
		'N15E046.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N15E046.hgt.zip',
		'N15E047.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N15E047.hgt.zip',
		'N15E048.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N15E048.hgt.zip',
		'N15E049.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N15E049.hgt.zip',
		'N15E050.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N15E050.hgt.zip',
		'N15E051.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N15E051.hgt.zip',
		'N15E052.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N15E052.hgt.zip',
		'N15E073.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N15E073.hgt.zip',
		'N15E074.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N15E074.hgt.zip',
		'N15E075.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N15E075.hgt.zip',
		'N15E076.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N15E076.hgt.zip',
		'N15E077.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N15E077.hgt.zip',
		'N15E078.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N15E078.hgt.zip',
		'N15E079.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N15E079.hgt.zip',
		'N15E080.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N15E080.hgt.zip',
		'N15E081.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N15E081.hgt.zip',
		'N15E094.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N15E094.hgt.zip',
		'N15E095.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N15E095.hgt.zip',
		'N15E097.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N15E097.hgt.zip',
		'N15E098.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N15E098.hgt.zip',
		'N15E099.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N15E099.hgt.zip',
		'N15E100.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N15E100.hgt.zip',
		'N15E101.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N15E101.hgt.zip',
		'N15E102.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N15E102.hgt.zip',
		'N15E103.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N15E103.hgt.zip',
		'N15E104.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N15E104.hgt.zip',
		'N15E105.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N15E105.hgt.zip',
		'N15E106.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N15E106.hgt.zip',
		'N15E107.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N15E107.hgt.zip',
		'N15E108.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N15E108.hgt.zip',
		'N15E109.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N15E109.hgt.zip',
		'N15E111.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N15E111.hgt.zip',
		'N15E119.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N15E119.hgt.zip',
		'N15E120.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N15E120.hgt.zip',
		'N15E121.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N15E121.hgt.zip',
		'N15E122.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N15E122.hgt.zip',
		'N15E145.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N15E145.hgt.zip',
		'N15W001.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N15W001.hgt.zip',
		'N15W002.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N15W002.hgt.zip',
		'N15W003.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N15W003.hgt.zip',
		'N15W004.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N15W004.hgt.zip',
		'N15W005.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N15W005.hgt.zip',
		'N15W006.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N15W006.hgt.zip',
		'N15W007.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N15W007.hgt.zip',
		'N15W008.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N15W008.hgt.zip',
		'N15W009.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N15W009.hgt.zip',
		'N15W010.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N15W010.hgt.zip',
		'N15W011.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N15W011.hgt.zip',
		'N15W012.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N15W012.hgt.zip',
		'N15W013.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N15W013.hgt.zip',
		'N15W014.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N15W014.hgt.zip',
		'N15W015.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N15W015.hgt.zip',
		'N15W016.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N15W016.hgt.zip',
		'N15W017.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N15W017.hgt.zip',
		'N15W018.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N15W018.hgt.zip',
		'N15W023.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N15W023.hgt.zip',
		'N15W024.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N15W024.hgt.zip',
		'N15W025.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N15W025.hgt.zip',
		'N15W062.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N15W062.hgt.zip',
		'N15W064.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N15W064.hgt.zip',
		'N15W079.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N15W079.hgt.zip',
		'N15W080.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N15W080.hgt.zip',
		'N15W083.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N15W083.hgt.zip',
		'N15W084.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N15W084.hgt.zip',
		'N15W085.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N15W085.hgt.zip',
		'N15W086.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N15W086.hgt.zip',
		'N15W087.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N15W087.hgt.zip',
		'N15W088.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N15W088.hgt.zip',
		'N15W089.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N15W089.hgt.zip',
		'N15W090.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N15W090.hgt.zip',
		'N15W091.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N15W091.hgt.zip',
		'N15W092.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N15W092.hgt.zip',
		'N15W093.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N15W093.hgt.zip',
		'N15W094.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N15W094.hgt.zip',
		'N15W096.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N15W096.hgt.zip',
		'N15W097.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N15W097.hgt.zip',
		'N15W098.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N15W098.hgt.zip',
		'N16E000.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N16E000.hgt.zip',
		'N16E001.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N16E001.hgt.zip',
		'N16E002.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N16E002.hgt.zip',
		'N16E003.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N16E003.hgt.zip',
		'N16E004.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N16E004.hgt.zip',
		'N16E005.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N16E005.hgt.zip',
		'N16E006.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N16E006.hgt.zip',
		'N16E007.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N16E007.hgt.zip',
		'N16E008.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N16E008.hgt.zip',
		'N16E009.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N16E009.hgt.zip',
		'N16E010.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N16E010.hgt.zip',
		'N16E011.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N16E011.hgt.zip',
		'N16E012.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N16E012.hgt.zip',
		'N16E013.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N16E013.hgt.zip',
		'N16E014.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N16E014.hgt.zip',
		'N16E015.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N16E015.hgt.zip',
		'N16E016.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N16E016.hgt.zip',
		'N16E017.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N16E017.hgt.zip',
		'N16E018.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N16E018.hgt.zip',
		'N16E019.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N16E019.hgt.zip',
		'N16E020.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N16E020.hgt.zip',
		'N16E021.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N16E021.hgt.zip',
		'N16E022.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N16E022.hgt.zip',
		'N16E023.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N16E023.hgt.zip',
		'N16E024.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N16E024.hgt.zip',
		'N16E025.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N16E025.hgt.zip',
		'N16E026.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N16E026.hgt.zip',
		'N16E027.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N16E027.hgt.zip',
		'N16E028.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N16E028.hgt.zip',
		'N16E029.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N16E029.hgt.zip',
		'N16E030.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N16E030.hgt.zip',
		'N16E031.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N16E031.hgt.zip',
		'N16E032.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N16E032.hgt.zip',
		'N16E033.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N16E033.hgt.zip',
		'N16E034.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N16E034.hgt.zip',
		'N16E035.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N16E035.hgt.zip',
		'N16E036.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N16E036.hgt.zip',
		'N16E037.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N16E037.hgt.zip',
		'N16E038.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N16E038.hgt.zip',
		'N16E039.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N16E039.hgt.zip',
		'N16E040.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N16E040.hgt.zip',
		'N16E041.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N16E041.hgt.zip',
		'N16E042.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N16E042.hgt.zip',
		'N16E043.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N16E043.hgt.zip',
		'N16E044.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N16E044.hgt.zip',
		'N16E045.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N16E045.hgt.zip',
		'N16E046.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N16E046.hgt.zip',
		'N16E047.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N16E047.hgt.zip',
		'N16E048.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N16E048.hgt.zip',
		'N16E049.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N16E049.hgt.zip',
		'N16E050.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N16E050.hgt.zip',
		'N16E051.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N16E051.hgt.zip',
		'N16E052.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N16E052.hgt.zip',
		'N16E053.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N16E053.hgt.zip',
		'N16E054.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N16E054.hgt.zip',
		'N16E055.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N16E055.hgt.zip',
		'N16E073.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N16E073.hgt.zip',
		'N16E074.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N16E074.hgt.zip',
		'N16E075.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N16E075.hgt.zip',
		'N16E076.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N16E076.hgt.zip',
		'N16E077.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N16E077.hgt.zip',
		'N16E078.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N16E078.hgt.zip',
		'N16E079.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N16E079.hgt.zip',
		'N16E080.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N16E080.hgt.zip',
		'N16E081.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N16E081.hgt.zip',
		'N16E082.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N16E082.hgt.zip',
		'N16E094.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N16E094.hgt.zip',
		'N16E095.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N16E095.hgt.zip',
		'N16E096.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N16E096.hgt.zip',
		'N16E097.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N16E097.hgt.zip',
		'N16E098.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N16E098.hgt.zip',
		'N16E099.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N16E099.hgt.zip',
		'N16E100.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N16E100.hgt.zip',
		'N16E101.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N16E101.hgt.zip',
		'N16E102.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N16E102.hgt.zip',
		'N16E103.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N16E103.hgt.zip',
		'N16E104.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N16E104.hgt.zip',
		'N16E105.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N16E105.hgt.zip',
		'N16E106.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N16E106.hgt.zip',
		'N16E107.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N16E107.hgt.zip',
		'N16E108.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N16E108.hgt.zip',
		'N16E111.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N16E111.hgt.zip',
		'N16E112.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N16E112.hgt.zip',
		'N16E119.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N16E119.hgt.zip',
		'N16E120.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N16E120.hgt.zip',
		'N16E121.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N16E121.hgt.zip',
		'N16E122.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N16E122.hgt.zip',
		'N16E145.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N16E145.hgt.zip',
		'N16E146.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N16E146.hgt.zip',
		'N16W001.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N16W001.hgt.zip',
		'N16W002.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N16W002.hgt.zip',
		'N16W003.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N16W003.hgt.zip',
		'N16W004.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N16W004.hgt.zip',
		'N16W005.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N16W005.hgt.zip',
		'N16W006.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N16W006.hgt.zip',
		'N16W007.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N16W007.hgt.zip',
		'N16W008.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N16W008.hgt.zip',
		'N16W009.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N16W009.hgt.zip',
		'N16W010.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N16W010.hgt.zip',
		'N16W011.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N16W011.hgt.zip',
		'N16W012.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N16W012.hgt.zip',
		'N16W013.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N16W013.hgt.zip',
		'N16W014.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N16W014.hgt.zip',
		'N16W015.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N16W015.hgt.zip',
		'N16W016.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N16W016.hgt.zip',
		'N16W017.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N16W017.hgt.zip',
		'N16W023.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N16W023.hgt.zip',
		'N16W025.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N16W025.hgt.zip',
		'N16W026.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N16W026.hgt.zip',
		'N16W062.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N16W062.hgt.zip',
		'N16W063.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N16W063.hgt.zip',
		'N16W086.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N16W086.hgt.zip',
		'N16W087.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N16W087.hgt.zip',
		'N16W088.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N16W088.hgt.zip',
		'N16W089.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N16W089.hgt.zip',
		'N16W090.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N16W090.hgt.zip',
		'N16W091.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N16W091.hgt.zip',
		'N16W092.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N16W092.hgt.zip',
		'N16W093.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N16W093.hgt.zip',
		'N16W094.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N16W094.hgt.zip',
		'N16W095.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N16W095.hgt.zip',
		'N16W096.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N16W096.hgt.zip',
		'N16W097.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N16W097.hgt.zip',
		'N16W098.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N16W098.hgt.zip',
		'N16W099.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N16W099.hgt.zip',
		'N16W100.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N16W100.hgt.zip',
		'N16W101.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N16W101.hgt.zip',
		'N16W170.hgt': 'https://srtm.kurviger.de/SRTM3//Islands//N16W170.hgt.zip',
		'N17E000.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N17E000.hgt.zip',
		'N17E001.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N17E001.hgt.zip',
		'N17E002.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N17E002.hgt.zip',
		'N17E003.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N17E003.hgt.zip',
		'N17E004.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N17E004.hgt.zip',
		'N17E005.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N17E005.hgt.zip',
		'N17E006.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N17E006.hgt.zip',
		'N17E007.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N17E007.hgt.zip',
		'N17E008.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N17E008.hgt.zip',
		'N17E009.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N17E009.hgt.zip',
		'N17E010.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N17E010.hgt.zip',
		'N17E011.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N17E011.hgt.zip',
		'N17E012.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N17E012.hgt.zip',
		'N17E013.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N17E013.hgt.zip',
		'N17E014.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N17E014.hgt.zip',
		'N17E015.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N17E015.hgt.zip',
		'N17E016.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N17E016.hgt.zip',
		'N17E017.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N17E017.hgt.zip',
		'N17E018.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N17E018.hgt.zip',
		'N17E019.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N17E019.hgt.zip',
		'N17E020.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N17E020.hgt.zip',
		'N17E021.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N17E021.hgt.zip',
		'N17E022.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N17E022.hgt.zip',
		'N17E023.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N17E023.hgt.zip',
		'N17E024.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N17E024.hgt.zip',
		'N17E025.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N17E025.hgt.zip',
		'N17E026.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N17E026.hgt.zip',
		'N17E027.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N17E027.hgt.zip',
		'N17E028.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N17E028.hgt.zip',
		'N17E029.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N17E029.hgt.zip',
		'N17E030.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N17E030.hgt.zip',
		'N17E031.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N17E031.hgt.zip',
		'N17E032.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N17E032.hgt.zip',
		'N17E033.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N17E033.hgt.zip',
		'N17E034.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N17E034.hgt.zip',
		'N17E035.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N17E035.hgt.zip',
		'N17E036.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N17E036.hgt.zip',
		'N17E037.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N17E037.hgt.zip',
		'N17E038.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N17E038.hgt.zip',
		'N17E039.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N17E039.hgt.zip',
		'N17E041.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N17E041.hgt.zip',
		'N17E042.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N17E042.hgt.zip',
		'N17E043.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N17E043.hgt.zip',
		'N17E044.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N17E044.hgt.zip',
		'N17E045.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N17E045.hgt.zip',
		'N17E046.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N17E046.hgt.zip',
		'N17E047.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N17E047.hgt.zip',
		'N17E048.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N17E048.hgt.zip',
		'N17E049.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N17E049.hgt.zip',
		'N17E050.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N17E050.hgt.zip',
		'N17E051.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N17E051.hgt.zip',
		'N17E052.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N17E052.hgt.zip',
		'N17E053.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N17E053.hgt.zip',
		'N17E054.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N17E054.hgt.zip',
		'N17E055.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N17E055.hgt.zip',
		'N17E056.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N17E056.hgt.zip',
		'N17E073.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N17E073.hgt.zip',
		'N17E074.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N17E074.hgt.zip',
		'N17E075.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N17E075.hgt.zip',
		'N17E076.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N17E076.hgt.zip',
		'N17E077.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N17E077.hgt.zip',
		'N17E078.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N17E078.hgt.zip',
		'N17E079.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N17E079.hgt.zip',
		'N17E080.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N17E080.hgt.zip',
		'N17E081.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N17E081.hgt.zip',
		'N17E082.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N17E082.hgt.zip',
		'N17E083.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N17E083.hgt.zip',
		'N17E094.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N17E094.hgt.zip',
		'N17E095.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N17E095.hgt.zip',
		'N17E096.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N17E096.hgt.zip',
		'N17E097.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N17E097.hgt.zip',
		'N17E098.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N17E098.hgt.zip',
		'N17E099.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N17E099.hgt.zip',
		'N17E100.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N17E100.hgt.zip',
		'N17E101.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N17E101.hgt.zip',
		'N17E102.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N17E102.hgt.zip',
		'N17E103.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N17E103.hgt.zip',
		'N17E104.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N17E104.hgt.zip',
		'N17E105.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N17E105.hgt.zip',
		'N17E106.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N17E106.hgt.zip',
		'N17E107.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N17E107.hgt.zip',
		'N17E120.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N17E120.hgt.zip',
		'N17E121.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N17E121.hgt.zip',
		'N17E122.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N17E122.hgt.zip',
		'N17E145.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N17E145.hgt.zip',
		'N17W001.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N17W001.hgt.zip',
		'N17W002.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N17W002.hgt.zip',
		'N17W003.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N17W003.hgt.zip',
		'N17W004.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N17W004.hgt.zip',
		'N17W005.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N17W005.hgt.zip',
		'N17W006.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N17W006.hgt.zip',
		'N17W007.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N17W007.hgt.zip',
		'N17W008.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N17W008.hgt.zip',
		'N17W009.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N17W009.hgt.zip',
		'N17W010.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N17W010.hgt.zip',
		'N17W011.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N17W011.hgt.zip',
		'N17W012.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N17W012.hgt.zip',
		'N17W013.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N17W013.hgt.zip',
		'N17W014.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N17W014.hgt.zip',
		'N17W015.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N17W015.hgt.zip',
		'N17W016.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N17W016.hgt.zip',
		'N17W017.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N17W017.hgt.zip',
		'N17W025.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N17W025.hgt.zip',
		'N17W026.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N17W026.hgt.zip',
		'N17W062.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N17W062.hgt.zip',
		'N17W063.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N17W063.hgt.zip',
		'N17W064.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N17W064.hgt.zip',
		'N17W065.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N17W065.hgt.zip',
		'N17W066.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N17W066.hgt.zip',
		'N17W067.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N17W067.hgt.zip',
		'N17W068.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N17W068.hgt.zip',
		'N17W072.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N17W072.hgt.zip',
		'N17W076.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N17W076.hgt.zip',
		'N17W077.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N17W077.hgt.zip',
		'N17W078.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N17W078.hgt.zip',
		'N17W084.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N17W084.hgt.zip',
		'N17W088.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N17W088.hgt.zip',
		'N17W089.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N17W089.hgt.zip',
		'N17W090.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N17W090.hgt.zip',
		'N17W091.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N17W091.hgt.zip',
		'N17W092.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N17W092.hgt.zip',
		'N17W093.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N17W093.hgt.zip',
		'N17W094.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N17W094.hgt.zip',
		'N17W095.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N17W095.hgt.zip',
		'N17W096.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N17W096.hgt.zip',
		'N17W097.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N17W097.hgt.zip',
		'N17W098.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N17W098.hgt.zip',
		'N17W099.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N17W099.hgt.zip',
		'N17W100.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N17W100.hgt.zip',
		'N17W101.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N17W101.hgt.zip',
		'N17W102.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N17W102.hgt.zip',
		'N17W103.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N17W103.hgt.zip',
		'N18E000.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N18E000.hgt.zip',
		'N18E001.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N18E001.hgt.zip',
		'N18E002.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N18E002.hgt.zip',
		'N18E003.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N18E003.hgt.zip',
		'N18E004.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N18E004.hgt.zip',
		'N18E005.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N18E005.hgt.zip',
		'N18E006.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N18E006.hgt.zip',
		'N18E007.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N18E007.hgt.zip',
		'N18E008.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N18E008.hgt.zip',
		'N18E009.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N18E009.hgt.zip',
		'N18E010.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N18E010.hgt.zip',
		'N18E011.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N18E011.hgt.zip',
		'N18E012.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N18E012.hgt.zip',
		'N18E013.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N18E013.hgt.zip',
		'N18E014.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N18E014.hgt.zip',
		'N18E015.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N18E015.hgt.zip',
		'N18E016.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N18E016.hgt.zip',
		'N18E017.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N18E017.hgt.zip',
		'N18E018.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N18E018.hgt.zip',
		'N18E019.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N18E019.hgt.zip',
		'N18E020.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N18E020.hgt.zip',
		'N18E021.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N18E021.hgt.zip',
		'N18E022.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N18E022.hgt.zip',
		'N18E023.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N18E023.hgt.zip',
		'N18E024.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N18E024.hgt.zip',
		'N18E025.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N18E025.hgt.zip',
		'N18E026.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N18E026.hgt.zip',
		'N18E027.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N18E027.hgt.zip',
		'N18E028.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N18E028.hgt.zip',
		'N18E029.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N18E029.hgt.zip',
		'N18E030.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N18E030.hgt.zip',
		'N18E031.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N18E031.hgt.zip',
		'N18E032.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N18E032.hgt.zip',
		'N18E033.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N18E033.hgt.zip',
		'N18E034.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N18E034.hgt.zip',
		'N18E035.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N18E035.hgt.zip',
		'N18E036.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N18E036.hgt.zip',
		'N18E037.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N18E037.hgt.zip',
		'N18E038.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N18E038.hgt.zip',
		'N18E040.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N18E040.hgt.zip',
		'N18E041.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N18E041.hgt.zip',
		'N18E042.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N18E042.hgt.zip',
		'N18E043.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N18E043.hgt.zip',
		'N18E044.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N18E044.hgt.zip',
		'N18E045.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N18E045.hgt.zip',
		'N18E046.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N18E046.hgt.zip',
		'N18E047.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N18E047.hgt.zip',
		'N18E048.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N18E048.hgt.zip',
		'N18E049.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N18E049.hgt.zip',
		'N18E050.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N18E050.hgt.zip',
		'N18E051.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N18E051.hgt.zip',
		'N18E052.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N18E052.hgt.zip',
		'N18E053.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N18E053.hgt.zip',
		'N18E054.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N18E054.hgt.zip',
		'N18E055.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N18E055.hgt.zip',
		'N18E056.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N18E056.hgt.zip',
		'N18E057.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N18E057.hgt.zip',
		'N18E072.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N18E072.hgt.zip',
		'N18E073.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N18E073.hgt.zip',
		'N18E074.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N18E074.hgt.zip',
		'N18E075.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N18E075.hgt.zip',
		'N18E076.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N18E076.hgt.zip',
		'N18E077.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N18E077.hgt.zip',
		'N18E078.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N18E078.hgt.zip',
		'N18E079.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N18E079.hgt.zip',
		'N18E080.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N18E080.hgt.zip',
		'N18E081.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N18E081.hgt.zip',
		'N18E082.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N18E082.hgt.zip',
		'N18E083.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N18E083.hgt.zip',
		'N18E084.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N18E084.hgt.zip',
		'N18E093.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N18E093.hgt.zip',
		'N18E094.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N18E094.hgt.zip',
		'N18E095.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N18E095.hgt.zip',
		'N18E096.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N18E096.hgt.zip',
		'N18E097.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N18E097.hgt.zip',
		'N18E098.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N18E098.hgt.zip',
		'N18E099.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N18E099.hgt.zip',
		'N18E100.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N18E100.hgt.zip',
		'N18E101.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N18E101.hgt.zip',
		'N18E102.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N18E102.hgt.zip',
		'N18E103.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N18E103.hgt.zip',
		'N18E104.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N18E104.hgt.zip',
		'N18E105.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N18E105.hgt.zip',
		'N18E106.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N18E106.hgt.zip',
		'N18E108.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N18E108.hgt.zip',
		'N18E109.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N18E109.hgt.zip',
		'N18E110.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N18E110.hgt.zip',
		'N18E120.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N18E120.hgt.zip',
		'N18E121.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N18E121.hgt.zip',
		'N18E122.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N18E122.hgt.zip',
		'N18E145.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N18E145.hgt.zip',
		'N18W001.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N18W001.hgt.zip',
		'N18W002.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N18W002.hgt.zip',
		'N18W003.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N18W003.hgt.zip',
		'N18W004.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N18W004.hgt.zip',
		'N18W005.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N18W005.hgt.zip',
		'N18W006.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N18W006.hgt.zip',
		'N18W007.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N18W007.hgt.zip',
		'N18W008.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N18W008.hgt.zip',
		'N18W009.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N18W009.hgt.zip',
		'N18W010.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N18W010.hgt.zip',
		'N18W011.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N18W011.hgt.zip',
		'N18W012.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N18W012.hgt.zip',
		'N18W013.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N18W013.hgt.zip',
		'N18W014.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N18W014.hgt.zip',
		'N18W015.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N18W015.hgt.zip',
		'N18W016.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N18W016.hgt.zip',
		'N18W017.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N18W017.hgt.zip',
		'N18W063.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N18W063.hgt.zip',
		'N18W064.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N18W064.hgt.zip',
		'N18W065.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N18W065.hgt.zip',
		'N18W066.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N18W066.hgt.zip',
		'N18W067.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N18W067.hgt.zip',
		'N18W068.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N18W068.hgt.zip',
		'N18W069.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N18W069.hgt.zip',
		'N18W070.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N18W070.hgt.zip',
		'N18W071.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N18W071.hgt.zip',
		'N18W072.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N18W072.hgt.zip',
		'N18W073.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N18W073.hgt.zip',
		'N18W074.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N18W074.hgt.zip',
		'N18W075.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N18W075.hgt.zip',
		'N18W076.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N18W076.hgt.zip',
		'N18W077.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N18W077.hgt.zip',
		'N18W078.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N18W078.hgt.zip',
		'N18W079.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N18W079.hgt.zip',
		'N18W088.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N18W088.hgt.zip',
		'N18W089.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N18W089.hgt.zip',
		'N18W090.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N18W090.hgt.zip',
		'N18W091.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N18W091.hgt.zip',
		'N18W092.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N18W092.hgt.zip',
		'N18W093.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N18W093.hgt.zip',
		'N18W094.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N18W094.hgt.zip',
		'N18W095.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N18W095.hgt.zip',
		'N18W096.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N18W096.hgt.zip',
		'N18W097.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N18W097.hgt.zip',
		'N18W098.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N18W098.hgt.zip',
		'N18W099.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N18W099.hgt.zip',
		'N18W100.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N18W100.hgt.zip',
		'N18W101.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N18W101.hgt.zip',
		'N18W102.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N18W102.hgt.zip',
		'N18W103.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N18W103.hgt.zip',
		'N18W104.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N18W104.hgt.zip',
		'N18W105.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N18W105.hgt.zip',
		'N18W111.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N18W111.hgt.zip',
		'N18W112.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N18W112.hgt.zip',
		'N18W115.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N18W115.hgt.zip',
		'N18W156.hgt': 'https://srtm.kurviger.de/SRTM3//Islands//N18W156.hgt.zip',
		'N19E000.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N19E000.hgt.zip',
		'N19E001.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N19E001.hgt.zip',
		'N19E002.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N19E002.hgt.zip',
		'N19E003.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N19E003.hgt.zip',
		'N19E004.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N19E004.hgt.zip',
		'N19E005.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N19E005.hgt.zip',
		'N19E006.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N19E006.hgt.zip',
		'N19E007.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N19E007.hgt.zip',
		'N19E008.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N19E008.hgt.zip',
		'N19E009.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N19E009.hgt.zip',
		'N19E010.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N19E010.hgt.zip',
		'N19E011.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N19E011.hgt.zip',
		'N19E012.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N19E012.hgt.zip',
		'N19E013.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N19E013.hgt.zip',
		'N19E014.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N19E014.hgt.zip',
		'N19E015.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N19E015.hgt.zip',
		'N19E016.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N19E016.hgt.zip',
		'N19E017.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N19E017.hgt.zip',
		'N19E018.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N19E018.hgt.zip',
		'N19E019.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N19E019.hgt.zip',
		'N19E020.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N19E020.hgt.zip',
		'N19E021.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N19E021.hgt.zip',
		'N19E022.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N19E022.hgt.zip',
		'N19E023.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N19E023.hgt.zip',
		'N19E024.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N19E024.hgt.zip',
		'N19E025.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N19E025.hgt.zip',
		'N19E026.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N19E026.hgt.zip',
		'N19E027.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N19E027.hgt.zip',
		'N19E028.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N19E028.hgt.zip',
		'N19E029.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N19E029.hgt.zip',
		'N19E030.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N19E030.hgt.zip',
		'N19E031.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N19E031.hgt.zip',
		'N19E032.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N19E032.hgt.zip',
		'N19E033.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N19E033.hgt.zip',
		'N19E034.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N19E034.hgt.zip',
		'N19E035.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N19E035.hgt.zip',
		'N19E036.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N19E036.hgt.zip',
		'N19E037.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N19E037.hgt.zip',
		'N19E038.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N19E038.hgt.zip',
		'N19E039.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N19E039.hgt.zip',
		'N19E040.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N19E040.hgt.zip',
		'N19E041.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N19E041.hgt.zip',
		'N19E042.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N19E042.hgt.zip',
		'N19E043.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N19E043.hgt.zip',
		'N19E044.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N19E044.hgt.zip',
		'N19E045.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N19E045.hgt.zip',
		'N19E046.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N19E046.hgt.zip',
		'N19E047.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N19E047.hgt.zip',
		'N19E048.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N19E048.hgt.zip',
		'N19E049.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N19E049.hgt.zip',
		'N19E050.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N19E050.hgt.zip',
		'N19E051.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N19E051.hgt.zip',
		'N19E052.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N19E052.hgt.zip',
		'N19E053.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N19E053.hgt.zip',
		'N19E054.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N19E054.hgt.zip',
		'N19E055.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N19E055.hgt.zip',
		'N19E056.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N19E056.hgt.zip',
		'N19E057.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N19E057.hgt.zip',
		'N19E072.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N19E072.hgt.zip',
		'N19E073.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N19E073.hgt.zip',
		'N19E074.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N19E074.hgt.zip',
		'N19E075.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N19E075.hgt.zip',
		'N19E076.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N19E076.hgt.zip',
		'N19E077.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N19E077.hgt.zip',
		'N19E078.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N19E078.hgt.zip',
		'N19E079.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N19E079.hgt.zip',
		'N19E080.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N19E080.hgt.zip',
		'N19E081.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N19E081.hgt.zip',
		'N19E082.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N19E082.hgt.zip',
		'N19E083.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N19E083.hgt.zip',
		'N19E084.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N19E084.hgt.zip',
		'N19E085.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N19E085.hgt.zip',
		'N19E086.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N19E086.hgt.zip',
		'N19E092.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N19E092.hgt.zip',
		'N19E093.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N19E093.hgt.zip',
		'N19E094.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N19E094.hgt.zip',
		'N19E095.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N19E095.hgt.zip',
		'N19E096.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N19E096.hgt.zip',
		'N19E097.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N19E097.hgt.zip',
		'N19E098.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N19E098.hgt.zip',
		'N19E099.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N19E099.hgt.zip',
		'N19E100.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N19E100.hgt.zip',
		'N19E101.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N19E101.hgt.zip',
		'N19E102.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N19E102.hgt.zip',
		'N19E103.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N19E103.hgt.zip',
		'N19E104.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N19E104.hgt.zip',
		'N19E105.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N19E105.hgt.zip',
		'N19E106.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N19E106.hgt.zip',
		'N19E108.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N19E108.hgt.zip',
		'N19E109.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N19E109.hgt.zip',
		'N19E110.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N19E110.hgt.zip',
		'N19E111.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N19E111.hgt.zip',
		'N19E121.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N19E121.hgt.zip',
		'N19E122.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N19E122.hgt.zip',
		'N19E145.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N19E145.hgt.zip',
		'N19E166.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N19E166.hgt.zip',
		'N19W001.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N19W001.hgt.zip',
		'N19W002.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N19W002.hgt.zip',
		'N19W003.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N19W003.hgt.zip',
		'N19W004.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N19W004.hgt.zip',
		'N19W005.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N19W005.hgt.zip',
		'N19W006.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N19W006.hgt.zip',
		'N19W007.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N19W007.hgt.zip',
		'N19W008.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N19W008.hgt.zip',
		'N19W009.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N19W009.hgt.zip',
		'N19W010.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N19W010.hgt.zip',
		'N19W011.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N19W011.hgt.zip',
		'N19W012.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N19W012.hgt.zip',
		'N19W013.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N19W013.hgt.zip',
		'N19W014.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N19W014.hgt.zip',
		'N19W015.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N19W015.hgt.zip',
		'N19W016.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N19W016.hgt.zip',
		'N19W017.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N19W017.hgt.zip',
		'N19W069.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N19W069.hgt.zip',
		'N19W070.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N19W070.hgt.zip',
		'N19W071.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N19W071.hgt.zip',
		'N19W072.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N19W072.hgt.zip',
		'N19W073.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N19W073.hgt.zip',
		'N19W074.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N19W074.hgt.zip',
		'N19W075.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N19W075.hgt.zip',
		'N19W076.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N19W076.hgt.zip',
		'N19W077.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N19W077.hgt.zip',
		'N19W078.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N19W078.hgt.zip',
		'N19W080.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N19W080.hgt.zip',
		'N19W081.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N19W081.hgt.zip',
		'N19W082.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N19W082.hgt.zip',
		'N19W088.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N19W088.hgt.zip',
		'N19W089.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N19W089.hgt.zip',
		'N19W090.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N19W090.hgt.zip',
		'N19W091.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N19W091.hgt.zip',
		'N19W092.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N19W092.hgt.zip',
		'N19W096.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N19W096.hgt.zip',
		'N19W097.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N19W097.hgt.zip',
		'N19W098.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N19W098.hgt.zip',
		'N19W099.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N19W099.hgt.zip',
		'N19W100.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N19W100.hgt.zip',
		'N19W101.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N19W101.hgt.zip',
		'N19W102.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N19W102.hgt.zip',
		'N19W103.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N19W103.hgt.zip',
		'N19W104.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N19W104.hgt.zip',
		'N19W105.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N19W105.hgt.zip',
		'N19W106.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N19W106.hgt.zip',
		'N19W111.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N19W111.hgt.zip',
		'N19W155.hgt': 'https://srtm.kurviger.de/SRTM3//Islands//N19W155.hgt.zip',
		'N19W156.hgt': 'https://srtm.kurviger.de/SRTM3//Islands//N19W156.hgt.zip',
		'N19W157.hgt': 'https://srtm.kurviger.de/SRTM3//Islands//N19W157.hgt.zip',
		'N20E000.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N20E000.hgt.zip',
		'N20E001.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N20E001.hgt.zip',
		'N20E002.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N20E002.hgt.zip',
		'N20E003.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N20E003.hgt.zip',
		'N20E004.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N20E004.hgt.zip',
		'N20E005.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N20E005.hgt.zip',
		'N20E006.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N20E006.hgt.zip',
		'N20E007.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N20E007.hgt.zip',
		'N20E008.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N20E008.hgt.zip',
		'N20E009.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N20E009.hgt.zip',
		'N20E010.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N20E010.hgt.zip',
		'N20E011.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N20E011.hgt.zip',
		'N20E012.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N20E012.hgt.zip',
		'N20E013.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N20E013.hgt.zip',
		'N20E014.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N20E014.hgt.zip',
		'N20E015.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N20E015.hgt.zip',
		'N20E016.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N20E016.hgt.zip',
		'N20E017.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N20E017.hgt.zip',
		'N20E018.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N20E018.hgt.zip',
		'N20E019.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N20E019.hgt.zip',
		'N20E020.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N20E020.hgt.zip',
		'N20E021.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N20E021.hgt.zip',
		'N20E022.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N20E022.hgt.zip',
		'N20E023.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N20E023.hgt.zip',
		'N20E024.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N20E024.hgt.zip',
		'N20E025.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N20E025.hgt.zip',
		'N20E026.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N20E026.hgt.zip',
		'N20E027.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N20E027.hgt.zip',
		'N20E028.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N20E028.hgt.zip',
		'N20E029.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N20E029.hgt.zip',
		'N20E030.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N20E030.hgt.zip',
		'N20E031.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N20E031.hgt.zip',
		'N20E032.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N20E032.hgt.zip',
		'N20E033.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N20E033.hgt.zip',
		'N20E034.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N20E034.hgt.zip',
		'N20E035.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N20E035.hgt.zip',
		'N20E036.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N20E036.hgt.zip',
		'N20E037.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N20E037.hgt.zip',
		'N20E039.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N20E039.hgt.zip',
		'N20E040.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N20E040.hgt.zip',
		'N20E041.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N20E041.hgt.zip',
		'N20E042.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N20E042.hgt.zip',
		'N20E043.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N20E043.hgt.zip',
		'N20E044.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N20E044.hgt.zip',
		'N20E045.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N20E045.hgt.zip',
		'N20E046.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N20E046.hgt.zip',
		'N20E047.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N20E047.hgt.zip',
		'N20E048.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N20E048.hgt.zip',
		'N20E049.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N20E049.hgt.zip',
		'N20E050.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N20E050.hgt.zip',
		'N20E051.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N20E051.hgt.zip',
		'N20E052.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N20E052.hgt.zip',
		'N20E053.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N20E053.hgt.zip',
		'N20E054.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N20E054.hgt.zip',
		'N20E055.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N20E055.hgt.zip',
		'N20E056.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N20E056.hgt.zip',
		'N20E057.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N20E057.hgt.zip',
		'N20E058.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N20E058.hgt.zip',
		'N20E070.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N20E070.hgt.zip',
		'N20E071.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N20E071.hgt.zip',
		'N20E072.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N20E072.hgt.zip',
		'N20E073.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N20E073.hgt.zip',
		'N20E074.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N20E074.hgt.zip',
		'N20E075.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N20E075.hgt.zip',
		'N20E076.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N20E076.hgt.zip',
		'N20E077.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N20E077.hgt.zip',
		'N20E078.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N20E078.hgt.zip',
		'N20E079.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N20E079.hgt.zip',
		'N20E080.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N20E080.hgt.zip',
		'N20E081.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N20E081.hgt.zip',
		'N20E082.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N20E082.hgt.zip',
		'N20E083.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N20E083.hgt.zip',
		'N20E084.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N20E084.hgt.zip',
		'N20E085.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N20E085.hgt.zip',
		'N20E086.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N20E086.hgt.zip',
		'N20E087.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N20E087.hgt.zip',
		'N20E092.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N20E092.hgt.zip',
		'N20E093.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N20E093.hgt.zip',
		'N20E094.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N20E094.hgt.zip',
		'N20E095.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N20E095.hgt.zip',
		'N20E096.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N20E096.hgt.zip',
		'N20E097.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N20E097.hgt.zip',
		'N20E098.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N20E098.hgt.zip',
		'N20E099.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N20E099.hgt.zip',
		'N20E100.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N20E100.hgt.zip',
		'N20E101.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N20E101.hgt.zip',
		'N20E102.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N20E102.hgt.zip',
		'N20E103.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N20E103.hgt.zip',
		'N20E104.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N20E104.hgt.zip',
		'N20E105.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N20E105.hgt.zip',
		'N20E106.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N20E106.hgt.zip',
		'N20E107.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N20E107.hgt.zip',
		'N20E109.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N20E109.hgt.zip',
		'N20E110.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N20E110.hgt.zip',
		'N20E116.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N20E116.hgt.zip',
		'N20E121.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N20E121.hgt.zip',
		'N20E122.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N20E122.hgt.zip',
		'N20E136.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N20E136.hgt.zip',
		'N20E144.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N20E144.hgt.zip',
		'N20E145.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N20E145.hgt.zip',
		'N20W001.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N20W001.hgt.zip',
		'N20W002.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N20W002.hgt.zip',
		'N20W003.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N20W003.hgt.zip',
		'N20W004.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N20W004.hgt.zip',
		'N20W005.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N20W005.hgt.zip',
		'N20W006.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N20W006.hgt.zip',
		'N20W007.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N20W007.hgt.zip',
		'N20W008.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N20W008.hgt.zip',
		'N20W009.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N20W009.hgt.zip',
		'N20W010.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N20W010.hgt.zip',
		'N20W011.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N20W011.hgt.zip',
		'N20W012.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N20W012.hgt.zip',
		'N20W013.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N20W013.hgt.zip',
		'N20W014.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N20W014.hgt.zip',
		'N20W015.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N20W015.hgt.zip',
		'N20W016.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N20W016.hgt.zip',
		'N20W017.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N20W017.hgt.zip',
		'N20W018.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N20W018.hgt.zip',
		'N20W073.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N20W073.hgt.zip',
		'N20W074.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N20W074.hgt.zip',
		'N20W075.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N20W075.hgt.zip',
		'N20W076.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N20W076.hgt.zip',
		'N20W077.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N20W077.hgt.zip',
		'N20W078.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N20W078.hgt.zip',
		'N20W079.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N20W079.hgt.zip',
		'N20W080.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N20W080.hgt.zip',
		'N20W087.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N20W087.hgt.zip',
		'N20W088.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N20W088.hgt.zip',
		'N20W089.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N20W089.hgt.zip',
		'N20W090.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N20W090.hgt.zip',
		'N20W091.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N20W091.hgt.zip',
		'N20W092.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N20W092.hgt.zip',
		'N20W093.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N20W093.hgt.zip',
		'N20W097.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N20W097.hgt.zip',
		'N20W098.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N20W098.hgt.zip',
		'N20W099.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N20W099.hgt.zip',
		'N20W100.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N20W100.hgt.zip',
		'N20W101.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N20W101.hgt.zip',
		'N20W102.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N20W102.hgt.zip',
		'N20W103.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N20W103.hgt.zip',
		'N20W104.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N20W104.hgt.zip',
		'N20W105.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N20W105.hgt.zip',
		'N20W106.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N20W106.hgt.zip',
		'N20W156.hgt': 'https://srtm.kurviger.de/SRTM3//Islands//N20W156.hgt.zip',
		'N20W157.hgt': 'https://srtm.kurviger.de/SRTM3//Islands//N20W157.hgt.zip',
		'N20W158.hgt': 'https://srtm.kurviger.de/SRTM3//Islands//N20W158.hgt.zip',
		'N21E000.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N21E000.hgt.zip',
		'N21E001.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N21E001.hgt.zip',
		'N21E002.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N21E002.hgt.zip',
		'N21E003.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N21E003.hgt.zip',
		'N21E004.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N21E004.hgt.zip',
		'N21E005.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N21E005.hgt.zip',
		'N21E006.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N21E006.hgt.zip',
		'N21E007.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N21E007.hgt.zip',
		'N21E008.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N21E008.hgt.zip',
		'N21E009.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N21E009.hgt.zip',
		'N21E010.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N21E010.hgt.zip',
		'N21E011.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N21E011.hgt.zip',
		'N21E012.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N21E012.hgt.zip',
		'N21E013.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N21E013.hgt.zip',
		'N21E014.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N21E014.hgt.zip',
		'N21E015.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N21E015.hgt.zip',
		'N21E016.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N21E016.hgt.zip',
		'N21E017.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N21E017.hgt.zip',
		'N21E018.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N21E018.hgt.zip',
		'N21E019.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N21E019.hgt.zip',
		'N21E020.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N21E020.hgt.zip',
		'N21E021.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N21E021.hgt.zip',
		'N21E022.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N21E022.hgt.zip',
		'N21E023.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N21E023.hgt.zip',
		'N21E024.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N21E024.hgt.zip',
		'N21E025.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N21E025.hgt.zip',
		'N21E026.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N21E026.hgt.zip',
		'N21E027.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N21E027.hgt.zip',
		'N21E028.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N21E028.hgt.zip',
		'N21E029.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N21E029.hgt.zip',
		'N21E030.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N21E030.hgt.zip',
		'N21E031.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N21E031.hgt.zip',
		'N21E032.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N21E032.hgt.zip',
		'N21E033.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N21E033.hgt.zip',
		'N21E034.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N21E034.hgt.zip',
		'N21E035.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N21E035.hgt.zip',
		'N21E036.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N21E036.hgt.zip',
		'N21E037.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N21E037.hgt.zip',
		'N21E038.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N21E038.hgt.zip',
		'N21E039.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N21E039.hgt.zip',
		'N21E040.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N21E040.hgt.zip',
		'N21E041.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N21E041.hgt.zip',
		'N21E042.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N21E042.hgt.zip',
		'N21E043.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N21E043.hgt.zip',
		'N21E044.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N21E044.hgt.zip',
		'N21E045.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N21E045.hgt.zip',
		'N21E046.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N21E046.hgt.zip',
		'N21E047.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N21E047.hgt.zip',
		'N21E048.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N21E048.hgt.zip',
		'N21E049.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N21E049.hgt.zip',
		'N21E050.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N21E050.hgt.zip',
		'N21E051.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N21E051.hgt.zip',
		'N21E052.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N21E052.hgt.zip',
		'N21E053.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N21E053.hgt.zip',
		'N21E054.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N21E054.hgt.zip',
		'N21E055.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N21E055.hgt.zip',
		'N21E056.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N21E056.hgt.zip',
		'N21E057.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N21E057.hgt.zip',
		'N21E058.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N21E058.hgt.zip',
		'N21E059.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N21E059.hgt.zip',
		'N21E069.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N21E069.hgt.zip',
		'N21E070.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N21E070.hgt.zip',
		'N21E071.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N21E071.hgt.zip',
		'N21E072.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N21E072.hgt.zip',
		'N21E073.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N21E073.hgt.zip',
		'N21E074.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N21E074.hgt.zip',
		'N21E075.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N21E075.hgt.zip',
		'N21E076.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N21E076.hgt.zip',
		'N21E077.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N21E077.hgt.zip',
		'N21E078.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N21E078.hgt.zip',
		'N21E079.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N21E079.hgt.zip',
		'N21E080.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N21E080.hgt.zip',
		'N21E081.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N21E081.hgt.zip',
		'N21E082.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N21E082.hgt.zip',
		'N21E083.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N21E083.hgt.zip',
		'N21E084.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N21E084.hgt.zip',
		'N21E085.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N21E085.hgt.zip',
		'N21E086.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N21E086.hgt.zip',
		'N21E087.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N21E087.hgt.zip',
		'N21E088.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N21E088.hgt.zip',
		'N21E089.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N21E089.hgt.zip',
		'N21E090.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N21E090.hgt.zip',
		'N21E091.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N21E091.hgt.zip',
		'N21E092.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N21E092.hgt.zip',
		'N21E093.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N21E093.hgt.zip',
		'N21E094.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N21E094.hgt.zip',
		'N21E095.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N21E095.hgt.zip',
		'N21E096.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N21E096.hgt.zip',
		'N21E097.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N21E097.hgt.zip',
		'N21E098.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N21E098.hgt.zip',
		'N21E099.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N21E099.hgt.zip',
		'N21E100.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N21E100.hgt.zip',
		'N21E101.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N21E101.hgt.zip',
		'N21E102.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N21E102.hgt.zip',
		'N21E103.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N21E103.hgt.zip',
		'N21E104.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N21E104.hgt.zip',
		'N21E105.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N21E105.hgt.zip',
		'N21E106.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N21E106.hgt.zip',
		'N21E107.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N21E107.hgt.zip',
		'N21E108.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N21E108.hgt.zip',
		'N21E109.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N21E109.hgt.zip',
		'N21E110.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N21E110.hgt.zip',
		'N21E111.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N21E111.hgt.zip',
		'N21E112.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N21E112.hgt.zip',
		'N21E113.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N21E113.hgt.zip',
		'N21E114.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N21E114.hgt.zip',
		'N21E120.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N21E120.hgt.zip',
		'N21E121.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N21E121.hgt.zip',
		'N21W001.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N21W001.hgt.zip',
		'N21W002.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N21W002.hgt.zip',
		'N21W003.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N21W003.hgt.zip',
		'N21W004.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N21W004.hgt.zip',
		'N21W005.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N21W005.hgt.zip',
		'N21W006.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N21W006.hgt.zip',
		'N21W007.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N21W007.hgt.zip',
		'N21W008.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N21W008.hgt.zip',
		'N21W009.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N21W009.hgt.zip',
		'N21W010.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N21W010.hgt.zip',
		'N21W011.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N21W011.hgt.zip',
		'N21W012.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N21W012.hgt.zip',
		'N21W013.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N21W013.hgt.zip',
		'N21W014.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N21W014.hgt.zip',
		'N21W015.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N21W015.hgt.zip',
		'N21W016.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N21W016.hgt.zip',
		'N21W017.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N21W017.hgt.zip',
		'N21W018.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N21W018.hgt.zip',
		'N21W072.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N21W072.hgt.zip',
		'N21W073.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N21W073.hgt.zip',
		'N21W074.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N21W074.hgt.zip',
		'N21W076.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N21W076.hgt.zip',
		'N21W077.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N21W077.hgt.zip',
		'N21W078.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N21W078.hgt.zip',
		'N21W079.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N21W079.hgt.zip',
		'N21W080.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N21W080.hgt.zip',
		'N21W081.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N21W081.hgt.zip',
		'N21W082.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N21W082.hgt.zip',
		'N21W083.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N21W083.hgt.zip',
		'N21W084.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N21W084.hgt.zip',
		'N21W085.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N21W085.hgt.zip',
		'N21W087.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N21W087.hgt.zip',
		'N21W088.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N21W088.hgt.zip',
		'N21W089.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N21W089.hgt.zip',
		'N21W090.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N21W090.hgt.zip',
		'N21W091.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N21W091.hgt.zip',
		'N21W098.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N21W098.hgt.zip',
		'N21W099.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N21W099.hgt.zip',
		'N21W100.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N21W100.hgt.zip',
		'N21W101.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N21W101.hgt.zip',
		'N21W102.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N21W102.hgt.zip',
		'N21W103.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N21W103.hgt.zip',
		'N21W104.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N21W104.hgt.zip',
		'N21W105.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N21W105.hgt.zip',
		'N21W106.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N21W106.hgt.zip',
		'N21W107.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N21W107.hgt.zip',
		'N21W157.hgt': 'https://srtm.kurviger.de/SRTM3//Islands//N21W157.hgt.zip',
		'N21W158.hgt': 'https://srtm.kurviger.de/SRTM3//Islands//N21W158.hgt.zip',
		'N21W159.hgt': 'https://srtm.kurviger.de/SRTM3//Islands//N21W159.hgt.zip',
		'N21W160.hgt': 'https://srtm.kurviger.de/SRTM3//Islands//N21W160.hgt.zip',
		'N21W161.hgt': 'https://srtm.kurviger.de/SRTM3//Islands//N21W161.hgt.zip',
		'N22E000.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N22E000.hgt.zip',
		'N22E001.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N22E001.hgt.zip',
		'N22E002.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N22E002.hgt.zip',
		'N22E003.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N22E003.hgt.zip',
		'N22E004.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N22E004.hgt.zip',
		'N22E005.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N22E005.hgt.zip',
		'N22E006.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N22E006.hgt.zip',
		'N22E007.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N22E007.hgt.zip',
		'N22E008.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N22E008.hgt.zip',
		'N22E009.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N22E009.hgt.zip',
		'N22E010.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N22E010.hgt.zip',
		'N22E011.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N22E011.hgt.zip',
		'N22E012.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N22E012.hgt.zip',
		'N22E013.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N22E013.hgt.zip',
		'N22E014.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N22E014.hgt.zip',
		'N22E015.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N22E015.hgt.zip',
		'N22E016.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N22E016.hgt.zip',
		'N22E017.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N22E017.hgt.zip',
		'N22E018.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N22E018.hgt.zip',
		'N22E019.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N22E019.hgt.zip',
		'N22E020.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N22E020.hgt.zip',
		'N22E021.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N22E021.hgt.zip',
		'N22E022.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N22E022.hgt.zip',
		'N22E023.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N22E023.hgt.zip',
		'N22E024.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N22E024.hgt.zip',
		'N22E025.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N22E025.hgt.zip',
		'N22E026.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N22E026.hgt.zip',
		'N22E027.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N22E027.hgt.zip',
		'N22E028.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N22E028.hgt.zip',
		'N22E029.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N22E029.hgt.zip',
		'N22E030.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N22E030.hgt.zip',
		'N22E031.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N22E031.hgt.zip',
		'N22E032.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N22E032.hgt.zip',
		'N22E033.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N22E033.hgt.zip',
		'N22E034.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N22E034.hgt.zip',
		'N22E035.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N22E035.hgt.zip',
		'N22E036.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N22E036.hgt.zip',
		'N22E038.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N22E038.hgt.zip',
		'N22E039.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N22E039.hgt.zip',
		'N22E040.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N22E040.hgt.zip',
		'N22E041.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N22E041.hgt.zip',
		'N22E042.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N22E042.hgt.zip',
		'N22E043.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N22E043.hgt.zip',
		'N22E044.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N22E044.hgt.zip',
		'N22E045.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N22E045.hgt.zip',
		'N22E046.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N22E046.hgt.zip',
		'N22E047.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N22E047.hgt.zip',
		'N22E048.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N22E048.hgt.zip',
		'N22E049.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N22E049.hgt.zip',
		'N22E050.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N22E050.hgt.zip',
		'N22E051.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N22E051.hgt.zip',
		'N22E052.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N22E052.hgt.zip',
		'N22E053.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N22E053.hgt.zip',
		'N22E054.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N22E054.hgt.zip',
		'N22E055.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N22E055.hgt.zip',
		'N22E056.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N22E056.hgt.zip',
		'N22E057.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N22E057.hgt.zip',
		'N22E058.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N22E058.hgt.zip',
		'N22E059.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N22E059.hgt.zip',
		'N22E068.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N22E068.hgt.zip',
		'N22E069.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N22E069.hgt.zip',
		'N22E070.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N22E070.hgt.zip',
		'N22E071.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N22E071.hgt.zip',
		'N22E072.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N22E072.hgt.zip',
		'N22E073.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N22E073.hgt.zip',
		'N22E074.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N22E074.hgt.zip',
		'N22E075.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N22E075.hgt.zip',
		'N22E076.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N22E076.hgt.zip',
		'N22E077.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N22E077.hgt.zip',
		'N22E078.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N22E078.hgt.zip',
		'N22E079.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N22E079.hgt.zip',
		'N22E080.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N22E080.hgt.zip',
		'N22E081.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N22E081.hgt.zip',
		'N22E082.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N22E082.hgt.zip',
		'N22E083.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N22E083.hgt.zip',
		'N22E084.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N22E084.hgt.zip',
		'N22E085.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N22E085.hgt.zip',
		'N22E086.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N22E086.hgt.zip',
		'N22E087.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N22E087.hgt.zip',
		'N22E088.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N22E088.hgt.zip',
		'N22E089.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N22E089.hgt.zip',
		'N22E090.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N22E090.hgt.zip',
		'N22E091.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N22E091.hgt.zip',
		'N22E092.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N22E092.hgt.zip',
		'N22E093.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N22E093.hgt.zip',
		'N22E094.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N22E094.hgt.zip',
		'N22E095.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N22E095.hgt.zip',
		'N22E096.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N22E096.hgt.zip',
		'N22E097.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N22E097.hgt.zip',
		'N22E098.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N22E098.hgt.zip',
		'N22E099.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N22E099.hgt.zip',
		'N22E100.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N22E100.hgt.zip',
		'N22E101.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N22E101.hgt.zip',
		'N22E102.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N22E102.hgt.zip',
		'N22E103.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N22E103.hgt.zip',
		'N22E104.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N22E104.hgt.zip',
		'N22E105.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N22E105.hgt.zip',
		'N22E106.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N22E106.hgt.zip',
		'N22E107.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N22E107.hgt.zip',
		'N22E108.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N22E108.hgt.zip',
		'N22E109.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N22E109.hgt.zip',
		'N22E110.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N22E110.hgt.zip',
		'N22E111.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N22E111.hgt.zip',
		'N22E112.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N22E112.hgt.zip',
		'N22E113.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N22E113.hgt.zip',
		'N22E114.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N22E114.hgt.zip',
		'N22E115.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N22E115.hgt.zip',
		'N22E116.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N22E116.hgt.zip',
		'N22E120.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N22E120.hgt.zip',
		'N22E121.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N22E121.hgt.zip',
		'N22W001.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N22W001.hgt.zip',
		'N22W002.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N22W002.hgt.zip',
		'N22W003.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N22W003.hgt.zip',
		'N22W004.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N22W004.hgt.zip',
		'N22W005.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N22W005.hgt.zip',
		'N22W006.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N22W006.hgt.zip',
		'N22W007.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N22W007.hgt.zip',
		'N22W008.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N22W008.hgt.zip',
		'N22W009.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N22W009.hgt.zip',
		'N22W010.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N22W010.hgt.zip',
		'N22W011.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N22W011.hgt.zip',
		'N22W012.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N22W012.hgt.zip',
		'N22W013.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N22W013.hgt.zip',
		'N22W014.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N22W014.hgt.zip',
		'N22W015.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N22W015.hgt.zip',
		'N22W016.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N22W016.hgt.zip',
		'N22W017.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N22W017.hgt.zip',
		'N22W073.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N22W073.hgt.zip',
		'N22W074.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N22W074.hgt.zip',
		'N22W075.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N22W075.hgt.zip',
		'N22W076.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N22W076.hgt.zip',
		'N22W078.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N22W078.hgt.zip',
		'N22W079.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N22W079.hgt.zip',
		'N22W080.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N22W080.hgt.zip',
		'N22W081.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N22W081.hgt.zip',
		'N22W082.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N22W082.hgt.zip',
		'N22W083.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N22W083.hgt.zip',
		'N22W084.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N22W084.hgt.zip',
		'N22W085.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N22W085.hgt.zip',
		'N22W090.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N22W090.hgt.zip',
		'N22W092.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N22W092.hgt.zip',
		'N22W098.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N22W098.hgt.zip',
		'N22W099.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N22W099.hgt.zip',
		'N22W100.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N22W100.hgt.zip',
		'N22W101.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N22W101.hgt.zip',
		'N22W102.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N22W102.hgt.zip',
		'N22W103.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N22W103.hgt.zip',
		'N22W104.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N22W104.hgt.zip',
		'N22W105.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N22W105.hgt.zip',
		'N22W106.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N22W106.hgt.zip',
		'N22W107.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N22W107.hgt.zip',
		'N22W110.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N22W110.hgt.zip',
		'N22W111.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N22W111.hgt.zip',
		'N22W160.hgt': 'https://srtm.kurviger.de/SRTM3//Islands//N22W160.hgt.zip',
		'N22W161.hgt': 'https://srtm.kurviger.de/SRTM3//Islands//N22W161.hgt.zip',
		'N23E000.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N23E000.hgt.zip',
		'N23E001.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N23E001.hgt.zip',
		'N23E002.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N23E002.hgt.zip',
		'N23E003.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N23E003.hgt.zip',
		'N23E004.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N23E004.hgt.zip',
		'N23E005.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N23E005.hgt.zip',
		'N23E006.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N23E006.hgt.zip',
		'N23E007.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N23E007.hgt.zip',
		'N23E008.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N23E008.hgt.zip',
		'N23E009.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N23E009.hgt.zip',
		'N23E010.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N23E010.hgt.zip',
		'N23E011.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N23E011.hgt.zip',
		'N23E012.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N23E012.hgt.zip',
		'N23E013.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N23E013.hgt.zip',
		'N23E014.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N23E014.hgt.zip',
		'N23E015.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N23E015.hgt.zip',
		'N23E016.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N23E016.hgt.zip',
		'N23E017.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N23E017.hgt.zip',
		'N23E018.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N23E018.hgt.zip',
		'N23E019.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N23E019.hgt.zip',
		'N23E020.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N23E020.hgt.zip',
		'N23E021.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N23E021.hgt.zip',
		'N23E022.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N23E022.hgt.zip',
		'N23E023.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N23E023.hgt.zip',
		'N23E024.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N23E024.hgt.zip',
		'N23E025.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N23E025.hgt.zip',
		'N23E026.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N23E026.hgt.zip',
		'N23E027.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N23E027.hgt.zip',
		'N23E028.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N23E028.hgt.zip',
		'N23E029.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N23E029.hgt.zip',
		'N23E030.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N23E030.hgt.zip',
		'N23E031.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N23E031.hgt.zip',
		'N23E032.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N23E032.hgt.zip',
		'N23E033.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N23E033.hgt.zip',
		'N23E034.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N23E034.hgt.zip',
		'N23E035.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N23E035.hgt.zip',
		'N23E036.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N23E036.hgt.zip',
		'N23E038.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N23E038.hgt.zip',
		'N23E039.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N23E039.hgt.zip',
		'N23E040.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N23E040.hgt.zip',
		'N23E041.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N23E041.hgt.zip',
		'N23E042.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N23E042.hgt.zip',
		'N23E043.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N23E043.hgt.zip',
		'N23E044.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N23E044.hgt.zip',
		'N23E045.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N23E045.hgt.zip',
		'N23E046.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N23E046.hgt.zip',
		'N23E047.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N23E047.hgt.zip',
		'N23E048.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N23E048.hgt.zip',
		'N23E049.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N23E049.hgt.zip',
		'N23E050.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N23E050.hgt.zip',
		'N23E051.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N23E051.hgt.zip',
		'N23E052.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N23E052.hgt.zip',
		'N23E053.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N23E053.hgt.zip',
		'N23E054.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N23E054.hgt.zip',
		'N23E055.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N23E055.hgt.zip',
		'N23E056.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N23E056.hgt.zip',
		'N23E057.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N23E057.hgt.zip',
		'N23E058.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N23E058.hgt.zip',
		'N23E059.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N23E059.hgt.zip',
		'N23E067.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N23E067.hgt.zip',
		'N23E068.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N23E068.hgt.zip',
		'N23E069.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N23E069.hgt.zip',
		'N23E070.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N23E070.hgt.zip',
		'N23E071.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N23E071.hgt.zip',
		'N23E072.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N23E072.hgt.zip',
		'N23E073.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N23E073.hgt.zip',
		'N23E074.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N23E074.hgt.zip',
		'N23E075.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N23E075.hgt.zip',
		'N23E076.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N23E076.hgt.zip',
		'N23E077.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N23E077.hgt.zip',
		'N23E078.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N23E078.hgt.zip',
		'N23E079.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N23E079.hgt.zip',
		'N23E080.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N23E080.hgt.zip',
		'N23E081.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N23E081.hgt.zip',
		'N23E082.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N23E082.hgt.zip',
		'N23E083.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N23E083.hgt.zip',
		'N23E084.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N23E084.hgt.zip',
		'N23E085.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N23E085.hgt.zip',
		'N23E086.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N23E086.hgt.zip',
		'N23E087.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N23E087.hgt.zip',
		'N23E088.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N23E088.hgt.zip',
		'N23E089.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N23E089.hgt.zip',
		'N23E090.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N23E090.hgt.zip',
		'N23E091.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N23E091.hgt.zip',
		'N23E092.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N23E092.hgt.zip',
		'N23E093.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N23E093.hgt.zip',
		'N23E094.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N23E094.hgt.zip',
		'N23E095.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N23E095.hgt.zip',
		'N23E096.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N23E096.hgt.zip',
		'N23E097.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N23E097.hgt.zip',
		'N23E098.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N23E098.hgt.zip',
		'N23E099.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N23E099.hgt.zip',
		'N23E100.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N23E100.hgt.zip',
		'N23E101.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N23E101.hgt.zip',
		'N23E102.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N23E102.hgt.zip',
		'N23E103.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N23E103.hgt.zip',
		'N23E104.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N23E104.hgt.zip',
		'N23E105.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N23E105.hgt.zip',
		'N23E106.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N23E106.hgt.zip',
		'N23E107.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N23E107.hgt.zip',
		'N23E108.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N23E108.hgt.zip',
		'N23E109.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N23E109.hgt.zip',
		'N23E110.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N23E110.hgt.zip',
		'N23E111.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N23E111.hgt.zip',
		'N23E112.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N23E112.hgt.zip',
		'N23E113.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N23E113.hgt.zip',
		'N23E114.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N23E114.hgt.zip',
		'N23E115.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N23E115.hgt.zip',
		'N23E116.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N23E116.hgt.zip',
		'N23E117.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N23E117.hgt.zip',
		'N23E119.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N23E119.hgt.zip',
		'N23E120.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N23E120.hgt.zip',
		'N23E121.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N23E121.hgt.zip',
		'N23W001.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N23W001.hgt.zip',
		'N23W002.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N23W002.hgt.zip',
		'N23W003.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N23W003.hgt.zip',
		'N23W004.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N23W004.hgt.zip',
		'N23W005.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N23W005.hgt.zip',
		'N23W006.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N23W006.hgt.zip',
		'N23W007.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N23W007.hgt.zip',
		'N23W008.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N23W008.hgt.zip',
		'N23W009.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N23W009.hgt.zip',
		'N23W010.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N23W010.hgt.zip',
		'N23W011.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N23W011.hgt.zip',
		'N23W012.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N23W012.hgt.zip',
		'N23W013.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N23W013.hgt.zip',
		'N23W014.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N23W014.hgt.zip',
		'N23W015.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N23W015.hgt.zip',
		'N23W016.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N23W016.hgt.zip',
		'N23W017.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N23W017.hgt.zip',
		'N23W074.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N23W074.hgt.zip',
		'N23W075.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N23W075.hgt.zip',
		'N23W076.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N23W076.hgt.zip',
		'N23W077.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N23W077.hgt.zip',
		'N23W078.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N23W078.hgt.zip',
		'N23W080.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N23W080.hgt.zip',
		'N23W081.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N23W081.hgt.zip',
		'N23W082.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N23W082.hgt.zip',
		'N23W083.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N23W083.hgt.zip',
		'N23W084.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N23W084.hgt.zip',
		'N23W098.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N23W098.hgt.zip',
		'N23W099.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N23W099.hgt.zip',
		'N23W100.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N23W100.hgt.zip',
		'N23W101.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N23W101.hgt.zip',
		'N23W102.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N23W102.hgt.zip',
		'N23W103.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N23W103.hgt.zip',
		'N23W104.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N23W104.hgt.zip',
		'N23W105.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N23W105.hgt.zip',
		'N23W106.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N23W106.hgt.zip',
		'N23W107.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N23W107.hgt.zip',
		'N23W108.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N23W108.hgt.zip',
		'N23W110.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N23W110.hgt.zip',
		'N23W111.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N23W111.hgt.zip',
		'N23W162.hgt': 'https://srtm.kurviger.de/SRTM3//Islands//N23W162.hgt.zip',
		'N23W165.hgt': 'https://srtm.kurviger.de/SRTM3//Islands//N23W165.hgt.zip',
		'N23W167.hgt': 'https://srtm.kurviger.de/SRTM3//Islands//N23W167.hgt.zip',
		'N24E000.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N24E000.hgt.zip',
		'N24E001.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N24E001.hgt.zip',
		'N24E002.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N24E002.hgt.zip',
		'N24E003.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N24E003.hgt.zip',
		'N24E004.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N24E004.hgt.zip',
		'N24E005.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N24E005.hgt.zip',
		'N24E006.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N24E006.hgt.zip',
		'N24E007.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N24E007.hgt.zip',
		'N24E008.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N24E008.hgt.zip',
		'N24E009.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N24E009.hgt.zip',
		'N24E010.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N24E010.hgt.zip',
		'N24E011.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N24E011.hgt.zip',
		'N24E012.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N24E012.hgt.zip',
		'N24E013.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N24E013.hgt.zip',
		'N24E014.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N24E014.hgt.zip',
		'N24E015.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N24E015.hgt.zip',
		'N24E016.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N24E016.hgt.zip',
		'N24E017.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N24E017.hgt.zip',
		'N24E018.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N24E018.hgt.zip',
		'N24E019.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N24E019.hgt.zip',
		'N24E020.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N24E020.hgt.zip',
		'N24E021.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N24E021.hgt.zip',
		'N24E022.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N24E022.hgt.zip',
		'N24E023.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N24E023.hgt.zip',
		'N24E024.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N24E024.hgt.zip',
		'N24E025.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N24E025.hgt.zip',
		'N24E026.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N24E026.hgt.zip',
		'N24E027.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N24E027.hgt.zip',
		'N24E028.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N24E028.hgt.zip',
		'N24E029.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N24E029.hgt.zip',
		'N24E030.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N24E030.hgt.zip',
		'N24E031.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N24E031.hgt.zip',
		'N24E032.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N24E032.hgt.zip',
		'N24E033.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N24E033.hgt.zip',
		'N24E034.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N24E034.hgt.zip',
		'N24E035.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N24E035.hgt.zip',
		'N24E037.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N24E037.hgt.zip',
		'N24E038.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N24E038.hgt.zip',
		'N24E039.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N24E039.hgt.zip',
		'N24E040.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N24E040.hgt.zip',
		'N24E041.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N24E041.hgt.zip',
		'N24E042.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N24E042.hgt.zip',
		'N24E043.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N24E043.hgt.zip',
		'N24E044.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N24E044.hgt.zip',
		'N24E045.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N24E045.hgt.zip',
		'N24E046.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N24E046.hgt.zip',
		'N24E047.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N24E047.hgt.zip',
		'N24E048.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N24E048.hgt.zip',
		'N24E049.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N24E049.hgt.zip',
		'N24E050.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N24E050.hgt.zip',
		'N24E051.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N24E051.hgt.zip',
		'N24E052.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N24E052.hgt.zip',
		'N24E053.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N24E053.hgt.zip',
		'N24E054.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N24E054.hgt.zip',
		'N24E055.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N24E055.hgt.zip',
		'N24E056.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N24E056.hgt.zip',
		'N24E057.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N24E057.hgt.zip',
		'N24E066.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N24E066.hgt.zip',
		'N24E067.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N24E067.hgt.zip',
		'N24E068.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N24E068.hgt.zip',
		'N24E069.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N24E069.hgt.zip',
		'N24E070.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N24E070.hgt.zip',
		'N24E071.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N24E071.hgt.zip',
		'N24E072.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N24E072.hgt.zip',
		'N24E073.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N24E073.hgt.zip',
		'N24E074.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N24E074.hgt.zip',
		'N24E075.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N24E075.hgt.zip',
		'N24E076.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N24E076.hgt.zip',
		'N24E077.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N24E077.hgt.zip',
		'N24E078.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N24E078.hgt.zip',
		'N24E079.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N24E079.hgt.zip',
		'N24E080.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N24E080.hgt.zip',
		'N24E081.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N24E081.hgt.zip',
		'N24E082.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N24E082.hgt.zip',
		'N24E083.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N24E083.hgt.zip',
		'N24E084.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N24E084.hgt.zip',
		'N24E085.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N24E085.hgt.zip',
		'N24E086.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N24E086.hgt.zip',
		'N24E087.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N24E087.hgt.zip',
		'N24E088.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N24E088.hgt.zip',
		'N24E089.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N24E089.hgt.zip',
		'N24E090.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N24E090.hgt.zip',
		'N24E091.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N24E091.hgt.zip',
		'N24E092.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N24E092.hgt.zip',
		'N24E093.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N24E093.hgt.zip',
		'N24E094.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N24E094.hgt.zip',
		'N24E095.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N24E095.hgt.zip',
		'N24E096.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N24E096.hgt.zip',
		'N24E097.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N24E097.hgt.zip',
		'N24E098.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N24E098.hgt.zip',
		'N24E099.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N24E099.hgt.zip',
		'N24E100.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N24E100.hgt.zip',
		'N24E101.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N24E101.hgt.zip',
		'N24E102.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N24E102.hgt.zip',
		'N24E103.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N24E103.hgt.zip',
		'N24E104.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N24E104.hgt.zip',
		'N24E105.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N24E105.hgt.zip',
		'N24E106.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N24E106.hgt.zip',
		'N24E107.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N24E107.hgt.zip',
		'N24E108.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N24E108.hgt.zip',
		'N24E109.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N24E109.hgt.zip',
		'N24E110.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N24E110.hgt.zip',
		'N24E111.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N24E111.hgt.zip',
		'N24E112.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N24E112.hgt.zip',
		'N24E113.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N24E113.hgt.zip',
		'N24E114.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N24E114.hgt.zip',
		'N24E115.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N24E115.hgt.zip',
		'N24E116.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N24E116.hgt.zip',
		'N24E117.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N24E117.hgt.zip',
		'N24E118.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N24E118.hgt.zip',
		'N24E119.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N24E119.hgt.zip',
		'N24E120.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N24E120.hgt.zip',
		'N24E121.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N24E121.hgt.zip',
		'N24E122.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N24E122.hgt.zip',
		'N24E123.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N24E123.hgt.zip',
		'N24E124.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N24E124.hgt.zip',
		'N24E125.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N24E125.hgt.zip',
		'N24E131.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N24E131.hgt.zip',
		'N24E141.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N24E141.hgt.zip',
		'N24E153.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N24E153.hgt.zip',
		'N24W001.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N24W001.hgt.zip',
		'N24W002.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N24W002.hgt.zip',
		'N24W003.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N24W003.hgt.zip',
		'N24W004.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N24W004.hgt.zip',
		'N24W005.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N24W005.hgt.zip',
		'N24W006.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N24W006.hgt.zip',
		'N24W007.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N24W007.hgt.zip',
		'N24W008.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N24W008.hgt.zip',
		'N24W009.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N24W009.hgt.zip',
		'N24W010.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N24W010.hgt.zip',
		'N24W011.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N24W011.hgt.zip',
		'N24W012.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N24W012.hgt.zip',
		'N24W013.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N24W013.hgt.zip',
		'N24W014.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N24W014.hgt.zip',
		'N24W015.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N24W015.hgt.zip',
		'N24W016.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N24W016.hgt.zip',
		'N24W075.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N24W075.hgt.zip',
		'N24W076.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N24W076.hgt.zip',
		'N24W077.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N24W077.hgt.zip',
		'N24W078.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N24W078.hgt.zip',
		'N24W079.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N24W079.hgt.zip',
		'N24W080.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N24W080.hgt.zip',
		'N24W081.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N24W081.hgt.zip',
		'N24W082.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N24W082.hgt.zip',
		'N24W083.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N24W083.hgt.zip',
		'N24W098.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N24W098.hgt.zip',
		'N24W099.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N24W099.hgt.zip',
		'N24W100.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N24W100.hgt.zip',
		'N24W101.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N24W101.hgt.zip',
		'N24W102.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N24W102.hgt.zip',
		'N24W103.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N24W103.hgt.zip',
		'N24W104.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N24W104.hgt.zip',
		'N24W105.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N24W105.hgt.zip',
		'N24W106.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N24W106.hgt.zip',
		'N24W107.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N24W107.hgt.zip',
		'N24W108.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N24W108.hgt.zip',
		'N24W109.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N24W109.hgt.zip',
		'N24W110.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N24W110.hgt.zip',
		'N24W111.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N24W111.hgt.zip',
		'N24W112.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N24W112.hgt.zip',
		'N24W113.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N24W113.hgt.zip',
		'N24W116.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N24W116.hgt.zip',
		'N24W168.hgt': 'https://srtm.kurviger.de/SRTM3//Islands//N24W168.hgt.zip',
		'N25E000.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N25E000.hgt.zip',
		'N25E001.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N25E001.hgt.zip',
		'N25E002.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N25E002.hgt.zip',
		'N25E003.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N25E003.hgt.zip',
		'N25E004.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N25E004.hgt.zip',
		'N25E005.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N25E005.hgt.zip',
		'N25E006.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N25E006.hgt.zip',
		'N25E007.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N25E007.hgt.zip',
		'N25E008.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N25E008.hgt.zip',
		'N25E009.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N25E009.hgt.zip',
		'N25E010.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N25E010.hgt.zip',
		'N25E011.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N25E011.hgt.zip',
		'N25E012.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N25E012.hgt.zip',
		'N25E013.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N25E013.hgt.zip',
		'N25E014.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N25E014.hgt.zip',
		'N25E015.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N25E015.hgt.zip',
		'N25E016.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N25E016.hgt.zip',
		'N25E017.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N25E017.hgt.zip',
		'N25E018.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N25E018.hgt.zip',
		'N25E019.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N25E019.hgt.zip',
		'N25E020.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N25E020.hgt.zip',
		'N25E021.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N25E021.hgt.zip',
		'N25E022.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N25E022.hgt.zip',
		'N25E023.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N25E023.hgt.zip',
		'N25E024.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N25E024.hgt.zip',
		'N25E025.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N25E025.hgt.zip',
		'N25E026.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N25E026.hgt.zip',
		'N25E027.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N25E027.hgt.zip',
		'N25E028.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N25E028.hgt.zip',
		'N25E029.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N25E029.hgt.zip',
		'N25E030.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N25E030.hgt.zip',
		'N25E031.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N25E031.hgt.zip',
		'N25E032.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N25E032.hgt.zip',
		'N25E033.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N25E033.hgt.zip',
		'N25E034.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N25E034.hgt.zip',
		'N25E036.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N25E036.hgt.zip',
		'N25E037.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N25E037.hgt.zip',
		'N25E038.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N25E038.hgt.zip',
		'N25E039.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N25E039.hgt.zip',
		'N25E040.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N25E040.hgt.zip',
		'N25E041.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N25E041.hgt.zip',
		'N25E042.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N25E042.hgt.zip',
		'N25E043.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N25E043.hgt.zip',
		'N25E044.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N25E044.hgt.zip',
		'N25E045.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N25E045.hgt.zip',
		'N25E046.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N25E046.hgt.zip',
		'N25E047.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N25E047.hgt.zip',
		'N25E048.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N25E048.hgt.zip',
		'N25E049.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N25E049.hgt.zip',
		'N25E050.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N25E050.hgt.zip',
		'N25E051.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N25E051.hgt.zip',
		'N25E052.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N25E052.hgt.zip',
		'N25E054.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N25E054.hgt.zip',
		'N25E055.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N25E055.hgt.zip',
		'N25E056.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N25E056.hgt.zip',
		'N25E057.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N25E057.hgt.zip',
		'N25E058.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N25E058.hgt.zip',
		'N25E059.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N25E059.hgt.zip',
		'N25E060.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N25E060.hgt.zip',
		'N25E061.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N25E061.hgt.zip',
		'N25E062.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N25E062.hgt.zip',
		'N25E063.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N25E063.hgt.zip',
		'N25E064.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N25E064.hgt.zip',
		'N25E065.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N25E065.hgt.zip',
		'N25E066.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N25E066.hgt.zip',
		'N25E067.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N25E067.hgt.zip',
		'N25E068.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N25E068.hgt.zip',
		'N25E069.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N25E069.hgt.zip',
		'N25E070.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N25E070.hgt.zip',
		'N25E071.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N25E071.hgt.zip',
		'N25E072.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N25E072.hgt.zip',
		'N25E073.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N25E073.hgt.zip',
		'N25E074.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N25E074.hgt.zip',
		'N25E075.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N25E075.hgt.zip',
		'N25E076.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N25E076.hgt.zip',
		'N25E077.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N25E077.hgt.zip',
		'N25E078.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N25E078.hgt.zip',
		'N25E079.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N25E079.hgt.zip',
		'N25E080.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N25E080.hgt.zip',
		'N25E081.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N25E081.hgt.zip',
		'N25E082.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N25E082.hgt.zip',
		'N25E083.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N25E083.hgt.zip',
		'N25E084.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N25E084.hgt.zip',
		'N25E085.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N25E085.hgt.zip',
		'N25E086.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N25E086.hgt.zip',
		'N25E087.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N25E087.hgt.zip',
		'N25E088.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N25E088.hgt.zip',
		'N25E089.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N25E089.hgt.zip',
		'N25E090.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N25E090.hgt.zip',
		'N25E091.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N25E091.hgt.zip',
		'N25E092.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N25E092.hgt.zip',
		'N25E093.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N25E093.hgt.zip',
		'N25E094.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N25E094.hgt.zip',
		'N25E095.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N25E095.hgt.zip',
		'N25E096.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N25E096.hgt.zip',
		'N25E097.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N25E097.hgt.zip',
		'N25E098.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N25E098.hgt.zip',
		'N25E099.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N25E099.hgt.zip',
		'N25E100.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N25E100.hgt.zip',
		'N25E101.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N25E101.hgt.zip',
		'N25E102.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N25E102.hgt.zip',
		'N25E103.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N25E103.hgt.zip',
		'N25E104.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N25E104.hgt.zip',
		'N25E105.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N25E105.hgt.zip',
		'N25E106.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N25E106.hgt.zip',
		'N25E107.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N25E107.hgt.zip',
		'N25E108.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N25E108.hgt.zip',
		'N25E109.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N25E109.hgt.zip',
		'N25E110.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N25E110.hgt.zip',
		'N25E111.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N25E111.hgt.zip',
		'N25E112.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N25E112.hgt.zip',
		'N25E113.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N25E113.hgt.zip',
		'N25E114.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N25E114.hgt.zip',
		'N25E115.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N25E115.hgt.zip',
		'N25E116.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N25E116.hgt.zip',
		'N25E117.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N25E117.hgt.zip',
		'N25E118.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N25E118.hgt.zip',
		'N25E119.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N25E119.hgt.zip',
		'N25E121.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N25E121.hgt.zip',
		'N25E122.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N25E122.hgt.zip',
		'N25E123.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N25E123.hgt.zip',
		'N25E124.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N25E124.hgt.zip',
		'N25E131.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N25E131.hgt.zip',
		'N25E141.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N25E141.hgt.zip',
		'N25W001.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N25W001.hgt.zip',
		'N25W002.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N25W002.hgt.zip',
		'N25W003.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N25W003.hgt.zip',
		'N25W004.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N25W004.hgt.zip',
		'N25W005.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N25W005.hgt.zip',
		'N25W006.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N25W006.hgt.zip',
		'N25W007.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N25W007.hgt.zip',
		'N25W008.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N25W008.hgt.zip',
		'N25W009.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N25W009.hgt.zip',
		'N25W010.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N25W010.hgt.zip',
		'N25W011.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N25W011.hgt.zip',
		'N25W012.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N25W012.hgt.zip',
		'N25W013.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N25W013.hgt.zip',
		'N25W014.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N25W014.hgt.zip',
		'N25W015.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N25W015.hgt.zip',
		'N25W077.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N25W077.hgt.zip',
		'N25W078.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N25W078.hgt.zip',
		'N25W079.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N25W079.hgt.zip',
		'N25W080.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N25W080.hgt.zip',
		'N25W081.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N25W081.hgt.zip',
		'N25W082.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N25W082.hgt.zip',
		'N25W098.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N25W098.hgt.zip',
		'N25W099.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N25W099.hgt.zip',
		'N25W100.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N25W100.hgt.zip',
		'N25W101.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N25W101.hgt.zip',
		'N25W102.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N25W102.hgt.zip',
		'N25W103.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N25W103.hgt.zip',
		'N25W104.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N25W104.hgt.zip',
		'N25W105.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N25W105.hgt.zip',
		'N25W106.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N25W106.hgt.zip',
		'N25W107.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N25W107.hgt.zip',
		'N25W108.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N25W108.hgt.zip',
		'N25W109.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N25W109.hgt.zip',
		'N25W110.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N25W110.hgt.zip',
		'N25W111.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N25W111.hgt.zip',
		'N25W112.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N25W112.hgt.zip',
		'N25W113.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N25W113.hgt.zip',
		'N25W168.hgt': 'https://srtm.kurviger.de/SRTM3//Islands//N25W168.hgt.zip',
		'N25W172.hgt': 'https://srtm.kurviger.de/SRTM3//Islands//N25W172.hgt.zip',
		'N26E000.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N26E000.hgt.zip',
		'N26E001.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N26E001.hgt.zip',
		'N26E002.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N26E002.hgt.zip',
		'N26E003.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N26E003.hgt.zip',
		'N26E004.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N26E004.hgt.zip',
		'N26E005.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N26E005.hgt.zip',
		'N26E006.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N26E006.hgt.zip',
		'N26E007.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N26E007.hgt.zip',
		'N26E008.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N26E008.hgt.zip',
		'N26E009.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N26E009.hgt.zip',
		'N26E010.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N26E010.hgt.zip',
		'N26E011.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N26E011.hgt.zip',
		'N26E012.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N26E012.hgt.zip',
		'N26E013.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N26E013.hgt.zip',
		'N26E014.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N26E014.hgt.zip',
		'N26E015.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N26E015.hgt.zip',
		'N26E016.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N26E016.hgt.zip',
		'N26E017.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N26E017.hgt.zip',
		'N26E018.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N26E018.hgt.zip',
		'N26E019.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N26E019.hgt.zip',
		'N26E020.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N26E020.hgt.zip',
		'N26E021.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N26E021.hgt.zip',
		'N26E022.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N26E022.hgt.zip',
		'N26E023.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N26E023.hgt.zip',
		'N26E024.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N26E024.hgt.zip',
		'N26E025.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N26E025.hgt.zip',
		'N26E026.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N26E026.hgt.zip',
		'N26E027.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N26E027.hgt.zip',
		'N26E028.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N26E028.hgt.zip',
		'N26E029.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N26E029.hgt.zip',
		'N26E030.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N26E030.hgt.zip',
		'N26E031.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N26E031.hgt.zip',
		'N26E032.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N26E032.hgt.zip',
		'N26E033.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N26E033.hgt.zip',
		'N26E034.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N26E034.hgt.zip',
		'N26E035.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N26E035.hgt.zip',
		'N26E036.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N26E036.hgt.zip',
		'N26E037.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N26E037.hgt.zip',
		'N26E038.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N26E038.hgt.zip',
		'N26E039.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N26E039.hgt.zip',
		'N26E040.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N26E040.hgt.zip',
		'N26E041.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N26E041.hgt.zip',
		'N26E042.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N26E042.hgt.zip',
		'N26E043.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N26E043.hgt.zip',
		'N26E044.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N26E044.hgt.zip',
		'N26E045.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N26E045.hgt.zip',
		'N26E046.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N26E046.hgt.zip',
		'N26E047.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N26E047.hgt.zip',
		'N26E048.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N26E048.hgt.zip',
		'N26E049.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N26E049.hgt.zip',
		'N26E050.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N26E050.hgt.zip',
		'N26E051.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N26E051.hgt.zip',
		'N26E053.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N26E053.hgt.zip',
		'N26E054.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N26E054.hgt.zip',
		'N26E055.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N26E055.hgt.zip',
		'N26E056.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N26E056.hgt.zip',
		'N26E057.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N26E057.hgt.zip',
		'N26E058.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N26E058.hgt.zip',
		'N26E059.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N26E059.hgt.zip',
		'N26E060.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N26E060.hgt.zip',
		'N26E061.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N26E061.hgt.zip',
		'N26E062.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N26E062.hgt.zip',
		'N26E063.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N26E063.hgt.zip',
		'N26E064.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N26E064.hgt.zip',
		'N26E065.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N26E065.hgt.zip',
		'N26E066.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N26E066.hgt.zip',
		'N26E067.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N26E067.hgt.zip',
		'N26E068.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N26E068.hgt.zip',
		'N26E069.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N26E069.hgt.zip',
		'N26E070.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N26E070.hgt.zip',
		'N26E071.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N26E071.hgt.zip',
		'N26E072.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N26E072.hgt.zip',
		'N26E073.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N26E073.hgt.zip',
		'N26E074.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N26E074.hgt.zip',
		'N26E075.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N26E075.hgt.zip',
		'N26E076.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N26E076.hgt.zip',
		'N26E077.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N26E077.hgt.zip',
		'N26E078.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N26E078.hgt.zip',
		'N26E079.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N26E079.hgt.zip',
		'N26E080.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N26E080.hgt.zip',
		'N26E081.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N26E081.hgt.zip',
		'N26E082.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N26E082.hgt.zip',
		'N26E083.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N26E083.hgt.zip',
		'N26E084.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N26E084.hgt.zip',
		'N26E085.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N26E085.hgt.zip',
		'N26E086.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N26E086.hgt.zip',
		'N26E087.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N26E087.hgt.zip',
		'N26E088.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N26E088.hgt.zip',
		'N26E089.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N26E089.hgt.zip',
		'N26E090.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N26E090.hgt.zip',
		'N26E091.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N26E091.hgt.zip',
		'N26E092.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N26E092.hgt.zip',
		'N26E093.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N26E093.hgt.zip',
		'N26E094.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N26E094.hgt.zip',
		'N26E095.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N26E095.hgt.zip',
		'N26E096.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N26E096.hgt.zip',
		'N26E097.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N26E097.hgt.zip',
		'N26E098.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N26E098.hgt.zip',
		'N26E099.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N26E099.hgt.zip',
		'N26E100.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N26E100.hgt.zip',
		'N26E101.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N26E101.hgt.zip',
		'N26E102.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N26E102.hgt.zip',
		'N26E103.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N26E103.hgt.zip',
		'N26E104.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N26E104.hgt.zip',
		'N26E105.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N26E105.hgt.zip',
		'N26E106.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N26E106.hgt.zip',
		'N26E107.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N26E107.hgt.zip',
		'N26E108.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N26E108.hgt.zip',
		'N26E109.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N26E109.hgt.zip',
		'N26E110.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N26E110.hgt.zip',
		'N26E111.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N26E111.hgt.zip',
		'N26E112.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N26E112.hgt.zip',
		'N26E113.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N26E113.hgt.zip',
		'N26E114.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N26E114.hgt.zip',
		'N26E115.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N26E115.hgt.zip',
		'N26E116.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N26E116.hgt.zip',
		'N26E117.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N26E117.hgt.zip',
		'N26E118.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N26E118.hgt.zip',
		'N26E119.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N26E119.hgt.zip',
		'N26E120.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N26E120.hgt.zip',
		'N26E126.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N26E126.hgt.zip',
		'N26E127.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N26E127.hgt.zip',
		'N26E128.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N26E128.hgt.zip',
		'N26E142.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N26E142.hgt.zip',
		'N26W001.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N26W001.hgt.zip',
		'N26W002.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N26W002.hgt.zip',
		'N26W003.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N26W003.hgt.zip',
		'N26W004.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N26W004.hgt.zip',
		'N26W005.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N26W005.hgt.zip',
		'N26W006.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N26W006.hgt.zip',
		'N26W007.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N26W007.hgt.zip',
		'N26W008.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N26W008.hgt.zip',
		'N26W009.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N26W009.hgt.zip',
		'N26W010.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N26W010.hgt.zip',
		'N26W011.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N26W011.hgt.zip',
		'N26W012.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N26W012.hgt.zip',
		'N26W013.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N26W013.hgt.zip',
		'N26W014.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N26W014.hgt.zip',
		'N26W015.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N26W015.hgt.zip',
		'N26W077.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N26W077.hgt.zip',
		'N26W078.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N26W078.hgt.zip',
		'N26W079.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N26W079.hgt.zip',
		'N26W080.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N26W080.hgt.zip',
		'N26W081.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N26W081.hgt.zip',
		'N26W082.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N26W082.hgt.zip',
		'N26W083.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N26W083.hgt.zip',
		'N26W098.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N26W098.hgt.zip',
		'N26W099.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N26W099.hgt.zip',
		'N26W100.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N26W100.hgt.zip',
		'N26W101.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N26W101.hgt.zip',
		'N26W102.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N26W102.hgt.zip',
		'N26W103.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N26W103.hgt.zip',
		'N26W104.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N26W104.hgt.zip',
		'N26W105.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N26W105.hgt.zip',
		'N26W106.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N26W106.hgt.zip',
		'N26W107.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N26W107.hgt.zip',
		'N26W108.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N26W108.hgt.zip',
		'N26W109.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N26W109.hgt.zip',
		'N26W110.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N26W110.hgt.zip',
		'N26W112.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N26W112.hgt.zip',
		'N26W113.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N26W113.hgt.zip',
		'N26W114.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N26W114.hgt.zip',
		'N26W115.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N26W115.hgt.zip',
		'N26W174.hgt': 'https://srtm.kurviger.de/SRTM3//Islands//N26W174.hgt.zip',
		'N27E000.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N27E000.hgt.zip',
		'N27E001.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N27E001.hgt.zip',
		'N27E002.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N27E002.hgt.zip',
		'N27E003.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N27E003.hgt.zip',
		'N27E004.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N27E004.hgt.zip',
		'N27E005.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N27E005.hgt.zip',
		'N27E006.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N27E006.hgt.zip',
		'N27E007.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N27E007.hgt.zip',
		'N27E008.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N27E008.hgt.zip',
		'N27E009.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N27E009.hgt.zip',
		'N27E010.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N27E010.hgt.zip',
		'N27E011.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N27E011.hgt.zip',
		'N27E012.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N27E012.hgt.zip',
		'N27E013.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N27E013.hgt.zip',
		'N27E014.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N27E014.hgt.zip',
		'N27E015.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N27E015.hgt.zip',
		'N27E016.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N27E016.hgt.zip',
		'N27E017.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N27E017.hgt.zip',
		'N27E018.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N27E018.hgt.zip',
		'N27E019.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N27E019.hgt.zip',
		'N27E020.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N27E020.hgt.zip',
		'N27E021.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N27E021.hgt.zip',
		'N27E022.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N27E022.hgt.zip',
		'N27E023.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N27E023.hgt.zip',
		'N27E024.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N27E024.hgt.zip',
		'N27E025.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N27E025.hgt.zip',
		'N27E026.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N27E026.hgt.zip',
		'N27E027.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N27E027.hgt.zip',
		'N27E028.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N27E028.hgt.zip',
		'N27E029.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N27E029.hgt.zip',
		'N27E030.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N27E030.hgt.zip',
		'N27E031.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N27E031.hgt.zip',
		'N27E032.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N27E032.hgt.zip',
		'N27E033.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N27E033.hgt.zip',
		'N27E034.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N27E034.hgt.zip',
		'N27E035.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N27E035.hgt.zip',
		'N27E036.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N27E036.hgt.zip',
		'N27E037.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N27E037.hgt.zip',
		'N27E038.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N27E038.hgt.zip',
		'N27E039.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N27E039.hgt.zip',
		'N27E040.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N27E040.hgt.zip',
		'N27E041.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N27E041.hgt.zip',
		'N27E042.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N27E042.hgt.zip',
		'N27E043.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N27E043.hgt.zip',
		'N27E044.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N27E044.hgt.zip',
		'N27E045.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N27E045.hgt.zip',
		'N27E046.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N27E046.hgt.zip',
		'N27E047.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N27E047.hgt.zip',
		'N27E048.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N27E048.hgt.zip',
		'N27E049.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N27E049.hgt.zip',
		'N27E050.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N27E050.hgt.zip',
		'N27E051.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N27E051.hgt.zip',
		'N27E052.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N27E052.hgt.zip',
		'N27E053.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N27E053.hgt.zip',
		'N27E054.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N27E054.hgt.zip',
		'N27E055.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N27E055.hgt.zip',
		'N27E056.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N27E056.hgt.zip',
		'N27E057.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N27E057.hgt.zip',
		'N27E058.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N27E058.hgt.zip',
		'N27E059.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N27E059.hgt.zip',
		'N27E060.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N27E060.hgt.zip',
		'N27E061.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N27E061.hgt.zip',
		'N27E062.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N27E062.hgt.zip',
		'N27E063.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N27E063.hgt.zip',
		'N27E064.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N27E064.hgt.zip',
		'N27E065.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N27E065.hgt.zip',
		'N27E066.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N27E066.hgt.zip',
		'N27E067.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N27E067.hgt.zip',
		'N27E068.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N27E068.hgt.zip',
		'N27E069.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N27E069.hgt.zip',
		'N27E070.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N27E070.hgt.zip',
		'N27E071.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N27E071.hgt.zip',
		'N27E072.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N27E072.hgt.zip',
		'N27E073.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N27E073.hgt.zip',
		'N27E074.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N27E074.hgt.zip',
		'N27E075.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N27E075.hgt.zip',
		'N27E076.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N27E076.hgt.zip',
		'N27E077.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N27E077.hgt.zip',
		'N27E078.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N27E078.hgt.zip',
		'N27E079.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N27E079.hgt.zip',
		'N27E080.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N27E080.hgt.zip',
		'N27E081.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N27E081.hgt.zip',
		'N27E082.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N27E082.hgt.zip',
		'N27E083.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N27E083.hgt.zip',
		'N27E084.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N27E084.hgt.zip',
		'N27E085.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N27E085.hgt.zip',
		'N27E086.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N27E086.hgt.zip',
		'N27E087.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N27E087.hgt.zip',
		'N27E088.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N27E088.hgt.zip',
		'N27E089.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N27E089.hgt.zip',
		'N27E090.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N27E090.hgt.zip',
		'N27E091.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N27E091.hgt.zip',
		'N27E092.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N27E092.hgt.zip',
		'N27E093.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N27E093.hgt.zip',
		'N27E094.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N27E094.hgt.zip',
		'N27E095.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N27E095.hgt.zip',
		'N27E096.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N27E096.hgt.zip',
		'N27E097.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N27E097.hgt.zip',
		'N27E098.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N27E098.hgt.zip',
		'N27E099.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N27E099.hgt.zip',
		'N27E100.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N27E100.hgt.zip',
		'N27E101.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N27E101.hgt.zip',
		'N27E102.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N27E102.hgt.zip',
		'N27E103.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N27E103.hgt.zip',
		'N27E104.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N27E104.hgt.zip',
		'N27E105.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N27E105.hgt.zip',
		'N27E106.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N27E106.hgt.zip',
		'N27E107.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N27E107.hgt.zip',
		'N27E108.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N27E108.hgt.zip',
		'N27E109.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N27E109.hgt.zip',
		'N27E110.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N27E110.hgt.zip',
		'N27E111.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N27E111.hgt.zip',
		'N27E112.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N27E112.hgt.zip',
		'N27E113.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N27E113.hgt.zip',
		'N27E114.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N27E114.hgt.zip',
		'N27E115.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N27E115.hgt.zip',
		'N27E116.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N27E116.hgt.zip',
		'N27E117.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N27E117.hgt.zip',
		'N27E118.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N27E118.hgt.zip',
		'N27E119.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N27E119.hgt.zip',
		'N27E120.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N27E120.hgt.zip',
		'N27E121.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N27E121.hgt.zip',
		'N27E127.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N27E127.hgt.zip',
		'N27E128.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N27E128.hgt.zip',
		'N27E129.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N27E129.hgt.zip',
		'N27E140.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N27E140.hgt.zip',
		'N27E142.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N27E142.hgt.zip',
		'N27W001.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N27W001.hgt.zip',
		'N27W002.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N27W002.hgt.zip',
		'N27W003.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N27W003.hgt.zip',
		'N27W004.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N27W004.hgt.zip',
		'N27W005.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N27W005.hgt.zip',
		'N27W006.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N27W006.hgt.zip',
		'N27W007.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N27W007.hgt.zip',
		'N27W008.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N27W008.hgt.zip',
		'N27W009.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N27W009.hgt.zip',
		'N27W010.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N27W010.hgt.zip',
		'N27W011.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N27W011.hgt.zip',
		'N27W012.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N27W012.hgt.zip',
		'N27W013.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N27W013.hgt.zip',
		'N27W014.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N27W014.hgt.zip',
		'N27W016.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N27W016.hgt.zip',
		'N27W017.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N27W017.hgt.zip',
		'N27W018.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N27W018.hgt.zip',
		'N27W019.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N27W019.hgt.zip',
		'N27W078.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N27W078.hgt.zip',
		'N27W079.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N27W079.hgt.zip',
		'N27W081.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N27W081.hgt.zip',
		'N27W082.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N27W082.hgt.zip',
		'N27W083.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N27W083.hgt.zip',
		'N27W097.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N27W097.hgt.zip',
		'N27W098.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N27W098.hgt.zip',
		'N27W099.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N27W099.hgt.zip',
		'N27W100.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N27W100.hgt.zip',
		'N27W101.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N27W101.hgt.zip',
		'N27W102.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N27W102.hgt.zip',
		'N27W103.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N27W103.hgt.zip',
		'N27W104.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N27W104.hgt.zip',
		'N27W105.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N27W105.hgt.zip',
		'N27W106.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N27W106.hgt.zip',
		'N27W107.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N27W107.hgt.zip',
		'N27W108.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N27W108.hgt.zip',
		'N27W109.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N27W109.hgt.zip',
		'N27W110.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N27W110.hgt.zip',
		'N27W111.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N27W111.hgt.zip',
		'N27W112.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N27W112.hgt.zip',
		'N27W113.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N27W113.hgt.zip',
		'N27W114.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N27W114.hgt.zip',
		'N27W115.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N27W115.hgt.zip',
		'N27W116.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N27W116.hgt.zip',
		'N27W176.hgt': 'https://srtm.kurviger.de/SRTM3//Islands//N27W176.hgt.zip',
		'N28E000.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N28E000.hgt.zip',
		'N28E001.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N28E001.hgt.zip',
		'N28E002.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N28E002.hgt.zip',
		'N28E003.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N28E003.hgt.zip',
		'N28E004.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N28E004.hgt.zip',
		'N28E005.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N28E005.hgt.zip',
		'N28E006.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N28E006.hgt.zip',
		'N28E007.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N28E007.hgt.zip',
		'N28E008.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N28E008.hgt.zip',
		'N28E009.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N28E009.hgt.zip',
		'N28E010.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N28E010.hgt.zip',
		'N28E011.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N28E011.hgt.zip',
		'N28E012.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N28E012.hgt.zip',
		'N28E013.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N28E013.hgt.zip',
		'N28E014.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N28E014.hgt.zip',
		'N28E015.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N28E015.hgt.zip',
		'N28E016.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N28E016.hgt.zip',
		'N28E017.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N28E017.hgt.zip',
		'N28E018.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N28E018.hgt.zip',
		'N28E019.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N28E019.hgt.zip',
		'N28E020.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N28E020.hgt.zip',
		'N28E021.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N28E021.hgt.zip',
		'N28E022.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N28E022.hgt.zip',
		'N28E023.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N28E023.hgt.zip',
		'N28E024.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N28E024.hgt.zip',
		'N28E025.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N28E025.hgt.zip',
		'N28E026.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N28E026.hgt.zip',
		'N28E027.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N28E027.hgt.zip',
		'N28E028.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N28E028.hgt.zip',
		'N28E029.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N28E029.hgt.zip',
		'N28E030.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N28E030.hgt.zip',
		'N28E031.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N28E031.hgt.zip',
		'N28E032.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N28E032.hgt.zip',
		'N28E033.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N28E033.hgt.zip',
		'N28E034.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N28E034.hgt.zip',
		'N28E035.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N28E035.hgt.zip',
		'N28E036.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N28E036.hgt.zip',
		'N28E037.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N28E037.hgt.zip',
		'N28E038.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N28E038.hgt.zip',
		'N28E039.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N28E039.hgt.zip',
		'N28E040.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N28E040.hgt.zip',
		'N28E041.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N28E041.hgt.zip',
		'N28E042.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N28E042.hgt.zip',
		'N28E043.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N28E043.hgt.zip',
		'N28E044.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N28E044.hgt.zip',
		'N28E045.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N28E045.hgt.zip',
		'N28E046.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N28E046.hgt.zip',
		'N28E047.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N28E047.hgt.zip',
		'N28E048.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N28E048.hgt.zip',
		'N28E050.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N28E050.hgt.zip',
		'N28E051.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N28E051.hgt.zip',
		'N28E052.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N28E052.hgt.zip',
		'N28E053.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N28E053.hgt.zip',
		'N28E054.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N28E054.hgt.zip',
		'N28E055.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N28E055.hgt.zip',
		'N28E056.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N28E056.hgt.zip',
		'N28E057.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N28E057.hgt.zip',
		'N28E058.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N28E058.hgt.zip',
		'N28E059.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N28E059.hgt.zip',
		'N28E060.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N28E060.hgt.zip',
		'N28E061.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N28E061.hgt.zip',
		'N28E062.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N28E062.hgt.zip',
		'N28E063.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N28E063.hgt.zip',
		'N28E064.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N28E064.hgt.zip',
		'N28E065.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N28E065.hgt.zip',
		'N28E066.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N28E066.hgt.zip',
		'N28E067.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N28E067.hgt.zip',
		'N28E068.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N28E068.hgt.zip',
		'N28E069.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N28E069.hgt.zip',
		'N28E070.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N28E070.hgt.zip',
		'N28E071.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N28E071.hgt.zip',
		'N28E072.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N28E072.hgt.zip',
		'N28E073.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N28E073.hgt.zip',
		'N28E074.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N28E074.hgt.zip',
		'N28E075.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N28E075.hgt.zip',
		'N28E076.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N28E076.hgt.zip',
		'N28E077.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N28E077.hgt.zip',
		'N28E078.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N28E078.hgt.zip',
		'N28E079.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N28E079.hgt.zip',
		'N28E080.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N28E080.hgt.zip',
		'N28E081.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N28E081.hgt.zip',
		'N28E082.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N28E082.hgt.zip',
		'N28E083.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N28E083.hgt.zip',
		'N28E084.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N28E084.hgt.zip',
		'N28E085.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N28E085.hgt.zip',
		'N28E086.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N28E086.hgt.zip',
		'N28E087.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N28E087.hgt.zip',
		'N28E088.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N28E088.hgt.zip',
		'N28E089.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N28E089.hgt.zip',
		'N28E090.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N28E090.hgt.zip',
		'N28E091.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N28E091.hgt.zip',
		'N28E092.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N28E092.hgt.zip',
		'N28E093.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N28E093.hgt.zip',
		'N28E094.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N28E094.hgt.zip',
		'N28E095.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N28E095.hgt.zip',
		'N28E096.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N28E096.hgt.zip',
		'N28E097.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N28E097.hgt.zip',
		'N28E098.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N28E098.hgt.zip',
		'N28E099.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N28E099.hgt.zip',
		'N28E100.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N28E100.hgt.zip',
		'N28E101.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N28E101.hgt.zip',
		'N28E102.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N28E102.hgt.zip',
		'N28E103.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N28E103.hgt.zip',
		'N28E104.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N28E104.hgt.zip',
		'N28E105.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N28E105.hgt.zip',
		'N28E106.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N28E106.hgt.zip',
		'N28E107.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N28E107.hgt.zip',
		'N28E108.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N28E108.hgt.zip',
		'N28E109.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N28E109.hgt.zip',
		'N28E110.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N28E110.hgt.zip',
		'N28E111.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N28E111.hgt.zip',
		'N28E112.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N28E112.hgt.zip',
		'N28E113.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N28E113.hgt.zip',
		'N28E114.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N28E114.hgt.zip',
		'N28E115.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N28E115.hgt.zip',
		'N28E116.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N28E116.hgt.zip',
		'N28E117.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N28E117.hgt.zip',
		'N28E118.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N28E118.hgt.zip',
		'N28E119.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N28E119.hgt.zip',
		'N28E120.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N28E120.hgt.zip',
		'N28E121.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N28E121.hgt.zip',
		'N28E122.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N28E122.hgt.zip',
		'N28E128.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N28E128.hgt.zip',
		'N28E129.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N28E129.hgt.zip',
		'N28E130.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N28E130.hgt.zip',
		'N28W001.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N28W001.hgt.zip',
		'N28W002.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N28W002.hgt.zip',
		'N28W003.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N28W003.hgt.zip',
		'N28W004.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N28W004.hgt.zip',
		'N28W005.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N28W005.hgt.zip',
		'N28W006.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N28W006.hgt.zip',
		'N28W007.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N28W007.hgt.zip',
		'N28W008.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N28W008.hgt.zip',
		'N28W009.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N28W009.hgt.zip',
		'N28W010.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N28W010.hgt.zip',
		'N28W011.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N28W011.hgt.zip',
		'N28W012.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N28W012.hgt.zip',
		'N28W013.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N28W013.hgt.zip',
		'N28W014.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N28W014.hgt.zip',
		'N28W015.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N28W015.hgt.zip',
		'N28W016.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N28W016.hgt.zip',
		'N28W017.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N28W017.hgt.zip',
		'N28W018.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N28W018.hgt.zip',
		'N28W019.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N28W019.hgt.zip',
		'N28W081.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N28W081.hgt.zip',
		'N28W082.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N28W082.hgt.zip',
		'N28W083.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N28W083.hgt.zip',
		'N28W090.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N28W090.hgt.zip',
		'N28W096.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N28W096.hgt.zip',
		'N28W097.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N28W097.hgt.zip',
		'N28W098.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N28W098.hgt.zip',
		'N28W099.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N28W099.hgt.zip',
		'N28W100.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N28W100.hgt.zip',
		'N28W101.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N28W101.hgt.zip',
		'N28W102.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N28W102.hgt.zip',
		'N28W103.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N28W103.hgt.zip',
		'N28W104.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N28W104.hgt.zip',
		'N28W105.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N28W105.hgt.zip',
		'N28W106.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N28W106.hgt.zip',
		'N28W107.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N28W107.hgt.zip',
		'N28W108.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N28W108.hgt.zip',
		'N28W109.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N28W109.hgt.zip',
		'N28W110.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N28W110.hgt.zip',
		'N28W111.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N28W111.hgt.zip',
		'N28W112.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N28W112.hgt.zip',
		'N28W113.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N28W113.hgt.zip',
		'N28W114.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N28W114.hgt.zip',
		'N28W115.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N28W115.hgt.zip',
		'N28W116.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N28W116.hgt.zip',
		'N28W119.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N28W119.hgt.zip',
		'N28W178.hgt': 'https://srtm.kurviger.de/SRTM3//Islands//N28W178.hgt.zip',
		'N28W179.hgt': 'https://srtm.kurviger.de/SRTM3//Islands//N28W179.hgt.zip',
		'N29E000.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N29E000.hgt.zip',
		'N29E001.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N29E001.hgt.zip',
		'N29E002.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N29E002.hgt.zip',
		'N29E003.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N29E003.hgt.zip',
		'N29E004.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N29E004.hgt.zip',
		'N29E005.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N29E005.hgt.zip',
		'N29E006.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N29E006.hgt.zip',
		'N29E007.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N29E007.hgt.zip',
		'N29E008.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N29E008.hgt.zip',
		'N29E009.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N29E009.hgt.zip',
		'N29E010.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N29E010.hgt.zip',
		'N29E011.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N29E011.hgt.zip',
		'N29E012.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N29E012.hgt.zip',
		'N29E013.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N29E013.hgt.zip',
		'N29E014.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N29E014.hgt.zip',
		'N29E015.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N29E015.hgt.zip',
		'N29E016.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N29E016.hgt.zip',
		'N29E017.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N29E017.hgt.zip',
		'N29E018.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N29E018.hgt.zip',
		'N29E019.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N29E019.hgt.zip',
		'N29E020.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N29E020.hgt.zip',
		'N29E021.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N29E021.hgt.zip',
		'N29E022.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N29E022.hgt.zip',
		'N29E023.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N29E023.hgt.zip',
		'N29E024.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N29E024.hgt.zip',
		'N29E025.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N29E025.hgt.zip',
		'N29E026.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N29E026.hgt.zip',
		'N29E027.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N29E027.hgt.zip',
		'N29E028.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N29E028.hgt.zip',
		'N29E029.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N29E029.hgt.zip',
		'N29E030.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N29E030.hgt.zip',
		'N29E031.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N29E031.hgt.zip',
		'N29E032.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N29E032.hgt.zip',
		'N29E033.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N29E033.hgt.zip',
		'N29E034.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N29E034.hgt.zip',
		'N29E035.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N29E035.hgt.zip',
		'N29E036.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N29E036.hgt.zip',
		'N29E037.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N29E037.hgt.zip',
		'N29E038.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N29E038.hgt.zip',
		'N29E039.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N29E039.hgt.zip',
		'N29E040.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N29E040.hgt.zip',
		'N29E041.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N29E041.hgt.zip',
		'N29E042.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N29E042.hgt.zip',
		'N29E043.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N29E043.hgt.zip',
		'N29E044.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N29E044.hgt.zip',
		'N29E045.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N29E045.hgt.zip',
		'N29E046.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N29E046.hgt.zip',
		'N29E047.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N29E047.hgt.zip',
		'N29E048.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N29E048.hgt.zip',
		'N29E049.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N29E049.hgt.zip',
		'N29E050.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N29E050.hgt.zip',
		'N29E051.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N29E051.hgt.zip',
		'N29E052.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N29E052.hgt.zip',
		'N29E053.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N29E053.hgt.zip',
		'N29E054.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N29E054.hgt.zip',
		'N29E055.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N29E055.hgt.zip',
		'N29E056.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N29E056.hgt.zip',
		'N29E057.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N29E057.hgt.zip',
		'N29E058.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N29E058.hgt.zip',
		'N29E059.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N29E059.hgt.zip',
		'N29E060.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N29E060.hgt.zip',
		'N29E061.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N29E061.hgt.zip',
		'N29E062.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N29E062.hgt.zip',
		'N29E063.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N29E063.hgt.zip',
		'N29E064.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N29E064.hgt.zip',
		'N29E065.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N29E065.hgt.zip',
		'N29E066.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N29E066.hgt.zip',
		'N29E067.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N29E067.hgt.zip',
		'N29E068.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N29E068.hgt.zip',
		'N29E069.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N29E069.hgt.zip',
		'N29E070.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N29E070.hgt.zip',
		'N29E071.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N29E071.hgt.zip',
		'N29E072.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N29E072.hgt.zip',
		'N29E073.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N29E073.hgt.zip',
		'N29E074.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N29E074.hgt.zip',
		'N29E075.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N29E075.hgt.zip',
		'N29E076.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N29E076.hgt.zip',
		'N29E077.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N29E077.hgt.zip',
		'N29E078.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N29E078.hgt.zip',
		'N29E079.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N29E079.hgt.zip',
		'N29E080.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N29E080.hgt.zip',
		'N29E081.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N29E081.hgt.zip',
		'N29E082.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N29E082.hgt.zip',
		'N29E083.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N29E083.hgt.zip',
		'N29E084.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N29E084.hgt.zip',
		'N29E085.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N29E085.hgt.zip',
		'N29E086.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N29E086.hgt.zip',
		'N29E087.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N29E087.hgt.zip',
		'N29E088.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N29E088.hgt.zip',
		'N29E089.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N29E089.hgt.zip',
		'N29E090.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N29E090.hgt.zip',
		'N29E091.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N29E091.hgt.zip',
		'N29E092.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N29E092.hgt.zip',
		'N29E093.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N29E093.hgt.zip',
		'N29E094.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N29E094.hgt.zip',
		'N29E095.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N29E095.hgt.zip',
		'N29E096.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N29E096.hgt.zip',
		'N29E097.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N29E097.hgt.zip',
		'N29E098.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N29E098.hgt.zip',
		'N29E099.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N29E099.hgt.zip',
		'N29E100.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N29E100.hgt.zip',
		'N29E101.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N29E101.hgt.zip',
		'N29E102.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N29E102.hgt.zip',
		'N29E103.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N29E103.hgt.zip',
		'N29E104.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N29E104.hgt.zip',
		'N29E105.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N29E105.hgt.zip',
		'N29E106.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N29E106.hgt.zip',
		'N29E107.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N29E107.hgt.zip',
		'N29E108.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N29E108.hgt.zip',
		'N29E109.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N29E109.hgt.zip',
		'N29E110.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N29E110.hgt.zip',
		'N29E111.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N29E111.hgt.zip',
		'N29E112.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N29E112.hgt.zip',
		'N29E113.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N29E113.hgt.zip',
		'N29E114.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N29E114.hgt.zip',
		'N29E115.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N29E115.hgt.zip',
		'N29E116.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N29E116.hgt.zip',
		'N29E117.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N29E117.hgt.zip',
		'N29E118.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N29E118.hgt.zip',
		'N29E119.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N29E119.hgt.zip',
		'N29E120.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N29E120.hgt.zip',
		'N29E121.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N29E121.hgt.zip',
		'N29E122.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N29E122.hgt.zip',
		'N29E129.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N29E129.hgt.zip',
		'N29E140.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N29E140.hgt.zip',
		'N29W001.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N29W001.hgt.zip',
		'N29W002.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N29W002.hgt.zip',
		'N29W003.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N29W003.hgt.zip',
		'N29W004.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N29W004.hgt.zip',
		'N29W005.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N29W005.hgt.zip',
		'N29W006.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N29W006.hgt.zip',
		'N29W007.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N29W007.hgt.zip',
		'N29W008.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N29W008.hgt.zip',
		'N29W009.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N29W009.hgt.zip',
		'N29W010.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N29W010.hgt.zip',
		'N29W011.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N29W011.hgt.zip',
		'N29W014.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N29W014.hgt.zip',
		'N29W081.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N29W081.hgt.zip',
		'N29W082.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N29W082.hgt.zip',
		'N29W083.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N29W083.hgt.zip',
		'N29W084.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N29W084.hgt.zip',
		'N29W085.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N29W085.hgt.zip',
		'N29W086.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N29W086.hgt.zip',
		'N29W089.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N29W089.hgt.zip',
		'N29W090.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N29W090.hgt.zip',
		'N29W091.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N29W091.hgt.zip',
		'N29W092.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N29W092.hgt.zip',
		'N29W093.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N29W093.hgt.zip',
		'N29W094.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N29W094.hgt.zip',
		'N29W095.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N29W095.hgt.zip',
		'N29W096.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N29W096.hgt.zip',
		'N29W097.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N29W097.hgt.zip',
		'N29W098.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N29W098.hgt.zip',
		'N29W099.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N29W099.hgt.zip',
		'N29W100.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N29W100.hgt.zip',
		'N29W101.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N29W101.hgt.zip',
		'N29W102.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N29W102.hgt.zip',
		'N29W103.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N29W103.hgt.zip',
		'N29W104.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N29W104.hgt.zip',
		'N29W105.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N29W105.hgt.zip',
		'N29W106.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N29W106.hgt.zip',
		'N29W107.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N29W107.hgt.zip',
		'N29W108.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N29W108.hgt.zip',
		'N29W109.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N29W109.hgt.zip',
		'N29W110.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N29W110.hgt.zip',
		'N29W111.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N29W111.hgt.zip',
		'N29W112.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N29W112.hgt.zip',
		'N29W113.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N29W113.hgt.zip',
		'N29W114.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N29W114.hgt.zip',
		'N29W115.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N29W115.hgt.zip',
		'N29W116.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N29W116.hgt.zip',
		'N29W119.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N29W119.hgt.zip',
		'N30E000.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N30E000.hgt.zip',
		'N30E001.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N30E001.hgt.zip',
		'N30E002.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N30E002.hgt.zip',
		'N30E003.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N30E003.hgt.zip',
		'N30E004.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N30E004.hgt.zip',
		'N30E005.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N30E005.hgt.zip',
		'N30E006.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N30E006.hgt.zip',
		'N30E007.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N30E007.hgt.zip',
		'N30E008.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N30E008.hgt.zip',
		'N30E009.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N30E009.hgt.zip',
		'N30E010.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N30E010.hgt.zip',
		'N30E011.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N30E011.hgt.zip',
		'N30E012.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N30E012.hgt.zip',
		'N30E013.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N30E013.hgt.zip',
		'N30E014.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N30E014.hgt.zip',
		'N30E015.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N30E015.hgt.zip',
		'N30E016.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N30E016.hgt.zip',
		'N30E017.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N30E017.hgt.zip',
		'N30E018.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N30E018.hgt.zip',
		'N30E019.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N30E019.hgt.zip',
		'N30E020.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N30E020.hgt.zip',
		'N30E021.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N30E021.hgt.zip',
		'N30E022.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N30E022.hgt.zip',
		'N30E023.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N30E023.hgt.zip',
		'N30E024.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N30E024.hgt.zip',
		'N30E025.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N30E025.hgt.zip',
		'N30E026.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N30E026.hgt.zip',
		'N30E027.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N30E027.hgt.zip',
		'N30E028.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N30E028.hgt.zip',
		'N30E029.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N30E029.hgt.zip',
		'N30E030.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N30E030.hgt.zip',
		'N30E031.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N30E031.hgt.zip',
		'N30E032.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N30E032.hgt.zip',
		'N30E033.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N30E033.hgt.zip',
		'N30E034.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N30E034.hgt.zip',
		'N30E035.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N30E035.hgt.zip',
		'N30E036.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N30E036.hgt.zip',
		'N30E037.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N30E037.hgt.zip',
		'N30E038.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N30E038.hgt.zip',
		'N30E039.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N30E039.hgt.zip',
		'N30E040.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N30E040.hgt.zip',
		'N30E041.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N30E041.hgt.zip',
		'N30E042.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N30E042.hgt.zip',
		'N30E043.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N30E043.hgt.zip',
		'N30E044.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N30E044.hgt.zip',
		'N30E045.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N30E045.hgt.zip',
		'N30E046.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N30E046.hgt.zip',
		'N30E047.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N30E047.hgt.zip',
		'N30E048.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N30E048.hgt.zip',
		'N30E049.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N30E049.hgt.zip',
		'N30E050.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N30E050.hgt.zip',
		'N30E051.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N30E051.hgt.zip',
		'N30E052.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N30E052.hgt.zip',
		'N30E053.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N30E053.hgt.zip',
		'N30E054.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N30E054.hgt.zip',
		'N30E055.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N30E055.hgt.zip',
		'N30E056.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N30E056.hgt.zip',
		'N30E057.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N30E057.hgt.zip',
		'N30E058.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N30E058.hgt.zip',
		'N30E059.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N30E059.hgt.zip',
		'N30E060.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N30E060.hgt.zip',
		'N30E061.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N30E061.hgt.zip',
		'N30E062.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N30E062.hgt.zip',
		'N30E063.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N30E063.hgt.zip',
		'N30E064.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N30E064.hgt.zip',
		'N30E065.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N30E065.hgt.zip',
		'N30E066.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N30E066.hgt.zip',
		'N30E067.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N30E067.hgt.zip',
		'N30E068.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N30E068.hgt.zip',
		'N30E069.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N30E069.hgt.zip',
		'N30E070.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N30E070.hgt.zip',
		'N30E071.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N30E071.hgt.zip',
		'N30E072.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N30E072.hgt.zip',
		'N30E073.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N30E073.hgt.zip',
		'N30E074.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N30E074.hgt.zip',
		'N30E075.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N30E075.hgt.zip',
		'N30E076.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N30E076.hgt.zip',
		'N30E077.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N30E077.hgt.zip',
		'N30E078.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N30E078.hgt.zip',
		'N30E079.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N30E079.hgt.zip',
		'N30E080.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N30E080.hgt.zip',
		'N30E081.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N30E081.hgt.zip',
		'N30E082.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N30E082.hgt.zip',
		'N30E083.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N30E083.hgt.zip',
		'N30E084.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N30E084.hgt.zip',
		'N30E085.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N30E085.hgt.zip',
		'N30E086.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N30E086.hgt.zip',
		'N30E087.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N30E087.hgt.zip',
		'N30E088.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N30E088.hgt.zip',
		'N30E089.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N30E089.hgt.zip',
		'N30E090.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N30E090.hgt.zip',
		'N30E091.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N30E091.hgt.zip',
		'N30E092.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N30E092.hgt.zip',
		'N30E093.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N30E093.hgt.zip',
		'N30E094.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N30E094.hgt.zip',
		'N30E095.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N30E095.hgt.zip',
		'N30E096.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N30E096.hgt.zip',
		'N30E097.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N30E097.hgt.zip',
		'N30E098.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N30E098.hgt.zip',
		'N30E099.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N30E099.hgt.zip',
		'N30E100.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N30E100.hgt.zip',
		'N30E101.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N30E101.hgt.zip',
		'N30E102.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N30E102.hgt.zip',
		'N30E103.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N30E103.hgt.zip',
		'N30E104.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N30E104.hgt.zip',
		'N30E105.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N30E105.hgt.zip',
		'N30E106.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N30E106.hgt.zip',
		'N30E107.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N30E107.hgt.zip',
		'N30E108.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N30E108.hgt.zip',
		'N30E109.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N30E109.hgt.zip',
		'N30E110.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N30E110.hgt.zip',
		'N30E111.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N30E111.hgt.zip',
		'N30E112.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N30E112.hgt.zip',
		'N30E113.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N30E113.hgt.zip',
		'N30E114.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N30E114.hgt.zip',
		'N30E115.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N30E115.hgt.zip',
		'N30E116.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N30E116.hgt.zip',
		'N30E117.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N30E117.hgt.zip',
		'N30E118.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N30E118.hgt.zip',
		'N30E119.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N30E119.hgt.zip',
		'N30E120.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N30E120.hgt.zip',
		'N30E121.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N30E121.hgt.zip',
		'N30E122.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N30E122.hgt.zip',
		'N30E129.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N30E129.hgt.zip',
		'N30E130.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N30E130.hgt.zip',
		'N30E131.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N30E131.hgt.zip',
		'N30E140.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N30E140.hgt.zip',
		'N30W001.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N30W001.hgt.zip',
		'N30W002.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N30W002.hgt.zip',
		'N30W003.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N30W003.hgt.zip',
		'N30W004.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N30W004.hgt.zip',
		'N30W005.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N30W005.hgt.zip',
		'N30W006.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N30W006.hgt.zip',
		'N30W007.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N30W007.hgt.zip',
		'N30W008.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N30W008.hgt.zip',
		'N30W009.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N30W009.hgt.zip',
		'N30W010.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N30W010.hgt.zip',
		'N30W016.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N30W016.hgt.zip',
		'N30W017.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N30W017.hgt.zip',
		'N30W082.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N30W082.hgt.zip',
		'N30W083.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N30W083.hgt.zip',
		'N30W084.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N30W084.hgt.zip',
		'N30W085.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N30W085.hgt.zip',
		'N30W086.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N30W086.hgt.zip',
		'N30W087.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N30W087.hgt.zip',
		'N30W088.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N30W088.hgt.zip',
		'N30W089.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N30W089.hgt.zip',
		'N30W090.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N30W090.hgt.zip',
		'N30W091.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N30W091.hgt.zip',
		'N30W092.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N30W092.hgt.zip',
		'N30W093.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N30W093.hgt.zip',
		'N30W094.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N30W094.hgt.zip',
		'N30W095.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N30W095.hgt.zip',
		'N30W096.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N30W096.hgt.zip',
		'N30W097.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N30W097.hgt.zip',
		'N30W098.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N30W098.hgt.zip',
		'N30W099.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N30W099.hgt.zip',
		'N30W100.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N30W100.hgt.zip',
		'N30W101.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N30W101.hgt.zip',
		'N30W102.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N30W102.hgt.zip',
		'N30W103.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N30W103.hgt.zip',
		'N30W104.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N30W104.hgt.zip',
		'N30W105.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N30W105.hgt.zip',
		'N30W106.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N30W106.hgt.zip',
		'N30W107.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N30W107.hgt.zip',
		'N30W108.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N30W108.hgt.zip',
		'N30W109.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N30W109.hgt.zip',
		'N30W110.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N30W110.hgt.zip',
		'N30W111.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N30W111.hgt.zip',
		'N30W112.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N30W112.hgt.zip',
		'N30W113.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N30W113.hgt.zip',
		'N30W114.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N30W114.hgt.zip',
		'N30W115.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N30W115.hgt.zip',
		'N30W116.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N30W116.hgt.zip',
		'N30W117.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N30W117.hgt.zip',
		'N31E000.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N31E000.hgt.zip',
		'N31E001.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N31E001.hgt.zip',
		'N31E002.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N31E002.hgt.zip',
		'N31E003.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N31E003.hgt.zip',
		'N31E004.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N31E004.hgt.zip',
		'N31E005.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N31E005.hgt.zip',
		'N31E006.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N31E006.hgt.zip',
		'N31E007.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N31E007.hgt.zip',
		'N31E008.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N31E008.hgt.zip',
		'N31E009.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N31E009.hgt.zip',
		'N31E010.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N31E010.hgt.zip',
		'N31E011.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N31E011.hgt.zip',
		'N31E012.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N31E012.hgt.zip',
		'N31E013.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N31E013.hgt.zip',
		'N31E014.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N31E014.hgt.zip',
		'N31E015.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N31E015.hgt.zip',
		'N31E016.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N31E016.hgt.zip',
		'N31E017.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N31E017.hgt.zip',
		'N31E019.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N31E019.hgt.zip',
		'N31E020.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N31E020.hgt.zip',
		'N31E021.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N31E021.hgt.zip',
		'N31E022.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N31E022.hgt.zip',
		'N31E023.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N31E023.hgt.zip',
		'N31E024.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N31E024.hgt.zip',
		'N31E025.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N31E025.hgt.zip',
		'N31E026.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N31E026.hgt.zip',
		'N31E027.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N31E027.hgt.zip',
		'N31E028.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N31E028.hgt.zip',
		'N31E029.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N31E029.hgt.zip',
		'N31E030.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N31E030.hgt.zip',
		'N31E031.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N31E031.hgt.zip',
		'N31E032.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N31E032.hgt.zip',
		'N31E033.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N31E033.hgt.zip',
		'N31E034.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N31E034.hgt.zip',
		'N31E035.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N31E035.hgt.zip',
		'N31E036.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N31E036.hgt.zip',
		'N31E037.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N31E037.hgt.zip',
		'N31E038.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N31E038.hgt.zip',
		'N31E039.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N31E039.hgt.zip',
		'N31E040.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N31E040.hgt.zip',
		'N31E041.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N31E041.hgt.zip',
		'N31E042.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N31E042.hgt.zip',
		'N31E043.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N31E043.hgt.zip',
		'N31E044.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N31E044.hgt.zip',
		'N31E045.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N31E045.hgt.zip',
		'N31E046.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N31E046.hgt.zip',
		'N31E047.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N31E047.hgt.zip',
		'N31E048.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N31E048.hgt.zip',
		'N31E049.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N31E049.hgt.zip',
		'N31E050.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N31E050.hgt.zip',
		'N31E051.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N31E051.hgt.zip',
		'N31E052.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N31E052.hgt.zip',
		'N31E053.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N31E053.hgt.zip',
		'N31E054.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N31E054.hgt.zip',
		'N31E055.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N31E055.hgt.zip',
		'N31E056.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N31E056.hgt.zip',
		'N31E057.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N31E057.hgt.zip',
		'N31E058.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N31E058.hgt.zip',
		'N31E059.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N31E059.hgt.zip',
		'N31E060.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N31E060.hgt.zip',
		'N31E061.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N31E061.hgt.zip',
		'N31E062.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N31E062.hgt.zip',
		'N31E063.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N31E063.hgt.zip',
		'N31E064.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N31E064.hgt.zip',
		'N31E065.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N31E065.hgt.zip',
		'N31E066.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N31E066.hgt.zip',
		'N31E067.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N31E067.hgt.zip',
		'N31E068.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N31E068.hgt.zip',
		'N31E069.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N31E069.hgt.zip',
		'N31E070.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N31E070.hgt.zip',
		'N31E071.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N31E071.hgt.zip',
		'N31E072.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N31E072.hgt.zip',
		'N31E073.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N31E073.hgt.zip',
		'N31E074.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N31E074.hgt.zip',
		'N31E075.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N31E075.hgt.zip',
		'N31E076.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N31E076.hgt.zip',
		'N31E077.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N31E077.hgt.zip',
		'N31E078.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N31E078.hgt.zip',
		'N31E079.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N31E079.hgt.zip',
		'N31E080.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N31E080.hgt.zip',
		'N31E081.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N31E081.hgt.zip',
		'N31E082.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N31E082.hgt.zip',
		'N31E083.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N31E083.hgt.zip',
		'N31E084.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N31E084.hgt.zip',
		'N31E085.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N31E085.hgt.zip',
		'N31E086.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N31E086.hgt.zip',
		'N31E087.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N31E087.hgt.zip',
		'N31E088.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N31E088.hgt.zip',
		'N31E089.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N31E089.hgt.zip',
		'N31E090.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N31E090.hgt.zip',
		'N31E091.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N31E091.hgt.zip',
		'N31E092.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N31E092.hgt.zip',
		'N31E093.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N31E093.hgt.zip',
		'N31E094.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N31E094.hgt.zip',
		'N31E095.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N31E095.hgt.zip',
		'N31E096.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N31E096.hgt.zip',
		'N31E097.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N31E097.hgt.zip',
		'N31E098.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N31E098.hgt.zip',
		'N31E099.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N31E099.hgt.zip',
		'N31E100.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N31E100.hgt.zip',
		'N31E101.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N31E101.hgt.zip',
		'N31E102.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N31E102.hgt.zip',
		'N31E103.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N31E103.hgt.zip',
		'N31E104.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N31E104.hgt.zip',
		'N31E105.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N31E105.hgt.zip',
		'N31E106.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N31E106.hgt.zip',
		'N31E107.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N31E107.hgt.zip',
		'N31E108.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N31E108.hgt.zip',
		'N31E109.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N31E109.hgt.zip',
		'N31E110.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N31E110.hgt.zip',
		'N31E111.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N31E111.hgt.zip',
		'N31E112.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N31E112.hgt.zip',
		'N31E113.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N31E113.hgt.zip',
		'N31E114.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N31E114.hgt.zip',
		'N31E115.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N31E115.hgt.zip',
		'N31E116.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N31E116.hgt.zip',
		'N31E117.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N31E117.hgt.zip',
		'N31E118.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N31E118.hgt.zip',
		'N31E119.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N31E119.hgt.zip',
		'N31E120.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N31E120.hgt.zip',
		'N31E121.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N31E121.hgt.zip',
		'N31E122.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N31E122.hgt.zip',
		'N31E128.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N31E128.hgt.zip',
		'N31E129.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N31E129.hgt.zip',
		'N31E130.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N31E130.hgt.zip',
		'N31E131.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N31E131.hgt.zip',
		'N31E139.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N31E139.hgt.zip',
		'N31E140.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N31E140.hgt.zip',
		'N31W001.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N31W001.hgt.zip',
		'N31W002.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N31W002.hgt.zip',
		'N31W003.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N31W003.hgt.zip',
		'N31W004.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N31W004.hgt.zip',
		'N31W005.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N31W005.hgt.zip',
		'N31W006.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N31W006.hgt.zip',
		'N31W007.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N31W007.hgt.zip',
		'N31W008.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N31W008.hgt.zip',
		'N31W009.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N31W009.hgt.zip',
		'N31W010.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N31W010.hgt.zip',
		'N31W081.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N31W081.hgt.zip',
		'N31W082.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N31W082.hgt.zip',
		'N31W083.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N31W083.hgt.zip',
		'N31W084.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N31W084.hgt.zip',
		'N31W085.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N31W085.hgt.zip',
		'N31W086.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N31W086.hgt.zip',
		'N31W087.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N31W087.hgt.zip',
		'N31W088.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N31W088.hgt.zip',
		'N31W089.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N31W089.hgt.zip',
		'N31W090.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N31W090.hgt.zip',
		'N31W091.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N31W091.hgt.zip',
		'N31W092.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N31W092.hgt.zip',
		'N31W093.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N31W093.hgt.zip',
		'N31W094.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N31W094.hgt.zip',
		'N31W095.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N31W095.hgt.zip',
		'N31W096.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N31W096.hgt.zip',
		'N31W097.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N31W097.hgt.zip',
		'N31W098.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N31W098.hgt.zip',
		'N31W099.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N31W099.hgt.zip',
		'N31W100.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N31W100.hgt.zip',
		'N31W101.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N31W101.hgt.zip',
		'N31W102.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N31W102.hgt.zip',
		'N31W103.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N31W103.hgt.zip',
		'N31W104.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N31W104.hgt.zip',
		'N31W105.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N31W105.hgt.zip',
		'N31W106.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N31W106.hgt.zip',
		'N31W107.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N31W107.hgt.zip',
		'N31W108.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N31W108.hgt.zip',
		'N31W109.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N31W109.hgt.zip',
		'N31W110.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N31W110.hgt.zip',
		'N31W111.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N31W111.hgt.zip',
		'N31W112.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N31W112.hgt.zip',
		'N31W113.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N31W113.hgt.zip',
		'N31W114.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N31W114.hgt.zip',
		'N31W115.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N31W115.hgt.zip',
		'N31W116.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N31W116.hgt.zip',
		'N31W117.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N31W117.hgt.zip',
		'N32E000.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N32E000.hgt.zip',
		'N32E001.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N32E001.hgt.zip',
		'N32E002.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N32E002.hgt.zip',
		'N32E003.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N32E003.hgt.zip',
		'N32E004.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N32E004.hgt.zip',
		'N32E005.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N32E005.hgt.zip',
		'N32E006.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N32E006.hgt.zip',
		'N32E007.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N32E007.hgt.zip',
		'N32E008.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N32E008.hgt.zip',
		'N32E009.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N32E009.hgt.zip',
		'N32E010.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N32E010.hgt.zip',
		'N32E011.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N32E011.hgt.zip',
		'N32E012.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N32E012.hgt.zip',
		'N32E013.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N32E013.hgt.zip',
		'N32E014.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N32E014.hgt.zip',
		'N32E015.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N32E015.hgt.zip',
		'N32E019.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N32E019.hgt.zip',
		'N32E020.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N32E020.hgt.zip',
		'N32E021.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N32E021.hgt.zip',
		'N32E022.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N32E022.hgt.zip',
		'N32E023.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N32E023.hgt.zip',
		'N32E024.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N32E024.hgt.zip',
		'N32E034.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N32E034.hgt.zip',
		'N32E035.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N32E035.hgt.zip',
		'N32E036.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N32E036.hgt.zip',
		'N32E037.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N32E037.hgt.zip',
		'N32E038.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N32E038.hgt.zip',
		'N32E039.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N32E039.hgt.zip',
		'N32E040.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N32E040.hgt.zip',
		'N32E041.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N32E041.hgt.zip',
		'N32E042.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N32E042.hgt.zip',
		'N32E043.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N32E043.hgt.zip',
		'N32E044.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N32E044.hgt.zip',
		'N32E045.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N32E045.hgt.zip',
		'N32E046.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N32E046.hgt.zip',
		'N32E047.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N32E047.hgt.zip',
		'N32E048.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N32E048.hgt.zip',
		'N32E049.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N32E049.hgt.zip',
		'N32E050.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N32E050.hgt.zip',
		'N32E051.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N32E051.hgt.zip',
		'N32E052.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N32E052.hgt.zip',
		'N32E053.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N32E053.hgt.zip',
		'N32E054.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N32E054.hgt.zip',
		'N32E055.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N32E055.hgt.zip',
		'N32E056.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N32E056.hgt.zip',
		'N32E057.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N32E057.hgt.zip',
		'N32E058.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N32E058.hgt.zip',
		'N32E059.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N32E059.hgt.zip',
		'N32E060.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N32E060.hgt.zip',
		'N32E061.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N32E061.hgt.zip',
		'N32E062.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N32E062.hgt.zip',
		'N32E063.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N32E063.hgt.zip',
		'N32E064.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N32E064.hgt.zip',
		'N32E065.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N32E065.hgt.zip',
		'N32E066.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N32E066.hgt.zip',
		'N32E067.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N32E067.hgt.zip',
		'N32E068.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N32E068.hgt.zip',
		'N32E069.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N32E069.hgt.zip',
		'N32E070.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N32E070.hgt.zip',
		'N32E071.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N32E071.hgt.zip',
		'N32E072.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N32E072.hgt.zip',
		'N32E073.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N32E073.hgt.zip',
		'N32E074.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N32E074.hgt.zip',
		'N32E075.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N32E075.hgt.zip',
		'N32E076.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N32E076.hgt.zip',
		'N32E077.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N32E077.hgt.zip',
		'N32E078.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N32E078.hgt.zip',
		'N32E079.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N32E079.hgt.zip',
		'N32E080.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N32E080.hgt.zip',
		'N32E081.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N32E081.hgt.zip',
		'N32E082.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N32E082.hgt.zip',
		'N32E083.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N32E083.hgt.zip',
		'N32E084.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N32E084.hgt.zip',
		'N32E085.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N32E085.hgt.zip',
		'N32E086.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N32E086.hgt.zip',
		'N32E087.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N32E087.hgt.zip',
		'N32E088.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N32E088.hgt.zip',
		'N32E089.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N32E089.hgt.zip',
		'N32E090.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N32E090.hgt.zip',
		'N32E091.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N32E091.hgt.zip',
		'N32E092.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N32E092.hgt.zip',
		'N32E093.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N32E093.hgt.zip',
		'N32E094.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N32E094.hgt.zip',
		'N32E095.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N32E095.hgt.zip',
		'N32E096.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N32E096.hgt.zip',
		'N32E097.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N32E097.hgt.zip',
		'N32E098.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N32E098.hgt.zip',
		'N32E099.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N32E099.hgt.zip',
		'N32E100.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N32E100.hgt.zip',
		'N32E101.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N32E101.hgt.zip',
		'N32E102.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N32E102.hgt.zip',
		'N32E103.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N32E103.hgt.zip',
		'N32E104.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N32E104.hgt.zip',
		'N32E105.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N32E105.hgt.zip',
		'N32E106.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N32E106.hgt.zip',
		'N32E107.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N32E107.hgt.zip',
		'N32E108.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N32E108.hgt.zip',
		'N32E109.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N32E109.hgt.zip',
		'N32E110.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N32E110.hgt.zip',
		'N32E111.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N32E111.hgt.zip',
		'N32E112.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N32E112.hgt.zip',
		'N32E113.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N32E113.hgt.zip',
		'N32E114.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N32E114.hgt.zip',
		'N32E115.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N32E115.hgt.zip',
		'N32E116.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N32E116.hgt.zip',
		'N32E117.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N32E117.hgt.zip',
		'N32E118.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N32E118.hgt.zip',
		'N32E119.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N32E119.hgt.zip',
		'N32E120.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N32E120.hgt.zip',
		'N32E121.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N32E121.hgt.zip',
		'N32E128.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N32E128.hgt.zip',
		'N32E129.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N32E129.hgt.zip',
		'N32E130.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N32E130.hgt.zip',
		'N32E131.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N32E131.hgt.zip',
		'N32E132.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N32E132.hgt.zip',
		'N32E133.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N32E133.hgt.zip',
		'N32E139.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N32E139.hgt.zip',
		'N32W001.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N32W001.hgt.zip',
		'N32W002.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N32W002.hgt.zip',
		'N32W003.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N32W003.hgt.zip',
		'N32W004.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N32W004.hgt.zip',
		'N32W005.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N32W005.hgt.zip',
		'N32W006.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N32W006.hgt.zip',
		'N32W007.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N32W007.hgt.zip',
		'N32W008.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N32W008.hgt.zip',
		'N32W009.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N32W009.hgt.zip',
		'N32W010.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N32W010.hgt.zip',
		'N32W017.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N32W017.hgt.zip',
		'N32W018.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N32W018.hgt.zip',
		'N32W065.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N32W065.hgt.zip',
		'N32W080.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N32W080.hgt.zip',
		'N32W081.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N32W081.hgt.zip',
		'N32W082.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N32W082.hgt.zip',
		'N32W083.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N32W083.hgt.zip',
		'N32W084.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N32W084.hgt.zip',
		'N32W085.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N32W085.hgt.zip',
		'N32W086.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N32W086.hgt.zip',
		'N32W087.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N32W087.hgt.zip',
		'N32W088.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N32W088.hgt.zip',
		'N32W089.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N32W089.hgt.zip',
		'N32W090.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N32W090.hgt.zip',
		'N32W091.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N32W091.hgt.zip',
		'N32W092.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N32W092.hgt.zip',
		'N32W093.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N32W093.hgt.zip',
		'N32W094.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N32W094.hgt.zip',
		'N32W095.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N32W095.hgt.zip',
		'N32W096.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N32W096.hgt.zip',
		'N32W097.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N32W097.hgt.zip',
		'N32W098.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N32W098.hgt.zip',
		'N32W099.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N32W099.hgt.zip',
		'N32W100.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N32W100.hgt.zip',
		'N32W101.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N32W101.hgt.zip',
		'N32W102.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N32W102.hgt.zip',
		'N32W103.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N32W103.hgt.zip',
		'N32W104.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N32W104.hgt.zip',
		'N32W105.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N32W105.hgt.zip',
		'N32W106.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N32W106.hgt.zip',
		'N32W107.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N32W107.hgt.zip',
		'N32W108.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N32W108.hgt.zip',
		'N32W109.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N32W109.hgt.zip',
		'N32W110.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N32W110.hgt.zip',
		'N32W111.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N32W111.hgt.zip',
		'N32W112.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N32W112.hgt.zip',
		'N32W113.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N32W113.hgt.zip',
		'N32W114.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N32W114.hgt.zip',
		'N32W115.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N32W115.hgt.zip',
		'N32W116.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N32W116.hgt.zip',
		'N32W117.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N32W117.hgt.zip',
		'N32W118.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N32W118.hgt.zip',
		'N32W119.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N32W119.hgt.zip',
		'N33E000.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N33E000.hgt.zip',
		'N33E001.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N33E001.hgt.zip',
		'N33E002.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N33E002.hgt.zip',
		'N33E003.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N33E003.hgt.zip',
		'N33E004.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N33E004.hgt.zip',
		'N33E005.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N33E005.hgt.zip',
		'N33E006.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N33E006.hgt.zip',
		'N33E007.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N33E007.hgt.zip',
		'N33E008.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N33E008.hgt.zip',
		'N33E009.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N33E009.hgt.zip',
		'N33E010.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N33E010.hgt.zip',
		'N33E011.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N33E011.hgt.zip',
		'N33E035.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N33E035.hgt.zip',
		'N33E036.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N33E036.hgt.zip',
		'N33E037.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N33E037.hgt.zip',
		'N33E038.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N33E038.hgt.zip',
		'N33E039.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N33E039.hgt.zip',
		'N33E040.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N33E040.hgt.zip',
		'N33E041.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N33E041.hgt.zip',
		'N33E042.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N33E042.hgt.zip',
		'N33E043.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N33E043.hgt.zip',
		'N33E044.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N33E044.hgt.zip',
		'N33E045.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N33E045.hgt.zip',
		'N33E046.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N33E046.hgt.zip',
		'N33E047.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N33E047.hgt.zip',
		'N33E048.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N33E048.hgt.zip',
		'N33E049.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N33E049.hgt.zip',
		'N33E050.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N33E050.hgt.zip',
		'N33E051.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N33E051.hgt.zip',
		'N33E052.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N33E052.hgt.zip',
		'N33E053.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N33E053.hgt.zip',
		'N33E054.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N33E054.hgt.zip',
		'N33E055.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N33E055.hgt.zip',
		'N33E056.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N33E056.hgt.zip',
		'N33E057.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N33E057.hgt.zip',
		'N33E058.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N33E058.hgt.zip',
		'N33E059.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N33E059.hgt.zip',
		'N33E060.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N33E060.hgt.zip',
		'N33E061.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N33E061.hgt.zip',
		'N33E062.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N33E062.hgt.zip',
		'N33E063.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N33E063.hgt.zip',
		'N33E064.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N33E064.hgt.zip',
		'N33E065.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N33E065.hgt.zip',
		'N33E066.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N33E066.hgt.zip',
		'N33E067.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N33E067.hgt.zip',
		'N33E068.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N33E068.hgt.zip',
		'N33E069.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N33E069.hgt.zip',
		'N33E070.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N33E070.hgt.zip',
		'N33E071.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N33E071.hgt.zip',
		'N33E072.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N33E072.hgt.zip',
		'N33E073.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N33E073.hgt.zip',
		'N33E074.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N33E074.hgt.zip',
		'N33E075.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N33E075.hgt.zip',
		'N33E076.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N33E076.hgt.zip',
		'N33E077.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N33E077.hgt.zip',
		'N33E078.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N33E078.hgt.zip',
		'N33E079.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N33E079.hgt.zip',
		'N33E080.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N33E080.hgt.zip',
		'N33E081.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N33E081.hgt.zip',
		'N33E082.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N33E082.hgt.zip',
		'N33E083.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N33E083.hgt.zip',
		'N33E084.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N33E084.hgt.zip',
		'N33E085.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N33E085.hgt.zip',
		'N33E086.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N33E086.hgt.zip',
		'N33E087.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N33E087.hgt.zip',
		'N33E088.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N33E088.hgt.zip',
		'N33E089.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N33E089.hgt.zip',
		'N33E090.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N33E090.hgt.zip',
		'N33E091.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N33E091.hgt.zip',
		'N33E092.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N33E092.hgt.zip',
		'N33E093.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N33E093.hgt.zip',
		'N33E094.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N33E094.hgt.zip',
		'N33E095.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N33E095.hgt.zip',
		'N33E096.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N33E096.hgt.zip',
		'N33E097.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N33E097.hgt.zip',
		'N33E098.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N33E098.hgt.zip',
		'N33E099.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N33E099.hgt.zip',
		'N33E100.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N33E100.hgt.zip',
		'N33E101.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N33E101.hgt.zip',
		'N33E102.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N33E102.hgt.zip',
		'N33E103.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N33E103.hgt.zip',
		'N33E104.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N33E104.hgt.zip',
		'N33E105.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N33E105.hgt.zip',
		'N33E106.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N33E106.hgt.zip',
		'N33E107.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N33E107.hgt.zip',
		'N33E108.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N33E108.hgt.zip',
		'N33E109.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N33E109.hgt.zip',
		'N33E110.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N33E110.hgt.zip',
		'N33E111.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N33E111.hgt.zip',
		'N33E112.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N33E112.hgt.zip',
		'N33E113.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N33E113.hgt.zip',
		'N33E114.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N33E114.hgt.zip',
		'N33E115.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N33E115.hgt.zip',
		'N33E116.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N33E116.hgt.zip',
		'N33E117.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N33E117.hgt.zip',
		'N33E118.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N33E118.hgt.zip',
		'N33E119.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N33E119.hgt.zip',
		'N33E120.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N33E120.hgt.zip',
		'N33E126.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N33E126.hgt.zip',
		'N33E128.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N33E128.hgt.zip',
		'N33E129.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N33E129.hgt.zip',
		'N33E130.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N33E130.hgt.zip',
		'N33E131.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N33E131.hgt.zip',
		'N33E132.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N33E132.hgt.zip',
		'N33E133.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N33E133.hgt.zip',
		'N33E134.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N33E134.hgt.zip',
		'N33E135.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N33E135.hgt.zip',
		'N33E136.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N33E136.hgt.zip',
		'N33E138.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N33E138.hgt.zip',
		'N33E139.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N33E139.hgt.zip',
		'N33W001.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N33W001.hgt.zip',
		'N33W002.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N33W002.hgt.zip',
		'N33W003.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N33W003.hgt.zip',
		'N33W004.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N33W004.hgt.zip',
		'N33W005.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N33W005.hgt.zip',
		'N33W006.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N33W006.hgt.zip',
		'N33W007.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N33W007.hgt.zip',
		'N33W008.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N33W008.hgt.zip',
		'N33W009.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N33W009.hgt.zip',
		'N33W017.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N33W017.hgt.zip',
		'N33W078.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N33W078.hgt.zip',
		'N33W079.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N33W079.hgt.zip',
		'N33W080.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N33W080.hgt.zip',
		'N33W081.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N33W081.hgt.zip',
		'N33W082.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N33W082.hgt.zip',
		'N33W083.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N33W083.hgt.zip',
		'N33W084.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N33W084.hgt.zip',
		'N33W085.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N33W085.hgt.zip',
		'N33W086.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N33W086.hgt.zip',
		'N33W087.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N33W087.hgt.zip',
		'N33W088.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N33W088.hgt.zip',
		'N33W089.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N33W089.hgt.zip',
		'N33W090.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N33W090.hgt.zip',
		'N33W091.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N33W091.hgt.zip',
		'N33W092.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N33W092.hgt.zip',
		'N33W093.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N33W093.hgt.zip',
		'N33W094.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N33W094.hgt.zip',
		'N33W095.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N33W095.hgt.zip',
		'N33W096.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N33W096.hgt.zip',
		'N33W097.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N33W097.hgt.zip',
		'N33W098.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N33W098.hgt.zip',
		'N33W099.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N33W099.hgt.zip',
		'N33W100.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N33W100.hgt.zip',
		'N33W101.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N33W101.hgt.zip',
		'N33W102.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N33W102.hgt.zip',
		'N33W103.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N33W103.hgt.zip',
		'N33W104.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N33W104.hgt.zip',
		'N33W105.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N33W105.hgt.zip',
		'N33W106.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N33W106.hgt.zip',
		'N33W107.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N33W107.hgt.zip',
		'N33W108.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N33W108.hgt.zip',
		'N33W109.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N33W109.hgt.zip',
		'N33W110.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N33W110.hgt.zip',
		'N33W111.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N33W111.hgt.zip',
		'N33W112.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N33W112.hgt.zip',
		'N33W113.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N33W113.hgt.zip',
		'N33W114.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N33W114.hgt.zip',
		'N33W115.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N33W115.hgt.zip',
		'N33W116.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N33W116.hgt.zip',
		'N33W117.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N33W117.hgt.zip',
		'N33W118.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N33W118.hgt.zip',
		'N33W119.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N33W119.hgt.zip',
		'N33W120.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N33W120.hgt.zip',
		'N33W121.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N33W121.hgt.zip',
		'N34E000.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N34E000.hgt.zip',
		'N34E001.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N34E001.hgt.zip',
		'N34E002.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N34E002.hgt.zip',
		'N34E003.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N34E003.hgt.zip',
		'N34E004.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N34E004.hgt.zip',
		'N34E005.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N34E005.hgt.zip',
		'N34E006.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N34E006.hgt.zip',
		'N34E007.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N34E007.hgt.zip',
		'N34E008.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N34E008.hgt.zip',
		'N34E009.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N34E009.hgt.zip',
		'N34E010.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N34E010.hgt.zip',
		'N34E011.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N34E011.hgt.zip',
		'N34E023.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N34E023.hgt.zip',
		'N34E024.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N34E024.hgt.zip',
		'N34E025.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N34E025.hgt.zip',
		'N34E026.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N34E026.hgt.zip',
		'N34E032.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N34E032.hgt.zip',
		'N34E033.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N34E033.hgt.zip',
		'N34E034.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N34E034.hgt.zip',
		'N34E035.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N34E035.hgt.zip',
		'N34E036.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N34E036.hgt.zip',
		'N34E037.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N34E037.hgt.zip',
		'N34E038.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N34E038.hgt.zip',
		'N34E039.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N34E039.hgt.zip',
		'N34E040.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N34E040.hgt.zip',
		'N34E041.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N34E041.hgt.zip',
		'N34E042.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N34E042.hgt.zip',
		'N34E043.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N34E043.hgt.zip',
		'N34E044.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N34E044.hgt.zip',
		'N34E045.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N34E045.hgt.zip',
		'N34E046.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N34E046.hgt.zip',
		'N34E047.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N34E047.hgt.zip',
		'N34E048.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N34E048.hgt.zip',
		'N34E049.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N34E049.hgt.zip',
		'N34E050.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N34E050.hgt.zip',
		'N34E051.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N34E051.hgt.zip',
		'N34E052.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N34E052.hgt.zip',
		'N34E053.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N34E053.hgt.zip',
		'N34E054.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N34E054.hgt.zip',
		'N34E055.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N34E055.hgt.zip',
		'N34E056.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N34E056.hgt.zip',
		'N34E057.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N34E057.hgt.zip',
		'N34E058.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N34E058.hgt.zip',
		'N34E059.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N34E059.hgt.zip',
		'N34E060.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N34E060.hgt.zip',
		'N34E061.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N34E061.hgt.zip',
		'N34E062.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N34E062.hgt.zip',
		'N34E063.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N34E063.hgt.zip',
		'N34E064.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N34E064.hgt.zip',
		'N34E065.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N34E065.hgt.zip',
		'N34E066.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N34E066.hgt.zip',
		'N34E067.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N34E067.hgt.zip',
		'N34E068.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N34E068.hgt.zip',
		'N34E069.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N34E069.hgt.zip',
		'N34E070.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N34E070.hgt.zip',
		'N34E071.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N34E071.hgt.zip',
		'N34E072.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N34E072.hgt.zip',
		'N34E073.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N34E073.hgt.zip',
		'N34E074.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N34E074.hgt.zip',
		'N34E075.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N34E075.hgt.zip',
		'N34E076.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N34E076.hgt.zip',
		'N34E077.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N34E077.hgt.zip',
		'N34E078.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N34E078.hgt.zip',
		'N34E079.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N34E079.hgt.zip',
		'N34E080.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N34E080.hgt.zip',
		'N34E081.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N34E081.hgt.zip',
		'N34E082.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N34E082.hgt.zip',
		'N34E083.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N34E083.hgt.zip',
		'N34E084.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N34E084.hgt.zip',
		'N34E085.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N34E085.hgt.zip',
		'N34E086.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N34E086.hgt.zip',
		'N34E087.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N34E087.hgt.zip',
		'N34E088.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N34E088.hgt.zip',
		'N34E089.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N34E089.hgt.zip',
		'N34E090.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N34E090.hgt.zip',
		'N34E091.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N34E091.hgt.zip',
		'N34E092.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N34E092.hgt.zip',
		'N34E093.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N34E093.hgt.zip',
		'N34E094.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N34E094.hgt.zip',
		'N34E095.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N34E095.hgt.zip',
		'N34E096.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N34E096.hgt.zip',
		'N34E097.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N34E097.hgt.zip',
		'N34E098.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N34E098.hgt.zip',
		'N34E099.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N34E099.hgt.zip',
		'N34E100.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N34E100.hgt.zip',
		'N34E101.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N34E101.hgt.zip',
		'N34E102.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N34E102.hgt.zip',
		'N34E103.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N34E103.hgt.zip',
		'N34E104.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N34E104.hgt.zip',
		'N34E105.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N34E105.hgt.zip',
		'N34E106.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N34E106.hgt.zip',
		'N34E107.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N34E107.hgt.zip',
		'N34E108.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N34E108.hgt.zip',
		'N34E109.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N34E109.hgt.zip',
		'N34E110.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N34E110.hgt.zip',
		'N34E111.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N34E111.hgt.zip',
		'N34E112.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N34E112.hgt.zip',
		'N34E113.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N34E113.hgt.zip',
		'N34E114.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N34E114.hgt.zip',
		'N34E115.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N34E115.hgt.zip',
		'N34E116.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N34E116.hgt.zip',
		'N34E117.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N34E117.hgt.zip',
		'N34E118.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N34E118.hgt.zip',
		'N34E119.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N34E119.hgt.zip',
		'N34E120.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N34E120.hgt.zip',
		'N34E125.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N34E125.hgt.zip',
		'N34E126.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N34E126.hgt.zip',
		'N34E127.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N34E127.hgt.zip',
		'N34E128.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N34E128.hgt.zip',
		'N34E129.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N34E129.hgt.zip',
		'N34E130.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N34E130.hgt.zip',
		'N34E131.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N34E131.hgt.zip',
		'N34E132.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N34E132.hgt.zip',
		'N34E133.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N34E133.hgt.zip',
		'N34E134.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N34E134.hgt.zip',
		'N34E135.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N34E135.hgt.zip',
		'N34E136.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N34E136.hgt.zip',
		'N34E137.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N34E137.hgt.zip',
		'N34E138.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N34E138.hgt.zip',
		'N34E139.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N34E139.hgt.zip',
		'N34W001.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N34W001.hgt.zip',
		'N34W002.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N34W002.hgt.zip',
		'N34W003.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N34W003.hgt.zip',
		'N34W004.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N34W004.hgt.zip',
		'N34W005.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N34W005.hgt.zip',
		'N34W006.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N34W006.hgt.zip',
		'N34W007.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N34W007.hgt.zip',
		'N34W077.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N34W077.hgt.zip',
		'N34W078.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N34W078.hgt.zip',
		'N34W079.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N34W079.hgt.zip',
		'N34W080.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N34W080.hgt.zip',
		'N34W081.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N34W081.hgt.zip',
		'N34W082.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N34W082.hgt.zip',
		'N34W083.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N34W083.hgt.zip',
		'N34W084.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N34W084.hgt.zip',
		'N34W085.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N34W085.hgt.zip',
		'N34W086.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N34W086.hgt.zip',
		'N34W087.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N34W087.hgt.zip',
		'N34W088.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N34W088.hgt.zip',
		'N34W089.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N34W089.hgt.zip',
		'N34W090.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N34W090.hgt.zip',
		'N34W091.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N34W091.hgt.zip',
		'N34W092.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N34W092.hgt.zip',
		'N34W093.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N34W093.hgt.zip',
		'N34W094.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N34W094.hgt.zip',
		'N34W095.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N34W095.hgt.zip',
		'N34W096.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N34W096.hgt.zip',
		'N34W097.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N34W097.hgt.zip',
		'N34W098.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N34W098.hgt.zip',
		'N34W099.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N34W099.hgt.zip',
		'N34W100.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N34W100.hgt.zip',
		'N34W101.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N34W101.hgt.zip',
		'N34W102.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N34W102.hgt.zip',
		'N34W103.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N34W103.hgt.zip',
		'N34W104.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N34W104.hgt.zip',
		'N34W105.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N34W105.hgt.zip',
		'N34W106.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N34W106.hgt.zip',
		'N34W107.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N34W107.hgt.zip',
		'N34W108.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N34W108.hgt.zip',
		'N34W109.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N34W109.hgt.zip',
		'N34W110.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N34W110.hgt.zip',
		'N34W111.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N34W111.hgt.zip',
		'N34W112.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N34W112.hgt.zip',
		'N34W113.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N34W113.hgt.zip',
		'N34W114.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N34W114.hgt.zip',
		'N34W115.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N34W115.hgt.zip',
		'N34W116.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N34W116.hgt.zip',
		'N34W117.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N34W117.hgt.zip',
		'N34W118.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N34W118.hgt.zip',
		'N34W119.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N34W119.hgt.zip',
		'N34W120.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N34W120.hgt.zip',
		'N34W121.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N34W121.hgt.zip',
		'N35E000.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N35E000.hgt.zip',
		'N35E001.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N35E001.hgt.zip',
		'N35E002.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N35E002.hgt.zip',
		'N35E003.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N35E003.hgt.zip',
		'N35E004.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N35E004.hgt.zip',
		'N35E005.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N35E005.hgt.zip',
		'N35E006.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N35E006.hgt.zip',
		'N35E007.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N35E007.hgt.zip',
		'N35E008.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N35E008.hgt.zip',
		'N35E009.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N35E009.hgt.zip',
		'N35E010.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N35E010.hgt.zip',
		'N35E011.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N35E011.hgt.zip',
		'N35E012.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N35E012.hgt.zip',
		'N35E014.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N35E014.hgt.zip',
		'N35E023.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N35E023.hgt.zip',
		'N35E024.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N35E024.hgt.zip',
		'N35E025.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N35E025.hgt.zip',
		'N35E026.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N35E026.hgt.zip',
		'N35E027.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N35E027.hgt.zip',
		'N35E032.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N35E032.hgt.zip',
		'N35E033.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N35E033.hgt.zip',
		'N35E034.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N35E034.hgt.zip',
		'N35E035.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N35E035.hgt.zip',
		'N35E036.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N35E036.hgt.zip',
		'N35E037.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N35E037.hgt.zip',
		'N35E038.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N35E038.hgt.zip',
		'N35E039.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N35E039.hgt.zip',
		'N35E040.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N35E040.hgt.zip',
		'N35E041.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N35E041.hgt.zip',
		'N35E042.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N35E042.hgt.zip',
		'N35E043.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N35E043.hgt.zip',
		'N35E044.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N35E044.hgt.zip',
		'N35E045.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N35E045.hgt.zip',
		'N35E046.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N35E046.hgt.zip',
		'N35E047.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N35E047.hgt.zip',
		'N35E048.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N35E048.hgt.zip',
		'N35E049.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N35E049.hgt.zip',
		'N35E050.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N35E050.hgt.zip',
		'N35E051.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N35E051.hgt.zip',
		'N35E052.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N35E052.hgt.zip',
		'N35E053.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N35E053.hgt.zip',
		'N35E054.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N35E054.hgt.zip',
		'N35E055.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N35E055.hgt.zip',
		'N35E056.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N35E056.hgt.zip',
		'N35E057.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N35E057.hgt.zip',
		'N35E058.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N35E058.hgt.zip',
		'N35E059.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N35E059.hgt.zip',
		'N35E060.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N35E060.hgt.zip',
		'N35E061.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N35E061.hgt.zip',
		'N35E062.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N35E062.hgt.zip',
		'N35E063.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N35E063.hgt.zip',
		'N35E064.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N35E064.hgt.zip',
		'N35E065.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N35E065.hgt.zip',
		'N35E066.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N35E066.hgt.zip',
		'N35E067.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N35E067.hgt.zip',
		'N35E068.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N35E068.hgt.zip',
		'N35E069.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N35E069.hgt.zip',
		'N35E070.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N35E070.hgt.zip',
		'N35E071.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N35E071.hgt.zip',
		'N35E072.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N35E072.hgt.zip',
		'N35E073.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N35E073.hgt.zip',
		'N35E074.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N35E074.hgt.zip',
		'N35E075.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N35E075.hgt.zip',
		'N35E076.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N35E076.hgt.zip',
		'N35E077.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N35E077.hgt.zip',
		'N35E078.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N35E078.hgt.zip',
		'N35E079.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N35E079.hgt.zip',
		'N35E080.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N35E080.hgt.zip',
		'N35E081.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N35E081.hgt.zip',
		'N35E082.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N35E082.hgt.zip',
		'N35E083.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N35E083.hgt.zip',
		'N35E084.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N35E084.hgt.zip',
		'N35E085.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N35E085.hgt.zip',
		'N35E086.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N35E086.hgt.zip',
		'N35E087.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N35E087.hgt.zip',
		'N35E088.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N35E088.hgt.zip',
		'N35E089.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N35E089.hgt.zip',
		'N35E090.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N35E090.hgt.zip',
		'N35E091.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N35E091.hgt.zip',
		'N35E092.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N35E092.hgt.zip',
		'N35E093.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N35E093.hgt.zip',
		'N35E094.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N35E094.hgt.zip',
		'N35E095.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N35E095.hgt.zip',
		'N35E096.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N35E096.hgt.zip',
		'N35E097.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N35E097.hgt.zip',
		'N35E098.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N35E098.hgt.zip',
		'N35E099.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N35E099.hgt.zip',
		'N35E100.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N35E100.hgt.zip',
		'N35E101.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N35E101.hgt.zip',
		'N35E102.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N35E102.hgt.zip',
		'N35E103.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N35E103.hgt.zip',
		'N35E104.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N35E104.hgt.zip',
		'N35E105.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N35E105.hgt.zip',
		'N35E106.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N35E106.hgt.zip',
		'N35E107.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N35E107.hgt.zip',
		'N35E108.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N35E108.hgt.zip',
		'N35E109.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N35E109.hgt.zip',
		'N35E110.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N35E110.hgt.zip',
		'N35E111.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N35E111.hgt.zip',
		'N35E112.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N35E112.hgt.zip',
		'N35E113.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N35E113.hgt.zip',
		'N35E114.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N35E114.hgt.zip',
		'N35E115.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N35E115.hgt.zip',
		'N35E116.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N35E116.hgt.zip',
		'N35E117.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N35E117.hgt.zip',
		'N35E118.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N35E118.hgt.zip',
		'N35E119.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N35E119.hgt.zip',
		'N35E120.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N35E120.hgt.zip',
		'N35E125.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N35E125.hgt.zip',
		'N35E126.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N35E126.hgt.zip',
		'N35E127.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N35E127.hgt.zip',
		'N35E128.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N35E128.hgt.zip',
		'N35E129.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N35E129.hgt.zip',
		'N35E132.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N35E132.hgt.zip',
		'N35E133.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N35E133.hgt.zip',
		'N35E134.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N35E134.hgt.zip',
		'N35E135.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N35E135.hgt.zip',
		'N35E136.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N35E136.hgt.zip',
		'N35E137.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N35E137.hgt.zip',
		'N35E138.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N35E138.hgt.zip',
		'N35E139.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N35E139.hgt.zip',
		'N35E140.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N35E140.hgt.zip',
		'N35W001.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N35W001.hgt.zip',
		'N35W002.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N35W002.hgt.zip',
		'N35W003.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N35W003.hgt.zip',
		'N35W004.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N35W004.hgt.zip',
		'N35W005.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N35W005.hgt.zip',
		'N35W006.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N35W006.hgt.zip',
		'N35W007.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N35W007.hgt.zip',
		'N35W076.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N35W076.hgt.zip',
		'N35W077.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N35W077.hgt.zip',
		'N35W078.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N35W078.hgt.zip',
		'N35W079.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N35W079.hgt.zip',
		'N35W080.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N35W080.hgt.zip',
		'N35W081.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N35W081.hgt.zip',
		'N35W082.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N35W082.hgt.zip',
		'N35W083.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N35W083.hgt.zip',
		'N35W084.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N35W084.hgt.zip',
		'N35W085.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N35W085.hgt.zip',
		'N35W086.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N35W086.hgt.zip',
		'N35W087.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N35W087.hgt.zip',
		'N35W088.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N35W088.hgt.zip',
		'N35W089.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N35W089.hgt.zip',
		'N35W090.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N35W090.hgt.zip',
		'N35W091.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N35W091.hgt.zip',
		'N35W092.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N35W092.hgt.zip',
		'N35W093.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N35W093.hgt.zip',
		'N35W094.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N35W094.hgt.zip',
		'N35W095.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N35W095.hgt.zip',
		'N35W096.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N35W096.hgt.zip',
		'N35W097.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N35W097.hgt.zip',
		'N35W098.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N35W098.hgt.zip',
		'N35W099.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N35W099.hgt.zip',
		'N35W100.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N35W100.hgt.zip',
		'N35W101.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N35W101.hgt.zip',
		'N35W102.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N35W102.hgt.zip',
		'N35W103.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N35W103.hgt.zip',
		'N35W104.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N35W104.hgt.zip',
		'N35W105.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N35W105.hgt.zip',
		'N35W106.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N35W106.hgt.zip',
		'N35W107.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N35W107.hgt.zip',
		'N35W108.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N35W108.hgt.zip',
		'N35W109.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N35W109.hgt.zip',
		'N35W110.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N35W110.hgt.zip',
		'N35W111.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N35W111.hgt.zip',
		'N35W112.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N35W112.hgt.zip',
		'N35W113.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N35W113.hgt.zip',
		'N35W114.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N35W114.hgt.zip',
		'N35W115.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N35W115.hgt.zip',
		'N35W116.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N35W116.hgt.zip',
		'N35W117.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N35W117.hgt.zip',
		'N35W118.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N35W118.hgt.zip',
		'N35W119.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N35W119.hgt.zip',
		'N35W120.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N35W120.hgt.zip',
		'N35W121.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N35W121.hgt.zip',
		'N35W122.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N35W122.hgt.zip',
		'N36E000.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N36E000.hgt.zip',
		'N36E001.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N36E001.hgt.zip',
		'N36E002.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N36E002.hgt.zip',
		'N36E003.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N36E003.hgt.zip',
		'N36E004.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N36E004.hgt.zip',
		'N36E005.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N36E005.hgt.zip',
		'N36E006.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N36E006.hgt.zip',
		'N36E007.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N36E007.hgt.zip',
		'N36E008.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N36E008.hgt.zip',
		'N36E009.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N36E009.hgt.zip',
		'N36E010.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N36E010.hgt.zip',
		'N36E011.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N36E011.hgt.zip',
		'N36E012.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N36E012.hgt.zip',
		'N36E014.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N36E014.hgt.zip',
		'N36E015.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N36E015.hgt.zip',
		'N36E021.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N36E021.hgt.zip',
		'N36E022.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N36E022.hgt.zip',
		'N36E023.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N36E023.hgt.zip',
		'N36E024.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N36E024.hgt.zip',
		'N36E025.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N36E025.hgt.zip',
		'N36E026.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N36E026.hgt.zip',
		'N36E027.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N36E027.hgt.zip',
		'N36E028.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N36E028.hgt.zip',
		'N36E029.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N36E029.hgt.zip',
		'N36E030.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N36E030.hgt.zip',
		'N36E031.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N36E031.hgt.zip',
		'N36E032.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N36E032.hgt.zip',
		'N36E033.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N36E033.hgt.zip',
		'N36E034.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N36E034.hgt.zip',
		'N36E035.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N36E035.hgt.zip',
		'N36E036.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N36E036.hgt.zip',
		'N36E037.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N36E037.hgt.zip',
		'N36E038.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N36E038.hgt.zip',
		'N36E039.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N36E039.hgt.zip',
		'N36E040.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N36E040.hgt.zip',
		'N36E041.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N36E041.hgt.zip',
		'N36E042.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N36E042.hgt.zip',
		'N36E043.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N36E043.hgt.zip',
		'N36E044.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N36E044.hgt.zip',
		'N36E045.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N36E045.hgt.zip',
		'N36E046.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N36E046.hgt.zip',
		'N36E047.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N36E047.hgt.zip',
		'N36E048.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N36E048.hgt.zip',
		'N36E049.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N36E049.hgt.zip',
		'N36E050.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N36E050.hgt.zip',
		'N36E051.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N36E051.hgt.zip',
		'N36E052.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N36E052.hgt.zip',
		'N36E053.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N36E053.hgt.zip',
		'N36E054.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N36E054.hgt.zip',
		'N36E055.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N36E055.hgt.zip',
		'N36E056.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N36E056.hgt.zip',
		'N36E057.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N36E057.hgt.zip',
		'N36E058.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N36E058.hgt.zip',
		'N36E059.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N36E059.hgt.zip',
		'N36E060.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N36E060.hgt.zip',
		'N36E061.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N36E061.hgt.zip',
		'N36E062.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N36E062.hgt.zip',
		'N36E063.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N36E063.hgt.zip',
		'N36E064.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N36E064.hgt.zip',
		'N36E065.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N36E065.hgt.zip',
		'N36E066.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N36E066.hgt.zip',
		'N36E067.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N36E067.hgt.zip',
		'N36E068.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N36E068.hgt.zip',
		'N36E069.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N36E069.hgt.zip',
		'N36E070.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N36E070.hgt.zip',
		'N36E071.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N36E071.hgt.zip',
		'N36E072.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N36E072.hgt.zip',
		'N36E073.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N36E073.hgt.zip',
		'N36E074.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N36E074.hgt.zip',
		'N36E075.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N36E075.hgt.zip',
		'N36E076.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N36E076.hgt.zip',
		'N36E077.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N36E077.hgt.zip',
		'N36E078.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N36E078.hgt.zip',
		'N36E079.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N36E079.hgt.zip',
		'N36E080.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N36E080.hgt.zip',
		'N36E081.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N36E081.hgt.zip',
		'N36E082.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N36E082.hgt.zip',
		'N36E083.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N36E083.hgt.zip',
		'N36E084.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N36E084.hgt.zip',
		'N36E085.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N36E085.hgt.zip',
		'N36E086.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N36E086.hgt.zip',
		'N36E087.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N36E087.hgt.zip',
		'N36E088.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N36E088.hgt.zip',
		'N36E089.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N36E089.hgt.zip',
		'N36E090.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N36E090.hgt.zip',
		'N36E091.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N36E091.hgt.zip',
		'N36E092.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N36E092.hgt.zip',
		'N36E093.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N36E093.hgt.zip',
		'N36E094.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N36E094.hgt.zip',
		'N36E095.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N36E095.hgt.zip',
		'N36E096.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N36E096.hgt.zip',
		'N36E097.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N36E097.hgt.zip',
		'N36E098.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N36E098.hgt.zip',
		'N36E099.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N36E099.hgt.zip',
		'N36E100.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N36E100.hgt.zip',
		'N36E101.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N36E101.hgt.zip',
		'N36E102.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N36E102.hgt.zip',
		'N36E103.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N36E103.hgt.zip',
		'N36E104.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N36E104.hgt.zip',
		'N36E105.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N36E105.hgt.zip',
		'N36E106.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N36E106.hgt.zip',
		'N36E107.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N36E107.hgt.zip',
		'N36E108.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N36E108.hgt.zip',
		'N36E109.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N36E109.hgt.zip',
		'N36E110.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N36E110.hgt.zip',
		'N36E111.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N36E111.hgt.zip',
		'N36E112.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N36E112.hgt.zip',
		'N36E113.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N36E113.hgt.zip',
		'N36E114.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N36E114.hgt.zip',
		'N36E115.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N36E115.hgt.zip',
		'N36E116.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N36E116.hgt.zip',
		'N36E117.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N36E117.hgt.zip',
		'N36E118.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N36E118.hgt.zip',
		'N36E119.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N36E119.hgt.zip',
		'N36E120.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N36E120.hgt.zip',
		'N36E121.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N36E121.hgt.zip',
		'N36E122.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N36E122.hgt.zip',
		'N36E125.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N36E125.hgt.zip',
		'N36E126.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N36E126.hgt.zip',
		'N36E127.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N36E127.hgt.zip',
		'N36E128.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N36E128.hgt.zip',
		'N36E129.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N36E129.hgt.zip',
		'N36E132.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N36E132.hgt.zip',
		'N36E133.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N36E133.hgt.zip',
		'N36E135.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N36E135.hgt.zip',
		'N36E136.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N36E136.hgt.zip',
		'N36E137.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N36E137.hgt.zip',
		'N36E138.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N36E138.hgt.zip',
		'N36E139.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N36E139.hgt.zip',
		'N36E140.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N36E140.hgt.zip',
		'N36W002.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N36W002.hgt.zip',
		'N36W003.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N36W003.hgt.zip',
		'N36W004.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N36W004.hgt.zip',
		'N36W005.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N36W005.hgt.zip',
		'N36W006.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N36W006.hgt.zip',
		'N36W007.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N36W007.hgt.zip',
		'N36W008.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N36W008.hgt.zip',
		'N36W009.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N36W009.hgt.zip',
		'N36W026.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N36W026.hgt.zip',
		'N36W076.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N36W076.hgt.zip',
		'N36W077.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N36W077.hgt.zip',
		'N36W078.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N36W078.hgt.zip',
		'N36W079.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N36W079.hgt.zip',
		'N36W080.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N36W080.hgt.zip',
		'N36W081.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N36W081.hgt.zip',
		'N36W082.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N36W082.hgt.zip',
		'N36W083.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N36W083.hgt.zip',
		'N36W084.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N36W084.hgt.zip',
		'N36W085.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N36W085.hgt.zip',
		'N36W086.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N36W086.hgt.zip',
		'N36W087.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N36W087.hgt.zip',
		'N36W088.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N36W088.hgt.zip',
		'N36W089.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N36W089.hgt.zip',
		'N36W090.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N36W090.hgt.zip',
		'N36W091.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N36W091.hgt.zip',
		'N36W092.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N36W092.hgt.zip',
		'N36W093.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N36W093.hgt.zip',
		'N36W094.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N36W094.hgt.zip',
		'N36W095.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N36W095.hgt.zip',
		'N36W096.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N36W096.hgt.zip',
		'N36W097.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N36W097.hgt.zip',
		'N36W098.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N36W098.hgt.zip',
		'N36W099.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N36W099.hgt.zip',
		'N36W100.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N36W100.hgt.zip',
		'N36W101.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N36W101.hgt.zip',
		'N36W102.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N36W102.hgt.zip',
		'N36W103.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N36W103.hgt.zip',
		'N36W104.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N36W104.hgt.zip',
		'N36W105.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N36W105.hgt.zip',
		'N36W106.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N36W106.hgt.zip',
		'N36W107.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N36W107.hgt.zip',
		'N36W108.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N36W108.hgt.zip',
		'N36W109.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N36W109.hgt.zip',
		'N36W110.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N36W110.hgt.zip',
		'N36W111.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N36W111.hgt.zip',
		'N36W112.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N36W112.hgt.zip',
		'N36W113.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N36W113.hgt.zip',
		'N36W114.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N36W114.hgt.zip',
		'N36W115.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N36W115.hgt.zip',
		'N36W116.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N36W116.hgt.zip',
		'N36W117.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N36W117.hgt.zip',
		'N36W118.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N36W118.hgt.zip',
		'N36W119.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N36W119.hgt.zip',
		'N36W120.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N36W120.hgt.zip',
		'N36W121.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N36W121.hgt.zip',
		'N36W122.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N36W122.hgt.zip',
		'N36W123.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N36W123.hgt.zip',
		'N37E006.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N37E006.hgt.zip',
		'N37E007.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N37E007.hgt.zip',
		'N37E008.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N37E008.hgt.zip',
		'N37E009.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N37E009.hgt.zip',
		'N37E010.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N37E010.hgt.zip',
		'N37E011.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N37E011.hgt.zip',
		'N37E012.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N37E012.hgt.zip',
		'N37E013.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N37E013.hgt.zip',
		'N37E014.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N37E014.hgt.zip',
		'N37E015.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N37E015.hgt.zip',
		'N37E016.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N37E016.hgt.zip',
		'N37E020.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N37E020.hgt.zip',
		'N37E021.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N37E021.hgt.zip',
		'N37E022.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N37E022.hgt.zip',
		'N37E023.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N37E023.hgt.zip',
		'N37E024.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N37E024.hgt.zip',
		'N37E025.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N37E025.hgt.zip',
		'N37E026.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N37E026.hgt.zip',
		'N37E027.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N37E027.hgt.zip',
		'N37E028.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N37E028.hgt.zip',
		'N37E029.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N37E029.hgt.zip',
		'N37E030.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N37E030.hgt.zip',
		'N37E031.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N37E031.hgt.zip',
		'N37E032.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N37E032.hgt.zip',
		'N37E033.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N37E033.hgt.zip',
		'N37E034.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N37E034.hgt.zip',
		'N37E035.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N37E035.hgt.zip',
		'N37E036.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N37E036.hgt.zip',
		'N37E037.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N37E037.hgt.zip',
		'N37E038.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N37E038.hgt.zip',
		'N37E039.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N37E039.hgt.zip',
		'N37E040.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N37E040.hgt.zip',
		'N37E041.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N37E041.hgt.zip',
		'N37E042.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N37E042.hgt.zip',
		'N37E043.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N37E043.hgt.zip',
		'N37E044.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N37E044.hgt.zip',
		'N37E045.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N37E045.hgt.zip',
		'N37E046.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N37E046.hgt.zip',
		'N37E047.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N37E047.hgt.zip',
		'N37E048.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N37E048.hgt.zip',
		'N37E049.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N37E049.hgt.zip',
		'N37E050.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N37E050.hgt.zip',
		'N37E053.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N37E053.hgt.zip',
		'N37E054.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N37E054.hgt.zip',
		'N37E055.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N37E055.hgt.zip',
		'N37E056.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N37E056.hgt.zip',
		'N37E057.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N37E057.hgt.zip',
		'N37E058.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N37E058.hgt.zip',
		'N37E059.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N37E059.hgt.zip',
		'N37E060.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N37E060.hgt.zip',
		'N37E061.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N37E061.hgt.zip',
		'N37E062.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N37E062.hgt.zip',
		'N37E063.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N37E063.hgt.zip',
		'N37E064.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N37E064.hgt.zip',
		'N37E065.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N37E065.hgt.zip',
		'N37E066.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N37E066.hgt.zip',
		'N37E067.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N37E067.hgt.zip',
		'N37E068.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N37E068.hgt.zip',
		'N37E069.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N37E069.hgt.zip',
		'N37E070.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N37E070.hgt.zip',
		'N37E071.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N37E071.hgt.zip',
		'N37E072.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N37E072.hgt.zip',
		'N37E073.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N37E073.hgt.zip',
		'N37E074.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N37E074.hgt.zip',
		'N37E075.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N37E075.hgt.zip',
		'N37E076.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N37E076.hgt.zip',
		'N37E077.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N37E077.hgt.zip',
		'N37E078.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N37E078.hgt.zip',
		'N37E079.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N37E079.hgt.zip',
		'N37E080.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N37E080.hgt.zip',
		'N37E081.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N37E081.hgt.zip',
		'N37E082.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N37E082.hgt.zip',
		'N37E083.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N37E083.hgt.zip',
		'N37E084.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N37E084.hgt.zip',
		'N37E085.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N37E085.hgt.zip',
		'N37E086.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N37E086.hgt.zip',
		'N37E087.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N37E087.hgt.zip',
		'N37E088.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N37E088.hgt.zip',
		'N37E089.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N37E089.hgt.zip',
		'N37E090.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N37E090.hgt.zip',
		'N37E091.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N37E091.hgt.zip',
		'N37E092.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N37E092.hgt.zip',
		'N37E093.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N37E093.hgt.zip',
		'N37E094.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N37E094.hgt.zip',
		'N37E095.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N37E095.hgt.zip',
		'N37E096.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N37E096.hgt.zip',
		'N37E097.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N37E097.hgt.zip',
		'N37E098.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N37E098.hgt.zip',
		'N37E099.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N37E099.hgt.zip',
		'N37E100.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N37E100.hgt.zip',
		'N37E101.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N37E101.hgt.zip',
		'N37E102.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N37E102.hgt.zip',
		'N37E103.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N37E103.hgt.zip',
		'N37E104.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N37E104.hgt.zip',
		'N37E105.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N37E105.hgt.zip',
		'N37E106.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N37E106.hgt.zip',
		'N37E107.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N37E107.hgt.zip',
		'N37E108.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N37E108.hgt.zip',
		'N37E109.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N37E109.hgt.zip',
		'N37E110.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N37E110.hgt.zip',
		'N37E111.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N37E111.hgt.zip',
		'N37E112.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N37E112.hgt.zip',
		'N37E113.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N37E113.hgt.zip',
		'N37E114.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N37E114.hgt.zip',
		'N37E115.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N37E115.hgt.zip',
		'N37E116.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N37E116.hgt.zip',
		'N37E117.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N37E117.hgt.zip',
		'N37E118.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N37E118.hgt.zip',
		'N37E119.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N37E119.hgt.zip',
		'N37E120.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N37E120.hgt.zip',
		'N37E121.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N37E121.hgt.zip',
		'N37E122.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N37E122.hgt.zip',
		'N37E124.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N37E124.hgt.zip',
		'N37E125.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N37E125.hgt.zip',
		'N37E126.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N37E126.hgt.zip',
		'N37E127.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N37E127.hgt.zip',
		'N37E128.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N37E128.hgt.zip',
		'N37E129.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N37E129.hgt.zip',
		'N37E130.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N37E130.hgt.zip',
		'N37E131.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N37E131.hgt.zip',
		'N37E136.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N37E136.hgt.zip',
		'N37E137.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N37E137.hgt.zip',
		'N37E138.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N37E138.hgt.zip',
		'N37E139.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N37E139.hgt.zip',
		'N37E140.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N37E140.hgt.zip',
		'N37E141.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N37E141.hgt.zip',
		'N37W001.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N37W001.hgt.zip',
		'N37W002.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N37W002.hgt.zip',
		'N37W003.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N37W003.hgt.zip',
		'N37W004.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N37W004.hgt.zip',
		'N37W005.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N37W005.hgt.zip',
		'N37W006.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N37W006.hgt.zip',
		'N37W007.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N37W007.hgt.zip',
		'N37W008.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N37W008.hgt.zip',
		'N37W009.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N37W009.hgt.zip',
		'N37W025.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N37W025.hgt.zip',
		'N37W026.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N37W026.hgt.zip',
		'N37W076.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N37W076.hgt.zip',
		'N37W077.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N37W077.hgt.zip',
		'N37W078.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N37W078.hgt.zip',
		'N37W079.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N37W079.hgt.zip',
		'N37W080.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N37W080.hgt.zip',
		'N37W081.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N37W081.hgt.zip',
		'N37W082.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N37W082.hgt.zip',
		'N37W083.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N37W083.hgt.zip',
		'N37W084.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N37W084.hgt.zip',
		'N37W085.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N37W085.hgt.zip',
		'N37W086.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N37W086.hgt.zip',
		'N37W087.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N37W087.hgt.zip',
		'N37W088.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N37W088.hgt.zip',
		'N37W089.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N37W089.hgt.zip',
		'N37W090.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N37W090.hgt.zip',
		'N37W091.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N37W091.hgt.zip',
		'N37W092.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N37W092.hgt.zip',
		'N37W093.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N37W093.hgt.zip',
		'N37W094.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N37W094.hgt.zip',
		'N37W095.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N37W095.hgt.zip',
		'N37W096.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N37W096.hgt.zip',
		'N37W097.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N37W097.hgt.zip',
		'N37W098.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N37W098.hgt.zip',
		'N37W099.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N37W099.hgt.zip',
		'N37W100.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N37W100.hgt.zip',
		'N37W101.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N37W101.hgt.zip',
		'N37W102.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N37W102.hgt.zip',
		'N37W103.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N37W103.hgt.zip',
		'N37W104.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N37W104.hgt.zip',
		'N37W105.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N37W105.hgt.zip',
		'N37W106.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N37W106.hgt.zip',
		'N37W107.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N37W107.hgt.zip',
		'N37W108.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N37W108.hgt.zip',
		'N37W109.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N37W109.hgt.zip',
		'N37W110.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N37W110.hgt.zip',
		'N37W111.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N37W111.hgt.zip',
		'N37W112.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N37W112.hgt.zip',
		'N37W113.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N37W113.hgt.zip',
		'N37W114.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N37W114.hgt.zip',
		'N37W115.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N37W115.hgt.zip',
		'N37W116.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N37W116.hgt.zip',
		'N37W117.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N37W117.hgt.zip',
		'N37W118.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N37W118.hgt.zip',
		'N37W119.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N37W119.hgt.zip',
		'N37W120.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N37W120.hgt.zip',
		'N37W121.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N37W121.hgt.zip',
		'N37W122.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N37W122.hgt.zip',
		'N37W123.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N37W123.hgt.zip',
		'N37W124.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N37W124.hgt.zip',
		'N38E000.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N38E000.hgt.zip',
		'N38E001.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N38E001.hgt.zip',
		'N38E008.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N38E008.hgt.zip',
		'N38E009.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N38E009.hgt.zip',
		'N38E012.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N38E012.hgt.zip',
		'N38E013.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N38E013.hgt.zip',
		'N38E014.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N38E014.hgt.zip',
		'N38E015.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N38E015.hgt.zip',
		'N38E016.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N38E016.hgt.zip',
		'N38E017.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N38E017.hgt.zip',
		'N38E020.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N38E020.hgt.zip',
		'N38E021.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N38E021.hgt.zip',
		'N38E022.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N38E022.hgt.zip',
		'N38E023.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N38E023.hgt.zip',
		'N38E024.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N38E024.hgt.zip',
		'N38E025.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N38E025.hgt.zip',
		'N38E026.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N38E026.hgt.zip',
		'N38E027.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N38E027.hgt.zip',
		'N38E028.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N38E028.hgt.zip',
		'N38E029.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N38E029.hgt.zip',
		'N38E030.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N38E030.hgt.zip',
		'N38E031.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N38E031.hgt.zip',
		'N38E032.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N38E032.hgt.zip',
		'N38E033.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N38E033.hgt.zip',
		'N38E034.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N38E034.hgt.zip',
		'N38E035.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N38E035.hgt.zip',
		'N38E036.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N38E036.hgt.zip',
		'N38E037.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N38E037.hgt.zip',
		'N38E038.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N38E038.hgt.zip',
		'N38E039.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N38E039.hgt.zip',
		'N38E040.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N38E040.hgt.zip',
		'N38E041.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N38E041.hgt.zip',
		'N38E042.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N38E042.hgt.zip',
		'N38E043.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N38E043.hgt.zip',
		'N38E044.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N38E044.hgt.zip',
		'N38E045.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N38E045.hgt.zip',
		'N38E046.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N38E046.hgt.zip',
		'N38E047.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N38E047.hgt.zip',
		'N38E048.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N38E048.hgt.zip',
		'N38E049.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N38E049.hgt.zip',
		'N38E053.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N38E053.hgt.zip',
		'N38E054.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N38E054.hgt.zip',
		'N38E055.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N38E055.hgt.zip',
		'N38E056.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N38E056.hgt.zip',
		'N38E057.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N38E057.hgt.zip',
		'N38E058.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N38E058.hgt.zip',
		'N38E059.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N38E059.hgt.zip',
		'N38E060.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N38E060.hgt.zip',
		'N38E061.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N38E061.hgt.zip',
		'N38E062.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N38E062.hgt.zip',
		'N38E063.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N38E063.hgt.zip',
		'N38E064.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N38E064.hgt.zip',
		'N38E065.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N38E065.hgt.zip',
		'N38E066.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N38E066.hgt.zip',
		'N38E067.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N38E067.hgt.zip',
		'N38E068.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N38E068.hgt.zip',
		'N38E069.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N38E069.hgt.zip',
		'N38E070.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N38E070.hgt.zip',
		'N38E071.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N38E071.hgt.zip',
		'N38E072.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N38E072.hgt.zip',
		'N38E073.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N38E073.hgt.zip',
		'N38E074.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N38E074.hgt.zip',
		'N38E075.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N38E075.hgt.zip',
		'N38E076.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N38E076.hgt.zip',
		'N38E077.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N38E077.hgt.zip',
		'N38E078.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N38E078.hgt.zip',
		'N38E079.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N38E079.hgt.zip',
		'N38E080.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N38E080.hgt.zip',
		'N38E081.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N38E081.hgt.zip',
		'N38E082.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N38E082.hgt.zip',
		'N38E083.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N38E083.hgt.zip',
		'N38E084.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N38E084.hgt.zip',
		'N38E085.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N38E085.hgt.zip',
		'N38E086.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N38E086.hgt.zip',
		'N38E087.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N38E087.hgt.zip',
		'N38E088.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N38E088.hgt.zip',
		'N38E089.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N38E089.hgt.zip',
		'N38E090.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N38E090.hgt.zip',
		'N38E091.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N38E091.hgt.zip',
		'N38E092.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N38E092.hgt.zip',
		'N38E093.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N38E093.hgt.zip',
		'N38E094.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N38E094.hgt.zip',
		'N38E095.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N38E095.hgt.zip',
		'N38E096.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N38E096.hgt.zip',
		'N38E097.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N38E097.hgt.zip',
		'N38E098.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N38E098.hgt.zip',
		'N38E099.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N38E099.hgt.zip',
		'N38E100.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N38E100.hgt.zip',
		'N38E101.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N38E101.hgt.zip',
		'N38E102.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N38E102.hgt.zip',
		'N38E103.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N38E103.hgt.zip',
		'N38E104.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N38E104.hgt.zip',
		'N38E105.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N38E105.hgt.zip',
		'N38E106.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N38E106.hgt.zip',
		'N38E107.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N38E107.hgt.zip',
		'N38E108.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N38E108.hgt.zip',
		'N38E109.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N38E109.hgt.zip',
		'N38E110.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N38E110.hgt.zip',
		'N38E111.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N38E111.hgt.zip',
		'N38E112.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N38E112.hgt.zip',
		'N38E113.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N38E113.hgt.zip',
		'N38E114.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N38E114.hgt.zip',
		'N38E115.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N38E115.hgt.zip',
		'N38E116.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N38E116.hgt.zip',
		'N38E117.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N38E117.hgt.zip',
		'N38E118.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N38E118.hgt.zip',
		'N38E120.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N38E120.hgt.zip',
		'N38E121.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N38E121.hgt.zip',
		'N38E124.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N38E124.hgt.zip',
		'N38E125.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N38E125.hgt.zip',
		'N38E126.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N38E126.hgt.zip',
		'N38E127.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N38E127.hgt.zip',
		'N38E128.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N38E128.hgt.zip',
		'N38E138.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N38E138.hgt.zip',
		'N38E139.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N38E139.hgt.zip',
		'N38E140.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N38E140.hgt.zip',
		'N38E141.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N38E141.hgt.zip',
		'N38W001.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N38W001.hgt.zip',
		'N38W002.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N38W002.hgt.zip',
		'N38W003.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N38W003.hgt.zip',
		'N38W004.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N38W004.hgt.zip',
		'N38W005.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N38W005.hgt.zip',
		'N38W006.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N38W006.hgt.zip',
		'N38W007.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N38W007.hgt.zip',
		'N38W008.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N38W008.hgt.zip',
		'N38W009.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N38W009.hgt.zip',
		'N38W010.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N38W010.hgt.zip',
		'N38W028.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N38W028.hgt.zip',
		'N38W029.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N38W029.hgt.zip',
		'N38W075.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N38W075.hgt.zip',
		'N38W076.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N38W076.hgt.zip',
		'N38W077.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N38W077.hgt.zip',
		'N38W078.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N38W078.hgt.zip',
		'N38W079.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N38W079.hgt.zip',
		'N38W080.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N38W080.hgt.zip',
		'N38W081.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N38W081.hgt.zip',
		'N38W082.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N38W082.hgt.zip',
		'N38W083.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N38W083.hgt.zip',
		'N38W084.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N38W084.hgt.zip',
		'N38W085.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N38W085.hgt.zip',
		'N38W086.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N38W086.hgt.zip',
		'N38W087.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N38W087.hgt.zip',
		'N38W088.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N38W088.hgt.zip',
		'N38W089.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N38W089.hgt.zip',
		'N38W090.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N38W090.hgt.zip',
		'N38W091.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N38W091.hgt.zip',
		'N38W092.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N38W092.hgt.zip',
		'N38W093.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N38W093.hgt.zip',
		'N38W094.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N38W094.hgt.zip',
		'N38W095.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N38W095.hgt.zip',
		'N38W096.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N38W096.hgt.zip',
		'N38W097.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N38W097.hgt.zip',
		'N38W098.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N38W098.hgt.zip',
		'N38W099.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N38W099.hgt.zip',
		'N38W100.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N38W100.hgt.zip',
		'N38W101.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N38W101.hgt.zip',
		'N38W102.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N38W102.hgt.zip',
		'N38W103.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N38W103.hgt.zip',
		'N38W104.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N38W104.hgt.zip',
		'N38W105.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N38W105.hgt.zip',
		'N38W106.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N38W106.hgt.zip',
		'N38W107.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N38W107.hgt.zip',
		'N38W108.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N38W108.hgt.zip',
		'N38W109.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N38W109.hgt.zip',
		'N38W110.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N38W110.hgt.zip',
		'N38W111.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N38W111.hgt.zip',
		'N38W112.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N38W112.hgt.zip',
		'N38W113.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N38W113.hgt.zip',
		'N38W114.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N38W114.hgt.zip',
		'N38W115.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N38W115.hgt.zip',
		'N38W116.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N38W116.hgt.zip',
		'N38W117.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N38W117.hgt.zip',
		'N38W118.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N38W118.hgt.zip',
		'N38W119.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N38W119.hgt.zip',
		'N38W120.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N38W120.hgt.zip',
		'N38W121.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N38W121.hgt.zip',
		'N38W122.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N38W122.hgt.zip',
		'N38W123.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N38W123.hgt.zip',
		'N38W124.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N38W124.hgt.zip',
		'N39E000.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N39E000.hgt.zip',
		'N39E001.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N39E001.hgt.zip',
		'N39E002.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N39E002.hgt.zip',
		'N39E003.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N39E003.hgt.zip',
		'N39E004.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N39E004.hgt.zip',
		'N39E008.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N39E008.hgt.zip',
		'N39E009.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N39E009.hgt.zip',
		'N39E015.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N39E015.hgt.zip',
		'N39E016.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N39E016.hgt.zip',
		'N39E017.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N39E017.hgt.zip',
		'N39E018.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N39E018.hgt.zip',
		'N39E019.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N39E019.hgt.zip',
		'N39E020.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N39E020.hgt.zip',
		'N39E021.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N39E021.hgt.zip',
		'N39E022.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N39E022.hgt.zip',
		'N39E023.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N39E023.hgt.zip',
		'N39E024.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N39E024.hgt.zip',
		'N39E025.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N39E025.hgt.zip',
		'N39E026.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N39E026.hgt.zip',
		'N39E027.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N39E027.hgt.zip',
		'N39E028.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N39E028.hgt.zip',
		'N39E029.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N39E029.hgt.zip',
		'N39E030.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N39E030.hgt.zip',
		'N39E031.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N39E031.hgt.zip',
		'N39E032.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N39E032.hgt.zip',
		'N39E033.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N39E033.hgt.zip',
		'N39E034.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N39E034.hgt.zip',
		'N39E035.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N39E035.hgt.zip',
		'N39E036.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N39E036.hgt.zip',
		'N39E037.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N39E037.hgt.zip',
		'N39E038.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N39E038.hgt.zip',
		'N39E039.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N39E039.hgt.zip',
		'N39E040.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N39E040.hgt.zip',
		'N39E041.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N39E041.hgt.zip',
		'N39E042.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N39E042.hgt.zip',
		'N39E043.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N39E043.hgt.zip',
		'N39E044.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N39E044.hgt.zip',
		'N39E045.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N39E045.hgt.zip',
		'N39E046.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N39E046.hgt.zip',
		'N39E047.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N39E047.hgt.zip',
		'N39E048.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N39E048.hgt.zip',
		'N39E049.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N39E049.hgt.zip',
		'N39E052.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N39E052.hgt.zip',
		'N39E053.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N39E053.hgt.zip',
		'N39E054.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N39E054.hgt.zip',
		'N39E055.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N39E055.hgt.zip',
		'N39E056.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N39E056.hgt.zip',
		'N39E057.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N39E057.hgt.zip',
		'N39E058.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N39E058.hgt.zip',
		'N39E059.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N39E059.hgt.zip',
		'N39E060.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N39E060.hgt.zip',
		'N39E061.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N39E061.hgt.zip',
		'N39E062.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N39E062.hgt.zip',
		'N39E063.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N39E063.hgt.zip',
		'N39E064.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N39E064.hgt.zip',
		'N39E065.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N39E065.hgt.zip',
		'N39E066.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N39E066.hgt.zip',
		'N39E067.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N39E067.hgt.zip',
		'N39E068.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N39E068.hgt.zip',
		'N39E069.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N39E069.hgt.zip',
		'N39E070.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N39E070.hgt.zip',
		'N39E071.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N39E071.hgt.zip',
		'N39E072.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N39E072.hgt.zip',
		'N39E073.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N39E073.hgt.zip',
		'N39E074.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N39E074.hgt.zip',
		'N39E075.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N39E075.hgt.zip',
		'N39E076.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N39E076.hgt.zip',
		'N39E077.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N39E077.hgt.zip',
		'N39E078.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N39E078.hgt.zip',
		'N39E079.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N39E079.hgt.zip',
		'N39E080.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N39E080.hgt.zip',
		'N39E081.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N39E081.hgt.zip',
		'N39E082.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N39E082.hgt.zip',
		'N39E083.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N39E083.hgt.zip',
		'N39E084.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N39E084.hgt.zip',
		'N39E085.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N39E085.hgt.zip',
		'N39E086.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N39E086.hgt.zip',
		'N39E087.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N39E087.hgt.zip',
		'N39E088.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N39E088.hgt.zip',
		'N39E089.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N39E089.hgt.zip',
		'N39E090.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N39E090.hgt.zip',
		'N39E091.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N39E091.hgt.zip',
		'N39E092.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N39E092.hgt.zip',
		'N39E093.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N39E093.hgt.zip',
		'N39E094.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N39E094.hgt.zip',
		'N39E095.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N39E095.hgt.zip',
		'N39E096.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N39E096.hgt.zip',
		'N39E097.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N39E097.hgt.zip',
		'N39E098.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N39E098.hgt.zip',
		'N39E099.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N39E099.hgt.zip',
		'N39E100.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N39E100.hgt.zip',
		'N39E101.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N39E101.hgt.zip',
		'N39E102.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N39E102.hgt.zip',
		'N39E103.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N39E103.hgt.zip',
		'N39E104.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N39E104.hgt.zip',
		'N39E105.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N39E105.hgt.zip',
		'N39E106.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N39E106.hgt.zip',
		'N39E107.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N39E107.hgt.zip',
		'N39E108.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N39E108.hgt.zip',
		'N39E109.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N39E109.hgt.zip',
		'N39E110.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N39E110.hgt.zip',
		'N39E111.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N39E111.hgt.zip',
		'N39E112.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N39E112.hgt.zip',
		'N39E113.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N39E113.hgt.zip',
		'N39E114.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N39E114.hgt.zip',
		'N39E115.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N39E115.hgt.zip',
		'N39E116.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N39E116.hgt.zip',
		'N39E117.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N39E117.hgt.zip',
		'N39E118.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N39E118.hgt.zip',
		'N39E119.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N39E119.hgt.zip',
		'N39E121.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N39E121.hgt.zip',
		'N39E122.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N39E122.hgt.zip',
		'N39E123.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N39E123.hgt.zip',
		'N39E124.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N39E124.hgt.zip',
		'N39E125.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N39E125.hgt.zip',
		'N39E126.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N39E126.hgt.zip',
		'N39E127.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N39E127.hgt.zip',
		'N39E128.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N39E128.hgt.zip',
		'N39E139.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N39E139.hgt.zip',
		'N39E140.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N39E140.hgt.zip',
		'N39E141.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N39E141.hgt.zip',
		'N39E142.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N39E142.hgt.zip',
		'N39W001.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N39W001.hgt.zip',
		'N39W002.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N39W002.hgt.zip',
		'N39W003.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N39W003.hgt.zip',
		'N39W004.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N39W004.hgt.zip',
		'N39W005.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N39W005.hgt.zip',
		'N39W006.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N39W006.hgt.zip',
		'N39W007.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N39W007.hgt.zip',
		'N39W008.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N39W008.hgt.zip',
		'N39W009.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N39W009.hgt.zip',
		'N39W010.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N39W010.hgt.zip',
		'N39W028.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N39W028.hgt.zip',
		'N39W029.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N39W029.hgt.zip',
		'N39W032.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//N39W032.hgt.zip',
		'N39W075.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N39W075.hgt.zip',
		'N39W076.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N39W076.hgt.zip',
		'N39W077.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N39W077.hgt.zip',
		'N39W078.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N39W078.hgt.zip',
		'N39W079.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N39W079.hgt.zip',
		'N39W080.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N39W080.hgt.zip',
		'N39W081.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N39W081.hgt.zip',
		'N39W082.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N39W082.hgt.zip',
		'N39W083.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N39W083.hgt.zip',
		'N39W084.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N39W084.hgt.zip',
		'N39W085.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N39W085.hgt.zip',
		'N39W086.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N39W086.hgt.zip',
		'N39W087.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N39W087.hgt.zip',
		'N39W088.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N39W088.hgt.zip',
		'N39W089.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N39W089.hgt.zip',
		'N39W090.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N39W090.hgt.zip',
		'N39W091.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N39W091.hgt.zip',
		'N39W092.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N39W092.hgt.zip',
		'N39W093.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N39W093.hgt.zip',
		'N39W094.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N39W094.hgt.zip',
		'N39W095.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N39W095.hgt.zip',
		'N39W096.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N39W096.hgt.zip',
		'N39W097.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N39W097.hgt.zip',
		'N39W098.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N39W098.hgt.zip',
		'N39W099.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N39W099.hgt.zip',
		'N39W100.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N39W100.hgt.zip',
		'N39W101.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N39W101.hgt.zip',
		'N39W102.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N39W102.hgt.zip',
		'N39W103.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N39W103.hgt.zip',
		'N39W104.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N39W104.hgt.zip',
		'N39W105.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N39W105.hgt.zip',
		'N39W106.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N39W106.hgt.zip',
		'N39W107.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N39W107.hgt.zip',
		'N39W108.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N39W108.hgt.zip',
		'N39W109.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N39W109.hgt.zip',
		'N39W110.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N39W110.hgt.zip',
		'N39W111.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N39W111.hgt.zip',
		'N39W112.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N39W112.hgt.zip',
		'N39W113.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N39W113.hgt.zip',
		'N39W114.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N39W114.hgt.zip',
		'N39W115.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N39W115.hgt.zip',
		'N39W116.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N39W116.hgt.zip',
		'N39W117.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N39W117.hgt.zip',
		'N39W118.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N39W118.hgt.zip',
		'N39W119.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N39W119.hgt.zip',
		'N39W120.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N39W120.hgt.zip',
		'N39W121.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N39W121.hgt.zip',
		'N39W122.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N39W122.hgt.zip',
		'N39W123.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N39W123.hgt.zip',
		'N39W124.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N39W124.hgt.zip',
		'N39W125.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N39W125.hgt.zip',
		'N40E000.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N40E000.hgt.zip',
		'N40E003.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N40E003.hgt.zip',
		'N40E004.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N40E004.hgt.zip',
		'N40E008.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N40E008.hgt.zip',
		'N40E009.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N40E009.hgt.zip',
		'N40E012.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N40E012.hgt.zip',
		'N40E013.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N40E013.hgt.zip',
		'N40E014.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N40E014.hgt.zip',
		'N40E015.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N40E015.hgt.zip',
		'N40E016.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N40E016.hgt.zip',
		'N40E017.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N40E017.hgt.zip',
		'N40E018.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N40E018.hgt.zip',
		'N40E019.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N40E019.hgt.zip',
		'N40E020.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N40E020.hgt.zip',
		'N40E021.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N40E021.hgt.zip',
		'N40E022.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N40E022.hgt.zip',
		'N40E023.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N40E023.hgt.zip',
		'N40E024.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N40E024.hgt.zip',
		'N40E025.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N40E025.hgt.zip',
		'N40E026.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N40E026.hgt.zip',
		'N40E027.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N40E027.hgt.zip',
		'N40E028.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N40E028.hgt.zip',
		'N40E029.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N40E029.hgt.zip',
		'N40E030.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N40E030.hgt.zip',
		'N40E031.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N40E031.hgt.zip',
		'N40E032.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N40E032.hgt.zip',
		'N40E033.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N40E033.hgt.zip',
		'N40E034.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N40E034.hgt.zip',
		'N40E035.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N40E035.hgt.zip',
		'N40E036.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N40E036.hgt.zip',
		'N40E037.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N40E037.hgt.zip',
		'N40E038.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N40E038.hgt.zip',
		'N40E039.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N40E039.hgt.zip',
		'N40E040.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N40E040.hgt.zip',
		'N40E041.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N40E041.hgt.zip',
		'N40E042.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N40E042.hgt.zip',
		'N40E043.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N40E043.hgt.zip',
		'N40E044.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N40E044.hgt.zip',
		'N40E045.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N40E045.hgt.zip',
		'N40E046.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N40E046.hgt.zip',
		'N40E047.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N40E047.hgt.zip',
		'N40E048.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N40E048.hgt.zip',
		'N40E049.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N40E049.hgt.zip',
		'N40E050.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N40E050.hgt.zip',
		'N40E052.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N40E052.hgt.zip',
		'N40E053.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N40E053.hgt.zip',
		'N40E054.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N40E054.hgt.zip',
		'N40E055.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N40E055.hgt.zip',
		'N40E056.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N40E056.hgt.zip',
		'N40E057.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N40E057.hgt.zip',
		'N40E058.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N40E058.hgt.zip',
		'N40E059.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N40E059.hgt.zip',
		'N40E060.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N40E060.hgt.zip',
		'N40E061.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N40E061.hgt.zip',
		'N40E062.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N40E062.hgt.zip',
		'N40E063.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N40E063.hgt.zip',
		'N40E064.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N40E064.hgt.zip',
		'N40E065.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N40E065.hgt.zip',
		'N40E066.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N40E066.hgt.zip',
		'N40E067.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N40E067.hgt.zip',
		'N40E068.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N40E068.hgt.zip',
		'N40E069.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N40E069.hgt.zip',
		'N40E070.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N40E070.hgt.zip',
		'N40E071.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N40E071.hgt.zip',
		'N40E072.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N40E072.hgt.zip',
		'N40E073.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N40E073.hgt.zip',
		'N40E074.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N40E074.hgt.zip',
		'N40E075.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N40E075.hgt.zip',
		'N40E076.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N40E076.hgt.zip',
		'N40E077.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N40E077.hgt.zip',
		'N40E078.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N40E078.hgt.zip',
		'N40E079.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N40E079.hgt.zip',
		'N40E080.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N40E080.hgt.zip',
		'N40E081.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N40E081.hgt.zip',
		'N40E082.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N40E082.hgt.zip',
		'N40E083.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N40E083.hgt.zip',
		'N40E084.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N40E084.hgt.zip',
		'N40E085.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N40E085.hgt.zip',
		'N40E086.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N40E086.hgt.zip',
		'N40E087.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N40E087.hgt.zip',
		'N40E088.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N40E088.hgt.zip',
		'N40E089.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N40E089.hgt.zip',
		'N40E090.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N40E090.hgt.zip',
		'N40E091.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N40E091.hgt.zip',
		'N40E092.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N40E092.hgt.zip',
		'N40E093.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N40E093.hgt.zip',
		'N40E094.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N40E094.hgt.zip',
		'N40E095.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N40E095.hgt.zip',
		'N40E096.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N40E096.hgt.zip',
		'N40E097.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N40E097.hgt.zip',
		'N40E098.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N40E098.hgt.zip',
		'N40E099.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N40E099.hgt.zip',
		'N40E100.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N40E100.hgt.zip',
		'N40E101.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N40E101.hgt.zip',
		'N40E102.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N40E102.hgt.zip',
		'N40E103.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N40E103.hgt.zip',
		'N40E104.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N40E104.hgt.zip',
		'N40E105.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N40E105.hgt.zip',
		'N40E106.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N40E106.hgt.zip',
		'N40E107.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N40E107.hgt.zip',
		'N40E108.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N40E108.hgt.zip',
		'N40E109.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N40E109.hgt.zip',
		'N40E110.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N40E110.hgt.zip',
		'N40E111.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N40E111.hgt.zip',
		'N40E112.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N40E112.hgt.zip',
		'N40E113.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N40E113.hgt.zip',
		'N40E114.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N40E114.hgt.zip',
		'N40E115.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N40E115.hgt.zip',
		'N40E116.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N40E116.hgt.zip',
		'N40E117.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N40E117.hgt.zip',
		'N40E118.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N40E118.hgt.zip',
		'N40E119.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N40E119.hgt.zip',
		'N40E120.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N40E120.hgt.zip',
		'N40E121.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N40E121.hgt.zip',
		'N40E122.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N40E122.hgt.zip',
		'N40E123.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N40E123.hgt.zip',
		'N40E124.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N40E124.hgt.zip',
		'N40E125.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N40E125.hgt.zip',
		'N40E126.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N40E126.hgt.zip',
		'N40E127.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N40E127.hgt.zip',
		'N40E128.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N40E128.hgt.zip',
		'N40E129.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N40E129.hgt.zip',
		'N40E139.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N40E139.hgt.zip',
		'N40E140.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N40E140.hgt.zip',
		'N40E141.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N40E141.hgt.zip',
		'N40W001.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N40W001.hgt.zip',
		'N40W002.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N40W002.hgt.zip',
		'N40W003.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N40W003.hgt.zip',
		'N40W004.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N40W004.hgt.zip',
		'N40W005.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N40W005.hgt.zip',
		'N40W006.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N40W006.hgt.zip',
		'N40W007.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N40W007.hgt.zip',
		'N40W008.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N40W008.hgt.zip',
		'N40W009.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N40W009.hgt.zip',
		'N40W073.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N40W073.hgt.zip',
		'N40W074.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N40W074.hgt.zip',
		'N40W075.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N40W075.hgt.zip',
		'N40W076.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N40W076.hgt.zip',
		'N40W077.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N40W077.hgt.zip',
		'N40W078.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N40W078.hgt.zip',
		'N40W079.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N40W079.hgt.zip',
		'N40W080.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N40W080.hgt.zip',
		'N40W081.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N40W081.hgt.zip',
		'N40W082.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N40W082.hgt.zip',
		'N40W083.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N40W083.hgt.zip',
		'N40W084.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N40W084.hgt.zip',
		'N40W085.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N40W085.hgt.zip',
		'N40W086.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N40W086.hgt.zip',
		'N40W087.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N40W087.hgt.zip',
		'N40W088.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N40W088.hgt.zip',
		'N40W089.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N40W089.hgt.zip',
		'N40W090.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N40W090.hgt.zip',
		'N40W091.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N40W091.hgt.zip',
		'N40W092.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N40W092.hgt.zip',
		'N40W093.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N40W093.hgt.zip',
		'N40W094.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N40W094.hgt.zip',
		'N40W095.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N40W095.hgt.zip',
		'N40W096.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N40W096.hgt.zip',
		'N40W097.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N40W097.hgt.zip',
		'N40W098.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N40W098.hgt.zip',
		'N40W099.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N40W099.hgt.zip',
		'N40W100.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N40W100.hgt.zip',
		'N40W101.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N40W101.hgt.zip',
		'N40W102.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N40W102.hgt.zip',
		'N40W103.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N40W103.hgt.zip',
		'N40W104.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N40W104.hgt.zip',
		'N40W105.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N40W105.hgt.zip',
		'N40W106.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N40W106.hgt.zip',
		'N40W107.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N40W107.hgt.zip',
		'N40W108.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N40W108.hgt.zip',
		'N40W109.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N40W109.hgt.zip',
		'N40W110.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N40W110.hgt.zip',
		'N40W111.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N40W111.hgt.zip',
		'N40W112.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N40W112.hgt.zip',
		'N40W113.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N40W113.hgt.zip',
		'N40W114.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N40W114.hgt.zip',
		'N40W115.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N40W115.hgt.zip',
		'N40W116.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N40W116.hgt.zip',
		'N40W117.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N40W117.hgt.zip',
		'N40W118.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N40W118.hgt.zip',
		'N40W119.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N40W119.hgt.zip',
		'N40W120.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N40W120.hgt.zip',
		'N40W121.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N40W121.hgt.zip',
		'N40W122.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N40W122.hgt.zip',
		'N40W123.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N40W123.hgt.zip',
		'N40W124.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N40W124.hgt.zip',
		'N40W125.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N40W125.hgt.zip',
		'N41E000.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N41E000.hgt.zip',
		'N41E001.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N41E001.hgt.zip',
		'N41E002.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N41E002.hgt.zip',
		'N41E003.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N41E003.hgt.zip',
		'N41E008.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N41E008.hgt.zip',
		'N41E009.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N41E009.hgt.zip',
		'N41E011.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N41E011.hgt.zip',
		'N41E012.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N41E012.hgt.zip',
		'N41E013.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N41E013.hgt.zip',
		'N41E014.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N41E014.hgt.zip',
		'N41E015.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N41E015.hgt.zip',
		'N41E016.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N41E016.hgt.zip',
		'N41E017.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N41E017.hgt.zip',
		'N41E019.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N41E019.hgt.zip',
		'N41E020.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N41E020.hgt.zip',
		'N41E021.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N41E021.hgt.zip',
		'N41E022.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N41E022.hgt.zip',
		'N41E023.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N41E023.hgt.zip',
		'N41E024.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N41E024.hgt.zip',
		'N41E025.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N41E025.hgt.zip',
		'N41E026.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N41E026.hgt.zip',
		'N41E027.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N41E027.hgt.zip',
		'N41E028.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N41E028.hgt.zip',
		'N41E029.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N41E029.hgt.zip',
		'N41E030.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N41E030.hgt.zip',
		'N41E031.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N41E031.hgt.zip',
		'N41E032.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N41E032.hgt.zip',
		'N41E033.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N41E033.hgt.zip',
		'N41E034.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N41E034.hgt.zip',
		'N41E035.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N41E035.hgt.zip',
		'N41E036.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N41E036.hgt.zip',
		'N41E037.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N41E037.hgt.zip',
		'N41E038.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N41E038.hgt.zip',
		'N41E039.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N41E039.hgt.zip',
		'N41E040.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N41E040.hgt.zip',
		'N41E041.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N41E041.hgt.zip',
		'N41E042.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N41E042.hgt.zip',
		'N41E043.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N41E043.hgt.zip',
		'N41E044.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N41E044.hgt.zip',
		'N41E045.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N41E045.hgt.zip',
		'N41E046.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N41E046.hgt.zip',
		'N41E047.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N41E047.hgt.zip',
		'N41E048.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N41E048.hgt.zip',
		'N41E049.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N41E049.hgt.zip',
		'N41E052.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N41E052.hgt.zip',
		'N41E053.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N41E053.hgt.zip',
		'N41E054.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N41E054.hgt.zip',
		'N41E055.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N41E055.hgt.zip',
		'N41E056.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N41E056.hgt.zip',
		'N41E057.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N41E057.hgt.zip',
		'N41E058.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N41E058.hgt.zip',
		'N41E059.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N41E059.hgt.zip',
		'N41E060.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N41E060.hgt.zip',
		'N41E061.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N41E061.hgt.zip',
		'N41E062.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N41E062.hgt.zip',
		'N41E063.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N41E063.hgt.zip',
		'N41E064.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N41E064.hgt.zip',
		'N41E065.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N41E065.hgt.zip',
		'N41E066.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N41E066.hgt.zip',
		'N41E067.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N41E067.hgt.zip',
		'N41E068.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N41E068.hgt.zip',
		'N41E069.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N41E069.hgt.zip',
		'N41E070.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N41E070.hgt.zip',
		'N41E071.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N41E071.hgt.zip',
		'N41E072.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N41E072.hgt.zip',
		'N41E073.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N41E073.hgt.zip',
		'N41E074.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N41E074.hgt.zip',
		'N41E075.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N41E075.hgt.zip',
		'N41E076.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N41E076.hgt.zip',
		'N41E077.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N41E077.hgt.zip',
		'N41E078.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N41E078.hgt.zip',
		'N41E079.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N41E079.hgt.zip',
		'N41E080.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N41E080.hgt.zip',
		'N41E081.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N41E081.hgt.zip',
		'N41E082.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N41E082.hgt.zip',
		'N41E083.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N41E083.hgt.zip',
		'N41E084.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N41E084.hgt.zip',
		'N41E085.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N41E085.hgt.zip',
		'N41E086.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N41E086.hgt.zip',
		'N41E087.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N41E087.hgt.zip',
		'N41E088.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N41E088.hgt.zip',
		'N41E089.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N41E089.hgt.zip',
		'N41E090.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N41E090.hgt.zip',
		'N41E091.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N41E091.hgt.zip',
		'N41E092.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N41E092.hgt.zip',
		'N41E093.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N41E093.hgt.zip',
		'N41E094.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N41E094.hgt.zip',
		'N41E095.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N41E095.hgt.zip',
		'N41E096.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N41E096.hgt.zip',
		'N41E097.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N41E097.hgt.zip',
		'N41E098.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N41E098.hgt.zip',
		'N41E099.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N41E099.hgt.zip',
		'N41E100.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N41E100.hgt.zip',
		'N41E101.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N41E101.hgt.zip',
		'N41E102.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N41E102.hgt.zip',
		'N41E103.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N41E103.hgt.zip',
		'N41E104.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N41E104.hgt.zip',
		'N41E105.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N41E105.hgt.zip',
		'N41E106.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N41E106.hgt.zip',
		'N41E107.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N41E107.hgt.zip',
		'N41E108.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N41E108.hgt.zip',
		'N41E109.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N41E109.hgt.zip',
		'N41E110.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N41E110.hgt.zip',
		'N41E111.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N41E111.hgt.zip',
		'N41E112.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N41E112.hgt.zip',
		'N41E113.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N41E113.hgt.zip',
		'N41E114.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N41E114.hgt.zip',
		'N41E115.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N41E115.hgt.zip',
		'N41E116.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N41E116.hgt.zip',
		'N41E117.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N41E117.hgt.zip',
		'N41E118.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N41E118.hgt.zip',
		'N41E119.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N41E119.hgt.zip',
		'N41E120.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N41E120.hgt.zip',
		'N41E121.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N41E121.hgt.zip',
		'N41E122.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N41E122.hgt.zip',
		'N41E123.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N41E123.hgt.zip',
		'N41E124.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N41E124.hgt.zip',
		'N41E125.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N41E125.hgt.zip',
		'N41E126.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N41E126.hgt.zip',
		'N41E127.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N41E127.hgt.zip',
		'N41E128.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N41E128.hgt.zip',
		'N41E129.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N41E129.hgt.zip',
		'N41E130.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N41E130.hgt.zip',
		'N41E139.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N41E139.hgt.zip',
		'N41E140.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N41E140.hgt.zip',
		'N41E141.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N41E141.hgt.zip',
		'N41E143.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N41E143.hgt.zip',
		'N41W001.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N41W001.hgt.zip',
		'N41W002.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N41W002.hgt.zip',
		'N41W003.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N41W003.hgt.zip',
		'N41W004.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N41W004.hgt.zip',
		'N41W005.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N41W005.hgt.zip',
		'N41W006.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N41W006.hgt.zip',
		'N41W007.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N41W007.hgt.zip',
		'N41W008.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N41W008.hgt.zip',
		'N41W009.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N41W009.hgt.zip',
		'N41W070.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N41W070.hgt.zip',
		'N41W071.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N41W071.hgt.zip',
		'N41W072.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N41W072.hgt.zip',
		'N41W073.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N41W073.hgt.zip',
		'N41W074.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N41W074.hgt.zip',
		'N41W075.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N41W075.hgt.zip',
		'N41W076.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N41W076.hgt.zip',
		'N41W077.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N41W077.hgt.zip',
		'N41W078.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N41W078.hgt.zip',
		'N41W079.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N41W079.hgt.zip',
		'N41W080.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N41W080.hgt.zip',
		'N41W081.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N41W081.hgt.zip',
		'N41W082.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N41W082.hgt.zip',
		'N41W083.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N41W083.hgt.zip',
		'N41W084.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N41W084.hgt.zip',
		'N41W085.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N41W085.hgt.zip',
		'N41W086.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N41W086.hgt.zip',
		'N41W087.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N41W087.hgt.zip',
		'N41W088.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N41W088.hgt.zip',
		'N41W089.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N41W089.hgt.zip',
		'N41W090.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N41W090.hgt.zip',
		'N41W091.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N41W091.hgt.zip',
		'N41W092.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N41W092.hgt.zip',
		'N41W093.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N41W093.hgt.zip',
		'N41W094.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N41W094.hgt.zip',
		'N41W095.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N41W095.hgt.zip',
		'N41W096.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N41W096.hgt.zip',
		'N41W097.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N41W097.hgt.zip',
		'N41W098.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N41W098.hgt.zip',
		'N41W099.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N41W099.hgt.zip',
		'N41W100.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N41W100.hgt.zip',
		'N41W101.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N41W101.hgt.zip',
		'N41W102.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N41W102.hgt.zip',
		'N41W103.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N41W103.hgt.zip',
		'N41W104.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N41W104.hgt.zip',
		'N41W105.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N41W105.hgt.zip',
		'N41W106.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N41W106.hgt.zip',
		'N41W107.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N41W107.hgt.zip',
		'N41W108.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N41W108.hgt.zip',
		'N41W109.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N41W109.hgt.zip',
		'N41W110.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N41W110.hgt.zip',
		'N41W111.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N41W111.hgt.zip',
		'N41W112.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N41W112.hgt.zip',
		'N41W113.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N41W113.hgt.zip',
		'N41W114.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N41W114.hgt.zip',
		'N41W115.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N41W115.hgt.zip',
		'N41W116.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N41W116.hgt.zip',
		'N41W117.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N41W117.hgt.zip',
		'N41W118.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N41W118.hgt.zip',
		'N41W119.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N41W119.hgt.zip',
		'N41W120.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N41W120.hgt.zip',
		'N41W121.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N41W121.hgt.zip',
		'N41W122.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N41W122.hgt.zip',
		'N41W123.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N41W123.hgt.zip',
		'N41W124.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N41W124.hgt.zip',
		'N41W125.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N41W125.hgt.zip',
		'N42E000.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N42E000.hgt.zip',
		'N42E001.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N42E001.hgt.zip',
		'N42E002.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N42E002.hgt.zip',
		'N42E003.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N42E003.hgt.zip',
		'N42E006.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N42E006.hgt.zip',
		'N42E008.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N42E008.hgt.zip',
		'N42E009.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N42E009.hgt.zip',
		'N42E010.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N42E010.hgt.zip',
		'N42E011.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N42E011.hgt.zip',
		'N42E012.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N42E012.hgt.zip',
		'N42E013.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N42E013.hgt.zip',
		'N42E014.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N42E014.hgt.zip',
		'N42E015.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N42E015.hgt.zip',
		'N42E016.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N42E016.hgt.zip',
		'N42E017.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N42E017.hgt.zip',
		'N42E018.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N42E018.hgt.zip',
		'N42E019.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N42E019.hgt.zip',
		'N42E020.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N42E020.hgt.zip',
		'N42E021.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N42E021.hgt.zip',
		'N42E022.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N42E022.hgt.zip',
		'N42E023.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N42E023.hgt.zip',
		'N42E024.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N42E024.hgt.zip',
		'N42E025.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N42E025.hgt.zip',
		'N42E026.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N42E026.hgt.zip',
		'N42E027.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N42E027.hgt.zip',
		'N42E028.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N42E028.hgt.zip',
		'N42E033.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N42E033.hgt.zip',
		'N42E034.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N42E034.hgt.zip',
		'N42E035.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N42E035.hgt.zip',
		'N42E040.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N42E040.hgt.zip',
		'N42E041.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N42E041.hgt.zip',
		'N42E042.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N42E042.hgt.zip',
		'N42E043.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N42E043.hgt.zip',
		'N42E044.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N42E044.hgt.zip',
		'N42E045.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N42E045.hgt.zip',
		'N42E046.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N42E046.hgt.zip',
		'N42E047.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N42E047.hgt.zip',
		'N42E048.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N42E048.hgt.zip',
		'N42E051.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N42E051.hgt.zip',
		'N42E052.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N42E052.hgt.zip',
		'N42E053.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N42E053.hgt.zip',
		'N42E054.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N42E054.hgt.zip',
		'N42E055.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N42E055.hgt.zip',
		'N42E056.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N42E056.hgt.zip',
		'N42E057.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N42E057.hgt.zip',
		'N42E058.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N42E058.hgt.zip',
		'N42E059.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N42E059.hgt.zip',
		'N42E060.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N42E060.hgt.zip',
		'N42E061.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N42E061.hgt.zip',
		'N42E062.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N42E062.hgt.zip',
		'N42E063.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N42E063.hgt.zip',
		'N42E064.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N42E064.hgt.zip',
		'N42E065.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N42E065.hgt.zip',
		'N42E066.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N42E066.hgt.zip',
		'N42E067.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N42E067.hgt.zip',
		'N42E068.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N42E068.hgt.zip',
		'N42E069.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N42E069.hgt.zip',
		'N42E070.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N42E070.hgt.zip',
		'N42E071.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N42E071.hgt.zip',
		'N42E072.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N42E072.hgt.zip',
		'N42E073.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N42E073.hgt.zip',
		'N42E074.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N42E074.hgt.zip',
		'N42E075.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N42E075.hgt.zip',
		'N42E076.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N42E076.hgt.zip',
		'N42E077.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N42E077.hgt.zip',
		'N42E078.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N42E078.hgt.zip',
		'N42E079.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N42E079.hgt.zip',
		'N42E080.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N42E080.hgt.zip',
		'N42E081.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N42E081.hgt.zip',
		'N42E082.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N42E082.hgt.zip',
		'N42E083.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N42E083.hgt.zip',
		'N42E084.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N42E084.hgt.zip',
		'N42E085.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N42E085.hgt.zip',
		'N42E086.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N42E086.hgt.zip',
		'N42E087.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N42E087.hgt.zip',
		'N42E088.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N42E088.hgt.zip',
		'N42E089.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N42E089.hgt.zip',
		'N42E090.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N42E090.hgt.zip',
		'N42E091.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N42E091.hgt.zip',
		'N42E092.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N42E092.hgt.zip',
		'N42E093.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N42E093.hgt.zip',
		'N42E094.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N42E094.hgt.zip',
		'N42E095.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N42E095.hgt.zip',
		'N42E096.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N42E096.hgt.zip',
		'N42E097.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N42E097.hgt.zip',
		'N42E098.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N42E098.hgt.zip',
		'N42E099.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N42E099.hgt.zip',
		'N42E100.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N42E100.hgt.zip',
		'N42E101.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N42E101.hgt.zip',
		'N42E102.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N42E102.hgt.zip',
		'N42E103.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N42E103.hgt.zip',
		'N42E104.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N42E104.hgt.zip',
		'N42E105.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N42E105.hgt.zip',
		'N42E106.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N42E106.hgt.zip',
		'N42E107.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N42E107.hgt.zip',
		'N42E108.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N42E108.hgt.zip',
		'N42E109.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N42E109.hgt.zip',
		'N42E110.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N42E110.hgt.zip',
		'N42E111.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N42E111.hgt.zip',
		'N42E112.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N42E112.hgt.zip',
		'N42E113.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N42E113.hgt.zip',
		'N42E114.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N42E114.hgt.zip',
		'N42E115.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N42E115.hgt.zip',
		'N42E116.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N42E116.hgt.zip',
		'N42E117.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N42E117.hgt.zip',
		'N42E118.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N42E118.hgt.zip',
		'N42E119.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N42E119.hgt.zip',
		'N42E120.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N42E120.hgt.zip',
		'N42E121.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N42E121.hgt.zip',
		'N42E122.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N42E122.hgt.zip',
		'N42E123.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N42E123.hgt.zip',
		'N42E124.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N42E124.hgt.zip',
		'N42E125.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N42E125.hgt.zip',
		'N42E126.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N42E126.hgt.zip',
		'N42E127.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N42E127.hgt.zip',
		'N42E128.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N42E128.hgt.zip',
		'N42E129.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N42E129.hgt.zip',
		'N42E130.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N42E130.hgt.zip',
		'N42E131.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N42E131.hgt.zip',
		'N42E132.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N42E132.hgt.zip',
		'N42E133.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N42E133.hgt.zip',
		'N42E134.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N42E134.hgt.zip',
		'N42E139.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N42E139.hgt.zip',
		'N42E140.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N42E140.hgt.zip',
		'N42E141.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N42E141.hgt.zip',
		'N42E142.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N42E142.hgt.zip',
		'N42E143.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N42E143.hgt.zip',
		'N42E144.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N42E144.hgt.zip',
		'N42E145.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N42E145.hgt.zip',
		'N42W001.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N42W001.hgt.zip',
		'N42W002.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N42W002.hgt.zip',
		'N42W003.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N42W003.hgt.zip',
		'N42W004.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N42W004.hgt.zip',
		'N42W005.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N42W005.hgt.zip',
		'N42W006.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N42W006.hgt.zip',
		'N42W007.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N42W007.hgt.zip',
		'N42W008.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N42W008.hgt.zip',
		'N42W009.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N42W009.hgt.zip',
		'N42W010.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N42W010.hgt.zip',
		'N42W071.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N42W071.hgt.zip',
		'N42W072.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N42W072.hgt.zip',
		'N42W073.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N42W073.hgt.zip',
		'N42W074.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N42W074.hgt.zip',
		'N42W075.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N42W075.hgt.zip',
		'N42W076.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N42W076.hgt.zip',
		'N42W077.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N42W077.hgt.zip',
		'N42W078.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N42W078.hgt.zip',
		'N42W079.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N42W079.hgt.zip',
		'N42W080.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N42W080.hgt.zip',
		'N42W081.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N42W081.hgt.zip',
		'N42W082.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N42W082.hgt.zip',
		'N42W083.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N42W083.hgt.zip',
		'N42W084.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N42W084.hgt.zip',
		'N42W085.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N42W085.hgt.zip',
		'N42W086.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N42W086.hgt.zip',
		'N42W087.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N42W087.hgt.zip',
		'N42W088.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N42W088.hgt.zip',
		'N42W089.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N42W089.hgt.zip',
		'N42W090.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N42W090.hgt.zip',
		'N42W091.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N42W091.hgt.zip',
		'N42W092.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N42W092.hgt.zip',
		'N42W093.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N42W093.hgt.zip',
		'N42W094.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N42W094.hgt.zip',
		'N42W095.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N42W095.hgt.zip',
		'N42W096.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N42W096.hgt.zip',
		'N42W097.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N42W097.hgt.zip',
		'N42W098.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N42W098.hgt.zip',
		'N42W099.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N42W099.hgt.zip',
		'N42W100.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N42W100.hgt.zip',
		'N42W101.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N42W101.hgt.zip',
		'N42W102.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N42W102.hgt.zip',
		'N42W103.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N42W103.hgt.zip',
		'N42W104.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N42W104.hgt.zip',
		'N42W105.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N42W105.hgt.zip',
		'N42W106.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N42W106.hgt.zip',
		'N42W107.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N42W107.hgt.zip',
		'N42W108.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N42W108.hgt.zip',
		'N42W109.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N42W109.hgt.zip',
		'N42W110.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N42W110.hgt.zip',
		'N42W111.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N42W111.hgt.zip',
		'N42W112.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N42W112.hgt.zip',
		'N42W113.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N42W113.hgt.zip',
		'N42W114.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N42W114.hgt.zip',
		'N42W115.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N42W115.hgt.zip',
		'N42W116.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N42W116.hgt.zip',
		'N42W117.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N42W117.hgt.zip',
		'N42W118.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N42W118.hgt.zip',
		'N42W119.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N42W119.hgt.zip',
		'N42W120.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N42W120.hgt.zip',
		'N42W121.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N42W121.hgt.zip',
		'N42W122.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N42W122.hgt.zip',
		'N42W123.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N42W123.hgt.zip',
		'N42W124.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N42W124.hgt.zip',
		'N42W125.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N42W125.hgt.zip',
		'N43E000.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N43E000.hgt.zip',
		'N43E001.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N43E001.hgt.zip',
		'N43E002.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N43E002.hgt.zip',
		'N43E003.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N43E003.hgt.zip',
		'N43E004.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N43E004.hgt.zip',
		'N43E005.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N43E005.hgt.zip',
		'N43E006.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N43E006.hgt.zip',
		'N43E007.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N43E007.hgt.zip',
		'N43E008.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N43E008.hgt.zip',
		'N43E009.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N43E009.hgt.zip',
		'N43E010.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N43E010.hgt.zip',
		'N43E011.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N43E011.hgt.zip',
		'N43E012.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N43E012.hgt.zip',
		'N43E013.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N43E013.hgt.zip',
		'N43E015.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N43E015.hgt.zip',
		'N43E016.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N43E016.hgt.zip',
		'N43E017.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N43E017.hgt.zip',
		'N43E018.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N43E018.hgt.zip',
		'N43E019.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N43E019.hgt.zip',
		'N43E020.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N43E020.hgt.zip',
		'N43E021.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N43E021.hgt.zip',
		'N43E022.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N43E022.hgt.zip',
		'N43E023.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N43E023.hgt.zip',
		'N43E024.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N43E024.hgt.zip',
		'N43E025.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N43E025.hgt.zip',
		'N43E026.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N43E026.hgt.zip',
		'N43E027.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N43E027.hgt.zip',
		'N43E028.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N43E028.hgt.zip',
		'N43E039.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N43E039.hgt.zip',
		'N43E040.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N43E040.hgt.zip',
		'N43E041.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N43E041.hgt.zip',
		'N43E042.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N43E042.hgt.zip',
		'N43E043.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N43E043.hgt.zip',
		'N43E044.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N43E044.hgt.zip',
		'N43E045.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N43E045.hgt.zip',
		'N43E046.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N43E046.hgt.zip',
		'N43E047.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N43E047.hgt.zip',
		'N43E050.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N43E050.hgt.zip',
		'N43E051.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N43E051.hgt.zip',
		'N43E052.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N43E052.hgt.zip',
		'N43E053.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N43E053.hgt.zip',
		'N43E054.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N43E054.hgt.zip',
		'N43E055.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N43E055.hgt.zip',
		'N43E056.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N43E056.hgt.zip',
		'N43E057.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N43E057.hgt.zip',
		'N43E058.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N43E058.hgt.zip',
		'N43E059.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N43E059.hgt.zip',
		'N43E060.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N43E060.hgt.zip',
		'N43E061.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N43E061.hgt.zip',
		'N43E062.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N43E062.hgt.zip',
		'N43E063.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N43E063.hgt.zip',
		'N43E064.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N43E064.hgt.zip',
		'N43E065.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N43E065.hgt.zip',
		'N43E066.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N43E066.hgt.zip',
		'N43E067.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N43E067.hgt.zip',
		'N43E068.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N43E068.hgt.zip',
		'N43E069.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N43E069.hgt.zip',
		'N43E070.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N43E070.hgt.zip',
		'N43E071.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N43E071.hgt.zip',
		'N43E072.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N43E072.hgt.zip',
		'N43E073.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N43E073.hgt.zip',
		'N43E074.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N43E074.hgt.zip',
		'N43E075.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N43E075.hgt.zip',
		'N43E076.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N43E076.hgt.zip',
		'N43E077.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N43E077.hgt.zip',
		'N43E078.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N43E078.hgt.zip',
		'N43E079.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N43E079.hgt.zip',
		'N43E080.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N43E080.hgt.zip',
		'N43E081.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N43E081.hgt.zip',
		'N43E082.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N43E082.hgt.zip',
		'N43E083.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N43E083.hgt.zip',
		'N43E084.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N43E084.hgt.zip',
		'N43E085.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N43E085.hgt.zip',
		'N43E086.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N43E086.hgt.zip',
		'N43E087.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N43E087.hgt.zip',
		'N43E088.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N43E088.hgt.zip',
		'N43E089.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N43E089.hgt.zip',
		'N43E090.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N43E090.hgt.zip',
		'N43E091.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N43E091.hgt.zip',
		'N43E092.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N43E092.hgt.zip',
		'N43E093.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N43E093.hgt.zip',
		'N43E094.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N43E094.hgt.zip',
		'N43E095.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N43E095.hgt.zip',
		'N43E096.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N43E096.hgt.zip',
		'N43E097.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N43E097.hgt.zip',
		'N43E098.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N43E098.hgt.zip',
		'N43E099.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N43E099.hgt.zip',
		'N43E100.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N43E100.hgt.zip',
		'N43E101.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N43E101.hgt.zip',
		'N43E102.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N43E102.hgt.zip',
		'N43E103.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N43E103.hgt.zip',
		'N43E104.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N43E104.hgt.zip',
		'N43E105.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N43E105.hgt.zip',
		'N43E106.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N43E106.hgt.zip',
		'N43E107.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N43E107.hgt.zip',
		'N43E108.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N43E108.hgt.zip',
		'N43E109.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N43E109.hgt.zip',
		'N43E110.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N43E110.hgt.zip',
		'N43E111.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N43E111.hgt.zip',
		'N43E112.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N43E112.hgt.zip',
		'N43E113.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N43E113.hgt.zip',
		'N43E114.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N43E114.hgt.zip',
		'N43E115.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N43E115.hgt.zip',
		'N43E116.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N43E116.hgt.zip',
		'N43E117.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N43E117.hgt.zip',
		'N43E118.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N43E118.hgt.zip',
		'N43E119.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N43E119.hgt.zip',
		'N43E120.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N43E120.hgt.zip',
		'N43E121.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N43E121.hgt.zip',
		'N43E122.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N43E122.hgt.zip',
		'N43E123.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N43E123.hgt.zip',
		'N43E124.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N43E124.hgt.zip',
		'N43E125.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N43E125.hgt.zip',
		'N43E126.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N43E126.hgt.zip',
		'N43E127.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N43E127.hgt.zip',
		'N43E128.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N43E128.hgt.zip',
		'N43E129.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N43E129.hgt.zip',
		'N43E130.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N43E130.hgt.zip',
		'N43E131.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N43E131.hgt.zip',
		'N43E132.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N43E132.hgt.zip',
		'N43E133.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N43E133.hgt.zip',
		'N43E134.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N43E134.hgt.zip',
		'N43E135.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N43E135.hgt.zip',
		'N43E140.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N43E140.hgt.zip',
		'N43E141.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N43E141.hgt.zip',
		'N43E142.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N43E142.hgt.zip',
		'N43E143.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N43E143.hgt.zip',
		'N43E144.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N43E144.hgt.zip',
		'N43E145.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N43E145.hgt.zip',
		'N43E146.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N43E146.hgt.zip',
		'N43W001.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N43W001.hgt.zip',
		'N43W002.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N43W002.hgt.zip',
		'N43W003.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N43W003.hgt.zip',
		'N43W004.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N43W004.hgt.zip',
		'N43W005.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N43W005.hgt.zip',
		'N43W006.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N43W006.hgt.zip',
		'N43W007.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N43W007.hgt.zip',
		'N43W008.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N43W008.hgt.zip',
		'N43W009.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N43W009.hgt.zip',
		'N43W010.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N43W010.hgt.zip',
		'N43W060.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N43W060.hgt.zip',
		'N43W061.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N43W061.hgt.zip',
		'N43W065.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N43W065.hgt.zip',
		'N43W066.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N43W066.hgt.zip',
		'N43W067.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N43W067.hgt.zip',
		'N43W069.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N43W069.hgt.zip',
		'N43W070.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N43W070.hgt.zip',
		'N43W071.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N43W071.hgt.zip',
		'N43W072.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N43W072.hgt.zip',
		'N43W073.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N43W073.hgt.zip',
		'N43W074.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N43W074.hgt.zip',
		'N43W075.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N43W075.hgt.zip',
		'N43W076.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N43W076.hgt.zip',
		'N43W077.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N43W077.hgt.zip',
		'N43W078.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N43W078.hgt.zip',
		'N43W079.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N43W079.hgt.zip',
		'N43W080.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N43W080.hgt.zip',
		'N43W081.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N43W081.hgt.zip',
		'N43W082.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N43W082.hgt.zip',
		'N43W083.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N43W083.hgt.zip',
		'N43W084.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N43W084.hgt.zip',
		'N43W085.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N43W085.hgt.zip',
		'N43W086.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N43W086.hgt.zip',
		'N43W087.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N43W087.hgt.zip',
		'N43W088.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N43W088.hgt.zip',
		'N43W089.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N43W089.hgt.zip',
		'N43W090.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N43W090.hgt.zip',
		'N43W091.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N43W091.hgt.zip',
		'N43W092.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N43W092.hgt.zip',
		'N43W093.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N43W093.hgt.zip',
		'N43W094.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N43W094.hgt.zip',
		'N43W095.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N43W095.hgt.zip',
		'N43W096.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N43W096.hgt.zip',
		'N43W097.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N43W097.hgt.zip',
		'N43W098.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N43W098.hgt.zip',
		'N43W099.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N43W099.hgt.zip',
		'N43W100.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N43W100.hgt.zip',
		'N43W101.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N43W101.hgt.zip',
		'N43W102.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N43W102.hgt.zip',
		'N43W103.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N43W103.hgt.zip',
		'N43W104.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N43W104.hgt.zip',
		'N43W105.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N43W105.hgt.zip',
		'N43W106.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N43W106.hgt.zip',
		'N43W107.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N43W107.hgt.zip',
		'N43W108.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N43W108.hgt.zip',
		'N43W109.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N43W109.hgt.zip',
		'N43W110.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N43W110.hgt.zip',
		'N43W111.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N43W111.hgt.zip',
		'N43W112.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N43W112.hgt.zip',
		'N43W113.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N43W113.hgt.zip',
		'N43W114.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N43W114.hgt.zip',
		'N43W115.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N43W115.hgt.zip',
		'N43W116.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N43W116.hgt.zip',
		'N43W117.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N43W117.hgt.zip',
		'N43W118.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N43W118.hgt.zip',
		'N43W119.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N43W119.hgt.zip',
		'N43W120.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N43W120.hgt.zip',
		'N43W121.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N43W121.hgt.zip',
		'N43W122.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N43W122.hgt.zip',
		'N43W123.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N43W123.hgt.zip',
		'N43W124.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N43W124.hgt.zip',
		'N43W125.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N43W125.hgt.zip',
		'N44E000.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N44E000.hgt.zip',
		'N44E001.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N44E001.hgt.zip',
		'N44E002.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N44E002.hgt.zip',
		'N44E003.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N44E003.hgt.zip',
		'N44E004.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N44E004.hgt.zip',
		'N44E005.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N44E005.hgt.zip',
		'N44E006.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N44E006.hgt.zip',
		'N44E007.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N44E007.hgt.zip',
		'N44E008.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N44E008.hgt.zip',
		'N44E009.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N44E009.hgt.zip',
		'N44E010.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N44E010.hgt.zip',
		'N44E011.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N44E011.hgt.zip',
		'N44E012.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N44E012.hgt.zip',
		'N44E013.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N44E013.hgt.zip',
		'N44E014.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N44E014.hgt.zip',
		'N44E015.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N44E015.hgt.zip',
		'N44E016.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N44E016.hgt.zip',
		'N44E017.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N44E017.hgt.zip',
		'N44E018.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N44E018.hgt.zip',
		'N44E019.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N44E019.hgt.zip',
		'N44E020.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N44E020.hgt.zip',
		'N44E021.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N44E021.hgt.zip',
		'N44E022.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N44E022.hgt.zip',
		'N44E023.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N44E023.hgt.zip',
		'N44E024.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N44E024.hgt.zip',
		'N44E025.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N44E025.hgt.zip',
		'N44E026.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N44E026.hgt.zip',
		'N44E027.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N44E027.hgt.zip',
		'N44E028.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N44E028.hgt.zip',
		'N44E029.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N44E029.hgt.zip',
		'N44E033.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N44E033.hgt.zip',
		'N44E034.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N44E034.hgt.zip',
		'N44E035.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N44E035.hgt.zip',
		'N44E037.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N44E037.hgt.zip',
		'N44E038.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N44E038.hgt.zip',
		'N44E039.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N44E039.hgt.zip',
		'N44E040.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N44E040.hgt.zip',
		'N44E041.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N44E041.hgt.zip',
		'N44E042.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N44E042.hgt.zip',
		'N44E043.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N44E043.hgt.zip',
		'N44E044.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N44E044.hgt.zip',
		'N44E045.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N44E045.hgt.zip',
		'N44E046.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N44E046.hgt.zip',
		'N44E047.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N44E047.hgt.zip',
		'N44E050.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N44E050.hgt.zip',
		'N44E051.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N44E051.hgt.zip',
		'N44E052.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N44E052.hgt.zip',
		'N44E053.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N44E053.hgt.zip',
		'N44E054.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N44E054.hgt.zip',
		'N44E055.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N44E055.hgt.zip',
		'N44E056.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N44E056.hgt.zip',
		'N44E057.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N44E057.hgt.zip',
		'N44E058.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N44E058.hgt.zip',
		'N44E059.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N44E059.hgt.zip',
		'N44E060.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N44E060.hgt.zip',
		'N44E061.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N44E061.hgt.zip',
		'N44E062.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N44E062.hgt.zip',
		'N44E063.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N44E063.hgt.zip',
		'N44E064.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N44E064.hgt.zip',
		'N44E065.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N44E065.hgt.zip',
		'N44E066.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N44E066.hgt.zip',
		'N44E067.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N44E067.hgt.zip',
		'N44E068.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N44E068.hgt.zip',
		'N44E069.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N44E069.hgt.zip',
		'N44E070.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N44E070.hgt.zip',
		'N44E071.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N44E071.hgt.zip',
		'N44E072.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N44E072.hgt.zip',
		'N44E073.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N44E073.hgt.zip',
		'N44E074.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N44E074.hgt.zip',
		'N44E075.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N44E075.hgt.zip',
		'N44E076.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N44E076.hgt.zip',
		'N44E077.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N44E077.hgt.zip',
		'N44E078.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N44E078.hgt.zip',
		'N44E079.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N44E079.hgt.zip',
		'N44E080.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N44E080.hgt.zip',
		'N44E081.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N44E081.hgt.zip',
		'N44E082.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N44E082.hgt.zip',
		'N44E083.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N44E083.hgt.zip',
		'N44E084.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N44E084.hgt.zip',
		'N44E085.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N44E085.hgt.zip',
		'N44E086.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N44E086.hgt.zip',
		'N44E087.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N44E087.hgt.zip',
		'N44E088.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N44E088.hgt.zip',
		'N44E089.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N44E089.hgt.zip',
		'N44E090.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N44E090.hgt.zip',
		'N44E091.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N44E091.hgt.zip',
		'N44E092.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N44E092.hgt.zip',
		'N44E093.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N44E093.hgt.zip',
		'N44E094.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N44E094.hgt.zip',
		'N44E095.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N44E095.hgt.zip',
		'N44E096.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N44E096.hgt.zip',
		'N44E097.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N44E097.hgt.zip',
		'N44E098.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N44E098.hgt.zip',
		'N44E099.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N44E099.hgt.zip',
		'N44E100.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N44E100.hgt.zip',
		'N44E101.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N44E101.hgt.zip',
		'N44E102.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N44E102.hgt.zip',
		'N44E103.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N44E103.hgt.zip',
		'N44E104.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N44E104.hgt.zip',
		'N44E105.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N44E105.hgt.zip',
		'N44E106.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N44E106.hgt.zip',
		'N44E107.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N44E107.hgt.zip',
		'N44E108.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N44E108.hgt.zip',
		'N44E109.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N44E109.hgt.zip',
		'N44E110.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N44E110.hgt.zip',
		'N44E111.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N44E111.hgt.zip',
		'N44E112.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N44E112.hgt.zip',
		'N44E113.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N44E113.hgt.zip',
		'N44E114.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N44E114.hgt.zip',
		'N44E115.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N44E115.hgt.zip',
		'N44E116.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N44E116.hgt.zip',
		'N44E117.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N44E117.hgt.zip',
		'N44E118.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N44E118.hgt.zip',
		'N44E119.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N44E119.hgt.zip',
		'N44E120.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N44E120.hgt.zip',
		'N44E121.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N44E121.hgt.zip',
		'N44E122.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N44E122.hgt.zip',
		'N44E123.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N44E123.hgt.zip',
		'N44E124.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N44E124.hgt.zip',
		'N44E125.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N44E125.hgt.zip',
		'N44E126.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N44E126.hgt.zip',
		'N44E127.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N44E127.hgt.zip',
		'N44E128.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N44E128.hgt.zip',
		'N44E129.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N44E129.hgt.zip',
		'N44E130.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N44E130.hgt.zip',
		'N44E131.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N44E131.hgt.zip',
		'N44E132.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N44E132.hgt.zip',
		'N44E133.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N44E133.hgt.zip',
		'N44E134.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N44E134.hgt.zip',
		'N44E135.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N44E135.hgt.zip',
		'N44E136.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N44E136.hgt.zip',
		'N44E141.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N44E141.hgt.zip',
		'N44E142.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N44E142.hgt.zip',
		'N44E143.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N44E143.hgt.zip',
		'N44E144.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N44E144.hgt.zip',
		'N44E145.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N44E145.hgt.zip',
		'N44E146.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N44E146.hgt.zip',
		'N44E147.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N44E147.hgt.zip',
		'N44W001.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N44W001.hgt.zip',
		'N44W002.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N44W002.hgt.zip',
		'N44W060.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N44W060.hgt.zip',
		'N44W062.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N44W062.hgt.zip',
		'N44W063.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N44W063.hgt.zip',
		'N44W064.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N44W064.hgt.zip',
		'N44W065.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N44W065.hgt.zip',
		'N44W066.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N44W066.hgt.zip',
		'N44W067.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N44W067.hgt.zip',
		'N44W068.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N44W068.hgt.zip',
		'N44W069.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N44W069.hgt.zip',
		'N44W070.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N44W070.hgt.zip',
		'N44W071.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N44W071.hgt.zip',
		'N44W072.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N44W072.hgt.zip',
		'N44W073.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N44W073.hgt.zip',
		'N44W074.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N44W074.hgt.zip',
		'N44W075.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N44W075.hgt.zip',
		'N44W076.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N44W076.hgt.zip',
		'N44W077.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N44W077.hgt.zip',
		'N44W078.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N44W078.hgt.zip',
		'N44W079.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N44W079.hgt.zip',
		'N44W080.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N44W080.hgt.zip',
		'N44W081.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N44W081.hgt.zip',
		'N44W082.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N44W082.hgt.zip',
		'N44W083.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N44W083.hgt.zip',
		'N44W084.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N44W084.hgt.zip',
		'N44W085.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N44W085.hgt.zip',
		'N44W086.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N44W086.hgt.zip',
		'N44W087.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N44W087.hgt.zip',
		'N44W088.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N44W088.hgt.zip',
		'N44W089.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N44W089.hgt.zip',
		'N44W090.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N44W090.hgt.zip',
		'N44W091.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N44W091.hgt.zip',
		'N44W092.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N44W092.hgt.zip',
		'N44W093.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N44W093.hgt.zip',
		'N44W094.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N44W094.hgt.zip',
		'N44W095.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N44W095.hgt.zip',
		'N44W096.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N44W096.hgt.zip',
		'N44W097.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N44W097.hgt.zip',
		'N44W098.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N44W098.hgt.zip',
		'N44W099.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N44W099.hgt.zip',
		'N44W100.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N44W100.hgt.zip',
		'N44W101.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N44W101.hgt.zip',
		'N44W102.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N44W102.hgt.zip',
		'N44W103.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N44W103.hgt.zip',
		'N44W104.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N44W104.hgt.zip',
		'N44W105.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N44W105.hgt.zip',
		'N44W106.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N44W106.hgt.zip',
		'N44W107.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N44W107.hgt.zip',
		'N44W108.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N44W108.hgt.zip',
		'N44W109.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N44W109.hgt.zip',
		'N44W110.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N44W110.hgt.zip',
		'N44W111.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N44W111.hgt.zip',
		'N44W112.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N44W112.hgt.zip',
		'N44W113.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N44W113.hgt.zip',
		'N44W114.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N44W114.hgt.zip',
		'N44W115.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N44W115.hgt.zip',
		'N44W116.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N44W116.hgt.zip',
		'N44W117.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N44W117.hgt.zip',
		'N44W118.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N44W118.hgt.zip',
		'N44W119.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N44W119.hgt.zip',
		'N44W120.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N44W120.hgt.zip',
		'N44W121.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N44W121.hgt.zip',
		'N44W122.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N44W122.hgt.zip',
		'N44W123.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N44W123.hgt.zip',
		'N44W124.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N44W124.hgt.zip',
		'N44W125.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N44W125.hgt.zip',
		'N45E000.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45E000.hgt.zip',
		'N45E001.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45E001.hgt.zip',
		'N45E002.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45E002.hgt.zip',
		'N45E003.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45E003.hgt.zip',
		'N45E004.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45E004.hgt.zip',
		'N45E005.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45E005.hgt.zip',
		'N45E006.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45E006.hgt.zip',
		'N45E007.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45E007.hgt.zip',
		'N45E008.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45E008.hgt.zip',
		'N45E009.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45E009.hgt.zip',
		'N45E010.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45E010.hgt.zip',
		'N45E011.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45E011.hgt.zip',
		'N45E012.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45E012.hgt.zip',
		'N45E013.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45E013.hgt.zip',
		'N45E014.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45E014.hgt.zip',
		'N45E015.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45E015.hgt.zip',
		'N45E016.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45E016.hgt.zip',
		'N45E017.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45E017.hgt.zip',
		'N45E018.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45E018.hgt.zip',
		'N45E019.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45E019.hgt.zip',
		'N45E020.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45E020.hgt.zip',
		'N45E021.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45E021.hgt.zip',
		'N45E022.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45E022.hgt.zip',
		'N45E023.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45E023.hgt.zip',
		'N45E024.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45E024.hgt.zip',
		'N45E025.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45E025.hgt.zip',
		'N45E026.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45E026.hgt.zip',
		'N45E027.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45E027.hgt.zip',
		'N45E028.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45E028.hgt.zip',
		'N45E029.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45E029.hgt.zip',
		'N45E030.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45E030.hgt.zip',
		'N45E032.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45E032.hgt.zip',
		'N45E033.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45E033.hgt.zip',
		'N45E034.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45E034.hgt.zip',
		'N45E035.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45E035.hgt.zip',
		'N45E036.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45E036.hgt.zip',
		'N45E037.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45E037.hgt.zip',
		'N45E038.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45E038.hgt.zip',
		'N45E039.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45E039.hgt.zip',
		'N45E040.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45E040.hgt.zip',
		'N45E041.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45E041.hgt.zip',
		'N45E042.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45E042.hgt.zip',
		'N45E043.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45E043.hgt.zip',
		'N45E044.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45E044.hgt.zip',
		'N45E045.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45E045.hgt.zip',
		'N45E046.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45E046.hgt.zip',
		'N45E047.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45E047.hgt.zip',
		'N45E048.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45E048.hgt.zip',
		'N45E049.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45E049.hgt.zip',
		'N45E050.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45E050.hgt.zip',
		'N45E051.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45E051.hgt.zip',
		'N45E052.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45E052.hgt.zip',
		'N45E053.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45E053.hgt.zip',
		'N45E054.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45E054.hgt.zip',
		'N45E055.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45E055.hgt.zip',
		'N45E056.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45E056.hgt.zip',
		'N45E057.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45E057.hgt.zip',
		'N45E058.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45E058.hgt.zip',
		'N45E059.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45E059.hgt.zip',
		'N45E060.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45E060.hgt.zip',
		'N45E061.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45E061.hgt.zip',
		'N45E062.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45E062.hgt.zip',
		'N45E063.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45E063.hgt.zip',
		'N45E064.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45E064.hgt.zip',
		'N45E065.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45E065.hgt.zip',
		'N45E066.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45E066.hgt.zip',
		'N45E067.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45E067.hgt.zip',
		'N45E068.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45E068.hgt.zip',
		'N45E069.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45E069.hgt.zip',
		'N45E070.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45E070.hgt.zip',
		'N45E071.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45E071.hgt.zip',
		'N45E072.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45E072.hgt.zip',
		'N45E073.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45E073.hgt.zip',
		'N45E074.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45E074.hgt.zip',
		'N45E075.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45E075.hgt.zip',
		'N45E076.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45E076.hgt.zip',
		'N45E077.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45E077.hgt.zip',
		'N45E078.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45E078.hgt.zip',
		'N45E079.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45E079.hgt.zip',
		'N45E080.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45E080.hgt.zip',
		'N45E081.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45E081.hgt.zip',
		'N45E082.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45E082.hgt.zip',
		'N45E083.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45E083.hgt.zip',
		'N45E084.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45E084.hgt.zip',
		'N45E085.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45E085.hgt.zip',
		'N45E086.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45E086.hgt.zip',
		'N45E087.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45E087.hgt.zip',
		'N45E088.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45E088.hgt.zip',
		'N45E089.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45E089.hgt.zip',
		'N45E090.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45E090.hgt.zip',
		'N45E091.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45E091.hgt.zip',
		'N45E092.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45E092.hgt.zip',
		'N45E093.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45E093.hgt.zip',
		'N45E094.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45E094.hgt.zip',
		'N45E095.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45E095.hgt.zip',
		'N45E096.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45E096.hgt.zip',
		'N45E097.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45E097.hgt.zip',
		'N45E098.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45E098.hgt.zip',
		'N45E099.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45E099.hgt.zip',
		'N45E100.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45E100.hgt.zip',
		'N45E101.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45E101.hgt.zip',
		'N45E102.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45E102.hgt.zip',
		'N45E103.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45E103.hgt.zip',
		'N45E104.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45E104.hgt.zip',
		'N45E105.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45E105.hgt.zip',
		'N45E106.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45E106.hgt.zip',
		'N45E107.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45E107.hgt.zip',
		'N45E108.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45E108.hgt.zip',
		'N45E109.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45E109.hgt.zip',
		'N45E110.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45E110.hgt.zip',
		'N45E111.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45E111.hgt.zip',
		'N45E112.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45E112.hgt.zip',
		'N45E113.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45E113.hgt.zip',
		'N45E114.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45E114.hgt.zip',
		'N45E115.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45E115.hgt.zip',
		'N45E116.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45E116.hgt.zip',
		'N45E117.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45E117.hgt.zip',
		'N45E118.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45E118.hgt.zip',
		'N45E119.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45E119.hgt.zip',
		'N45E120.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45E120.hgt.zip',
		'N45E121.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45E121.hgt.zip',
		'N45E122.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45E122.hgt.zip',
		'N45E123.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45E123.hgt.zip',
		'N45E124.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45E124.hgt.zip',
		'N45E125.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45E125.hgt.zip',
		'N45E126.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45E126.hgt.zip',
		'N45E127.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45E127.hgt.zip',
		'N45E128.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45E128.hgt.zip',
		'N45E129.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45E129.hgt.zip',
		'N45E130.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45E130.hgt.zip',
		'N45E131.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45E131.hgt.zip',
		'N45E132.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45E132.hgt.zip',
		'N45E133.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45E133.hgt.zip',
		'N45E134.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45E134.hgt.zip',
		'N45E135.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45E135.hgt.zip',
		'N45E136.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45E136.hgt.zip',
		'N45E137.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45E137.hgt.zip',
		'N45E140.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45E140.hgt.zip',
		'N45E141.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45E141.hgt.zip',
		'N45E142.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45E142.hgt.zip',
		'N45E147.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45E147.hgt.zip',
		'N45E148.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45E148.hgt.zip',
		'N45E149.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45E149.hgt.zip',
		'N45E150.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45E150.hgt.zip',
		'N45W001.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45W001.hgt.zip',
		'N45W002.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N45W002.hgt.zip',
		'N45W060.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N45W060.hgt.zip',
		'N45W061.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N45W061.hgt.zip',
		'N45W062.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N45W062.hgt.zip',
		'N45W063.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N45W063.hgt.zip',
		'N45W064.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N45W064.hgt.zip',
		'N45W065.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N45W065.hgt.zip',
		'N45W066.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N45W066.hgt.zip',
		'N45W067.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N45W067.hgt.zip',
		'N45W068.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N45W068.hgt.zip',
		'N45W069.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N45W069.hgt.zip',
		'N45W070.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N45W070.hgt.zip',
		'N45W071.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N45W071.hgt.zip',
		'N45W072.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N45W072.hgt.zip',
		'N45W073.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N45W073.hgt.zip',
		'N45W074.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N45W074.hgt.zip',
		'N45W075.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N45W075.hgt.zip',
		'N45W076.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N45W076.hgt.zip',
		'N45W077.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N45W077.hgt.zip',
		'N45W078.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N45W078.hgt.zip',
		'N45W079.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N45W079.hgt.zip',
		'N45W080.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N45W080.hgt.zip',
		'N45W081.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N45W081.hgt.zip',
		'N45W082.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N45W082.hgt.zip',
		'N45W083.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N45W083.hgt.zip',
		'N45W084.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N45W084.hgt.zip',
		'N45W085.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N45W085.hgt.zip',
		'N45W086.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N45W086.hgt.zip',
		'N45W087.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N45W087.hgt.zip',
		'N45W088.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N45W088.hgt.zip',
		'N45W089.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N45W089.hgt.zip',
		'N45W090.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N45W090.hgt.zip',
		'N45W091.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N45W091.hgt.zip',
		'N45W092.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N45W092.hgt.zip',
		'N45W093.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N45W093.hgt.zip',
		'N45W094.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N45W094.hgt.zip',
		'N45W095.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N45W095.hgt.zip',
		'N45W096.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N45W096.hgt.zip',
		'N45W097.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N45W097.hgt.zip',
		'N45W098.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N45W098.hgt.zip',
		'N45W099.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N45W099.hgt.zip',
		'N45W100.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N45W100.hgt.zip',
		'N45W101.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N45W101.hgt.zip',
		'N45W102.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N45W102.hgt.zip',
		'N45W103.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N45W103.hgt.zip',
		'N45W104.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N45W104.hgt.zip',
		'N45W105.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N45W105.hgt.zip',
		'N45W106.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N45W106.hgt.zip',
		'N45W107.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N45W107.hgt.zip',
		'N45W108.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N45W108.hgt.zip',
		'N45W109.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N45W109.hgt.zip',
		'N45W110.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N45W110.hgt.zip',
		'N45W111.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N45W111.hgt.zip',
		'N45W112.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N45W112.hgt.zip',
		'N45W113.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N45W113.hgt.zip',
		'N45W114.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N45W114.hgt.zip',
		'N45W115.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N45W115.hgt.zip',
		'N45W116.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N45W116.hgt.zip',
		'N45W117.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N45W117.hgt.zip',
		'N45W118.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N45W118.hgt.zip',
		'N45W119.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N45W119.hgt.zip',
		'N45W120.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N45W120.hgt.zip',
		'N45W121.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N45W121.hgt.zip',
		'N45W122.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N45W122.hgt.zip',
		'N45W123.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N45W123.hgt.zip',
		'N45W124.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N45W124.hgt.zip',
		'N45W125.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N45W125.hgt.zip',
		'N46E000.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E000.hgt.zip',
		'N46E001.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E001.hgt.zip',
		'N46E002.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E002.hgt.zip',
		'N46E003.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E003.hgt.zip',
		'N46E004.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E004.hgt.zip',
		'N46E005.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E005.hgt.zip',
		'N46E006.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E006.hgt.zip',
		'N46E007.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E007.hgt.zip',
		'N46E008.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E008.hgt.zip',
		'N46E009.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E009.hgt.zip',
		'N46E010.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E010.hgt.zip',
		'N46E011.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E011.hgt.zip',
		'N46E012.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E012.hgt.zip',
		'N46E013.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E013.hgt.zip',
		'N46E014.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E014.hgt.zip',
		'N46E015.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E015.hgt.zip',
		'N46E016.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E016.hgt.zip',
		'N46E017.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E017.hgt.zip',
		'N46E018.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E018.hgt.zip',
		'N46E019.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E019.hgt.zip',
		'N46E020.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E020.hgt.zip',
		'N46E021.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E021.hgt.zip',
		'N46E022.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E022.hgt.zip',
		'N46E023.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E023.hgt.zip',
		'N46E024.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E024.hgt.zip',
		'N46E025.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E025.hgt.zip',
		'N46E026.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E026.hgt.zip',
		'N46E027.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E027.hgt.zip',
		'N46E028.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E028.hgt.zip',
		'N46E029.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E029.hgt.zip',
		'N46E030.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E030.hgt.zip',
		'N46E031.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E031.hgt.zip',
		'N46E032.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E032.hgt.zip',
		'N46E033.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E033.hgt.zip',
		'N46E034.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E034.hgt.zip',
		'N46E035.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E035.hgt.zip',
		'N46E036.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E036.hgt.zip',
		'N46E037.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E037.hgt.zip',
		'N46E038.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E038.hgt.zip',
		'N46E039.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E039.hgt.zip',
		'N46E040.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E040.hgt.zip',
		'N46E041.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E041.hgt.zip',
		'N46E042.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E042.hgt.zip',
		'N46E043.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E043.hgt.zip',
		'N46E044.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E044.hgt.zip',
		'N46E045.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E045.hgt.zip',
		'N46E046.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E046.hgt.zip',
		'N46E047.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E047.hgt.zip',
		'N46E048.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E048.hgt.zip',
		'N46E049.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E049.hgt.zip',
		'N46E050.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E050.hgt.zip',
		'N46E051.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E051.hgt.zip',
		'N46E052.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E052.hgt.zip',
		'N46E053.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E053.hgt.zip',
		'N46E054.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E054.hgt.zip',
		'N46E055.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E055.hgt.zip',
		'N46E056.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E056.hgt.zip',
		'N46E057.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E057.hgt.zip',
		'N46E058.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E058.hgt.zip',
		'N46E059.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E059.hgt.zip',
		'N46E060.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E060.hgt.zip',
		'N46E061.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E061.hgt.zip',
		'N46E062.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E062.hgt.zip',
		'N46E063.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E063.hgt.zip',
		'N46E064.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E064.hgt.zip',
		'N46E065.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E065.hgt.zip',
		'N46E066.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E066.hgt.zip',
		'N46E067.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E067.hgt.zip',
		'N46E068.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E068.hgt.zip',
		'N46E069.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E069.hgt.zip',
		'N46E070.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E070.hgt.zip',
		'N46E071.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E071.hgt.zip',
		'N46E072.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E072.hgt.zip',
		'N46E073.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E073.hgt.zip',
		'N46E074.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E074.hgt.zip',
		'N46E075.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E075.hgt.zip',
		'N46E076.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E076.hgt.zip',
		'N46E077.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E077.hgt.zip',
		'N46E078.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E078.hgt.zip',
		'N46E079.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E079.hgt.zip',
		'N46E080.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E080.hgt.zip',
		'N46E081.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E081.hgt.zip',
		'N46E082.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E082.hgt.zip',
		'N46E083.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E083.hgt.zip',
		'N46E084.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E084.hgt.zip',
		'N46E085.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E085.hgt.zip',
		'N46E086.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E086.hgt.zip',
		'N46E087.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E087.hgt.zip',
		'N46E088.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E088.hgt.zip',
		'N46E089.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E089.hgt.zip',
		'N46E090.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E090.hgt.zip',
		'N46E091.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E091.hgt.zip',
		'N46E092.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E092.hgt.zip',
		'N46E093.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E093.hgt.zip',
		'N46E094.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E094.hgt.zip',
		'N46E095.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E095.hgt.zip',
		'N46E096.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E096.hgt.zip',
		'N46E097.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E097.hgt.zip',
		'N46E098.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E098.hgt.zip',
		'N46E099.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E099.hgt.zip',
		'N46E100.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E100.hgt.zip',
		'N46E101.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E101.hgt.zip',
		'N46E102.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E102.hgt.zip',
		'N46E103.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E103.hgt.zip',
		'N46E104.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E104.hgt.zip',
		'N46E105.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E105.hgt.zip',
		'N46E106.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E106.hgt.zip',
		'N46E107.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E107.hgt.zip',
		'N46E108.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E108.hgt.zip',
		'N46E109.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E109.hgt.zip',
		'N46E110.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E110.hgt.zip',
		'N46E111.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E111.hgt.zip',
		'N46E112.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E112.hgt.zip',
		'N46E113.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E113.hgt.zip',
		'N46E114.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E114.hgt.zip',
		'N46E115.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E115.hgt.zip',
		'N46E116.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E116.hgt.zip',
		'N46E117.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E117.hgt.zip',
		'N46E118.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E118.hgt.zip',
		'N46E119.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E119.hgt.zip',
		'N46E120.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E120.hgt.zip',
		'N46E121.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E121.hgt.zip',
		'N46E122.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E122.hgt.zip',
		'N46E123.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E123.hgt.zip',
		'N46E124.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E124.hgt.zip',
		'N46E125.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E125.hgt.zip',
		'N46E126.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E126.hgt.zip',
		'N46E127.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E127.hgt.zip',
		'N46E128.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E128.hgt.zip',
		'N46E129.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E129.hgt.zip',
		'N46E130.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E130.hgt.zip',
		'N46E131.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E131.hgt.zip',
		'N46E132.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E132.hgt.zip',
		'N46E133.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E133.hgt.zip',
		'N46E134.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E134.hgt.zip',
		'N46E135.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E135.hgt.zip',
		'N46E136.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E136.hgt.zip',
		'N46E137.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E137.hgt.zip',
		'N46E138.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E138.hgt.zip',
		'N46E141.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E141.hgt.zip',
		'N46E142.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E142.hgt.zip',
		'N46E143.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E143.hgt.zip',
		'N46E149.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E149.hgt.zip',
		'N46E150.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E150.hgt.zip',
		'N46E151.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E151.hgt.zip',
		'N46E152.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46E152.hgt.zip',
		'N46W001.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46W001.hgt.zip',
		'N46W002.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46W002.hgt.zip',
		'N46W003.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N46W003.hgt.zip',
		'N46W053.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N46W053.hgt.zip',
		'N46W054.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N46W054.hgt.zip',
		'N46W055.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N46W055.hgt.zip',
		'N46W056.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N46W056.hgt.zip',
		'N46W057.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N46W057.hgt.zip',
		'N46W060.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N46W060.hgt.zip',
		'N46W061.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N46W061.hgt.zip',
		'N46W062.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N46W062.hgt.zip',
		'N46W063.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N46W063.hgt.zip',
		'N46W064.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N46W064.hgt.zip',
		'N46W065.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N46W065.hgt.zip',
		'N46W066.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N46W066.hgt.zip',
		'N46W067.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N46W067.hgt.zip',
		'N46W068.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N46W068.hgt.zip',
		'N46W069.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N46W069.hgt.zip',
		'N46W070.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N46W070.hgt.zip',
		'N46W071.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N46W071.hgt.zip',
		'N46W072.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N46W072.hgt.zip',
		'N46W073.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N46W073.hgt.zip',
		'N46W074.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N46W074.hgt.zip',
		'N46W075.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N46W075.hgt.zip',
		'N46W076.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N46W076.hgt.zip',
		'N46W077.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N46W077.hgt.zip',
		'N46W078.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N46W078.hgt.zip',
		'N46W079.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N46W079.hgt.zip',
		'N46W080.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N46W080.hgt.zip',
		'N46W081.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N46W081.hgt.zip',
		'N46W082.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N46W082.hgt.zip',
		'N46W083.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N46W083.hgt.zip',
		'N46W084.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N46W084.hgt.zip',
		'N46W085.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N46W085.hgt.zip',
		'N46W086.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N46W086.hgt.zip',
		'N46W087.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N46W087.hgt.zip',
		'N46W088.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N46W088.hgt.zip',
		'N46W089.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N46W089.hgt.zip',
		'N46W090.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N46W090.hgt.zip',
		'N46W091.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N46W091.hgt.zip',
		'N46W092.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N46W092.hgt.zip',
		'N46W093.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N46W093.hgt.zip',
		'N46W094.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N46W094.hgt.zip',
		'N46W095.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N46W095.hgt.zip',
		'N46W096.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N46W096.hgt.zip',
		'N46W097.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N46W097.hgt.zip',
		'N46W098.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N46W098.hgt.zip',
		'N46W099.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N46W099.hgt.zip',
		'N46W100.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N46W100.hgt.zip',
		'N46W101.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N46W101.hgt.zip',
		'N46W102.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N46W102.hgt.zip',
		'N46W103.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N46W103.hgt.zip',
		'N46W104.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N46W104.hgt.zip',
		'N46W105.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N46W105.hgt.zip',
		'N46W106.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N46W106.hgt.zip',
		'N46W107.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N46W107.hgt.zip',
		'N46W108.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N46W108.hgt.zip',
		'N46W109.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N46W109.hgt.zip',
		'N46W110.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N46W110.hgt.zip',
		'N46W111.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N46W111.hgt.zip',
		'N46W112.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N46W112.hgt.zip',
		'N46W113.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N46W113.hgt.zip',
		'N46W114.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N46W114.hgt.zip',
		'N46W115.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N46W115.hgt.zip',
		'N46W116.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N46W116.hgt.zip',
		'N46W117.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N46W117.hgt.zip',
		'N46W118.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N46W118.hgt.zip',
		'N46W119.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N46W119.hgt.zip',
		'N46W120.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N46W120.hgt.zip',
		'N46W121.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N46W121.hgt.zip',
		'N46W122.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N46W122.hgt.zip',
		'N46W123.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N46W123.hgt.zip',
		'N46W124.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N46W124.hgt.zip',
		'N46W125.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N46W125.hgt.zip',
		'N47E000.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E000.hgt.zip',
		'N47E001.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E001.hgt.zip',
		'N47E002.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E002.hgt.zip',
		'N47E003.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E003.hgt.zip',
		'N47E004.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E004.hgt.zip',
		'N47E005.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E005.hgt.zip',
		'N47E006.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E006.hgt.zip',
		'N47E007.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E007.hgt.zip',
		'N47E008.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E008.hgt.zip',
		'N47E009.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E009.hgt.zip',
		'N47E010.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E010.hgt.zip',
		'N47E011.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E011.hgt.zip',
		'N47E012.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E012.hgt.zip',
		'N47E013.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E013.hgt.zip',
		'N47E014.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E014.hgt.zip',
		'N47E015.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E015.hgt.zip',
		'N47E016.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E016.hgt.zip',
		'N47E017.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E017.hgt.zip',
		'N47E018.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E018.hgt.zip',
		'N47E019.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E019.hgt.zip',
		'N47E020.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E020.hgt.zip',
		'N47E021.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E021.hgt.zip',
		'N47E022.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E022.hgt.zip',
		'N47E023.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E023.hgt.zip',
		'N47E024.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E024.hgt.zip',
		'N47E025.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E025.hgt.zip',
		'N47E026.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E026.hgt.zip',
		'N47E027.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E027.hgt.zip',
		'N47E028.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E028.hgt.zip',
		'N47E029.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E029.hgt.zip',
		'N47E030.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E030.hgt.zip',
		'N47E031.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E031.hgt.zip',
		'N47E032.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E032.hgt.zip',
		'N47E033.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E033.hgt.zip',
		'N47E034.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E034.hgt.zip',
		'N47E035.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E035.hgt.zip',
		'N47E036.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E036.hgt.zip',
		'N47E037.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E037.hgt.zip',
		'N47E038.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E038.hgt.zip',
		'N47E039.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E039.hgt.zip',
		'N47E040.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E040.hgt.zip',
		'N47E041.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E041.hgt.zip',
		'N47E042.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E042.hgt.zip',
		'N47E043.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E043.hgt.zip',
		'N47E044.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E044.hgt.zip',
		'N47E045.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E045.hgt.zip',
		'N47E046.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E046.hgt.zip',
		'N47E047.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E047.hgt.zip',
		'N47E048.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E048.hgt.zip',
		'N47E049.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E049.hgt.zip',
		'N47E050.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E050.hgt.zip',
		'N47E051.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E051.hgt.zip',
		'N47E052.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E052.hgt.zip',
		'N47E053.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E053.hgt.zip',
		'N47E054.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E054.hgt.zip',
		'N47E055.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E055.hgt.zip',
		'N47E056.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E056.hgt.zip',
		'N47E057.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E057.hgt.zip',
		'N47E058.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E058.hgt.zip',
		'N47E059.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E059.hgt.zip',
		'N47E060.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E060.hgt.zip',
		'N47E061.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E061.hgt.zip',
		'N47E062.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E062.hgt.zip',
		'N47E063.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E063.hgt.zip',
		'N47E064.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E064.hgt.zip',
		'N47E065.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E065.hgt.zip',
		'N47E066.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E066.hgt.zip',
		'N47E067.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E067.hgt.zip',
		'N47E068.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E068.hgt.zip',
		'N47E069.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E069.hgt.zip',
		'N47E070.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E070.hgt.zip',
		'N47E071.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E071.hgt.zip',
		'N47E072.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E072.hgt.zip',
		'N47E073.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E073.hgt.zip',
		'N47E074.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E074.hgt.zip',
		'N47E075.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E075.hgt.zip',
		'N47E076.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E076.hgt.zip',
		'N47E077.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E077.hgt.zip',
		'N47E078.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E078.hgt.zip',
		'N47E079.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E079.hgt.zip',
		'N47E080.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E080.hgt.zip',
		'N47E081.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E081.hgt.zip',
		'N47E082.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E082.hgt.zip',
		'N47E083.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E083.hgt.zip',
		'N47E084.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E084.hgt.zip',
		'N47E085.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E085.hgt.zip',
		'N47E086.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E086.hgt.zip',
		'N47E087.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E087.hgt.zip',
		'N47E088.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E088.hgt.zip',
		'N47E089.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E089.hgt.zip',
		'N47E090.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E090.hgt.zip',
		'N47E091.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E091.hgt.zip',
		'N47E092.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E092.hgt.zip',
		'N47E093.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E093.hgt.zip',
		'N47E094.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E094.hgt.zip',
		'N47E095.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E095.hgt.zip',
		'N47E096.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E096.hgt.zip',
		'N47E097.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E097.hgt.zip',
		'N47E098.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E098.hgt.zip',
		'N47E099.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E099.hgt.zip',
		'N47E100.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E100.hgt.zip',
		'N47E101.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E101.hgt.zip',
		'N47E102.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E102.hgt.zip',
		'N47E103.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E103.hgt.zip',
		'N47E104.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E104.hgt.zip',
		'N47E105.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E105.hgt.zip',
		'N47E106.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E106.hgt.zip',
		'N47E107.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E107.hgt.zip',
		'N47E108.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E108.hgt.zip',
		'N47E109.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E109.hgt.zip',
		'N47E110.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E110.hgt.zip',
		'N47E111.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E111.hgt.zip',
		'N47E112.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E112.hgt.zip',
		'N47E113.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E113.hgt.zip',
		'N47E114.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E114.hgt.zip',
		'N47E115.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E115.hgt.zip',
		'N47E116.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E116.hgt.zip',
		'N47E117.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E117.hgt.zip',
		'N47E118.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E118.hgt.zip',
		'N47E119.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E119.hgt.zip',
		'N47E120.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E120.hgt.zip',
		'N47E121.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E121.hgt.zip',
		'N47E122.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E122.hgt.zip',
		'N47E123.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E123.hgt.zip',
		'N47E124.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E124.hgt.zip',
		'N47E125.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E125.hgt.zip',
		'N47E126.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E126.hgt.zip',
		'N47E127.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E127.hgt.zip',
		'N47E128.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E128.hgt.zip',
		'N47E129.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E129.hgt.zip',
		'N47E130.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E130.hgt.zip',
		'N47E131.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E131.hgt.zip',
		'N47E132.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E132.hgt.zip',
		'N47E133.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E133.hgt.zip',
		'N47E134.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E134.hgt.zip',
		'N47E135.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E135.hgt.zip',
		'N47E136.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E136.hgt.zip',
		'N47E137.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E137.hgt.zip',
		'N47E138.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E138.hgt.zip',
		'N47E139.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E139.hgt.zip',
		'N47E141.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E141.hgt.zip',
		'N47E142.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E142.hgt.zip',
		'N47E143.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E143.hgt.zip',
		'N47E152.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E152.hgt.zip',
		'N47E153.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47E153.hgt.zip',
		'N47W001.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47W001.hgt.zip',
		'N47W002.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47W002.hgt.zip',
		'N47W003.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47W003.hgt.zip',
		'N47W004.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47W004.hgt.zip',
		'N47W005.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N47W005.hgt.zip',
		'N47W053.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N47W053.hgt.zip',
		'N47W054.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N47W054.hgt.zip',
		'N47W055.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N47W055.hgt.zip',
		'N47W056.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N47W056.hgt.zip',
		'N47W057.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N47W057.hgt.zip',
		'N47W058.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N47W058.hgt.zip',
		'N47W059.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N47W059.hgt.zip',
		'N47W060.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N47W060.hgt.zip',
		'N47W061.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N47W061.hgt.zip',
		'N47W062.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N47W062.hgt.zip',
		'N47W063.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N47W063.hgt.zip',
		'N47W064.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N47W064.hgt.zip',
		'N47W065.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N47W065.hgt.zip',
		'N47W066.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N47W066.hgt.zip',
		'N47W067.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N47W067.hgt.zip',
		'N47W068.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N47W068.hgt.zip',
		'N47W069.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N47W069.hgt.zip',
		'N47W070.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N47W070.hgt.zip',
		'N47W071.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N47W071.hgt.zip',
		'N47W072.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N47W072.hgt.zip',
		'N47W073.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N47W073.hgt.zip',
		'N47W074.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N47W074.hgt.zip',
		'N47W075.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N47W075.hgt.zip',
		'N47W076.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N47W076.hgt.zip',
		'N47W077.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N47W077.hgt.zip',
		'N47W078.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N47W078.hgt.zip',
		'N47W079.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N47W079.hgt.zip',
		'N47W080.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N47W080.hgt.zip',
		'N47W081.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N47W081.hgt.zip',
		'N47W082.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N47W082.hgt.zip',
		'N47W083.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N47W083.hgt.zip',
		'N47W084.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N47W084.hgt.zip',
		'N47W085.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N47W085.hgt.zip',
		'N47W086.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N47W086.hgt.zip',
		'N47W088.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N47W088.hgt.zip',
		'N47W089.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N47W089.hgt.zip',
		'N47W090.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N47W090.hgt.zip',
		'N47W091.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N47W091.hgt.zip',
		'N47W092.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N47W092.hgt.zip',
		'N47W093.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N47W093.hgt.zip',
		'N47W094.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N47W094.hgt.zip',
		'N47W095.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N47W095.hgt.zip',
		'N47W096.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N47W096.hgt.zip',
		'N47W097.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N47W097.hgt.zip',
		'N47W098.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N47W098.hgt.zip',
		'N47W099.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N47W099.hgt.zip',
		'N47W100.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N47W100.hgt.zip',
		'N47W101.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N47W101.hgt.zip',
		'N47W102.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N47W102.hgt.zip',
		'N47W103.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N47W103.hgt.zip',
		'N47W104.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N47W104.hgt.zip',
		'N47W105.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N47W105.hgt.zip',
		'N47W106.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N47W106.hgt.zip',
		'N47W107.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N47W107.hgt.zip',
		'N47W108.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N47W108.hgt.zip',
		'N47W109.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N47W109.hgt.zip',
		'N47W110.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N47W110.hgt.zip',
		'N47W111.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N47W111.hgt.zip',
		'N47W112.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N47W112.hgt.zip',
		'N47W113.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N47W113.hgt.zip',
		'N47W114.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N47W114.hgt.zip',
		'N47W115.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N47W115.hgt.zip',
		'N47W116.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N47W116.hgt.zip',
		'N47W117.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N47W117.hgt.zip',
		'N47W118.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N47W118.hgt.zip',
		'N47W119.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N47W119.hgt.zip',
		'N47W120.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N47W120.hgt.zip',
		'N47W121.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N47W121.hgt.zip',
		'N47W122.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N47W122.hgt.zip',
		'N47W123.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N47W123.hgt.zip',
		'N47W124.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N47W124.hgt.zip',
		'N47W125.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N47W125.hgt.zip',
		'N48E000.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E000.hgt.zip',
		'N48E001.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E001.hgt.zip',
		'N48E002.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E002.hgt.zip',
		'N48E003.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E003.hgt.zip',
		'N48E004.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E004.hgt.zip',
		'N48E005.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E005.hgt.zip',
		'N48E006.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E006.hgt.zip',
		'N48E007.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E007.hgt.zip',
		'N48E008.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E008.hgt.zip',
		'N48E009.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E009.hgt.zip',
		'N48E010.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E010.hgt.zip',
		'N48E011.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E011.hgt.zip',
		'N48E012.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E012.hgt.zip',
		'N48E013.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E013.hgt.zip',
		'N48E014.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E014.hgt.zip',
		'N48E015.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E015.hgt.zip',
		'N48E016.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E016.hgt.zip',
		'N48E017.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E017.hgt.zip',
		'N48E018.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E018.hgt.zip',
		'N48E019.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E019.hgt.zip',
		'N48E020.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E020.hgt.zip',
		'N48E021.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E021.hgt.zip',
		'N48E022.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E022.hgt.zip',
		'N48E023.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E023.hgt.zip',
		'N48E024.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E024.hgt.zip',
		'N48E025.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E025.hgt.zip',
		'N48E026.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E026.hgt.zip',
		'N48E027.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E027.hgt.zip',
		'N48E028.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E028.hgt.zip',
		'N48E029.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E029.hgt.zip',
		'N48E030.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E030.hgt.zip',
		'N48E031.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E031.hgt.zip',
		'N48E032.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E032.hgt.zip',
		'N48E033.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E033.hgt.zip',
		'N48E034.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E034.hgt.zip',
		'N48E035.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E035.hgt.zip',
		'N48E036.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E036.hgt.zip',
		'N48E037.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E037.hgt.zip',
		'N48E038.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E038.hgt.zip',
		'N48E039.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E039.hgt.zip',
		'N48E040.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E040.hgt.zip',
		'N48E041.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E041.hgt.zip',
		'N48E042.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E042.hgt.zip',
		'N48E043.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E043.hgt.zip',
		'N48E044.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E044.hgt.zip',
		'N48E045.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E045.hgt.zip',
		'N48E046.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E046.hgt.zip',
		'N48E047.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E047.hgt.zip',
		'N48E048.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E048.hgt.zip',
		'N48E049.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E049.hgt.zip',
		'N48E050.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E050.hgt.zip',
		'N48E051.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E051.hgt.zip',
		'N48E052.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E052.hgt.zip',
		'N48E053.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E053.hgt.zip',
		'N48E054.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E054.hgt.zip',
		'N48E055.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E055.hgt.zip',
		'N48E056.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E056.hgt.zip',
		'N48E057.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E057.hgt.zip',
		'N48E058.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E058.hgt.zip',
		'N48E059.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E059.hgt.zip',
		'N48E060.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E060.hgt.zip',
		'N48E061.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E061.hgt.zip',
		'N48E062.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E062.hgt.zip',
		'N48E063.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E063.hgt.zip',
		'N48E064.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E064.hgt.zip',
		'N48E065.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E065.hgt.zip',
		'N48E066.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E066.hgt.zip',
		'N48E067.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E067.hgt.zip',
		'N48E068.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E068.hgt.zip',
		'N48E069.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E069.hgt.zip',
		'N48E070.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E070.hgt.zip',
		'N48E071.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E071.hgt.zip',
		'N48E072.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E072.hgt.zip',
		'N48E073.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E073.hgt.zip',
		'N48E074.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E074.hgt.zip',
		'N48E075.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E075.hgt.zip',
		'N48E076.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E076.hgt.zip',
		'N48E077.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E077.hgt.zip',
		'N48E078.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E078.hgt.zip',
		'N48E079.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E079.hgt.zip',
		'N48E080.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E080.hgt.zip',
		'N48E081.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E081.hgt.zip',
		'N48E082.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E082.hgt.zip',
		'N48E083.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E083.hgt.zip',
		'N48E084.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E084.hgt.zip',
		'N48E085.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E085.hgt.zip',
		'N48E086.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E086.hgt.zip',
		'N48E087.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E087.hgt.zip',
		'N48E088.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E088.hgt.zip',
		'N48E089.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E089.hgt.zip',
		'N48E090.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E090.hgt.zip',
		'N48E091.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E091.hgt.zip',
		'N48E092.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E092.hgt.zip',
		'N48E093.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E093.hgt.zip',
		'N48E094.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E094.hgt.zip',
		'N48E095.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E095.hgt.zip',
		'N48E096.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E096.hgt.zip',
		'N48E097.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E097.hgt.zip',
		'N48E098.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E098.hgt.zip',
		'N48E099.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E099.hgt.zip',
		'N48E100.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E100.hgt.zip',
		'N48E101.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E101.hgt.zip',
		'N48E102.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E102.hgt.zip',
		'N48E103.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E103.hgt.zip',
		'N48E104.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E104.hgt.zip',
		'N48E105.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E105.hgt.zip',
		'N48E106.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E106.hgt.zip',
		'N48E107.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E107.hgt.zip',
		'N48E108.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E108.hgt.zip',
		'N48E109.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E109.hgt.zip',
		'N48E110.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E110.hgt.zip',
		'N48E111.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E111.hgt.zip',
		'N48E112.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E112.hgt.zip',
		'N48E113.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E113.hgt.zip',
		'N48E114.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E114.hgt.zip',
		'N48E115.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E115.hgt.zip',
		'N48E116.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E116.hgt.zip',
		'N48E117.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E117.hgt.zip',
		'N48E118.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E118.hgt.zip',
		'N48E119.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E119.hgt.zip',
		'N48E120.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E120.hgt.zip',
		'N48E121.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E121.hgt.zip',
		'N48E122.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E122.hgt.zip',
		'N48E123.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E123.hgt.zip',
		'N48E124.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E124.hgt.zip',
		'N48E125.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E125.hgt.zip',
		'N48E126.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E126.hgt.zip',
		'N48E127.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E127.hgt.zip',
		'N48E128.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E128.hgt.zip',
		'N48E129.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E129.hgt.zip',
		'N48E130.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E130.hgt.zip',
		'N48E131.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E131.hgt.zip',
		'N48E132.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E132.hgt.zip',
		'N48E133.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E133.hgt.zip',
		'N48E134.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E134.hgt.zip',
		'N48E135.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E135.hgt.zip',
		'N48E136.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E136.hgt.zip',
		'N48E137.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E137.hgt.zip',
		'N48E138.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E138.hgt.zip',
		'N48E139.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E139.hgt.zip',
		'N48E140.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E140.hgt.zip',
		'N48E141.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E141.hgt.zip',
		'N48E142.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E142.hgt.zip',
		'N48E144.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E144.hgt.zip',
		'N48E153.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E153.hgt.zip',
		'N48E154.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48E154.hgt.zip',
		'N48W001.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48W001.hgt.zip',
		'N48W002.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48W002.hgt.zip',
		'N48W003.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48W003.hgt.zip',
		'N48W004.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48W004.hgt.zip',
		'N48W005.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48W005.hgt.zip',
		'N48W006.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N48W006.hgt.zip',
		'N48W053.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N48W053.hgt.zip',
		'N48W054.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N48W054.hgt.zip',
		'N48W055.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N48W055.hgt.zip',
		'N48W056.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N48W056.hgt.zip',
		'N48W057.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N48W057.hgt.zip',
		'N48W058.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N48W058.hgt.zip',
		'N48W059.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N48W059.hgt.zip',
		'N48W060.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N48W060.hgt.zip',
		'N48W065.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N48W065.hgt.zip',
		'N48W066.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N48W066.hgt.zip',
		'N48W067.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N48W067.hgt.zip',
		'N48W068.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N48W068.hgt.zip',
		'N48W069.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N48W069.hgt.zip',
		'N48W070.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N48W070.hgt.zip',
		'N48W071.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N48W071.hgt.zip',
		'N48W072.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N48W072.hgt.zip',
		'N48W073.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N48W073.hgt.zip',
		'N48W074.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N48W074.hgt.zip',
		'N48W075.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N48W075.hgt.zip',
		'N48W076.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N48W076.hgt.zip',
		'N48W077.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N48W077.hgt.zip',
		'N48W078.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N48W078.hgt.zip',
		'N48W079.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N48W079.hgt.zip',
		'N48W080.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N48W080.hgt.zip',
		'N48W081.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N48W081.hgt.zip',
		'N48W082.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N48W082.hgt.zip',
		'N48W083.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N48W083.hgt.zip',
		'N48W084.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N48W084.hgt.zip',
		'N48W085.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N48W085.hgt.zip',
		'N48W086.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N48W086.hgt.zip',
		'N48W087.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N48W087.hgt.zip',
		'N48W088.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N48W088.hgt.zip',
		'N48W089.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N48W089.hgt.zip',
		'N48W090.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N48W090.hgt.zip',
		'N48W091.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N48W091.hgt.zip',
		'N48W092.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N48W092.hgt.zip',
		'N48W093.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N48W093.hgt.zip',
		'N48W094.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N48W094.hgt.zip',
		'N48W095.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N48W095.hgt.zip',
		'N48W096.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N48W096.hgt.zip',
		'N48W097.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N48W097.hgt.zip',
		'N48W098.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N48W098.hgt.zip',
		'N48W099.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N48W099.hgt.zip',
		'N48W100.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N48W100.hgt.zip',
		'N48W101.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N48W101.hgt.zip',
		'N48W102.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N48W102.hgt.zip',
		'N48W103.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N48W103.hgt.zip',
		'N48W104.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N48W104.hgt.zip',
		'N48W105.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N48W105.hgt.zip',
		'N48W106.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N48W106.hgt.zip',
		'N48W107.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N48W107.hgt.zip',
		'N48W108.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N48W108.hgt.zip',
		'N48W109.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N48W109.hgt.zip',
		'N48W110.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N48W110.hgt.zip',
		'N48W111.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N48W111.hgt.zip',
		'N48W112.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N48W112.hgt.zip',
		'N48W113.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N48W113.hgt.zip',
		'N48W114.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N48W114.hgt.zip',
		'N48W115.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N48W115.hgt.zip',
		'N48W116.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N48W116.hgt.zip',
		'N48W117.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N48W117.hgt.zip',
		'N48W118.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N48W118.hgt.zip',
		'N48W119.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N48W119.hgt.zip',
		'N48W120.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N48W120.hgt.zip',
		'N48W121.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N48W121.hgt.zip',
		'N48W122.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N48W122.hgt.zip',
		'N48W123.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N48W123.hgt.zip',
		'N48W124.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N48W124.hgt.zip',
		'N48W125.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N48W125.hgt.zip',
		'N48W126.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N48W126.hgt.zip',
		'N49E000.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E000.hgt.zip',
		'N49E001.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E001.hgt.zip',
		'N49E002.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E002.hgt.zip',
		'N49E003.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E003.hgt.zip',
		'N49E004.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E004.hgt.zip',
		'N49E005.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E005.hgt.zip',
		'N49E006.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E006.hgt.zip',
		'N49E007.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E007.hgt.zip',
		'N49E008.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E008.hgt.zip',
		'N49E009.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E009.hgt.zip',
		'N49E010.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E010.hgt.zip',
		'N49E011.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E011.hgt.zip',
		'N49E012.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E012.hgt.zip',
		'N49E013.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E013.hgt.zip',
		'N49E014.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E014.hgt.zip',
		'N49E015.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E015.hgt.zip',
		'N49E016.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E016.hgt.zip',
		'N49E017.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E017.hgt.zip',
		'N49E018.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E018.hgt.zip',
		'N49E019.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E019.hgt.zip',
		'N49E020.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E020.hgt.zip',
		'N49E021.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E021.hgt.zip',
		'N49E022.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E022.hgt.zip',
		'N49E023.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E023.hgt.zip',
		'N49E024.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E024.hgt.zip',
		'N49E025.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E025.hgt.zip',
		'N49E026.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E026.hgt.zip',
		'N49E027.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E027.hgt.zip',
		'N49E028.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E028.hgt.zip',
		'N49E029.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E029.hgt.zip',
		'N49E030.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E030.hgt.zip',
		'N49E031.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E031.hgt.zip',
		'N49E032.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E032.hgt.zip',
		'N49E033.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E033.hgt.zip',
		'N49E034.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E034.hgt.zip',
		'N49E035.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E035.hgt.zip',
		'N49E036.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E036.hgt.zip',
		'N49E037.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E037.hgt.zip',
		'N49E038.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E038.hgt.zip',
		'N49E039.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E039.hgt.zip',
		'N49E040.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E040.hgt.zip',
		'N49E041.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E041.hgt.zip',
		'N49E042.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E042.hgt.zip',
		'N49E043.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E043.hgt.zip',
		'N49E044.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E044.hgt.zip',
		'N49E045.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E045.hgt.zip',
		'N49E046.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E046.hgt.zip',
		'N49E047.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E047.hgt.zip',
		'N49E048.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E048.hgt.zip',
		'N49E049.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E049.hgt.zip',
		'N49E050.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E050.hgt.zip',
		'N49E051.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E051.hgt.zip',
		'N49E052.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E052.hgt.zip',
		'N49E053.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E053.hgt.zip',
		'N49E054.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E054.hgt.zip',
		'N49E055.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E055.hgt.zip',
		'N49E056.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E056.hgt.zip',
		'N49E057.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E057.hgt.zip',
		'N49E058.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E058.hgt.zip',
		'N49E059.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E059.hgt.zip',
		'N49E060.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E060.hgt.zip',
		'N49E061.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E061.hgt.zip',
		'N49E062.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E062.hgt.zip',
		'N49E063.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E063.hgt.zip',
		'N49E064.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E064.hgt.zip',
		'N49E065.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E065.hgt.zip',
		'N49E066.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E066.hgt.zip',
		'N49E067.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E067.hgt.zip',
		'N49E068.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E068.hgt.zip',
		'N49E069.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E069.hgt.zip',
		'N49E070.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E070.hgt.zip',
		'N49E071.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E071.hgt.zip',
		'N49E072.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E072.hgt.zip',
		'N49E073.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E073.hgt.zip',
		'N49E074.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E074.hgt.zip',
		'N49E075.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E075.hgt.zip',
		'N49E076.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E076.hgt.zip',
		'N49E077.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E077.hgt.zip',
		'N49E078.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E078.hgt.zip',
		'N49E079.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E079.hgt.zip',
		'N49E080.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E080.hgt.zip',
		'N49E081.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E081.hgt.zip',
		'N49E082.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E082.hgt.zip',
		'N49E083.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E083.hgt.zip',
		'N49E084.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E084.hgt.zip',
		'N49E085.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E085.hgt.zip',
		'N49E086.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E086.hgt.zip',
		'N49E087.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E087.hgt.zip',
		'N49E088.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E088.hgt.zip',
		'N49E089.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E089.hgt.zip',
		'N49E090.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E090.hgt.zip',
		'N49E091.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E091.hgt.zip',
		'N49E092.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E092.hgt.zip',
		'N49E093.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E093.hgt.zip',
		'N49E094.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E094.hgt.zip',
		'N49E095.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E095.hgt.zip',
		'N49E096.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E096.hgt.zip',
		'N49E097.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E097.hgt.zip',
		'N49E098.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E098.hgt.zip',
		'N49E099.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E099.hgt.zip',
		'N49E100.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E100.hgt.zip',
		'N49E101.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E101.hgt.zip',
		'N49E102.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E102.hgt.zip',
		'N49E103.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E103.hgt.zip',
		'N49E104.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E104.hgt.zip',
		'N49E105.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E105.hgt.zip',
		'N49E106.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E106.hgt.zip',
		'N49E107.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E107.hgt.zip',
		'N49E108.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E108.hgt.zip',
		'N49E109.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E109.hgt.zip',
		'N49E110.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E110.hgt.zip',
		'N49E111.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E111.hgt.zip',
		'N49E112.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E112.hgt.zip',
		'N49E113.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E113.hgt.zip',
		'N49E114.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E114.hgt.zip',
		'N49E115.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E115.hgt.zip',
		'N49E116.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E116.hgt.zip',
		'N49E117.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E117.hgt.zip',
		'N49E118.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E118.hgt.zip',
		'N49E119.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E119.hgt.zip',
		'N49E120.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E120.hgt.zip',
		'N49E121.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E121.hgt.zip',
		'N49E122.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E122.hgt.zip',
		'N49E123.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E123.hgt.zip',
		'N49E124.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E124.hgt.zip',
		'N49E125.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E125.hgt.zip',
		'N49E126.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E126.hgt.zip',
		'N49E127.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E127.hgt.zip',
		'N49E128.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E128.hgt.zip',
		'N49E129.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E129.hgt.zip',
		'N49E130.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E130.hgt.zip',
		'N49E131.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E131.hgt.zip',
		'N49E132.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E132.hgt.zip',
		'N49E133.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E133.hgt.zip',
		'N49E134.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E134.hgt.zip',
		'N49E135.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E135.hgt.zip',
		'N49E136.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E136.hgt.zip',
		'N49E137.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E137.hgt.zip',
		'N49E138.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E138.hgt.zip',
		'N49E139.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E139.hgt.zip',
		'N49E140.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E140.hgt.zip',
		'N49E142.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E142.hgt.zip',
		'N49E143.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E143.hgt.zip',
		'N49E144.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E144.hgt.zip',
		'N49E154.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E154.hgt.zip',
		'N49E155.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49E155.hgt.zip',
		'N49W001.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49W001.hgt.zip',
		'N49W002.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49W002.hgt.zip',
		'N49W003.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49W003.hgt.zip',
		'N49W006.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49W006.hgt.zip',
		'N49W007.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N49W007.hgt.zip',
		'N49W054.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N49W054.hgt.zip',
		'N49W055.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N49W055.hgt.zip',
		'N49W056.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N49W056.hgt.zip',
		'N49W057.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N49W057.hgt.zip',
		'N49W058.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N49W058.hgt.zip',
		'N49W059.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N49W059.hgt.zip',
		'N49W062.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N49W062.hgt.zip',
		'N49W063.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N49W063.hgt.zip',
		'N49W064.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N49W064.hgt.zip',
		'N49W065.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N49W065.hgt.zip',
		'N49W066.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N49W066.hgt.zip',
		'N49W067.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N49W067.hgt.zip',
		'N49W068.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N49W068.hgt.zip',
		'N49W069.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N49W069.hgt.zip',
		'N49W070.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N49W070.hgt.zip',
		'N49W071.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N49W071.hgt.zip',
		'N49W072.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N49W072.hgt.zip',
		'N49W073.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N49W073.hgt.zip',
		'N49W074.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N49W074.hgt.zip',
		'N49W075.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N49W075.hgt.zip',
		'N49W076.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N49W076.hgt.zip',
		'N49W077.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N49W077.hgt.zip',
		'N49W078.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N49W078.hgt.zip',
		'N49W079.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N49W079.hgt.zip',
		'N49W080.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N49W080.hgt.zip',
		'N49W081.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N49W081.hgt.zip',
		'N49W082.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N49W082.hgt.zip',
		'N49W083.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N49W083.hgt.zip',
		'N49W084.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N49W084.hgt.zip',
		'N49W085.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N49W085.hgt.zip',
		'N49W086.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N49W086.hgt.zip',
		'N49W087.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N49W087.hgt.zip',
		'N49W088.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N49W088.hgt.zip',
		'N49W089.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N49W089.hgt.zip',
		'N49W090.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N49W090.hgt.zip',
		'N49W091.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N49W091.hgt.zip',
		'N49W092.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N49W092.hgt.zip',
		'N49W093.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N49W093.hgt.zip',
		'N49W094.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N49W094.hgt.zip',
		'N49W095.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N49W095.hgt.zip',
		'N49W096.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N49W096.hgt.zip',
		'N49W097.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N49W097.hgt.zip',
		'N49W098.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N49W098.hgt.zip',
		'N49W099.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N49W099.hgt.zip',
		'N49W100.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N49W100.hgt.zip',
		'N49W101.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N49W101.hgt.zip',
		'N49W102.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N49W102.hgt.zip',
		'N49W103.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N49W103.hgt.zip',
		'N49W104.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N49W104.hgt.zip',
		'N49W105.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N49W105.hgt.zip',
		'N49W106.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N49W106.hgt.zip',
		'N49W107.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N49W107.hgt.zip',
		'N49W108.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N49W108.hgt.zip',
		'N49W109.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N49W109.hgt.zip',
		'N49W110.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N49W110.hgt.zip',
		'N49W111.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N49W111.hgt.zip',
		'N49W112.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N49W112.hgt.zip',
		'N49W113.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N49W113.hgt.zip',
		'N49W114.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N49W114.hgt.zip',
		'N49W115.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N49W115.hgt.zip',
		'N49W116.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N49W116.hgt.zip',
		'N49W117.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N49W117.hgt.zip',
		'N49W118.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N49W118.hgt.zip',
		'N49W119.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N49W119.hgt.zip',
		'N49W120.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N49W120.hgt.zip',
		'N49W121.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N49W121.hgt.zip',
		'N49W122.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N49W122.hgt.zip',
		'N49W123.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N49W123.hgt.zip',
		'N49W124.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N49W124.hgt.zip',
		'N49W125.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N49W125.hgt.zip',
		'N49W126.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N49W126.hgt.zip',
		'N49W127.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N49W127.hgt.zip',
		'N49W128.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N49W128.hgt.zip',
		'N50E000.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E000.hgt.zip',
		'N50E001.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E001.hgt.zip',
		'N50E002.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E002.hgt.zip',
		'N50E003.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E003.hgt.zip',
		'N50E004.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E004.hgt.zip',
		'N50E005.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E005.hgt.zip',
		'N50E006.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E006.hgt.zip',
		'N50E007.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E007.hgt.zip',
		'N50E008.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E008.hgt.zip',
		'N50E009.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E009.hgt.zip',
		'N50E010.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E010.hgt.zip',
		'N50E011.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E011.hgt.zip',
		'N50E012.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E012.hgt.zip',
		'N50E013.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E013.hgt.zip',
		'N50E014.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E014.hgt.zip',
		'N50E015.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E015.hgt.zip',
		'N50E016.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E016.hgt.zip',
		'N50E017.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E017.hgt.zip',
		'N50E018.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E018.hgt.zip',
		'N50E019.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E019.hgt.zip',
		'N50E020.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E020.hgt.zip',
		'N50E021.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E021.hgt.zip',
		'N50E022.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E022.hgt.zip',
		'N50E023.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E023.hgt.zip',
		'N50E024.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E024.hgt.zip',
		'N50E025.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E025.hgt.zip',
		'N50E026.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E026.hgt.zip',
		'N50E027.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E027.hgt.zip',
		'N50E028.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E028.hgt.zip',
		'N50E029.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E029.hgt.zip',
		'N50E030.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E030.hgt.zip',
		'N50E031.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E031.hgt.zip',
		'N50E032.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E032.hgt.zip',
		'N50E033.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E033.hgt.zip',
		'N50E034.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E034.hgt.zip',
		'N50E035.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E035.hgt.zip',
		'N50E036.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E036.hgt.zip',
		'N50E037.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E037.hgt.zip',
		'N50E038.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E038.hgt.zip',
		'N50E039.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E039.hgt.zip',
		'N50E040.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E040.hgt.zip',
		'N50E041.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E041.hgt.zip',
		'N50E042.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E042.hgt.zip',
		'N50E043.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E043.hgt.zip',
		'N50E044.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E044.hgt.zip',
		'N50E045.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E045.hgt.zip',
		'N50E046.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E046.hgt.zip',
		'N50E047.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E047.hgt.zip',
		'N50E048.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E048.hgt.zip',
		'N50E049.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E049.hgt.zip',
		'N50E050.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E050.hgt.zip',
		'N50E051.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E051.hgt.zip',
		'N50E052.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E052.hgt.zip',
		'N50E053.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E053.hgt.zip',
		'N50E054.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E054.hgt.zip',
		'N50E055.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E055.hgt.zip',
		'N50E056.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E056.hgt.zip',
		'N50E057.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E057.hgt.zip',
		'N50E058.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E058.hgt.zip',
		'N50E059.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E059.hgt.zip',
		'N50E060.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E060.hgt.zip',
		'N50E061.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E061.hgt.zip',
		'N50E062.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E062.hgt.zip',
		'N50E063.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E063.hgt.zip',
		'N50E064.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E064.hgt.zip',
		'N50E065.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E065.hgt.zip',
		'N50E066.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E066.hgt.zip',
		'N50E067.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E067.hgt.zip',
		'N50E068.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E068.hgt.zip',
		'N50E069.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E069.hgt.zip',
		'N50E070.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E070.hgt.zip',
		'N50E071.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E071.hgt.zip',
		'N50E072.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E072.hgt.zip',
		'N50E073.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E073.hgt.zip',
		'N50E074.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E074.hgt.zip',
		'N50E075.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E075.hgt.zip',
		'N50E076.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E076.hgt.zip',
		'N50E077.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E077.hgt.zip',
		'N50E078.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E078.hgt.zip',
		'N50E079.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E079.hgt.zip',
		'N50E080.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E080.hgt.zip',
		'N50E081.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E081.hgt.zip',
		'N50E082.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E082.hgt.zip',
		'N50E083.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E083.hgt.zip',
		'N50E084.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E084.hgt.zip',
		'N50E085.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E085.hgt.zip',
		'N50E086.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E086.hgt.zip',
		'N50E087.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E087.hgt.zip',
		'N50E088.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E088.hgt.zip',
		'N50E089.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E089.hgt.zip',
		'N50E090.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E090.hgt.zip',
		'N50E091.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E091.hgt.zip',
		'N50E092.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E092.hgt.zip',
		'N50E093.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E093.hgt.zip',
		'N50E094.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E094.hgt.zip',
		'N50E095.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E095.hgt.zip',
		'N50E096.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E096.hgt.zip',
		'N50E097.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E097.hgt.zip',
		'N50E098.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E098.hgt.zip',
		'N50E099.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E099.hgt.zip',
		'N50E100.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E100.hgt.zip',
		'N50E101.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E101.hgt.zip',
		'N50E102.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E102.hgt.zip',
		'N50E103.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E103.hgt.zip',
		'N50E104.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E104.hgt.zip',
		'N50E105.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E105.hgt.zip',
		'N50E106.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E106.hgt.zip',
		'N50E107.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E107.hgt.zip',
		'N50E108.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E108.hgt.zip',
		'N50E109.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E109.hgt.zip',
		'N50E110.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E110.hgt.zip',
		'N50E111.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E111.hgt.zip',
		'N50E112.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E112.hgt.zip',
		'N50E113.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E113.hgt.zip',
		'N50E114.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E114.hgt.zip',
		'N50E115.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E115.hgt.zip',
		'N50E116.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E116.hgt.zip',
		'N50E117.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E117.hgt.zip',
		'N50E118.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E118.hgt.zip',
		'N50E119.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E119.hgt.zip',
		'N50E120.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E120.hgt.zip',
		'N50E121.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E121.hgt.zip',
		'N50E122.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E122.hgt.zip',
		'N50E123.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E123.hgt.zip',
		'N50E124.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E124.hgt.zip',
		'N50E125.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E125.hgt.zip',
		'N50E126.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E126.hgt.zip',
		'N50E127.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E127.hgt.zip',
		'N50E128.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E128.hgt.zip',
		'N50E129.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E129.hgt.zip',
		'N50E130.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E130.hgt.zip',
		'N50E131.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E131.hgt.zip',
		'N50E132.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E132.hgt.zip',
		'N50E133.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E133.hgt.zip',
		'N50E134.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E134.hgt.zip',
		'N50E135.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E135.hgt.zip',
		'N50E136.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E136.hgt.zip',
		'N50E137.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E137.hgt.zip',
		'N50E138.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E138.hgt.zip',
		'N50E139.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E139.hgt.zip',
		'N50E140.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E140.hgt.zip',
		'N50E142.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E142.hgt.zip',
		'N50E143.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E143.hgt.zip',
		'N50E154.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E154.hgt.zip',
		'N50E155.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E155.hgt.zip',
		'N50E156.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50E156.hgt.zip',
		'N50W001.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50W001.hgt.zip',
		'N50W002.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50W002.hgt.zip',
		'N50W003.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50W003.hgt.zip',
		'N50W004.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50W004.hgt.zip',
		'N50W005.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50W005.hgt.zip',
		'N50W006.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N50W006.hgt.zip',
		'N50W056.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N50W056.hgt.zip',
		'N50W057.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N50W057.hgt.zip',
		'N50W058.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N50W058.hgt.zip',
		'N50W059.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N50W059.hgt.zip',
		'N50W060.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N50W060.hgt.zip',
		'N50W061.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N50W061.hgt.zip',
		'N50W062.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N50W062.hgt.zip',
		'N50W063.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N50W063.hgt.zip',
		'N50W064.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N50W064.hgt.zip',
		'N50W065.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N50W065.hgt.zip',
		'N50W066.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N50W066.hgt.zip',
		'N50W067.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N50W067.hgt.zip',
		'N50W068.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N50W068.hgt.zip',
		'N50W069.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N50W069.hgt.zip',
		'N50W070.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N50W070.hgt.zip',
		'N50W071.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N50W071.hgt.zip',
		'N50W072.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N50W072.hgt.zip',
		'N50W073.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N50W073.hgt.zip',
		'N50W074.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N50W074.hgt.zip',
		'N50W075.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N50W075.hgt.zip',
		'N50W076.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N50W076.hgt.zip',
		'N50W077.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N50W077.hgt.zip',
		'N50W078.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N50W078.hgt.zip',
		'N50W079.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N50W079.hgt.zip',
		'N50W080.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N50W080.hgt.zip',
		'N50W081.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N50W081.hgt.zip',
		'N50W082.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N50W082.hgt.zip',
		'N50W083.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N50W083.hgt.zip',
		'N50W084.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N50W084.hgt.zip',
		'N50W085.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N50W085.hgt.zip',
		'N50W086.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N50W086.hgt.zip',
		'N50W087.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N50W087.hgt.zip',
		'N50W088.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N50W088.hgt.zip',
		'N50W089.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N50W089.hgt.zip',
		'N50W090.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N50W090.hgt.zip',
		'N50W091.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N50W091.hgt.zip',
		'N50W092.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N50W092.hgt.zip',
		'N50W093.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N50W093.hgt.zip',
		'N50W094.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N50W094.hgt.zip',
		'N50W095.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N50W095.hgt.zip',
		'N50W096.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N50W096.hgt.zip',
		'N50W097.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N50W097.hgt.zip',
		'N50W098.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N50W098.hgt.zip',
		'N50W099.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N50W099.hgt.zip',
		'N50W100.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N50W100.hgt.zip',
		'N50W101.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N50W101.hgt.zip',
		'N50W102.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N50W102.hgt.zip',
		'N50W103.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N50W103.hgt.zip',
		'N50W104.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N50W104.hgt.zip',
		'N50W105.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N50W105.hgt.zip',
		'N50W106.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N50W106.hgt.zip',
		'N50W107.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N50W107.hgt.zip',
		'N50W108.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N50W108.hgt.zip',
		'N50W109.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N50W109.hgt.zip',
		'N50W110.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N50W110.hgt.zip',
		'N50W111.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N50W111.hgt.zip',
		'N50W112.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N50W112.hgt.zip',
		'N50W113.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N50W113.hgt.zip',
		'N50W114.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N50W114.hgt.zip',
		'N50W115.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N50W115.hgt.zip',
		'N50W116.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N50W116.hgt.zip',
		'N50W117.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N50W117.hgt.zip',
		'N50W118.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N50W118.hgt.zip',
		'N50W119.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N50W119.hgt.zip',
		'N50W120.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N50W120.hgt.zip',
		'N50W121.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N50W121.hgt.zip',
		'N50W122.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N50W122.hgt.zip',
		'N50W123.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N50W123.hgt.zip',
		'N50W124.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N50W124.hgt.zip',
		'N50W125.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N50W125.hgt.zip',
		'N50W126.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N50W126.hgt.zip',
		'N50W127.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N50W127.hgt.zip',
		'N50W128.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N50W128.hgt.zip',
		'N50W129.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N50W129.hgt.zip',
		'N50W130.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N50W130.hgt.zip',
		'N51E000.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E000.hgt.zip',
		'N51E001.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E001.hgt.zip',
		'N51E002.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E002.hgt.zip',
		'N51E003.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E003.hgt.zip',
		'N51E004.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E004.hgt.zip',
		'N51E005.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E005.hgt.zip',
		'N51E006.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E006.hgt.zip',
		'N51E007.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E007.hgt.zip',
		'N51E008.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E008.hgt.zip',
		'N51E009.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E009.hgt.zip',
		'N51E010.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E010.hgt.zip',
		'N51E011.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E011.hgt.zip',
		'N51E012.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E012.hgt.zip',
		'N51E013.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E013.hgt.zip',
		'N51E014.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E014.hgt.zip',
		'N51E015.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E015.hgt.zip',
		'N51E016.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E016.hgt.zip',
		'N51E017.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E017.hgt.zip',
		'N51E018.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E018.hgt.zip',
		'N51E019.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E019.hgt.zip',
		'N51E020.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E020.hgt.zip',
		'N51E021.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E021.hgt.zip',
		'N51E022.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E022.hgt.zip',
		'N51E023.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E023.hgt.zip',
		'N51E024.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E024.hgt.zip',
		'N51E025.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E025.hgt.zip',
		'N51E026.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E026.hgt.zip',
		'N51E027.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E027.hgt.zip',
		'N51E028.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E028.hgt.zip',
		'N51E029.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E029.hgt.zip',
		'N51E030.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E030.hgt.zip',
		'N51E031.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E031.hgt.zip',
		'N51E032.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E032.hgt.zip',
		'N51E033.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E033.hgt.zip',
		'N51E034.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E034.hgt.zip',
		'N51E035.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E035.hgt.zip',
		'N51E036.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E036.hgt.zip',
		'N51E037.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E037.hgt.zip',
		'N51E038.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E038.hgt.zip',
		'N51E039.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E039.hgt.zip',
		'N51E040.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E040.hgt.zip',
		'N51E041.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E041.hgt.zip',
		'N51E042.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E042.hgt.zip',
		'N51E043.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E043.hgt.zip',
		'N51E044.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E044.hgt.zip',
		'N51E045.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E045.hgt.zip',
		'N51E046.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E046.hgt.zip',
		'N51E047.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E047.hgt.zip',
		'N51E048.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E048.hgt.zip',
		'N51E049.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E049.hgt.zip',
		'N51E050.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E050.hgt.zip',
		'N51E051.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E051.hgt.zip',
		'N51E052.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E052.hgt.zip',
		'N51E053.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E053.hgt.zip',
		'N51E054.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E054.hgt.zip',
		'N51E055.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E055.hgt.zip',
		'N51E056.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E056.hgt.zip',
		'N51E057.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E057.hgt.zip',
		'N51E058.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E058.hgt.zip',
		'N51E059.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E059.hgt.zip',
		'N51E060.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E060.hgt.zip',
		'N51E061.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E061.hgt.zip',
		'N51E062.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E062.hgt.zip',
		'N51E063.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E063.hgt.zip',
		'N51E064.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E064.hgt.zip',
		'N51E065.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E065.hgt.zip',
		'N51E066.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E066.hgt.zip',
		'N51E067.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E067.hgt.zip',
		'N51E068.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E068.hgt.zip',
		'N51E069.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E069.hgt.zip',
		'N51E070.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E070.hgt.zip',
		'N51E071.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E071.hgt.zip',
		'N51E072.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E072.hgt.zip',
		'N51E073.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E073.hgt.zip',
		'N51E074.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E074.hgt.zip',
		'N51E075.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E075.hgt.zip',
		'N51E076.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E076.hgt.zip',
		'N51E077.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E077.hgt.zip',
		'N51E078.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E078.hgt.zip',
		'N51E079.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E079.hgt.zip',
		'N51E080.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E080.hgt.zip',
		'N51E081.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E081.hgt.zip',
		'N51E082.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E082.hgt.zip',
		'N51E083.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E083.hgt.zip',
		'N51E084.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E084.hgt.zip',
		'N51E085.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E085.hgt.zip',
		'N51E086.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E086.hgt.zip',
		'N51E087.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E087.hgt.zip',
		'N51E088.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E088.hgt.zip',
		'N51E089.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E089.hgt.zip',
		'N51E090.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E090.hgt.zip',
		'N51E091.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E091.hgt.zip',
		'N51E092.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E092.hgt.zip',
		'N51E093.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E093.hgt.zip',
		'N51E094.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E094.hgt.zip',
		'N51E095.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E095.hgt.zip',
		'N51E096.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E096.hgt.zip',
		'N51E097.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E097.hgt.zip',
		'N51E098.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E098.hgt.zip',
		'N51E099.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E099.hgt.zip',
		'N51E100.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E100.hgt.zip',
		'N51E101.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E101.hgt.zip',
		'N51E102.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E102.hgt.zip',
		'N51E103.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E103.hgt.zip',
		'N51E104.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E104.hgt.zip',
		'N51E105.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E105.hgt.zip',
		'N51E106.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E106.hgt.zip',
		'N51E107.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E107.hgt.zip',
		'N51E108.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E108.hgt.zip',
		'N51E109.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E109.hgt.zip',
		'N51E110.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E110.hgt.zip',
		'N51E111.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E111.hgt.zip',
		'N51E112.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E112.hgt.zip',
		'N51E113.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E113.hgt.zip',
		'N51E114.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E114.hgt.zip',
		'N51E115.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E115.hgt.zip',
		'N51E116.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E116.hgt.zip',
		'N51E117.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E117.hgt.zip',
		'N51E118.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E118.hgt.zip',
		'N51E119.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E119.hgt.zip',
		'N51E120.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E120.hgt.zip',
		'N51E121.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E121.hgt.zip',
		'N51E122.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E122.hgt.zip',
		'N51E123.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E123.hgt.zip',
		'N51E124.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E124.hgt.zip',
		'N51E125.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E125.hgt.zip',
		'N51E126.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E126.hgt.zip',
		'N51E127.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E127.hgt.zip',
		'N51E128.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E128.hgt.zip',
		'N51E129.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E129.hgt.zip',
		'N51E130.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E130.hgt.zip',
		'N51E131.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E131.hgt.zip',
		'N51E132.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E132.hgt.zip',
		'N51E133.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E133.hgt.zip',
		'N51E134.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E134.hgt.zip',
		'N51E135.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E135.hgt.zip',
		'N51E136.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E136.hgt.zip',
		'N51E137.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E137.hgt.zip',
		'N51E138.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E138.hgt.zip',
		'N51E139.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E139.hgt.zip',
		'N51E140.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E140.hgt.zip',
		'N51E141.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E141.hgt.zip',
		'N51E142.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E142.hgt.zip',
		'N51E143.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E143.hgt.zip',
		'N51E156.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E156.hgt.zip',
		'N51E157.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E157.hgt.zip',
		'N51E158.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51E158.hgt.zip',
		'N51E177.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N51E177.hgt.zip',
		'N51E178.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N51E178.hgt.zip',
		'N51E179.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N51E179.hgt.zip',
		'N51W001.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51W001.hgt.zip',
		'N51W002.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51W002.hgt.zip',
		'N51W003.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51W003.hgt.zip',
		'N51W004.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51W004.hgt.zip',
		'N51W005.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51W005.hgt.zip',
		'N51W006.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51W006.hgt.zip',
		'N51W008.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51W008.hgt.zip',
		'N51W009.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51W009.hgt.zip',
		'N51W010.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51W010.hgt.zip',
		'N51W011.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N51W011.hgt.zip',
		'N51W056.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N51W056.hgt.zip',
		'N51W057.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N51W057.hgt.zip',
		'N51W058.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N51W058.hgt.zip',
		'N51W059.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N51W059.hgt.zip',
		'N51W060.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N51W060.hgt.zip',
		'N51W061.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N51W061.hgt.zip',
		'N51W062.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N51W062.hgt.zip',
		'N51W063.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N51W063.hgt.zip',
		'N51W064.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N51W064.hgt.zip',
		'N51W065.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N51W065.hgt.zip',
		'N51W066.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N51W066.hgt.zip',
		'N51W067.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N51W067.hgt.zip',
		'N51W068.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N51W068.hgt.zip',
		'N51W069.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N51W069.hgt.zip',
		'N51W070.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N51W070.hgt.zip',
		'N51W071.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N51W071.hgt.zip',
		'N51W072.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N51W072.hgt.zip',
		'N51W073.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N51W073.hgt.zip',
		'N51W074.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N51W074.hgt.zip',
		'N51W075.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N51W075.hgt.zip',
		'N51W076.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N51W076.hgt.zip',
		'N51W077.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N51W077.hgt.zip',
		'N51W078.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N51W078.hgt.zip',
		'N51W079.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N51W079.hgt.zip',
		'N51W080.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N51W080.hgt.zip',
		'N51W081.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N51W081.hgt.zip',
		'N51W082.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N51W082.hgt.zip',
		'N51W083.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N51W083.hgt.zip',
		'N51W084.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N51W084.hgt.zip',
		'N51W085.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N51W085.hgt.zip',
		'N51W086.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N51W086.hgt.zip',
		'N51W087.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N51W087.hgt.zip',
		'N51W088.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N51W088.hgt.zip',
		'N51W089.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N51W089.hgt.zip',
		'N51W090.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N51W090.hgt.zip',
		'N51W091.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N51W091.hgt.zip',
		'N51W092.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N51W092.hgt.zip',
		'N51W093.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N51W093.hgt.zip',
		'N51W094.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N51W094.hgt.zip',
		'N51W095.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N51W095.hgt.zip',
		'N51W096.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N51W096.hgt.zip',
		'N51W097.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N51W097.hgt.zip',
		'N51W098.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N51W098.hgt.zip',
		'N51W099.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N51W099.hgt.zip',
		'N51W100.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N51W100.hgt.zip',
		'N51W101.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N51W101.hgt.zip',
		'N51W102.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N51W102.hgt.zip',
		'N51W103.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N51W103.hgt.zip',
		'N51W104.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N51W104.hgt.zip',
		'N51W105.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N51W105.hgt.zip',
		'N51W106.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N51W106.hgt.zip',
		'N51W107.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N51W107.hgt.zip',
		'N51W108.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N51W108.hgt.zip',
		'N51W109.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N51W109.hgt.zip',
		'N51W110.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N51W110.hgt.zip',
		'N51W111.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N51W111.hgt.zip',
		'N51W112.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N51W112.hgt.zip',
		'N51W113.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N51W113.hgt.zip',
		'N51W114.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N51W114.hgt.zip',
		'N51W115.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N51W115.hgt.zip',
		'N51W116.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N51W116.hgt.zip',
		'N51W117.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N51W117.hgt.zip',
		'N51W118.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N51W118.hgt.zip',
		'N51W119.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N51W119.hgt.zip',
		'N51W120.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N51W120.hgt.zip',
		'N51W121.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N51W121.hgt.zip',
		'N51W122.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N51W122.hgt.zip',
		'N51W123.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N51W123.hgt.zip',
		'N51W124.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N51W124.hgt.zip',
		'N51W125.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N51W125.hgt.zip',
		'N51W126.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N51W126.hgt.zip',
		'N51W127.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N51W127.hgt.zip',
		'N51W128.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N51W128.hgt.zip',
		'N51W129.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N51W129.hgt.zip',
		'N51W131.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N51W131.hgt.zip',
		'N51W132.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N51W132.hgt.zip',
		'N51W176.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N51W176.hgt.zip',
		'N51W177.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N51W177.hgt.zip',
		'N51W178.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N51W178.hgt.zip',
		'N51W179.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N51W179.hgt.zip',
		'N51W180.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N51W180.hgt.zip',
		'N52E000.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E000.hgt.zip',
		'N52E001.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E001.hgt.zip',
		'N52E004.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E004.hgt.zip',
		'N52E005.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E005.hgt.zip',
		'N52E006.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E006.hgt.zip',
		'N52E007.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E007.hgt.zip',
		'N52E008.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E008.hgt.zip',
		'N52E009.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E009.hgt.zip',
		'N52E010.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E010.hgt.zip',
		'N52E011.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E011.hgt.zip',
		'N52E012.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E012.hgt.zip',
		'N52E013.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E013.hgt.zip',
		'N52E014.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E014.hgt.zip',
		'N52E015.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E015.hgt.zip',
		'N52E016.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E016.hgt.zip',
		'N52E017.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E017.hgt.zip',
		'N52E018.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E018.hgt.zip',
		'N52E019.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E019.hgt.zip',
		'N52E020.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E020.hgt.zip',
		'N52E021.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E021.hgt.zip',
		'N52E022.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E022.hgt.zip',
		'N52E023.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E023.hgt.zip',
		'N52E024.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E024.hgt.zip',
		'N52E025.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E025.hgt.zip',
		'N52E026.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E026.hgt.zip',
		'N52E027.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E027.hgt.zip',
		'N52E028.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E028.hgt.zip',
		'N52E029.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E029.hgt.zip',
		'N52E030.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E030.hgt.zip',
		'N52E031.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E031.hgt.zip',
		'N52E032.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E032.hgt.zip',
		'N52E033.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E033.hgt.zip',
		'N52E034.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E034.hgt.zip',
		'N52E035.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E035.hgt.zip',
		'N52E036.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E036.hgt.zip',
		'N52E037.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E037.hgt.zip',
		'N52E038.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E038.hgt.zip',
		'N52E039.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E039.hgt.zip',
		'N52E040.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E040.hgt.zip',
		'N52E041.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E041.hgt.zip',
		'N52E042.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E042.hgt.zip',
		'N52E043.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E043.hgt.zip',
		'N52E044.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E044.hgt.zip',
		'N52E045.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E045.hgt.zip',
		'N52E046.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E046.hgt.zip',
		'N52E047.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E047.hgt.zip',
		'N52E048.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E048.hgt.zip',
		'N52E049.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E049.hgt.zip',
		'N52E050.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E050.hgt.zip',
		'N52E051.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E051.hgt.zip',
		'N52E052.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E052.hgt.zip',
		'N52E053.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E053.hgt.zip',
		'N52E054.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E054.hgt.zip',
		'N52E055.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E055.hgt.zip',
		'N52E056.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E056.hgt.zip',
		'N52E057.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E057.hgt.zip',
		'N52E058.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E058.hgt.zip',
		'N52E059.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E059.hgt.zip',
		'N52E060.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E060.hgt.zip',
		'N52E061.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E061.hgt.zip',
		'N52E062.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E062.hgt.zip',
		'N52E063.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E063.hgt.zip',
		'N52E064.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E064.hgt.zip',
		'N52E065.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E065.hgt.zip',
		'N52E066.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E066.hgt.zip',
		'N52E067.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E067.hgt.zip',
		'N52E068.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E068.hgt.zip',
		'N52E069.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E069.hgt.zip',
		'N52E070.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E070.hgt.zip',
		'N52E071.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E071.hgt.zip',
		'N52E072.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E072.hgt.zip',
		'N52E073.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E073.hgt.zip',
		'N52E074.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E074.hgt.zip',
		'N52E075.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E075.hgt.zip',
		'N52E076.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E076.hgt.zip',
		'N52E077.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E077.hgt.zip',
		'N52E078.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E078.hgt.zip',
		'N52E079.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E079.hgt.zip',
		'N52E080.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E080.hgt.zip',
		'N52E081.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E081.hgt.zip',
		'N52E082.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E082.hgt.zip',
		'N52E083.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E083.hgt.zip',
		'N52E084.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E084.hgt.zip',
		'N52E085.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E085.hgt.zip',
		'N52E086.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E086.hgt.zip',
		'N52E087.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E087.hgt.zip',
		'N52E088.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E088.hgt.zip',
		'N52E089.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E089.hgt.zip',
		'N52E090.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E090.hgt.zip',
		'N52E091.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E091.hgt.zip',
		'N52E092.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E092.hgt.zip',
		'N52E093.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E093.hgt.zip',
		'N52E094.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E094.hgt.zip',
		'N52E095.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E095.hgt.zip',
		'N52E096.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E096.hgt.zip',
		'N52E097.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E097.hgt.zip',
		'N52E098.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E098.hgt.zip',
		'N52E099.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E099.hgt.zip',
		'N52E100.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E100.hgt.zip',
		'N52E101.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E101.hgt.zip',
		'N52E102.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E102.hgt.zip',
		'N52E103.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E103.hgt.zip',
		'N52E104.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E104.hgt.zip',
		'N52E105.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E105.hgt.zip',
		'N52E106.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E106.hgt.zip',
		'N52E107.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E107.hgt.zip',
		'N52E108.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E108.hgt.zip',
		'N52E109.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E109.hgt.zip',
		'N52E110.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E110.hgt.zip',
		'N52E111.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E111.hgt.zip',
		'N52E112.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E112.hgt.zip',
		'N52E113.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E113.hgt.zip',
		'N52E114.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E114.hgt.zip',
		'N52E115.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E115.hgt.zip',
		'N52E116.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E116.hgt.zip',
		'N52E117.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E117.hgt.zip',
		'N52E118.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E118.hgt.zip',
		'N52E119.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E119.hgt.zip',
		'N52E120.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E120.hgt.zip',
		'N52E121.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E121.hgt.zip',
		'N52E122.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E122.hgt.zip',
		'N52E123.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E123.hgt.zip',
		'N52E124.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E124.hgt.zip',
		'N52E125.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E125.hgt.zip',
		'N52E126.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E126.hgt.zip',
		'N52E127.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E127.hgt.zip',
		'N52E128.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E128.hgt.zip',
		'N52E129.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E129.hgt.zip',
		'N52E130.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E130.hgt.zip',
		'N52E131.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E131.hgt.zip',
		'N52E132.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E132.hgt.zip',
		'N52E133.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E133.hgt.zip',
		'N52E134.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E134.hgt.zip',
		'N52E135.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E135.hgt.zip',
		'N52E136.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E136.hgt.zip',
		'N52E137.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E137.hgt.zip',
		'N52E138.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E138.hgt.zip',
		'N52E139.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E139.hgt.zip',
		'N52E140.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E140.hgt.zip',
		'N52E141.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E141.hgt.zip',
		'N52E142.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E142.hgt.zip',
		'N52E143.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E143.hgt.zip',
		'N52E156.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E156.hgt.zip',
		'N52E157.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E157.hgt.zip',
		'N52E158.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52E158.hgt.zip',
		'N52E172.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N52E172.hgt.zip',
		'N52E173.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N52E173.hgt.zip',
		'N52E174.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N52E174.hgt.zip',
		'N52E175.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N52E175.hgt.zip',
		'N52E177.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N52E177.hgt.zip',
		'N52E178.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N52E178.hgt.zip',
		'N52E179.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N52E179.hgt.zip',
		'N52W001.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52W001.hgt.zip',
		'N52W002.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52W002.hgt.zip',
		'N52W003.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52W003.hgt.zip',
		'N52W004.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52W004.hgt.zip',
		'N52W005.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52W005.hgt.zip',
		'N52W006.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52W006.hgt.zip',
		'N52W007.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52W007.hgt.zip',
		'N52W008.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52W008.hgt.zip',
		'N52W009.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52W009.hgt.zip',
		'N52W010.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52W010.hgt.zip',
		'N52W011.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N52W011.hgt.zip',
		'N52W056.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N52W056.hgt.zip',
		'N52W057.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N52W057.hgt.zip',
		'N52W058.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N52W058.hgt.zip',
		'N52W059.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N52W059.hgt.zip',
		'N52W060.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N52W060.hgt.zip',
		'N52W061.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N52W061.hgt.zip',
		'N52W062.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N52W062.hgt.zip',
		'N52W063.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N52W063.hgt.zip',
		'N52W064.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N52W064.hgt.zip',
		'N52W065.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N52W065.hgt.zip',
		'N52W066.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N52W066.hgt.zip',
		'N52W067.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N52W067.hgt.zip',
		'N52W068.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N52W068.hgt.zip',
		'N52W069.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N52W069.hgt.zip',
		'N52W070.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N52W070.hgt.zip',
		'N52W071.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N52W071.hgt.zip',
		'N52W072.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N52W072.hgt.zip',
		'N52W073.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N52W073.hgt.zip',
		'N52W074.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N52W074.hgt.zip',
		'N52W075.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N52W075.hgt.zip',
		'N52W076.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N52W076.hgt.zip',
		'N52W077.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N52W077.hgt.zip',
		'N52W078.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N52W078.hgt.zip',
		'N52W079.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N52W079.hgt.zip',
		'N52W080.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N52W080.hgt.zip',
		'N52W081.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N52W081.hgt.zip',
		'N52W082.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N52W082.hgt.zip',
		'N52W083.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N52W083.hgt.zip',
		'N52W084.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N52W084.hgt.zip',
		'N52W085.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N52W085.hgt.zip',
		'N52W086.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N52W086.hgt.zip',
		'N52W087.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N52W087.hgt.zip',
		'N52W088.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N52W088.hgt.zip',
		'N52W089.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N52W089.hgt.zip',
		'N52W090.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N52W090.hgt.zip',
		'N52W091.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N52W091.hgt.zip',
		'N52W092.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N52W092.hgt.zip',
		'N52W093.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N52W093.hgt.zip',
		'N52W094.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N52W094.hgt.zip',
		'N52W095.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N52W095.hgt.zip',
		'N52W096.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N52W096.hgt.zip',
		'N52W097.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N52W097.hgt.zip',
		'N52W098.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N52W098.hgt.zip',
		'N52W099.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N52W099.hgt.zip',
		'N52W100.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N52W100.hgt.zip',
		'N52W101.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N52W101.hgt.zip',
		'N52W102.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N52W102.hgt.zip',
		'N52W103.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N52W103.hgt.zip',
		'N52W104.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N52W104.hgt.zip',
		'N52W105.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N52W105.hgt.zip',
		'N52W106.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N52W106.hgt.zip',
		'N52W107.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N52W107.hgt.zip',
		'N52W108.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N52W108.hgt.zip',
		'N52W109.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N52W109.hgt.zip',
		'N52W110.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N52W110.hgt.zip',
		'N52W111.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N52W111.hgt.zip',
		'N52W112.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N52W112.hgt.zip',
		'N52W113.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N52W113.hgt.zip',
		'N52W114.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N52W114.hgt.zip',
		'N52W115.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N52W115.hgt.zip',
		'N52W116.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N52W116.hgt.zip',
		'N52W117.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N52W117.hgt.zip',
		'N52W118.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N52W118.hgt.zip',
		'N52W119.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N52W119.hgt.zip',
		'N52W120.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N52W120.hgt.zip',
		'N52W121.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N52W121.hgt.zip',
		'N52W122.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N52W122.hgt.zip',
		'N52W123.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N52W123.hgt.zip',
		'N52W124.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N52W124.hgt.zip',
		'N52W125.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N52W125.hgt.zip',
		'N52W126.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N52W126.hgt.zip',
		'N52W127.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N52W127.hgt.zip',
		'N52W128.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N52W128.hgt.zip',
		'N52W129.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N52W129.hgt.zip',
		'N52W130.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N52W130.hgt.zip',
		'N52W131.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N52W131.hgt.zip',
		'N52W132.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N52W132.hgt.zip',
		'N52W133.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N52W133.hgt.zip',
		'N52W169.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N52W169.hgt.zip',
		'N52W170.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N52W170.hgt.zip',
		'N52W171.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N52W171.hgt.zip',
		'N52W172.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N52W172.hgt.zip',
		'N52W173.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N52W173.hgt.zip',
		'N52W174.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N52W174.hgt.zip',
		'N52W175.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N52W175.hgt.zip',
		'N52W176.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N52W176.hgt.zip',
		'N52W177.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N52W177.hgt.zip',
		'N53E000.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E000.hgt.zip',
		'N53E004.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E004.hgt.zip',
		'N53E005.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E005.hgt.zip',
		'N53E006.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E006.hgt.zip',
		'N53E007.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E007.hgt.zip',
		'N53E008.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E008.hgt.zip',
		'N53E009.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E009.hgt.zip',
		'N53E010.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E010.hgt.zip',
		'N53E011.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E011.hgt.zip',
		'N53E012.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E012.hgt.zip',
		'N53E013.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E013.hgt.zip',
		'N53E014.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E014.hgt.zip',
		'N53E015.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E015.hgt.zip',
		'N53E016.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E016.hgt.zip',
		'N53E017.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E017.hgt.zip',
		'N53E018.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E018.hgt.zip',
		'N53E019.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E019.hgt.zip',
		'N53E020.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E020.hgt.zip',
		'N53E021.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E021.hgt.zip',
		'N53E022.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E022.hgt.zip',
		'N53E023.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E023.hgt.zip',
		'N53E024.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E024.hgt.zip',
		'N53E025.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E025.hgt.zip',
		'N53E026.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E026.hgt.zip',
		'N53E027.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E027.hgt.zip',
		'N53E028.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E028.hgt.zip',
		'N53E029.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E029.hgt.zip',
		'N53E030.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E030.hgt.zip',
		'N53E031.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E031.hgt.zip',
		'N53E032.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E032.hgt.zip',
		'N53E033.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E033.hgt.zip',
		'N53E034.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E034.hgt.zip',
		'N53E035.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E035.hgt.zip',
		'N53E036.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E036.hgt.zip',
		'N53E037.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E037.hgt.zip',
		'N53E038.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E038.hgt.zip',
		'N53E039.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E039.hgt.zip',
		'N53E040.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E040.hgt.zip',
		'N53E041.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E041.hgt.zip',
		'N53E042.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E042.hgt.zip',
		'N53E043.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E043.hgt.zip',
		'N53E044.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E044.hgt.zip',
		'N53E045.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E045.hgt.zip',
		'N53E046.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E046.hgt.zip',
		'N53E047.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E047.hgt.zip',
		'N53E048.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E048.hgt.zip',
		'N53E049.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E049.hgt.zip',
		'N53E050.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E050.hgt.zip',
		'N53E051.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E051.hgt.zip',
		'N53E052.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E052.hgt.zip',
		'N53E053.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E053.hgt.zip',
		'N53E054.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E054.hgt.zip',
		'N53E055.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E055.hgt.zip',
		'N53E056.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E056.hgt.zip',
		'N53E057.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E057.hgt.zip',
		'N53E058.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E058.hgt.zip',
		'N53E059.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E059.hgt.zip',
		'N53E060.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E060.hgt.zip',
		'N53E061.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E061.hgt.zip',
		'N53E062.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E062.hgt.zip',
		'N53E063.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E063.hgt.zip',
		'N53E064.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E064.hgt.zip',
		'N53E065.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E065.hgt.zip',
		'N53E066.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E066.hgt.zip',
		'N53E067.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E067.hgt.zip',
		'N53E068.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E068.hgt.zip',
		'N53E069.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E069.hgt.zip',
		'N53E070.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E070.hgt.zip',
		'N53E071.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E071.hgt.zip',
		'N53E072.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E072.hgt.zip',
		'N53E073.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E073.hgt.zip',
		'N53E074.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E074.hgt.zip',
		'N53E075.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E075.hgt.zip',
		'N53E076.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E076.hgt.zip',
		'N53E077.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E077.hgt.zip',
		'N53E078.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E078.hgt.zip',
		'N53E079.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E079.hgt.zip',
		'N53E080.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E080.hgt.zip',
		'N53E081.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E081.hgt.zip',
		'N53E082.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E082.hgt.zip',
		'N53E083.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E083.hgt.zip',
		'N53E084.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E084.hgt.zip',
		'N53E085.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E085.hgt.zip',
		'N53E086.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E086.hgt.zip',
		'N53E087.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E087.hgt.zip',
		'N53E088.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E088.hgt.zip',
		'N53E089.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E089.hgt.zip',
		'N53E090.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E090.hgt.zip',
		'N53E091.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E091.hgt.zip',
		'N53E092.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E092.hgt.zip',
		'N53E093.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E093.hgt.zip',
		'N53E094.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E094.hgt.zip',
		'N53E095.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E095.hgt.zip',
		'N53E096.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E096.hgt.zip',
		'N53E097.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E097.hgt.zip',
		'N53E098.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E098.hgt.zip',
		'N53E099.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E099.hgt.zip',
		'N53E100.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E100.hgt.zip',
		'N53E101.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E101.hgt.zip',
		'N53E102.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E102.hgt.zip',
		'N53E103.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E103.hgt.zip',
		'N53E104.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E104.hgt.zip',
		'N53E105.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E105.hgt.zip',
		'N53E106.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E106.hgt.zip',
		'N53E107.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E107.hgt.zip',
		'N53E108.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E108.hgt.zip',
		'N53E109.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E109.hgt.zip',
		'N53E110.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E110.hgt.zip',
		'N53E111.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E111.hgt.zip',
		'N53E112.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E112.hgt.zip',
		'N53E113.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E113.hgt.zip',
		'N53E114.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E114.hgt.zip',
		'N53E115.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E115.hgt.zip',
		'N53E116.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E116.hgt.zip',
		'N53E117.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E117.hgt.zip',
		'N53E118.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E118.hgt.zip',
		'N53E119.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E119.hgt.zip',
		'N53E120.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E120.hgt.zip',
		'N53E121.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E121.hgt.zip',
		'N53E122.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E122.hgt.zip',
		'N53E123.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E123.hgt.zip',
		'N53E124.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E124.hgt.zip',
		'N53E125.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E125.hgt.zip',
		'N53E126.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E126.hgt.zip',
		'N53E127.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E127.hgt.zip',
		'N53E128.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E128.hgt.zip',
		'N53E129.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E129.hgt.zip',
		'N53E130.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E130.hgt.zip',
		'N53E131.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E131.hgt.zip',
		'N53E132.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E132.hgt.zip',
		'N53E133.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E133.hgt.zip',
		'N53E134.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E134.hgt.zip',
		'N53E135.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E135.hgt.zip',
		'N53E136.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E136.hgt.zip',
		'N53E137.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E137.hgt.zip',
		'N53E138.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E138.hgt.zip',
		'N53E139.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E139.hgt.zip',
		'N53E140.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E140.hgt.zip',
		'N53E141.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E141.hgt.zip',
		'N53E142.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E142.hgt.zip',
		'N53E143.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E143.hgt.zip',
		'N53E155.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E155.hgt.zip',
		'N53E156.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E156.hgt.zip',
		'N53E157.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E157.hgt.zip',
		'N53E158.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E158.hgt.zip',
		'N53E159.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E159.hgt.zip',
		'N53E160.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53E160.hgt.zip',
		'N53E172.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N53E172.hgt.zip',
		'N53W001.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53W001.hgt.zip',
		'N53W002.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53W002.hgt.zip',
		'N53W003.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53W003.hgt.zip',
		'N53W004.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53W004.hgt.zip',
		'N53W005.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53W005.hgt.zip',
		'N53W006.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53W006.hgt.zip',
		'N53W007.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53W007.hgt.zip',
		'N53W008.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53W008.hgt.zip',
		'N53W009.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53W009.hgt.zip',
		'N53W010.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53W010.hgt.zip',
		'N53W011.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N53W011.hgt.zip',
		'N53W056.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N53W056.hgt.zip',
		'N53W057.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N53W057.hgt.zip',
		'N53W058.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N53W058.hgt.zip',
		'N53W059.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N53W059.hgt.zip',
		'N53W060.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N53W060.hgt.zip',
		'N53W061.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N53W061.hgt.zip',
		'N53W062.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N53W062.hgt.zip',
		'N53W063.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N53W063.hgt.zip',
		'N53W064.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N53W064.hgt.zip',
		'N53W065.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N53W065.hgt.zip',
		'N53W066.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N53W066.hgt.zip',
		'N53W067.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N53W067.hgt.zip',
		'N53W068.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N53W068.hgt.zip',
		'N53W069.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N53W069.hgt.zip',
		'N53W070.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N53W070.hgt.zip',
		'N53W071.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N53W071.hgt.zip',
		'N53W072.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N53W072.hgt.zip',
		'N53W073.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N53W073.hgt.zip',
		'N53W074.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N53W074.hgt.zip',
		'N53W075.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N53W075.hgt.zip',
		'N53W076.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N53W076.hgt.zip',
		'N53W077.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N53W077.hgt.zip',
		'N53W078.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N53W078.hgt.zip',
		'N53W079.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N53W079.hgt.zip',
		'N53W080.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N53W080.hgt.zip',
		'N53W081.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N53W081.hgt.zip',
		'N53W082.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N53W082.hgt.zip',
		'N53W083.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N53W083.hgt.zip',
		'N53W084.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N53W084.hgt.zip',
		'N53W085.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N53W085.hgt.zip',
		'N53W086.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N53W086.hgt.zip',
		'N53W087.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N53W087.hgt.zip',
		'N53W088.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N53W088.hgt.zip',
		'N53W089.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N53W089.hgt.zip',
		'N53W090.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N53W090.hgt.zip',
		'N53W091.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N53W091.hgt.zip',
		'N53W092.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N53W092.hgt.zip',
		'N53W093.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N53W093.hgt.zip',
		'N53W094.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N53W094.hgt.zip',
		'N53W095.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N53W095.hgt.zip',
		'N53W096.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N53W096.hgt.zip',
		'N53W097.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N53W097.hgt.zip',
		'N53W098.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N53W098.hgt.zip',
		'N53W099.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N53W099.hgt.zip',
		'N53W100.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N53W100.hgt.zip',
		'N53W101.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N53W101.hgt.zip',
		'N53W102.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N53W102.hgt.zip',
		'N53W103.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N53W103.hgt.zip',
		'N53W104.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N53W104.hgt.zip',
		'N53W105.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N53W105.hgt.zip',
		'N53W106.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N53W106.hgt.zip',
		'N53W107.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N53W107.hgt.zip',
		'N53W108.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N53W108.hgt.zip',
		'N53W109.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N53W109.hgt.zip',
		'N53W110.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N53W110.hgt.zip',
		'N53W111.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N53W111.hgt.zip',
		'N53W112.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N53W112.hgt.zip',
		'N53W113.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N53W113.hgt.zip',
		'N53W114.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N53W114.hgt.zip',
		'N53W115.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N53W115.hgt.zip',
		'N53W116.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N53W116.hgt.zip',
		'N53W117.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N53W117.hgt.zip',
		'N53W118.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N53W118.hgt.zip',
		'N53W119.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N53W119.hgt.zip',
		'N53W120.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N53W120.hgt.zip',
		'N53W121.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N53W121.hgt.zip',
		'N53W122.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N53W122.hgt.zip',
		'N53W123.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N53W123.hgt.zip',
		'N53W124.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N53W124.hgt.zip',
		'N53W125.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N53W125.hgt.zip',
		'N53W126.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N53W126.hgt.zip',
		'N53W127.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N53W127.hgt.zip',
		'N53W128.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N53W128.hgt.zip',
		'N53W129.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N53W129.hgt.zip',
		'N53W130.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N53W130.hgt.zip',
		'N53W131.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N53W131.hgt.zip',
		'N53W132.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N53W132.hgt.zip',
		'N53W133.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N53W133.hgt.zip',
		'N53W134.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N53W134.hgt.zip',
		'N53W167.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N53W167.hgt.zip',
		'N53W168.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N53W168.hgt.zip',
		'N53W169.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N53W169.hgt.zip',
		'N53W170.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N53W170.hgt.zip',
		'N54E007.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E007.hgt.zip',
		'N54E008.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E008.hgt.zip',
		'N54E009.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E009.hgt.zip',
		'N54E010.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E010.hgt.zip',
		'N54E011.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E011.hgt.zip',
		'N54E012.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E012.hgt.zip',
		'N54E013.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E013.hgt.zip',
		'N54E014.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E014.hgt.zip',
		'N54E015.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E015.hgt.zip',
		'N54E016.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E016.hgt.zip',
		'N54E017.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E017.hgt.zip',
		'N54E018.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E018.hgt.zip',
		'N54E019.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E019.hgt.zip',
		'N54E020.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E020.hgt.zip',
		'N54E021.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E021.hgt.zip',
		'N54E022.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E022.hgt.zip',
		'N54E023.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E023.hgt.zip',
		'N54E024.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E024.hgt.zip',
		'N54E025.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E025.hgt.zip',
		'N54E026.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E026.hgt.zip',
		'N54E027.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E027.hgt.zip',
		'N54E028.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E028.hgt.zip',
		'N54E029.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E029.hgt.zip',
		'N54E030.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E030.hgt.zip',
		'N54E031.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E031.hgt.zip',
		'N54E032.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E032.hgt.zip',
		'N54E033.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E033.hgt.zip',
		'N54E034.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E034.hgt.zip',
		'N54E035.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E035.hgt.zip',
		'N54E036.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E036.hgt.zip',
		'N54E037.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E037.hgt.zip',
		'N54E038.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E038.hgt.zip',
		'N54E039.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E039.hgt.zip',
		'N54E040.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E040.hgt.zip',
		'N54E041.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E041.hgt.zip',
		'N54E042.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E042.hgt.zip',
		'N54E043.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E043.hgt.zip',
		'N54E044.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E044.hgt.zip',
		'N54E045.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E045.hgt.zip',
		'N54E046.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E046.hgt.zip',
		'N54E047.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E047.hgt.zip',
		'N54E048.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E048.hgt.zip',
		'N54E049.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E049.hgt.zip',
		'N54E050.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E050.hgt.zip',
		'N54E051.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E051.hgt.zip',
		'N54E052.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E052.hgt.zip',
		'N54E053.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E053.hgt.zip',
		'N54E054.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E054.hgt.zip',
		'N54E055.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E055.hgt.zip',
		'N54E056.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E056.hgt.zip',
		'N54E057.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E057.hgt.zip',
		'N54E058.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E058.hgt.zip',
		'N54E059.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E059.hgt.zip',
		'N54E060.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E060.hgt.zip',
		'N54E061.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E061.hgt.zip',
		'N54E062.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E062.hgt.zip',
		'N54E063.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E063.hgt.zip',
		'N54E064.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E064.hgt.zip',
		'N54E065.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E065.hgt.zip',
		'N54E066.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E066.hgt.zip',
		'N54E067.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E067.hgt.zip',
		'N54E068.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E068.hgt.zip',
		'N54E069.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E069.hgt.zip',
		'N54E070.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E070.hgt.zip',
		'N54E071.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E071.hgt.zip',
		'N54E072.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E072.hgt.zip',
		'N54E073.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E073.hgt.zip',
		'N54E074.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E074.hgt.zip',
		'N54E075.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E075.hgt.zip',
		'N54E076.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E076.hgt.zip',
		'N54E077.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E077.hgt.zip',
		'N54E078.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E078.hgt.zip',
		'N54E079.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E079.hgt.zip',
		'N54E080.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E080.hgt.zip',
		'N54E081.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E081.hgt.zip',
		'N54E082.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E082.hgt.zip',
		'N54E083.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E083.hgt.zip',
		'N54E084.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E084.hgt.zip',
		'N54E085.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E085.hgt.zip',
		'N54E086.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E086.hgt.zip',
		'N54E087.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E087.hgt.zip',
		'N54E088.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E088.hgt.zip',
		'N54E089.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E089.hgt.zip',
		'N54E090.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E090.hgt.zip',
		'N54E091.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E091.hgt.zip',
		'N54E092.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E092.hgt.zip',
		'N54E093.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E093.hgt.zip',
		'N54E094.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E094.hgt.zip',
		'N54E095.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E095.hgt.zip',
		'N54E096.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E096.hgt.zip',
		'N54E097.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E097.hgt.zip',
		'N54E098.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E098.hgt.zip',
		'N54E099.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E099.hgt.zip',
		'N54E100.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E100.hgt.zip',
		'N54E101.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E101.hgt.zip',
		'N54E102.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E102.hgt.zip',
		'N54E103.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E103.hgt.zip',
		'N54E104.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E104.hgt.zip',
		'N54E105.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E105.hgt.zip',
		'N54E106.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E106.hgt.zip',
		'N54E107.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E107.hgt.zip',
		'N54E108.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E108.hgt.zip',
		'N54E109.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E109.hgt.zip',
		'N54E110.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E110.hgt.zip',
		'N54E111.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E111.hgt.zip',
		'N54E112.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E112.hgt.zip',
		'N54E113.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E113.hgt.zip',
		'N54E114.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E114.hgt.zip',
		'N54E115.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E115.hgt.zip',
		'N54E116.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E116.hgt.zip',
		'N54E117.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E117.hgt.zip',
		'N54E118.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E118.hgt.zip',
		'N54E119.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E119.hgt.zip',
		'N54E120.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E120.hgt.zip',
		'N54E121.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E121.hgt.zip',
		'N54E122.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E122.hgt.zip',
		'N54E123.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E123.hgt.zip',
		'N54E124.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E124.hgt.zip',
		'N54E125.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E125.hgt.zip',
		'N54E126.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E126.hgt.zip',
		'N54E127.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E127.hgt.zip',
		'N54E128.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E128.hgt.zip',
		'N54E129.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E129.hgt.zip',
		'N54E130.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E130.hgt.zip',
		'N54E131.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E131.hgt.zip',
		'N54E132.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E132.hgt.zip',
		'N54E133.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E133.hgt.zip',
		'N54E134.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E134.hgt.zip',
		'N54E135.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E135.hgt.zip',
		'N54E136.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E136.hgt.zip',
		'N54E137.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E137.hgt.zip',
		'N54E138.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E138.hgt.zip',
		'N54E139.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E139.hgt.zip',
		'N54E140.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E140.hgt.zip',
		'N54E142.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E142.hgt.zip',
		'N54E155.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E155.hgt.zip',
		'N54E156.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E156.hgt.zip',
		'N54E157.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E157.hgt.zip',
		'N54E158.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E158.hgt.zip',
		'N54E159.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E159.hgt.zip',
		'N54E160.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E160.hgt.zip',
		'N54E161.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E161.hgt.zip',
		'N54E162.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E162.hgt.zip',
		'N54E166.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E166.hgt.zip',
		'N54E167.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E167.hgt.zip',
		'N54E168.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54E168.hgt.zip',
		'N54W001.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54W001.hgt.zip',
		'N54W002.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54W002.hgt.zip',
		'N54W003.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54W003.hgt.zip',
		'N54W004.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54W004.hgt.zip',
		'N54W005.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54W005.hgt.zip',
		'N54W006.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54W006.hgt.zip',
		'N54W007.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54W007.hgt.zip',
		'N54W008.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54W008.hgt.zip',
		'N54W009.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54W009.hgt.zip',
		'N54W010.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54W010.hgt.zip',
		'N54W011.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N54W011.hgt.zip',
		'N54W057.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N54W057.hgt.zip',
		'N54W058.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N54W058.hgt.zip',
		'N54W059.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N54W059.hgt.zip',
		'N54W060.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N54W060.hgt.zip',
		'N54W061.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N54W061.hgt.zip',
		'N54W062.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N54W062.hgt.zip',
		'N54W063.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N54W063.hgt.zip',
		'N54W064.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N54W064.hgt.zip',
		'N54W065.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N54W065.hgt.zip',
		'N54W066.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N54W066.hgt.zip',
		'N54W067.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N54W067.hgt.zip',
		'N54W068.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N54W068.hgt.zip',
		'N54W069.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N54W069.hgt.zip',
		'N54W070.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N54W070.hgt.zip',
		'N54W071.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N54W071.hgt.zip',
		'N54W072.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N54W072.hgt.zip',
		'N54W073.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N54W073.hgt.zip',
		'N54W074.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N54W074.hgt.zip',
		'N54W075.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N54W075.hgt.zip',
		'N54W076.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N54W076.hgt.zip',
		'N54W077.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N54W077.hgt.zip',
		'N54W078.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N54W078.hgt.zip',
		'N54W079.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N54W079.hgt.zip',
		'N54W080.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N54W080.hgt.zip',
		'N54W081.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N54W081.hgt.zip',
		'N54W082.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N54W082.hgt.zip',
		'N54W083.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N54W083.hgt.zip',
		'N54W084.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N54W084.hgt.zip',
		'N54W085.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N54W085.hgt.zip',
		'N54W086.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N54W086.hgt.zip',
		'N54W087.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N54W087.hgt.zip',
		'N54W088.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N54W088.hgt.zip',
		'N54W089.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N54W089.hgt.zip',
		'N54W090.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N54W090.hgt.zip',
		'N54W091.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N54W091.hgt.zip',
		'N54W092.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N54W092.hgt.zip',
		'N54W093.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N54W093.hgt.zip',
		'N54W094.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N54W094.hgt.zip',
		'N54W095.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N54W095.hgt.zip',
		'N54W096.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N54W096.hgt.zip',
		'N54W097.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N54W097.hgt.zip',
		'N54W098.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N54W098.hgt.zip',
		'N54W099.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N54W099.hgt.zip',
		'N54W100.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N54W100.hgt.zip',
		'N54W101.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N54W101.hgt.zip',
		'N54W102.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N54W102.hgt.zip',
		'N54W103.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N54W103.hgt.zip',
		'N54W104.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N54W104.hgt.zip',
		'N54W105.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N54W105.hgt.zip',
		'N54W106.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N54W106.hgt.zip',
		'N54W107.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N54W107.hgt.zip',
		'N54W108.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N54W108.hgt.zip',
		'N54W109.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N54W109.hgt.zip',
		'N54W110.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N54W110.hgt.zip',
		'N54W111.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N54W111.hgt.zip',
		'N54W112.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N54W112.hgt.zip',
		'N54W113.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N54W113.hgt.zip',
		'N54W114.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N54W114.hgt.zip',
		'N54W115.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N54W115.hgt.zip',
		'N54W116.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N54W116.hgt.zip',
		'N54W117.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N54W117.hgt.zip',
		'N54W118.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N54W118.hgt.zip',
		'N54W119.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N54W119.hgt.zip',
		'N54W120.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N54W120.hgt.zip',
		'N54W121.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N54W121.hgt.zip',
		'N54W122.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N54W122.hgt.zip',
		'N54W123.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N54W123.hgt.zip',
		'N54W124.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N54W124.hgt.zip',
		'N54W125.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N54W125.hgt.zip',
		'N54W126.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N54W126.hgt.zip',
		'N54W127.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N54W127.hgt.zip',
		'N54W128.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N54W128.hgt.zip',
		'N54W129.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N54W129.hgt.zip',
		'N54W130.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N54W130.hgt.zip',
		'N54W131.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N54W131.hgt.zip',
		'N54W132.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N54W132.hgt.zip',
		'N54W133.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N54W133.hgt.zip',
		'N54W134.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N54W134.hgt.zip',
		'N54W160.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N54W160.hgt.zip',
		'N54W161.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N54W161.hgt.zip',
		'N54W162.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N54W162.hgt.zip',
		'N54W163.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N54W163.hgt.zip',
		'N54W164.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N54W164.hgt.zip',
		'N54W165.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N54W165.hgt.zip',
		'N54W166.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N54W166.hgt.zip',
		'N54W167.hgt': 'https://srtm.kurviger.de/SRTM3//North_America//N54W167.hgt.zip',
		'N55E008.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55E008.hgt.zip',
		'N55E009.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55E009.hgt.zip',
		'N55E010.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55E010.hgt.zip',
		'N55E011.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55E011.hgt.zip',
		'N55E012.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55E012.hgt.zip',
		'N55E013.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55E013.hgt.zip',
		'N55E014.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55E014.hgt.zip',
		'N55E015.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55E015.hgt.zip',
		'N55E020.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55E020.hgt.zip',
		'N55E021.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55E021.hgt.zip',
		'N55E022.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55E022.hgt.zip',
		'N55E023.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55E023.hgt.zip',
		'N55E024.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55E024.hgt.zip',
		'N55E025.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55E025.hgt.zip',
		'N55E026.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55E026.hgt.zip',
		'N55E027.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55E027.hgt.zip',
		'N55E028.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55E028.hgt.zip',
		'N55E029.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55E029.hgt.zip',
		'N55E030.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55E030.hgt.zip',
		'N55E031.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55E031.hgt.zip',
		'N55E032.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55E032.hgt.zip',
		'N55E033.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55E033.hgt.zip',
		'N55E034.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55E034.hgt.zip',
		'N55E035.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55E035.hgt.zip',
		'N55E036.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55E036.hgt.zip',
		'N55E037.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55E037.hgt.zip',
		'N55E038.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55E038.hgt.zip',
		'N55E039.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55E039.hgt.zip',
		'N55E040.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55E040.hgt.zip',
		'N55E041.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55E041.hgt.zip',
		'N55E042.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55E042.hgt.zip',
		'N55E043.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55E043.hgt.zip',
		'N55E044.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55E044.hgt.zip',
		'N55E045.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55E045.hgt.zip',
		'N55E046.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55E046.hgt.zip',
		'N55E047.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55E047.hgt.zip',
		'N55E048.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55E048.hgt.zip',
		'N55E049.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55E049.hgt.zip',
		'N55E050.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55E050.hgt.zip',
		'N55E051.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55E051.hgt.zip',
		'N55E052.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55E052.hgt.zip',
		'N55E053.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55E053.hgt.zip',
		'N55E054.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55E054.hgt.zip',
		'N55E055.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55E055.hgt.zip',
		'N55E056.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55E056.hgt.zip',
		'N55E057.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55E057.hgt.zip',
		'N55E058.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55E058.hgt.zip',
		'N55E059.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55E059.hgt.zip',
		'N55E060.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55E060.hgt.zip',
		'N55E061.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55E061.hgt.zip',
		'N55E062.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55E062.hgt.zip',
		'N55E063.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55E063.hgt.zip',
		'N55E064.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55E064.hgt.zip',
		'N55E065.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55E065.hgt.zip',
		'N55E066.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55E066.hgt.zip',
		'N55E067.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55E067.hgt.zip',
		'N55E068.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55E068.hgt.zip',
		'N55E069.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55E069.hgt.zip',
		'N55E070.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55E070.hgt.zip',
		'N55E071.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55E071.hgt.zip',
		'N55E072.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55E072.hgt.zip',
		'N55E073.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55E073.hgt.zip',
		'N55E074.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55E074.hgt.zip',
		'N55E075.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55E075.hgt.zip',
		'N55E076.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55E076.hgt.zip',
		'N55E077.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55E077.hgt.zip',
		'N55E078.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55E078.hgt.zip',
		'N55E079.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55E079.hgt.zip',
		'N55E080.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55E080.hgt.zip',
		'N55E081.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55E081.hgt.zip',
		'N55E082.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55E082.hgt.zip',
		'N55E083.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55E083.hgt.zip',
		'N55E084.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55E084.hgt.zip',
		'N55E085.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55E085.hgt.zip',
		'N55E086.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55E086.hgt.zip',
		'N55E087.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55E087.hgt.zip',
		'N55E088.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55E088.hgt.zip',
		'N55E089.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55E089.hgt.zip',
		'N55E090.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55E090.hgt.zip',
		'N55E091.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55E091.hgt.zip',
		'N55E092.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55E092.hgt.zip',
		'N55E093.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55E093.hgt.zip',
		'N55E094.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55E094.hgt.zip',
		'N55E095.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55E095.hgt.zip',
		'N55E096.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55E096.hgt.zip',
		'N55E097.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55E097.hgt.zip',
		'N55E098.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55E098.hgt.zip',
		'N55E099.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55E099.hgt.zip',
		'N55E100.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55E100.hgt.zip',
		'N55E101.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55E101.hgt.zip',
		'N55E102.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55E102.hgt.zip',
		'N55E103.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55E103.hgt.zip',
		'N55E104.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55E104.hgt.zip',
		'N55E105.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55E105.hgt.zip',
		'N55E106.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55E106.hgt.zip',
		'N55E107.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55E107.hgt.zip',
		'N55E108.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55E108.hgt.zip',
		'N55E109.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55E109.hgt.zip',
		'N55E110.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55E110.hgt.zip',
		'N55E111.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55E111.hgt.zip',
		'N55E112.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55E112.hgt.zip',
		'N55E113.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55E113.hgt.zip',
		'N55E114.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55E114.hgt.zip',
		'N55E115.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55E115.hgt.zip',
		'N55E116.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55E116.hgt.zip',
		'N55E117.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55E117.hgt.zip',
		'N55E118.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55E118.hgt.zip',
		'N55E119.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55E119.hgt.zip',
		'N55E120.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55E120.hgt.zip',
		'N55E121.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55E121.hgt.zip',
		'N55E122.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55E122.hgt.zip',
		'N55E123.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55E123.hgt.zip',
		'N55E124.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55E124.hgt.zip',
		'N55E125.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55E125.hgt.zip',
		'N55E126.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55E126.hgt.zip',
		'N55E127.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55E127.hgt.zip',
		'N55E128.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55E128.hgt.zip',
		'N55E129.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55E129.hgt.zip',
		'N55E130.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55E130.hgt.zip',
		'N55E131.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55E131.hgt.zip',
		'N55E132.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55E132.hgt.zip',
		'N55E133.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55E133.hgt.zip',
		'N55E134.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55E134.hgt.zip',
		'N55E135.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55E135.hgt.zip',
		'N55E136.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55E136.hgt.zip',
		'N55E137.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55E137.hgt.zip',
		'N55E138.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55E138.hgt.zip',
		'N55E155.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55E155.hgt.zip',
		'N55E156.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55E156.hgt.zip',
		'N55E157.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55E157.hgt.zip',
		'N55E158.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55E158.hgt.zip',
		'N55E159.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55E159.hgt.zip',
		'N55E160.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55E160.hgt.zip',
		'N55E161.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55E161.hgt.zip',
		'N55E162.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55E162.hgt.zip',
		'N55E165.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55E165.hgt.zip',
		'N55E166.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55E166.hgt.zip',
		'N55W002.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55W002.hgt.zip',
		'N55W003.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55W003.hgt.zip',
		'N55W004.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55W004.hgt.zip',
		'N55W005.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55W005.hgt.zip',
		'N55W006.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55W006.hgt.zip',
		'N55W007.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55W007.hgt.zip',
		'N55W008.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55W008.hgt.zip',
		'N55W009.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N55W009.hgt.zip',
		'N56E008.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56E008.hgt.zip',
		'N56E009.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56E009.hgt.zip',
		'N56E010.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56E010.hgt.zip',
		'N56E011.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56E011.hgt.zip',
		'N56E012.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56E012.hgt.zip',
		'N56E013.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56E013.hgt.zip',
		'N56E014.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56E014.hgt.zip',
		'N56E015.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56E015.hgt.zip',
		'N56E016.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56E016.hgt.zip',
		'N56E018.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56E018.hgt.zip',
		'N56E020.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56E020.hgt.zip',
		'N56E021.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56E021.hgt.zip',
		'N56E022.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56E022.hgt.zip',
		'N56E023.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56E023.hgt.zip',
		'N56E024.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56E024.hgt.zip',
		'N56E025.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56E025.hgt.zip',
		'N56E026.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56E026.hgt.zip',
		'N56E027.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56E027.hgt.zip',
		'N56E028.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56E028.hgt.zip',
		'N56E029.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56E029.hgt.zip',
		'N56E030.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56E030.hgt.zip',
		'N56E031.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56E031.hgt.zip',
		'N56E032.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56E032.hgt.zip',
		'N56E033.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56E033.hgt.zip',
		'N56E034.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56E034.hgt.zip',
		'N56E035.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56E035.hgt.zip',
		'N56E036.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56E036.hgt.zip',
		'N56E037.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56E037.hgt.zip',
		'N56E038.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56E038.hgt.zip',
		'N56E039.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56E039.hgt.zip',
		'N56E040.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56E040.hgt.zip',
		'N56E041.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56E041.hgt.zip',
		'N56E042.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56E042.hgt.zip',
		'N56E043.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56E043.hgt.zip',
		'N56E044.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56E044.hgt.zip',
		'N56E045.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56E045.hgt.zip',
		'N56E046.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56E046.hgt.zip',
		'N56E047.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56E047.hgt.zip',
		'N56E048.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56E048.hgt.zip',
		'N56E049.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56E049.hgt.zip',
		'N56E050.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56E050.hgt.zip',
		'N56E051.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56E051.hgt.zip',
		'N56E052.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56E052.hgt.zip',
		'N56E053.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56E053.hgt.zip',
		'N56E054.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56E054.hgt.zip',
		'N56E055.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56E055.hgt.zip',
		'N56E056.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56E056.hgt.zip',
		'N56E057.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56E057.hgt.zip',
		'N56E058.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56E058.hgt.zip',
		'N56E059.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56E059.hgt.zip',
		'N56E060.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56E060.hgt.zip',
		'N56E061.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56E061.hgt.zip',
		'N56E062.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56E062.hgt.zip',
		'N56E063.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56E063.hgt.zip',
		'N56E064.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56E064.hgt.zip',
		'N56E065.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56E065.hgt.zip',
		'N56E066.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56E066.hgt.zip',
		'N56E067.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56E067.hgt.zip',
		'N56E068.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56E068.hgt.zip',
		'N56E069.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56E069.hgt.zip',
		'N56E070.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56E070.hgt.zip',
		'N56E071.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56E071.hgt.zip',
		'N56E072.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56E072.hgt.zip',
		'N56E073.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56E073.hgt.zip',
		'N56E074.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56E074.hgt.zip',
		'N56E075.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56E075.hgt.zip',
		'N56E076.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56E076.hgt.zip',
		'N56E077.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56E077.hgt.zip',
		'N56E078.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56E078.hgt.zip',
		'N56E079.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56E079.hgt.zip',
		'N56E080.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56E080.hgt.zip',
		'N56E081.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56E081.hgt.zip',
		'N56E082.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56E082.hgt.zip',
		'N56E083.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56E083.hgt.zip',
		'N56E084.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56E084.hgt.zip',
		'N56E085.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56E085.hgt.zip',
		'N56E086.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56E086.hgt.zip',
		'N56E087.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56E087.hgt.zip',
		'N56E088.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56E088.hgt.zip',
		'N56E089.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56E089.hgt.zip',
		'N56E090.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56E090.hgt.zip',
		'N56E091.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56E091.hgt.zip',
		'N56E092.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56E092.hgt.zip',
		'N56E093.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56E093.hgt.zip',
		'N56E094.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56E094.hgt.zip',
		'N56E095.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56E095.hgt.zip',
		'N56E096.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56E096.hgt.zip',
		'N56E097.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56E097.hgt.zip',
		'N56E098.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56E098.hgt.zip',
		'N56E099.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56E099.hgt.zip',
		'N56E100.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56E100.hgt.zip',
		'N56E101.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56E101.hgt.zip',
		'N56E102.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56E102.hgt.zip',
		'N56E103.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56E103.hgt.zip',
		'N56E104.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56E104.hgt.zip',
		'N56E105.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56E105.hgt.zip',
		'N56E106.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56E106.hgt.zip',
		'N56E107.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56E107.hgt.zip',
		'N56E108.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56E108.hgt.zip',
		'N56E109.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56E109.hgt.zip',
		'N56E110.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56E110.hgt.zip',
		'N56E111.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56E111.hgt.zip',
		'N56E112.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56E112.hgt.zip',
		'N56E113.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56E113.hgt.zip',
		'N56E114.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56E114.hgt.zip',
		'N56E115.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56E115.hgt.zip',
		'N56E116.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56E116.hgt.zip',
		'N56E117.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56E117.hgt.zip',
		'N56E118.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56E118.hgt.zip',
		'N56E119.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56E119.hgt.zip',
		'N56E120.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56E120.hgt.zip',
		'N56E121.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56E121.hgt.zip',
		'N56E122.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56E122.hgt.zip',
		'N56E123.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56E123.hgt.zip',
		'N56E124.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56E124.hgt.zip',
		'N56E125.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56E125.hgt.zip',
		'N56E126.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56E126.hgt.zip',
		'N56E127.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56E127.hgt.zip',
		'N56E128.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56E128.hgt.zip',
		'N56E129.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56E129.hgt.zip',
		'N56E130.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56E130.hgt.zip',
		'N56E131.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56E131.hgt.zip',
		'N56E132.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56E132.hgt.zip',
		'N56E133.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56E133.hgt.zip',
		'N56E134.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56E134.hgt.zip',
		'N56E135.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56E135.hgt.zip',
		'N56E136.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56E136.hgt.zip',
		'N56E137.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56E137.hgt.zip',
		'N56E138.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56E138.hgt.zip',
		'N56E143.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56E143.hgt.zip',
		'N56E155.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56E155.hgt.zip',
		'N56E156.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56E156.hgt.zip',
		'N56E157.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56E157.hgt.zip',
		'N56E158.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56E158.hgt.zip',
		'N56E159.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56E159.hgt.zip',
		'N56E160.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56E160.hgt.zip',
		'N56E161.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56E161.hgt.zip',
		'N56E162.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56E162.hgt.zip',
		'N56E163.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56E163.hgt.zip',
		'N56W003.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56W003.hgt.zip',
		'N56W004.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56W004.hgt.zip',
		'N56W005.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56W005.hgt.zip',
		'N56W006.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56W006.hgt.zip',
		'N56W007.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56W007.hgt.zip',
		'N56W008.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N56W008.hgt.zip',
		'N57E006.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57E006.hgt.zip',
		'N57E007.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57E007.hgt.zip',
		'N57E008.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57E008.hgt.zip',
		'N57E009.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57E009.hgt.zip',
		'N57E010.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57E010.hgt.zip',
		'N57E011.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57E011.hgt.zip',
		'N57E012.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57E012.hgt.zip',
		'N57E013.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57E013.hgt.zip',
		'N57E014.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57E014.hgt.zip',
		'N57E015.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57E015.hgt.zip',
		'N57E016.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57E016.hgt.zip',
		'N57E017.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57E017.hgt.zip',
		'N57E018.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57E018.hgt.zip',
		'N57E019.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57E019.hgt.zip',
		'N57E021.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57E021.hgt.zip',
		'N57E022.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57E022.hgt.zip',
		'N57E023.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57E023.hgt.zip',
		'N57E024.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57E024.hgt.zip',
		'N57E025.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57E025.hgt.zip',
		'N57E026.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57E026.hgt.zip',
		'N57E027.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57E027.hgt.zip',
		'N57E028.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57E028.hgt.zip',
		'N57E029.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57E029.hgt.zip',
		'N57E030.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57E030.hgt.zip',
		'N57E031.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57E031.hgt.zip',
		'N57E032.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57E032.hgt.zip',
		'N57E033.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57E033.hgt.zip',
		'N57E034.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57E034.hgt.zip',
		'N57E035.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57E035.hgt.zip',
		'N57E036.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57E036.hgt.zip',
		'N57E037.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57E037.hgt.zip',
		'N57E038.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57E038.hgt.zip',
		'N57E039.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57E039.hgt.zip',
		'N57E040.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57E040.hgt.zip',
		'N57E041.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57E041.hgt.zip',
		'N57E042.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57E042.hgt.zip',
		'N57E043.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57E043.hgt.zip',
		'N57E044.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57E044.hgt.zip',
		'N57E045.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57E045.hgt.zip',
		'N57E046.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57E046.hgt.zip',
		'N57E047.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57E047.hgt.zip',
		'N57E048.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57E048.hgt.zip',
		'N57E049.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57E049.hgt.zip',
		'N57E050.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57E050.hgt.zip',
		'N57E051.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57E051.hgt.zip',
		'N57E052.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57E052.hgt.zip',
		'N57E053.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57E053.hgt.zip',
		'N57E054.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57E054.hgt.zip',
		'N57E055.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57E055.hgt.zip',
		'N57E056.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57E056.hgt.zip',
		'N57E057.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57E057.hgt.zip',
		'N57E058.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57E058.hgt.zip',
		'N57E059.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57E059.hgt.zip',
		'N57E060.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57E060.hgt.zip',
		'N57E061.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57E061.hgt.zip',
		'N57E062.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57E062.hgt.zip',
		'N57E063.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57E063.hgt.zip',
		'N57E064.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57E064.hgt.zip',
		'N57E065.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57E065.hgt.zip',
		'N57E066.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57E066.hgt.zip',
		'N57E067.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57E067.hgt.zip',
		'N57E068.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57E068.hgt.zip',
		'N57E069.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57E069.hgt.zip',
		'N57E070.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57E070.hgt.zip',
		'N57E071.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57E071.hgt.zip',
		'N57E072.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57E072.hgt.zip',
		'N57E073.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57E073.hgt.zip',
		'N57E074.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57E074.hgt.zip',
		'N57E075.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57E075.hgt.zip',
		'N57E076.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57E076.hgt.zip',
		'N57E077.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57E077.hgt.zip',
		'N57E078.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57E078.hgt.zip',
		'N57E079.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57E079.hgt.zip',
		'N57E080.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57E080.hgt.zip',
		'N57E081.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57E081.hgt.zip',
		'N57E082.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57E082.hgt.zip',
		'N57E083.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57E083.hgt.zip',
		'N57E084.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57E084.hgt.zip',
		'N57E085.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57E085.hgt.zip',
		'N57E086.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57E086.hgt.zip',
		'N57E087.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57E087.hgt.zip',
		'N57E088.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57E088.hgt.zip',
		'N57E089.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57E089.hgt.zip',
		'N57E090.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57E090.hgt.zip',
		'N57E091.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57E091.hgt.zip',
		'N57E092.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57E092.hgt.zip',
		'N57E093.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57E093.hgt.zip',
		'N57E094.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57E094.hgt.zip',
		'N57E095.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57E095.hgt.zip',
		'N57E096.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57E096.hgt.zip',
		'N57E097.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57E097.hgt.zip',
		'N57E098.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57E098.hgt.zip',
		'N57E099.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57E099.hgt.zip',
		'N57E100.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57E100.hgt.zip',
		'N57E101.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57E101.hgt.zip',
		'N57E102.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57E102.hgt.zip',
		'N57E103.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57E103.hgt.zip',
		'N57E104.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57E104.hgt.zip',
		'N57E105.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57E105.hgt.zip',
		'N57E106.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57E106.hgt.zip',
		'N57E107.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57E107.hgt.zip',
		'N57E108.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57E108.hgt.zip',
		'N57E109.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57E109.hgt.zip',
		'N57E110.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57E110.hgt.zip',
		'N57E111.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57E111.hgt.zip',
		'N57E112.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57E112.hgt.zip',
		'N57E113.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57E113.hgt.zip',
		'N57E114.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57E114.hgt.zip',
		'N57E115.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57E115.hgt.zip',
		'N57E116.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57E116.hgt.zip',
		'N57E117.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57E117.hgt.zip',
		'N57E118.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57E118.hgt.zip',
		'N57E119.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57E119.hgt.zip',
		'N57E120.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57E120.hgt.zip',
		'N57E121.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57E121.hgt.zip',
		'N57E122.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57E122.hgt.zip',
		'N57E123.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57E123.hgt.zip',
		'N57E124.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57E124.hgt.zip',
		'N57E125.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57E125.hgt.zip',
		'N57E126.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57E126.hgt.zip',
		'N57E127.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57E127.hgt.zip',
		'N57E128.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57E128.hgt.zip',
		'N57E129.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57E129.hgt.zip',
		'N57E130.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57E130.hgt.zip',
		'N57E131.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57E131.hgt.zip',
		'N57E132.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57E132.hgt.zip',
		'N57E133.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57E133.hgt.zip',
		'N57E134.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57E134.hgt.zip',
		'N57E135.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57E135.hgt.zip',
		'N57E136.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57E136.hgt.zip',
		'N57E137.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57E137.hgt.zip',
		'N57E138.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57E138.hgt.zip',
		'N57E139.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57E139.hgt.zip',
		'N57E140.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57E140.hgt.zip',
		'N57E156.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57E156.hgt.zip',
		'N57E157.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57E157.hgt.zip',
		'N57E158.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57E158.hgt.zip',
		'N57E159.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57E159.hgt.zip',
		'N57E160.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57E160.hgt.zip',
		'N57E161.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57E161.hgt.zip',
		'N57E162.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57E162.hgt.zip',
		'N57E163.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57E163.hgt.zip',
		'N57W002.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57W002.hgt.zip',
		'N57W003.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57W003.hgt.zip',
		'N57W004.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57W004.hgt.zip',
		'N57W005.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57W005.hgt.zip',
		'N57W006.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57W006.hgt.zip',
		'N57W007.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57W007.hgt.zip',
		'N57W008.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57W008.hgt.zip',
		'N57W009.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57W009.hgt.zip',
		'N57W014.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N57W014.hgt.zip',
		'N58E005.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E005.hgt.zip',
		'N58E006.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E006.hgt.zip',
		'N58E007.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E007.hgt.zip',
		'N58E008.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E008.hgt.zip',
		'N58E009.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E009.hgt.zip',
		'N58E010.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E010.hgt.zip',
		'N58E011.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E011.hgt.zip',
		'N58E012.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E012.hgt.zip',
		'N58E013.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E013.hgt.zip',
		'N58E014.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E014.hgt.zip',
		'N58E015.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E015.hgt.zip',
		'N58E016.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E016.hgt.zip',
		'N58E017.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E017.hgt.zip',
		'N58E018.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E018.hgt.zip',
		'N58E019.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E019.hgt.zip',
		'N58E021.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E021.hgt.zip',
		'N58E022.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E022.hgt.zip',
		'N58E023.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E023.hgt.zip',
		'N58E024.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E024.hgt.zip',
		'N58E025.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E025.hgt.zip',
		'N58E026.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E026.hgt.zip',
		'N58E027.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E027.hgt.zip',
		'N58E028.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E028.hgt.zip',
		'N58E029.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E029.hgt.zip',
		'N58E030.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E030.hgt.zip',
		'N58E031.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E031.hgt.zip',
		'N58E032.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E032.hgt.zip',
		'N58E033.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E033.hgt.zip',
		'N58E034.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E034.hgt.zip',
		'N58E035.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E035.hgt.zip',
		'N58E036.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E036.hgt.zip',
		'N58E037.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E037.hgt.zip',
		'N58E038.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E038.hgt.zip',
		'N58E039.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E039.hgt.zip',
		'N58E040.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E040.hgt.zip',
		'N58E041.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E041.hgt.zip',
		'N58E042.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E042.hgt.zip',
		'N58E043.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E043.hgt.zip',
		'N58E044.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E044.hgt.zip',
		'N58E045.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E045.hgt.zip',
		'N58E046.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E046.hgt.zip',
		'N58E047.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E047.hgt.zip',
		'N58E048.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E048.hgt.zip',
		'N58E049.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E049.hgt.zip',
		'N58E050.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E050.hgt.zip',
		'N58E051.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E051.hgt.zip',
		'N58E052.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E052.hgt.zip',
		'N58E053.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E053.hgt.zip',
		'N58E054.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E054.hgt.zip',
		'N58E055.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E055.hgt.zip',
		'N58E056.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E056.hgt.zip',
		'N58E057.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E057.hgt.zip',
		'N58E058.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E058.hgt.zip',
		'N58E059.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E059.hgt.zip',
		'N58E060.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E060.hgt.zip',
		'N58E061.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E061.hgt.zip',
		'N58E062.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E062.hgt.zip',
		'N58E063.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E063.hgt.zip',
		'N58E064.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E064.hgt.zip',
		'N58E065.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E065.hgt.zip',
		'N58E066.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E066.hgt.zip',
		'N58E067.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E067.hgt.zip',
		'N58E068.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E068.hgt.zip',
		'N58E069.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E069.hgt.zip',
		'N58E070.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E070.hgt.zip',
		'N58E071.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E071.hgt.zip',
		'N58E072.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E072.hgt.zip',
		'N58E073.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E073.hgt.zip',
		'N58E074.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E074.hgt.zip',
		'N58E075.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E075.hgt.zip',
		'N58E076.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E076.hgt.zip',
		'N58E077.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E077.hgt.zip',
		'N58E078.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E078.hgt.zip',
		'N58E079.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E079.hgt.zip',
		'N58E080.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E080.hgt.zip',
		'N58E081.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E081.hgt.zip',
		'N58E082.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E082.hgt.zip',
		'N58E083.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E083.hgt.zip',
		'N58E084.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E084.hgt.zip',
		'N58E085.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E085.hgt.zip',
		'N58E086.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E086.hgt.zip',
		'N58E087.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E087.hgt.zip',
		'N58E088.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E088.hgt.zip',
		'N58E089.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E089.hgt.zip',
		'N58E090.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E090.hgt.zip',
		'N58E091.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E091.hgt.zip',
		'N58E092.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E092.hgt.zip',
		'N58E093.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E093.hgt.zip',
		'N58E094.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E094.hgt.zip',
		'N58E095.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E095.hgt.zip',
		'N58E096.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E096.hgt.zip',
		'N58E097.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E097.hgt.zip',
		'N58E098.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E098.hgt.zip',
		'N58E099.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E099.hgt.zip',
		'N58E100.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E100.hgt.zip',
		'N58E101.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E101.hgt.zip',
		'N58E102.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E102.hgt.zip',
		'N58E103.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E103.hgt.zip',
		'N58E104.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E104.hgt.zip',
		'N58E105.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E105.hgt.zip',
		'N58E106.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E106.hgt.zip',
		'N58E107.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E107.hgt.zip',
		'N58E108.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E108.hgt.zip',
		'N58E109.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E109.hgt.zip',
		'N58E110.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E110.hgt.zip',
		'N58E111.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E111.hgt.zip',
		'N58E112.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E112.hgt.zip',
		'N58E113.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E113.hgt.zip',
		'N58E114.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E114.hgt.zip',
		'N58E115.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E115.hgt.zip',
		'N58E116.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E116.hgt.zip',
		'N58E117.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E117.hgt.zip',
		'N58E118.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E118.hgt.zip',
		'N58E119.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E119.hgt.zip',
		'N58E120.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E120.hgt.zip',
		'N58E121.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E121.hgt.zip',
		'N58E122.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E122.hgt.zip',
		'N58E123.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E123.hgt.zip',
		'N58E124.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E124.hgt.zip',
		'N58E125.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E125.hgt.zip',
		'N58E126.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E126.hgt.zip',
		'N58E127.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E127.hgt.zip',
		'N58E128.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E128.hgt.zip',
		'N58E129.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E129.hgt.zip',
		'N58E130.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E130.hgt.zip',
		'N58E131.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E131.hgt.zip',
		'N58E132.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E132.hgt.zip',
		'N58E133.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E133.hgt.zip',
		'N58E134.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E134.hgt.zip',
		'N58E135.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E135.hgt.zip',
		'N58E136.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E136.hgt.zip',
		'N58E137.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E137.hgt.zip',
		'N58E138.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E138.hgt.zip',
		'N58E139.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E139.hgt.zip',
		'N58E140.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E140.hgt.zip',
		'N58E141.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E141.hgt.zip',
		'N58E142.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E142.hgt.zip',
		'N58E150.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E150.hgt.zip',
		'N58E151.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E151.hgt.zip',
		'N58E152.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E152.hgt.zip',
		'N58E157.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E157.hgt.zip',
		'N58E158.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E158.hgt.zip',
		'N58E159.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E159.hgt.zip',
		'N58E160.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E160.hgt.zip',
		'N58E161.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E161.hgt.zip',
		'N58E162.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E162.hgt.zip',
		'N58E163.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E163.hgt.zip',
		'N58E164.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58E164.hgt.zip',
		'N58W003.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58W003.hgt.zip',
		'N58W004.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58W004.hgt.zip',
		'N58W005.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58W005.hgt.zip',
		'N58W006.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58W006.hgt.zip',
		'N58W007.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58W007.hgt.zip',
		'N58W008.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N58W008.hgt.zip',
		'N59E004.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E004.hgt.zip',
		'N59E005.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E005.hgt.zip',
		'N59E006.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E006.hgt.zip',
		'N59E007.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E007.hgt.zip',
		'N59E008.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E008.hgt.zip',
		'N59E009.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E009.hgt.zip',
		'N59E010.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E010.hgt.zip',
		'N59E011.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E011.hgt.zip',
		'N59E012.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E012.hgt.zip',
		'N59E013.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E013.hgt.zip',
		'N59E014.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E014.hgt.zip',
		'N59E015.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E015.hgt.zip',
		'N59E016.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E016.hgt.zip',
		'N59E017.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E017.hgt.zip',
		'N59E018.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E018.hgt.zip',
		'N59E019.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E019.hgt.zip',
		'N59E020.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E020.hgt.zip',
		'N59E021.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E021.hgt.zip',
		'N59E022.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E022.hgt.zip',
		'N59E023.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E023.hgt.zip',
		'N59E024.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E024.hgt.zip',
		'N59E025.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E025.hgt.zip',
		'N59E026.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E026.hgt.zip',
		'N59E027.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E027.hgt.zip',
		'N59E028.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E028.hgt.zip',
		'N59E029.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E029.hgt.zip',
		'N59E030.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E030.hgt.zip',
		'N59E031.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E031.hgt.zip',
		'N59E032.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E032.hgt.zip',
		'N59E033.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E033.hgt.zip',
		'N59E034.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E034.hgt.zip',
		'N59E035.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E035.hgt.zip',
		'N59E036.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E036.hgt.zip',
		'N59E037.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E037.hgt.zip',
		'N59E038.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E038.hgt.zip',
		'N59E039.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E039.hgt.zip',
		'N59E040.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E040.hgt.zip',
		'N59E041.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E041.hgt.zip',
		'N59E042.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E042.hgt.zip',
		'N59E043.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E043.hgt.zip',
		'N59E044.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E044.hgt.zip',
		'N59E045.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E045.hgt.zip',
		'N59E046.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E046.hgt.zip',
		'N59E047.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E047.hgt.zip',
		'N59E048.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E048.hgt.zip',
		'N59E049.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E049.hgt.zip',
		'N59E050.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E050.hgt.zip',
		'N59E051.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E051.hgt.zip',
		'N59E052.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E052.hgt.zip',
		'N59E053.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E053.hgt.zip',
		'N59E054.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E054.hgt.zip',
		'N59E055.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E055.hgt.zip',
		'N59E056.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E056.hgt.zip',
		'N59E057.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E057.hgt.zip',
		'N59E058.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E058.hgt.zip',
		'N59E059.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E059.hgt.zip',
		'N59E060.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E060.hgt.zip',
		'N59E061.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E061.hgt.zip',
		'N59E062.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E062.hgt.zip',
		'N59E063.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E063.hgt.zip',
		'N59E064.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E064.hgt.zip',
		'N59E065.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E065.hgt.zip',
		'N59E066.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E066.hgt.zip',
		'N59E067.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E067.hgt.zip',
		'N59E068.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E068.hgt.zip',
		'N59E069.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E069.hgt.zip',
		'N59E070.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E070.hgt.zip',
		'N59E071.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E071.hgt.zip',
		'N59E072.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E072.hgt.zip',
		'N59E073.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E073.hgt.zip',
		'N59E074.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E074.hgt.zip',
		'N59E075.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E075.hgt.zip',
		'N59E076.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E076.hgt.zip',
		'N59E077.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E077.hgt.zip',
		'N59E078.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E078.hgt.zip',
		'N59E079.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E079.hgt.zip',
		'N59E080.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E080.hgt.zip',
		'N59E081.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E081.hgt.zip',
		'N59E082.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E082.hgt.zip',
		'N59E083.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E083.hgt.zip',
		'N59E084.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E084.hgt.zip',
		'N59E085.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E085.hgt.zip',
		'N59E086.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E086.hgt.zip',
		'N59E087.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E087.hgt.zip',
		'N59E088.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E088.hgt.zip',
		'N59E089.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E089.hgt.zip',
		'N59E090.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E090.hgt.zip',
		'N59E091.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E091.hgt.zip',
		'N59E092.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E092.hgt.zip',
		'N59E093.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E093.hgt.zip',
		'N59E094.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E094.hgt.zip',
		'N59E095.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E095.hgt.zip',
		'N59E096.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E096.hgt.zip',
		'N59E097.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E097.hgt.zip',
		'N59E098.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E098.hgt.zip',
		'N59E099.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E099.hgt.zip',
		'N59E100.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E100.hgt.zip',
		'N59E101.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E101.hgt.zip',
		'N59E102.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E102.hgt.zip',
		'N59E103.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E103.hgt.zip',
		'N59E104.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E104.hgt.zip',
		'N59E105.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E105.hgt.zip',
		'N59E106.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E106.hgt.zip',
		'N59E107.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E107.hgt.zip',
		'N59E108.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E108.hgt.zip',
		'N59E109.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E109.hgt.zip',
		'N59E110.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E110.hgt.zip',
		'N59E111.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E111.hgt.zip',
		'N59E112.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E112.hgt.zip',
		'N59E113.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E113.hgt.zip',
		'N59E114.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E114.hgt.zip',
		'N59E115.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E115.hgt.zip',
		'N59E116.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E116.hgt.zip',
		'N59E117.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E117.hgt.zip',
		'N59E118.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E118.hgt.zip',
		'N59E119.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E119.hgt.zip',
		'N59E120.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E120.hgt.zip',
		'N59E121.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E121.hgt.zip',
		'N59E122.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E122.hgt.zip',
		'N59E123.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E123.hgt.zip',
		'N59E124.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E124.hgt.zip',
		'N59E125.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E125.hgt.zip',
		'N59E126.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E126.hgt.zip',
		'N59E127.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E127.hgt.zip',
		'N59E128.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E128.hgt.zip',
		'N59E129.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E129.hgt.zip',
		'N59E130.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E130.hgt.zip',
		'N59E131.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E131.hgt.zip',
		'N59E132.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E132.hgt.zip',
		'N59E133.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E133.hgt.zip',
		'N59E134.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E134.hgt.zip',
		'N59E135.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E135.hgt.zip',
		'N59E136.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E136.hgt.zip',
		'N59E137.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E137.hgt.zip',
		'N59E138.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E138.hgt.zip',
		'N59E139.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E139.hgt.zip',
		'N59E140.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E140.hgt.zip',
		'N59E141.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E141.hgt.zip',
		'N59E142.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E142.hgt.zip',
		'N59E143.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E143.hgt.zip',
		'N59E144.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E144.hgt.zip',
		'N59E145.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E145.hgt.zip',
		'N59E146.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E146.hgt.zip',
		'N59E147.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E147.hgt.zip',
		'N59E148.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E148.hgt.zip',
		'N59E149.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E149.hgt.zip',
		'N59E150.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E150.hgt.zip',
		'N59E151.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E151.hgt.zip',
		'N59E152.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E152.hgt.zip',
		'N59E153.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E153.hgt.zip',
		'N59E154.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E154.hgt.zip',
		'N59E155.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E155.hgt.zip',
		'N59E159.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E159.hgt.zip',
		'N59E160.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E160.hgt.zip',
		'N59E161.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E161.hgt.zip',
		'N59E162.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E162.hgt.zip',
		'N59E163.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E163.hgt.zip',
		'N59E164.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E164.hgt.zip',
		'N59E165.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E165.hgt.zip',
		'N59E166.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59E166.hgt.zip',
		'N59W002.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59W002.hgt.zip',
		'N59W003.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59W003.hgt.zip',
		'N59W004.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59W004.hgt.zip',
		'N59W005.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59W005.hgt.zip',
		'N59W006.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59W006.hgt.zip',
		'N59W007.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N59W007.hgt.zip',
		'N60E004.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E004.hgt.zip',
		'N60E005.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E005.hgt.zip',
		'N60E006.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E006.hgt.zip',
		'N60E007.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E007.hgt.zip',
		'N60E008.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E008.hgt.zip',
		'N60E009.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E009.hgt.zip',
		'N60E010.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E010.hgt.zip',
		'N60E011.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E011.hgt.zip',
		'N60E012.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E012.hgt.zip',
		'N60E013.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E013.hgt.zip',
		'N60E014.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E014.hgt.zip',
		'N60E015.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E015.hgt.zip',
		'N60E016.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E016.hgt.zip',
		'N60E017.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E017.hgt.zip',
		'N60E018.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E018.hgt.zip',
		'N60E019.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E019.hgt.zip',
		'N60E020.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E020.hgt.zip',
		'N60E021.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E021.hgt.zip',
		'N60E022.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E022.hgt.zip',
		'N60E023.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E023.hgt.zip',
		'N60E024.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E024.hgt.zip',
		'N60E025.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E025.hgt.zip',
		'N60E026.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E026.hgt.zip',
		'N60E027.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E027.hgt.zip',
		'N60E028.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E028.hgt.zip',
		'N60E029.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E029.hgt.zip',
		'N60E030.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E030.hgt.zip',
		'N60E031.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E031.hgt.zip',
		'N60E032.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E032.hgt.zip',
		'N60E033.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E033.hgt.zip',
		'N60E034.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E034.hgt.zip',
		'N60E035.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E035.hgt.zip',
		'N60E036.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E036.hgt.zip',
		'N60E037.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E037.hgt.zip',
		'N60E038.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E038.hgt.zip',
		'N60E039.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E039.hgt.zip',
		'N60E040.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E040.hgt.zip',
		'N60E041.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E041.hgt.zip',
		'N60E042.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E042.hgt.zip',
		'N60E043.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E043.hgt.zip',
		'N60E044.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E044.hgt.zip',
		'N60E045.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E045.hgt.zip',
		'N60E046.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E046.hgt.zip',
		'N60E047.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E047.hgt.zip',
		'N60E048.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E048.hgt.zip',
		'N60E049.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E049.hgt.zip',
		'N60E050.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E050.hgt.zip',
		'N60E051.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E051.hgt.zip',
		'N60E052.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E052.hgt.zip',
		'N60E053.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E053.hgt.zip',
		'N60E054.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E054.hgt.zip',
		'N60E055.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E055.hgt.zip',
		'N60E056.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E056.hgt.zip',
		'N60E057.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E057.hgt.zip',
		'N60E058.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E058.hgt.zip',
		'N60E059.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E059.hgt.zip',
		'N60E060.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E060.hgt.zip',
		'N60E061.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E061.hgt.zip',
		'N60E062.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E062.hgt.zip',
		'N60E063.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E063.hgt.zip',
		'N60E064.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E064.hgt.zip',
		'N60E065.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E065.hgt.zip',
		'N60E066.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E066.hgt.zip',
		'N60E067.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E067.hgt.zip',
		'N60E068.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E068.hgt.zip',
		'N60E069.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E069.hgt.zip',
		'N60E070.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E070.hgt.zip',
		'N60E071.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E071.hgt.zip',
		'N60E072.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E072.hgt.zip',
		'N60E073.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E073.hgt.zip',
		'N60E074.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E074.hgt.zip',
		'N60E075.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E075.hgt.zip',
		'N60E076.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E076.hgt.zip',
		'N60E077.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E077.hgt.zip',
		'N60E078.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E078.hgt.zip',
		'N60E079.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E079.hgt.zip',
		'N60E080.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E080.hgt.zip',
		'N60E081.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E081.hgt.zip',
		'N60E082.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E082.hgt.zip',
		'N60E083.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E083.hgt.zip',
		'N60E084.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E084.hgt.zip',
		'N60E085.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E085.hgt.zip',
		'N60E086.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E086.hgt.zip',
		'N60E087.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E087.hgt.zip',
		'N60E088.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E088.hgt.zip',
		'N60E089.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E089.hgt.zip',
		'N60E090.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E090.hgt.zip',
		'N60E091.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E091.hgt.zip',
		'N60E092.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E092.hgt.zip',
		'N60E093.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E093.hgt.zip',
		'N60E094.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E094.hgt.zip',
		'N60E095.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E095.hgt.zip',
		'N60E096.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E096.hgt.zip',
		'N60E097.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E097.hgt.zip',
		'N60E098.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E098.hgt.zip',
		'N60E099.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E099.hgt.zip',
		'N60E100.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E100.hgt.zip',
		'N60E101.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E101.hgt.zip',
		'N60E102.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E102.hgt.zip',
		'N60E103.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E103.hgt.zip',
		'N60E104.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E104.hgt.zip',
		'N60E105.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E105.hgt.zip',
		'N60E106.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E106.hgt.zip',
		'N60E107.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E107.hgt.zip',
		'N60E108.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E108.hgt.zip',
		'N60E109.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E109.hgt.zip',
		'N60E110.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E110.hgt.zip',
		'N60E111.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E111.hgt.zip',
		'N60E112.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E112.hgt.zip',
		'N60E113.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E113.hgt.zip',
		'N60E114.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E114.hgt.zip',
		'N60E115.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E115.hgt.zip',
		'N60E116.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E116.hgt.zip',
		'N60E117.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E117.hgt.zip',
		'N60E118.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E118.hgt.zip',
		'N60E119.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E119.hgt.zip',
		'N60E120.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E120.hgt.zip',
		'N60E121.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E121.hgt.zip',
		'N60E122.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E122.hgt.zip',
		'N60E123.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E123.hgt.zip',
		'N60E124.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E124.hgt.zip',
		'N60E125.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E125.hgt.zip',
		'N60E126.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E126.hgt.zip',
		'N60E127.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E127.hgt.zip',
		'N60E128.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E128.hgt.zip',
		'N60E129.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E129.hgt.zip',
		'N60E130.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E130.hgt.zip',
		'N60E131.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E131.hgt.zip',
		'N60E132.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E132.hgt.zip',
		'N60E133.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E133.hgt.zip',
		'N60E134.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E134.hgt.zip',
		'N60E135.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E135.hgt.zip',
		'N60E136.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E136.hgt.zip',
		'N60E137.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E137.hgt.zip',
		'N60E138.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E138.hgt.zip',
		'N60E139.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E139.hgt.zip',
		'N60E140.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E140.hgt.zip',
		'N60E141.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E141.hgt.zip',
		'N60E142.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E142.hgt.zip',
		'N60E143.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E143.hgt.zip',
		'N60E144.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E144.hgt.zip',
		'N60E145.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E145.hgt.zip',
		'N60E146.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E146.hgt.zip',
		'N60E147.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E147.hgt.zip',
		'N60E148.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E148.hgt.zip',
		'N60E149.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E149.hgt.zip',
		'N60E150.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E150.hgt.zip',
		'N60E151.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E151.hgt.zip',
		'N60E152.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E152.hgt.zip',
		'N60E153.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E153.hgt.zip',
		'N60E154.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E154.hgt.zip',
		'N60E155.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E155.hgt.zip',
		'N60E156.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E156.hgt.zip',
		'N60E159.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E159.hgt.zip',
		'N60E160.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E160.hgt.zip',
		'N60E161.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E161.hgt.zip',
		'N60E162.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E162.hgt.zip',
		'N60E163.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E163.hgt.zip',
		'N60E164.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E164.hgt.zip',
		'N60E165.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E165.hgt.zip',
		'N60E166.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E166.hgt.zip',
		'N60E167.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E167.hgt.zip',
		'N60E168.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E168.hgt.zip',
		'N60E169.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E169.hgt.zip',
		'N60E170.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E170.hgt.zip',
		'N60E171.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E171.hgt.zip',
		'N60E172.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60E172.hgt.zip',
		'N60W001.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60W001.hgt.zip',
		'N60W002.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60W002.hgt.zip',
		'N60W003.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//N60W003.hgt.zip',
		'S01E006.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S01E006.hgt.zip',
		'S01E008.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S01E008.hgt.zip',
		'S01E009.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S01E009.hgt.zip',
		'S01E010.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S01E010.hgt.zip',
		'S01E011.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S01E011.hgt.zip',
		'S01E012.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S01E012.hgt.zip',
		'S01E013.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S01E013.hgt.zip',
		'S01E014.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S01E014.hgt.zip',
		'S01E015.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S01E015.hgt.zip',
		'S01E016.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S01E016.hgt.zip',
		'S01E017.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S01E017.hgt.zip',
		'S01E018.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S01E018.hgt.zip',
		'S01E019.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S01E019.hgt.zip',
		'S01E020.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S01E020.hgt.zip',
		'S01E021.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S01E021.hgt.zip',
		'S01E022.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S01E022.hgt.zip',
		'S01E023.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S01E023.hgt.zip',
		'S01E024.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S01E024.hgt.zip',
		'S01E025.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S01E025.hgt.zip',
		'S01E026.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S01E026.hgt.zip',
		'S01E027.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S01E027.hgt.zip',
		'S01E028.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S01E028.hgt.zip',
		'S01E029.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S01E029.hgt.zip',
		'S01E030.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S01E030.hgt.zip',
		'S01E031.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S01E031.hgt.zip',
		'S01E032.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S01E032.hgt.zip',
		'S01E033.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S01E033.hgt.zip',
		'S01E034.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S01E034.hgt.zip',
		'S01E035.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S01E035.hgt.zip',
		'S01E036.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S01E036.hgt.zip',
		'S01E037.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S01E037.hgt.zip',
		'S01E038.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S01E038.hgt.zip',
		'S01E039.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S01E039.hgt.zip',
		'S01E040.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S01E040.hgt.zip',
		'S01E041.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S01E041.hgt.zip',
		'S01E042.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S01E042.hgt.zip',
		'S01E073.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S01E073.hgt.zip',
		'S01E098.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S01E098.hgt.zip',
		'S01E099.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S01E099.hgt.zip',
		'S01E100.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S01E100.hgt.zip',
		'S01E101.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S01E101.hgt.zip',
		'S01E102.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S01E102.hgt.zip',
		'S01E103.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S01E103.hgt.zip',
		'S01E104.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S01E104.hgt.zip',
		'S01E105.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S01E105.hgt.zip',
		'S01E109.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S01E109.hgt.zip',
		'S01E110.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S01E110.hgt.zip',
		'S01E111.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S01E111.hgt.zip',
		'S01E112.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S01E112.hgt.zip',
		'S01E113.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S01E113.hgt.zip',
		'S01E114.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S01E114.hgt.zip',
		'S01E115.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S01E115.hgt.zip',
		'S01E116.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S01E116.hgt.zip',
		'S01E117.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S01E117.hgt.zip',
		'S01E119.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S01E119.hgt.zip',
		'S01E120.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S01E120.hgt.zip',
		'S01E121.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S01E121.hgt.zip',
		'S01E122.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S01E122.hgt.zip',
		'S01E123.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S01E123.hgt.zip',
		'S01E127.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S01E127.hgt.zip',
		'S01E128.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S01E128.hgt.zip',
		'S01E129.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S01E129.hgt.zip',
		'S01E130.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S01E130.hgt.zip',
		'S01E131.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S01E131.hgt.zip',
		'S01E132.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S01E132.hgt.zip',
		'S01E133.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S01E133.hgt.zip',
		'S01E134.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S01E134.hgt.zip',
		'S01E135.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S01E135.hgt.zip',
		'S01E136.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S01E136.hgt.zip',
		'S01E145.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S01E145.hgt.zip',
		'S01E166.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S01E166.hgt.zip',
		'S01E169.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S01E169.hgt.zip',
		'S01E174.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S01E174.hgt.zip',
		'S01W047.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S01W047.hgt.zip',
		'S01W048.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S01W048.hgt.zip',
		'S01W049.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S01W049.hgt.zip',
		'S01W050.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S01W050.hgt.zip',
		'S01W051.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S01W051.hgt.zip',
		'S01W052.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S01W052.hgt.zip',
		'S01W053.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S01W053.hgt.zip',
		'S01W054.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S01W054.hgt.zip',
		'S01W055.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S01W055.hgt.zip',
		'S01W056.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S01W056.hgt.zip',
		'S01W057.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S01W057.hgt.zip',
		'S01W058.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S01W058.hgt.zip',
		'S01W059.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S01W059.hgt.zip',
		'S01W060.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S01W060.hgt.zip',
		'S01W061.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S01W061.hgt.zip',
		'S01W062.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S01W062.hgt.zip',
		'S01W063.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S01W063.hgt.zip',
		'S01W064.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S01W064.hgt.zip',
		'S01W065.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S01W065.hgt.zip',
		'S01W066.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S01W066.hgt.zip',
		'S01W067.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S01W067.hgt.zip',
		'S01W068.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S01W068.hgt.zip',
		'S01W069.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S01W069.hgt.zip',
		'S01W070.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S01W070.hgt.zip',
		'S01W071.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S01W071.hgt.zip',
		'S01W072.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S01W072.hgt.zip',
		'S01W073.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S01W073.hgt.zip',
		'S01W074.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S01W074.hgt.zip',
		'S01W075.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S01W075.hgt.zip',
		'S01W076.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S01W076.hgt.zip',
		'S01W077.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S01W077.hgt.zip',
		'S01W078.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S01W078.hgt.zip',
		'S01W079.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S01W079.hgt.zip',
		'S01W080.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S01W080.hgt.zip',
		'S01W081.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S01W081.hgt.zip',
		'S01W090.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S01W090.hgt.zip',
		'S01W091.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S01W091.hgt.zip',
		'S01W092.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S01W092.hgt.zip',
		'S01W161.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S01W161.hgt.zip',
		'S02E005.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S02E005.hgt.zip',
		'S02E008.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S02E008.hgt.zip',
		'S02E009.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S02E009.hgt.zip',
		'S02E010.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S02E010.hgt.zip',
		'S02E011.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S02E011.hgt.zip',
		'S02E012.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S02E012.hgt.zip',
		'S02E013.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S02E013.hgt.zip',
		'S02E014.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S02E014.hgt.zip',
		'S02E015.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S02E015.hgt.zip',
		'S02E016.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S02E016.hgt.zip',
		'S02E017.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S02E017.hgt.zip',
		'S02E018.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S02E018.hgt.zip',
		'S02E019.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S02E019.hgt.zip',
		'S02E020.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S02E020.hgt.zip',
		'S02E021.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S02E021.hgt.zip',
		'S02E022.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S02E022.hgt.zip',
		'S02E023.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S02E023.hgt.zip',
		'S02E024.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S02E024.hgt.zip',
		'S02E025.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S02E025.hgt.zip',
		'S02E026.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S02E026.hgt.zip',
		'S02E027.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S02E027.hgt.zip',
		'S02E028.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S02E028.hgt.zip',
		'S02E029.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S02E029.hgt.zip',
		'S02E030.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S02E030.hgt.zip',
		'S02E031.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S02E031.hgt.zip',
		'S02E032.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S02E032.hgt.zip',
		'S02E033.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S02E033.hgt.zip',
		'S02E034.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S02E034.hgt.zip',
		'S02E035.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S02E035.hgt.zip',
		'S02E036.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S02E036.hgt.zip',
		'S02E037.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S02E037.hgt.zip',
		'S02E038.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S02E038.hgt.zip',
		'S02E039.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S02E039.hgt.zip',
		'S02E040.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S02E040.hgt.zip',
		'S02E041.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S02E041.hgt.zip',
		'S02E042.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S02E042.hgt.zip',
		'S02E098.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S02E098.hgt.zip',
		'S02E099.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S02E099.hgt.zip',
		'S02E100.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S02E100.hgt.zip',
		'S02E101.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S02E101.hgt.zip',
		'S02E102.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S02E102.hgt.zip',
		'S02E103.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S02E103.hgt.zip',
		'S02E104.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S02E104.hgt.zip',
		'S02E105.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S02E105.hgt.zip',
		'S02E106.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S02E106.hgt.zip',
		'S02E108.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S02E108.hgt.zip',
		'S02E109.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S02E109.hgt.zip',
		'S02E110.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S02E110.hgt.zip',
		'S02E111.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S02E111.hgt.zip',
		'S02E112.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S02E112.hgt.zip',
		'S02E113.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S02E113.hgt.zip',
		'S02E114.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S02E114.hgt.zip',
		'S02E115.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S02E115.hgt.zip',
		'S02E116.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S02E116.hgt.zip',
		'S02E117.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S02E117.hgt.zip',
		'S02E119.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S02E119.hgt.zip',
		'S02E120.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S02E120.hgt.zip',
		'S02E121.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S02E121.hgt.zip',
		'S02E122.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S02E122.hgt.zip',
		'S02E123.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S02E123.hgt.zip',
		'S02E124.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S02E124.hgt.zip',
		'S02E125.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S02E125.hgt.zip',
		'S02E126.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S02E126.hgt.zip',
		'S02E127.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S02E127.hgt.zip',
		'S02E128.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S02E128.hgt.zip',
		'S02E129.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S02E129.hgt.zip',
		'S02E130.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S02E130.hgt.zip',
		'S02E131.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S02E131.hgt.zip',
		'S02E132.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S02E132.hgt.zip',
		'S02E133.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S02E133.hgt.zip',
		'S02E134.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S02E134.hgt.zip',
		'S02E135.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S02E135.hgt.zip',
		'S02E136.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S02E136.hgt.zip',
		'S02E137.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S02E137.hgt.zip',
		'S02E138.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S02E138.hgt.zip',
		'S02E139.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S02E139.hgt.zip',
		'S02E142.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S02E142.hgt.zip',
		'S02E143.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S02E143.hgt.zip',
		'S02E144.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S02E144.hgt.zip',
		'S02E145.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S02E145.hgt.zip',
		'S02E146.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S02E146.hgt.zip',
		'S02E147.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S02E147.hgt.zip',
		'S02E148.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S02E148.hgt.zip',
		'S02E149.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S02E149.hgt.zip',
		'S02E150.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S02E150.hgt.zip',
		'S02E174.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S02E174.hgt.zip',
		'S02E175.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S02E175.hgt.zip',
		'S02E176.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S02E176.hgt.zip',
		'S02W045.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S02W045.hgt.zip',
		'S02W046.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S02W046.hgt.zip',
		'S02W047.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S02W047.hgt.zip',
		'S02W048.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S02W048.hgt.zip',
		'S02W049.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S02W049.hgt.zip',
		'S02W050.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S02W050.hgt.zip',
		'S02W051.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S02W051.hgt.zip',
		'S02W052.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S02W052.hgt.zip',
		'S02W053.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S02W053.hgt.zip',
		'S02W054.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S02W054.hgt.zip',
		'S02W055.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S02W055.hgt.zip',
		'S02W056.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S02W056.hgt.zip',
		'S02W057.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S02W057.hgt.zip',
		'S02W058.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S02W058.hgt.zip',
		'S02W059.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S02W059.hgt.zip',
		'S02W060.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S02W060.hgt.zip',
		'S02W061.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S02W061.hgt.zip',
		'S02W062.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S02W062.hgt.zip',
		'S02W063.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S02W063.hgt.zip',
		'S02W064.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S02W064.hgt.zip',
		'S02W065.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S02W065.hgt.zip',
		'S02W066.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S02W066.hgt.zip',
		'S02W067.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S02W067.hgt.zip',
		'S02W068.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S02W068.hgt.zip',
		'S02W069.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S02W069.hgt.zip',
		'S02W070.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S02W070.hgt.zip',
		'S02W071.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S02W071.hgt.zip',
		'S02W072.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S02W072.hgt.zip',
		'S02W073.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S02W073.hgt.zip',
		'S02W074.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S02W074.hgt.zip',
		'S02W075.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S02W075.hgt.zip',
		'S02W076.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S02W076.hgt.zip',
		'S02W077.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S02W077.hgt.zip',
		'S02W078.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S02W078.hgt.zip',
		'S02W079.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S02W079.hgt.zip',
		'S02W080.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S02W080.hgt.zip',
		'S02W081.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S02W081.hgt.zip',
		'S02W082.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S02W082.hgt.zip',
		'S02W090.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S02W090.hgt.zip',
		'S02W091.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S02W091.hgt.zip',
		'S02W092.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S02W092.hgt.zip',
		'S03E009.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S03E009.hgt.zip',
		'S03E010.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S03E010.hgt.zip',
		'S03E011.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S03E011.hgt.zip',
		'S03E012.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S03E012.hgt.zip',
		'S03E013.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S03E013.hgt.zip',
		'S03E014.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S03E014.hgt.zip',
		'S03E015.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S03E015.hgt.zip',
		'S03E016.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S03E016.hgt.zip',
		'S03E017.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S03E017.hgt.zip',
		'S03E018.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S03E018.hgt.zip',
		'S03E019.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S03E019.hgt.zip',
		'S03E020.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S03E020.hgt.zip',
		'S03E021.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S03E021.hgt.zip',
		'S03E022.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S03E022.hgt.zip',
		'S03E023.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S03E023.hgt.zip',
		'S03E024.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S03E024.hgt.zip',
		'S03E025.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S03E025.hgt.zip',
		'S03E026.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S03E026.hgt.zip',
		'S03E027.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S03E027.hgt.zip',
		'S03E028.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S03E028.hgt.zip',
		'S03E029.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S03E029.hgt.zip',
		'S03E030.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S03E030.hgt.zip',
		'S03E031.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S03E031.hgt.zip',
		'S03E032.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S03E032.hgt.zip',
		'S03E033.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S03E033.hgt.zip',
		'S03E034.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S03E034.hgt.zip',
		'S03E035.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S03E035.hgt.zip',
		'S03E036.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S03E036.hgt.zip',
		'S03E037.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S03E037.hgt.zip',
		'S03E038.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S03E038.hgt.zip',
		'S03E039.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S03E039.hgt.zip',
		'S03E040.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S03E040.hgt.zip',
		'S03E041.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S03E041.hgt.zip',
		'S03E099.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S03E099.hgt.zip',
		'S03E100.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S03E100.hgt.zip',
		'S03E101.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S03E101.hgt.zip',
		'S03E102.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S03E102.hgt.zip',
		'S03E103.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S03E103.hgt.zip',
		'S03E104.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S03E104.hgt.zip',
		'S03E105.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S03E105.hgt.zip',
		'S03E106.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S03E106.hgt.zip',
		'S03E107.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S03E107.hgt.zip',
		'S03E108.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S03E108.hgt.zip',
		'S03E110.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S03E110.hgt.zip',
		'S03E111.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S03E111.hgt.zip',
		'S03E112.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S03E112.hgt.zip',
		'S03E113.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S03E113.hgt.zip',
		'S03E114.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S03E114.hgt.zip',
		'S03E115.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S03E115.hgt.zip',
		'S03E116.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S03E116.hgt.zip',
		'S03E117.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S03E117.hgt.zip',
		'S03E118.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S03E118.hgt.zip',
		'S03E119.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S03E119.hgt.zip',
		'S03E120.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S03E120.hgt.zip',
		'S03E121.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S03E121.hgt.zip',
		'S03E122.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S03E122.hgt.zip',
		'S03E123.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S03E123.hgt.zip',
		'S03E124.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S03E124.hgt.zip',
		'S03E125.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S03E125.hgt.zip',
		'S03E126.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S03E126.hgt.zip',
		'S03E127.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S03E127.hgt.zip',
		'S03E128.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S03E128.hgt.zip',
		'S03E129.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S03E129.hgt.zip',
		'S03E130.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S03E130.hgt.zip',
		'S03E131.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S03E131.hgt.zip',
		'S03E132.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S03E132.hgt.zip',
		'S03E133.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S03E133.hgt.zip',
		'S03E134.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S03E134.hgt.zip',
		'S03E135.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S03E135.hgt.zip',
		'S03E136.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S03E136.hgt.zip',
		'S03E137.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S03E137.hgt.zip',
		'S03E138.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S03E138.hgt.zip',
		'S03E139.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S03E139.hgt.zip',
		'S03E140.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S03E140.hgt.zip',
		'S03E141.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S03E141.hgt.zip',
		'S03E142.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S03E142.hgt.zip',
		'S03E145.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S03E145.hgt.zip',
		'S03E146.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S03E146.hgt.zip',
		'S03E147.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S03E147.hgt.zip',
		'S03E148.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S03E148.hgt.zip',
		'S03E149.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S03E149.hgt.zip',
		'S03E150.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S03E150.hgt.zip',
		'S03E151.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S03E151.hgt.zip',
		'S03E152.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S03E152.hgt.zip',
		'S03E175.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S03E175.hgt.zip',
		'S03E176.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S03E176.hgt.zip',
		'S03W040.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S03W040.hgt.zip',
		'S03W041.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S03W041.hgt.zip',
		'S03W042.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S03W042.hgt.zip',
		'S03W043.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S03W043.hgt.zip',
		'S03W044.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S03W044.hgt.zip',
		'S03W045.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S03W045.hgt.zip',
		'S03W046.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S03W046.hgt.zip',
		'S03W047.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S03W047.hgt.zip',
		'S03W048.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S03W048.hgt.zip',
		'S03W049.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S03W049.hgt.zip',
		'S03W050.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S03W050.hgt.zip',
		'S03W051.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S03W051.hgt.zip',
		'S03W052.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S03W052.hgt.zip',
		'S03W053.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S03W053.hgt.zip',
		'S03W054.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S03W054.hgt.zip',
		'S03W055.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S03W055.hgt.zip',
		'S03W056.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S03W056.hgt.zip',
		'S03W057.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S03W057.hgt.zip',
		'S03W058.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S03W058.hgt.zip',
		'S03W059.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S03W059.hgt.zip',
		'S03W060.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S03W060.hgt.zip',
		'S03W061.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S03W061.hgt.zip',
		'S03W062.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S03W062.hgt.zip',
		'S03W063.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S03W063.hgt.zip',
		'S03W064.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S03W064.hgt.zip',
		'S03W065.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S03W065.hgt.zip',
		'S03W066.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S03W066.hgt.zip',
		'S03W067.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S03W067.hgt.zip',
		'S03W068.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S03W068.hgt.zip',
		'S03W069.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S03W069.hgt.zip',
		'S03W070.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S03W070.hgt.zip',
		'S03W071.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S03W071.hgt.zip',
		'S03W072.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S03W072.hgt.zip',
		'S03W073.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S03W073.hgt.zip',
		'S03W074.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S03W074.hgt.zip',
		'S03W075.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S03W075.hgt.zip',
		'S03W076.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S03W076.hgt.zip',
		'S03W077.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S03W077.hgt.zip',
		'S03W078.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S03W078.hgt.zip',
		'S03W079.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S03W079.hgt.zip',
		'S03W080.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S03W080.hgt.zip',
		'S03W081.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S03W081.hgt.zip',
		'S03W082.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S03W082.hgt.zip',
		'S03W172.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S03W172.hgt.zip',
		'S04E010.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S04E010.hgt.zip',
		'S04E011.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S04E011.hgt.zip',
		'S04E012.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S04E012.hgt.zip',
		'S04E013.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S04E013.hgt.zip',
		'S04E014.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S04E014.hgt.zip',
		'S04E015.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S04E015.hgt.zip',
		'S04E016.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S04E016.hgt.zip',
		'S04E017.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S04E017.hgt.zip',
		'S04E018.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S04E018.hgt.zip',
		'S04E019.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S04E019.hgt.zip',
		'S04E020.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S04E020.hgt.zip',
		'S04E021.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S04E021.hgt.zip',
		'S04E022.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S04E022.hgt.zip',
		'S04E023.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S04E023.hgt.zip',
		'S04E024.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S04E024.hgt.zip',
		'S04E025.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S04E025.hgt.zip',
		'S04E026.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S04E026.hgt.zip',
		'S04E027.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S04E027.hgt.zip',
		'S04E028.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S04E028.hgt.zip',
		'S04E029.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S04E029.hgt.zip',
		'S04E030.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S04E030.hgt.zip',
		'S04E031.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S04E031.hgt.zip',
		'S04E032.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S04E032.hgt.zip',
		'S04E033.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S04E033.hgt.zip',
		'S04E034.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S04E034.hgt.zip',
		'S04E035.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S04E035.hgt.zip',
		'S04E036.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S04E036.hgt.zip',
		'S04E037.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S04E037.hgt.zip',
		'S04E038.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S04E038.hgt.zip',
		'S04E039.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S04E039.hgt.zip',
		'S04E040.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S04E040.hgt.zip',
		'S04E055.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S04E055.hgt.zip',
		'S04E100.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S04E100.hgt.zip',
		'S04E101.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S04E101.hgt.zip',
		'S04E102.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S04E102.hgt.zip',
		'S04E103.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S04E103.hgt.zip',
		'S04E104.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S04E104.hgt.zip',
		'S04E105.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S04E105.hgt.zip',
		'S04E106.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S04E106.hgt.zip',
		'S04E107.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S04E107.hgt.zip',
		'S04E108.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S04E108.hgt.zip',
		'S04E110.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S04E110.hgt.zip',
		'S04E111.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S04E111.hgt.zip',
		'S04E112.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S04E112.hgt.zip',
		'S04E113.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S04E113.hgt.zip',
		'S04E114.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S04E114.hgt.zip',
		'S04E115.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S04E115.hgt.zip',
		'S04E116.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S04E116.hgt.zip',
		'S04E117.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S04E117.hgt.zip',
		'S04E118.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S04E118.hgt.zip',
		'S04E119.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S04E119.hgt.zip',
		'S04E120.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S04E120.hgt.zip',
		'S04E121.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S04E121.hgt.zip',
		'S04E122.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S04E122.hgt.zip',
		'S04E123.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S04E123.hgt.zip',
		'S04E125.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S04E125.hgt.zip',
		'S04E126.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S04E126.hgt.zip',
		'S04E127.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S04E127.hgt.zip',
		'S04E128.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S04E128.hgt.zip',
		'S04E129.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S04E129.hgt.zip',
		'S04E130.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S04E130.hgt.zip',
		'S04E131.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S04E131.hgt.zip',
		'S04E132.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S04E132.hgt.zip',
		'S04E133.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S04E133.hgt.zip',
		'S04E134.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S04E134.hgt.zip',
		'S04E135.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S04E135.hgt.zip',
		'S04E136.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S04E136.hgt.zip',
		'S04E137.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S04E137.hgt.zip',
		'S04E138.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S04E138.hgt.zip',
		'S04E139.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S04E139.hgt.zip',
		'S04E140.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S04E140.hgt.zip',
		'S04E141.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S04E141.hgt.zip',
		'S04E142.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S04E142.hgt.zip',
		'S04E143.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S04E143.hgt.zip',
		'S04E144.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S04E144.hgt.zip',
		'S04E150.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S04E150.hgt.zip',
		'S04E151.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S04E151.hgt.zip',
		'S04E152.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S04E152.hgt.zip',
		'S04E153.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S04E153.hgt.zip',
		'S04E154.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S04E154.hgt.zip',
		'S04W033.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S04W033.hgt.zip',
		'S04W034.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S04W034.hgt.zip',
		'S04W039.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S04W039.hgt.zip',
		'S04W040.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S04W040.hgt.zip',
		'S04W041.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S04W041.hgt.zip',
		'S04W042.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S04W042.hgt.zip',
		'S04W043.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S04W043.hgt.zip',
		'S04W044.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S04W044.hgt.zip',
		'S04W045.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S04W045.hgt.zip',
		'S04W046.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S04W046.hgt.zip',
		'S04W047.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S04W047.hgt.zip',
		'S04W048.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S04W048.hgt.zip',
		'S04W049.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S04W049.hgt.zip',
		'S04W050.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S04W050.hgt.zip',
		'S04W051.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S04W051.hgt.zip',
		'S04W052.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S04W052.hgt.zip',
		'S04W053.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S04W053.hgt.zip',
		'S04W054.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S04W054.hgt.zip',
		'S04W055.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S04W055.hgt.zip',
		'S04W056.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S04W056.hgt.zip',
		'S04W057.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S04W057.hgt.zip',
		'S04W058.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S04W058.hgt.zip',
		'S04W059.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S04W059.hgt.zip',
		'S04W060.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S04W060.hgt.zip',
		'S04W061.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S04W061.hgt.zip',
		'S04W062.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S04W062.hgt.zip',
		'S04W063.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S04W063.hgt.zip',
		'S04W064.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S04W064.hgt.zip',
		'S04W065.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S04W065.hgt.zip',
		'S04W066.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S04W066.hgt.zip',
		'S04W067.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S04W067.hgt.zip',
		'S04W068.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S04W068.hgt.zip',
		'S04W069.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S04W069.hgt.zip',
		'S04W070.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S04W070.hgt.zip',
		'S04W071.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S04W071.hgt.zip',
		'S04W072.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S04W072.hgt.zip',
		'S04W073.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S04W073.hgt.zip',
		'S04W074.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S04W074.hgt.zip',
		'S04W075.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S04W075.hgt.zip',
		'S04W076.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S04W076.hgt.zip',
		'S04W077.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S04W077.hgt.zip',
		'S04W078.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S04W078.hgt.zip',
		'S04W079.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S04W079.hgt.zip',
		'S04W080.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S04W080.hgt.zip',
		'S04W081.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S04W081.hgt.zip',
		'S04W155.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S04W155.hgt.zip',
		'S04W171.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S04W171.hgt.zip',
		'S04W172.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S04W172.hgt.zip',
		'S04W175.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S04W175.hgt.zip',
		'S05E011.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S05E011.hgt.zip',
		'S05E012.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S05E012.hgt.zip',
		'S05E013.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S05E013.hgt.zip',
		'S05E014.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S05E014.hgt.zip',
		'S05E015.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S05E015.hgt.zip',
		'S05E016.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S05E016.hgt.zip',
		'S05E017.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S05E017.hgt.zip',
		'S05E018.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S05E018.hgt.zip',
		'S05E019.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S05E019.hgt.zip',
		'S05E020.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S05E020.hgt.zip',
		'S05E021.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S05E021.hgt.zip',
		'S05E022.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S05E022.hgt.zip',
		'S05E023.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S05E023.hgt.zip',
		'S05E024.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S05E024.hgt.zip',
		'S05E025.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S05E025.hgt.zip',
		'S05E026.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S05E026.hgt.zip',
		'S05E027.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S05E027.hgt.zip',
		'S05E028.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S05E028.hgt.zip',
		'S05E029.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S05E029.hgt.zip',
		'S05E030.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S05E030.hgt.zip',
		'S05E031.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S05E031.hgt.zip',
		'S05E032.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S05E032.hgt.zip',
		'S05E033.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S05E033.hgt.zip',
		'S05E034.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S05E034.hgt.zip',
		'S05E035.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S05E035.hgt.zip',
		'S05E036.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S05E036.hgt.zip',
		'S05E037.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S05E037.hgt.zip',
		'S05E038.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S05E038.hgt.zip',
		'S05E039.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S05E039.hgt.zip',
		'S05E053.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S05E053.hgt.zip',
		'S05E055.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S05E055.hgt.zip',
		'S05E101.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S05E101.hgt.zip',
		'S05E102.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S05E102.hgt.zip',
		'S05E103.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S05E103.hgt.zip',
		'S05E104.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S05E104.hgt.zip',
		'S05E105.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S05E105.hgt.zip',
		'S05E114.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S05E114.hgt.zip',
		'S05E115.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S05E115.hgt.zip',
		'S05E116.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S05E116.hgt.zip',
		'S05E119.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S05E119.hgt.zip',
		'S05E120.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S05E120.hgt.zip',
		'S05E121.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S05E121.hgt.zip',
		'S05E122.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S05E122.hgt.zip',
		'S05E123.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S05E123.hgt.zip',
		'S05E129.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S05E129.hgt.zip',
		'S05E130.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S05E130.hgt.zip',
		'S05E131.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S05E131.hgt.zip',
		'S05E132.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S05E132.hgt.zip',
		'S05E133.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S05E133.hgt.zip',
		'S05E134.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S05E134.hgt.zip',
		'S05E135.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S05E135.hgt.zip',
		'S05E136.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S05E136.hgt.zip',
		'S05E137.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S05E137.hgt.zip',
		'S05E138.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S05E138.hgt.zip',
		'S05E139.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S05E139.hgt.zip',
		'S05E140.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S05E140.hgt.zip',
		'S05E141.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S05E141.hgt.zip',
		'S05E142.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S05E142.hgt.zip',
		'S05E143.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S05E143.hgt.zip',
		'S05E144.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S05E144.hgt.zip',
		'S05E145.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S05E145.hgt.zip',
		'S05E146.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S05E146.hgt.zip',
		'S05E149.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S05E149.hgt.zip',
		'S05E150.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S05E150.hgt.zip',
		'S05E151.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S05E151.hgt.zip',
		'S05E152.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S05E152.hgt.zip',
		'S05E153.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S05E153.hgt.zip',
		'S05E154.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S05E154.hgt.zip',
		'S05E155.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S05E155.hgt.zip',
		'S05E156.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S05E156.hgt.zip',
		'S05E157.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S05E157.hgt.zip',
		'S05E159.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S05E159.hgt.zip',
		'S05W037.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S05W037.hgt.zip',
		'S05W038.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S05W038.hgt.zip',
		'S05W039.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S05W039.hgt.zip',
		'S05W040.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S05W040.hgt.zip',
		'S05W041.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S05W041.hgt.zip',
		'S05W042.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S05W042.hgt.zip',
		'S05W043.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S05W043.hgt.zip',
		'S05W044.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S05W044.hgt.zip',
		'S05W045.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S05W045.hgt.zip',
		'S05W046.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S05W046.hgt.zip',
		'S05W047.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S05W047.hgt.zip',
		'S05W048.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S05W048.hgt.zip',
		'S05W049.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S05W049.hgt.zip',
		'S05W050.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S05W050.hgt.zip',
		'S05W051.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S05W051.hgt.zip',
		'S05W052.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S05W052.hgt.zip',
		'S05W053.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S05W053.hgt.zip',
		'S05W054.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S05W054.hgt.zip',
		'S05W055.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S05W055.hgt.zip',
		'S05W056.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S05W056.hgt.zip',
		'S05W057.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S05W057.hgt.zip',
		'S05W058.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S05W058.hgt.zip',
		'S05W059.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S05W059.hgt.zip',
		'S05W060.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S05W060.hgt.zip',
		'S05W061.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S05W061.hgt.zip',
		'S05W062.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S05W062.hgt.zip',
		'S05W063.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S05W063.hgt.zip',
		'S05W064.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S05W064.hgt.zip',
		'S05W065.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S05W065.hgt.zip',
		'S05W066.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S05W066.hgt.zip',
		'S05W067.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S05W067.hgt.zip',
		'S05W068.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S05W068.hgt.zip',
		'S05W069.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S05W069.hgt.zip',
		'S05W070.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S05W070.hgt.zip',
		'S05W071.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S05W071.hgt.zip',
		'S05W072.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S05W072.hgt.zip',
		'S05W073.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S05W073.hgt.zip',
		'S05W074.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S05W074.hgt.zip',
		'S05W075.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S05W075.hgt.zip',
		'S05W076.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S05W076.hgt.zip',
		'S05W077.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S05W077.hgt.zip',
		'S05W078.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S05W078.hgt.zip',
		'S05W079.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S05W079.hgt.zip',
		'S05W080.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S05W080.hgt.zip',
		'S05W081.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S05W081.hgt.zip',
		'S05W082.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S05W082.hgt.zip',
		'S05W155.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S05W155.hgt.zip',
		'S05W172.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S05W172.hgt.zip',
		'S05W173.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S05W173.hgt.zip',
		'S05W175.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S05W175.hgt.zip',
		'S06E011.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S06E011.hgt.zip',
		'S06E012.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S06E012.hgt.zip',
		'S06E013.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S06E013.hgt.zip',
		'S06E014.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S06E014.hgt.zip',
		'S06E015.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S06E015.hgt.zip',
		'S06E016.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S06E016.hgt.zip',
		'S06E017.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S06E017.hgt.zip',
		'S06E018.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S06E018.hgt.zip',
		'S06E019.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S06E019.hgt.zip',
		'S06E020.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S06E020.hgt.zip',
		'S06E021.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S06E021.hgt.zip',
		'S06E022.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S06E022.hgt.zip',
		'S06E023.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S06E023.hgt.zip',
		'S06E024.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S06E024.hgt.zip',
		'S06E025.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S06E025.hgt.zip',
		'S06E026.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S06E026.hgt.zip',
		'S06E027.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S06E027.hgt.zip',
		'S06E028.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S06E028.hgt.zip',
		'S06E029.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S06E029.hgt.zip',
		'S06E030.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S06E030.hgt.zip',
		'S06E031.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S06E031.hgt.zip',
		'S06E032.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S06E032.hgt.zip',
		'S06E033.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S06E033.hgt.zip',
		'S06E034.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S06E034.hgt.zip',
		'S06E035.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S06E035.hgt.zip',
		'S06E036.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S06E036.hgt.zip',
		'S06E037.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S06E037.hgt.zip',
		'S06E038.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S06E038.hgt.zip',
		'S06E039.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S06E039.hgt.zip',
		'S06E053.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S06E053.hgt.zip',
		'S06E055.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S06E055.hgt.zip',
		'S06E071.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S06E071.hgt.zip',
		'S06E072.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S06E072.hgt.zip',
		'S06E102.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S06E102.hgt.zip',
		'S06E103.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S06E103.hgt.zip',
		'S06E104.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S06E104.hgt.zip',
		'S06E105.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S06E105.hgt.zip',
		'S06E106.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S06E106.hgt.zip',
		'S06E107.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S06E107.hgt.zip',
		'S06E108.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S06E108.hgt.zip',
		'S06E110.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S06E110.hgt.zip',
		'S06E112.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S06E112.hgt.zip',
		'S06E114.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S06E114.hgt.zip',
		'S06E117.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S06E117.hgt.zip',
		'S06E118.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S06E118.hgt.zip',
		'S06E119.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S06E119.hgt.zip',
		'S06E120.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S06E120.hgt.zip',
		'S06E121.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S06E121.hgt.zip',
		'S06E122.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S06E122.hgt.zip',
		'S06E123.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S06E123.hgt.zip',
		'S06E124.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S06E124.hgt.zip',
		'S06E127.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S06E127.hgt.zip',
		'S06E130.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S06E130.hgt.zip',
		'S06E131.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S06E131.hgt.zip',
		'S06E132.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S06E132.hgt.zip',
		'S06E133.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S06E133.hgt.zip',
		'S06E134.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S06E134.hgt.zip',
		'S06E137.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S06E137.hgt.zip',
		'S06E138.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S06E138.hgt.zip',
		'S06E139.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S06E139.hgt.zip',
		'S06E140.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S06E140.hgt.zip',
		'S06E141.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S06E141.hgt.zip',
		'S06E142.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S06E142.hgt.zip',
		'S06E143.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S06E143.hgt.zip',
		'S06E144.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S06E144.hgt.zip',
		'S06E145.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S06E145.hgt.zip',
		'S06E146.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S06E146.hgt.zip',
		'S06E147.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S06E147.hgt.zip',
		'S06E148.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S06E148.hgt.zip',
		'S06E149.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S06E149.hgt.zip',
		'S06E150.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S06E150.hgt.zip',
		'S06E151.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S06E151.hgt.zip',
		'S06E152.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S06E152.hgt.zip',
		'S06E154.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S06E154.hgt.zip',
		'S06E155.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S06E155.hgt.zip',
		'S06E159.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S06E159.hgt.zip',
		'S06E176.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S06E176.hgt.zip',
		'S06W036.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S06W036.hgt.zip',
		'S06W037.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S06W037.hgt.zip',
		'S06W038.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S06W038.hgt.zip',
		'S06W039.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S06W039.hgt.zip',
		'S06W040.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S06W040.hgt.zip',
		'S06W041.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S06W041.hgt.zip',
		'S06W042.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S06W042.hgt.zip',
		'S06W043.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S06W043.hgt.zip',
		'S06W044.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S06W044.hgt.zip',
		'S06W045.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S06W045.hgt.zip',
		'S06W046.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S06W046.hgt.zip',
		'S06W047.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S06W047.hgt.zip',
		'S06W048.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S06W048.hgt.zip',
		'S06W049.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S06W049.hgt.zip',
		'S06W050.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S06W050.hgt.zip',
		'S06W051.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S06W051.hgt.zip',
		'S06W052.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S06W052.hgt.zip',
		'S06W053.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S06W053.hgt.zip',
		'S06W054.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S06W054.hgt.zip',
		'S06W055.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S06W055.hgt.zip',
		'S06W056.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S06W056.hgt.zip',
		'S06W057.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S06W057.hgt.zip',
		'S06W058.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S06W058.hgt.zip',
		'S06W059.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S06W059.hgt.zip',
		'S06W060.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S06W060.hgt.zip',
		'S06W061.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S06W061.hgt.zip',
		'S06W062.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S06W062.hgt.zip',
		'S06W063.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S06W063.hgt.zip',
		'S06W064.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S06W064.hgt.zip',
		'S06W065.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S06W065.hgt.zip',
		'S06W066.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S06W066.hgt.zip',
		'S06W067.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S06W067.hgt.zip',
		'S06W068.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S06W068.hgt.zip',
		'S06W069.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S06W069.hgt.zip',
		'S06W070.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S06W070.hgt.zip',
		'S06W071.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S06W071.hgt.zip',
		'S06W072.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S06W072.hgt.zip',
		'S06W073.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S06W073.hgt.zip',
		'S06W074.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S06W074.hgt.zip',
		'S06W075.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S06W075.hgt.zip',
		'S06W076.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S06W076.hgt.zip',
		'S06W077.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S06W077.hgt.zip',
		'S06W078.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S06W078.hgt.zip',
		'S06W079.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S06W079.hgt.zip',
		'S06W080.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S06W080.hgt.zip',
		'S06W081.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S06W081.hgt.zip',
		'S06W082.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S06W082.hgt.zip',
		'S06W156.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S06W156.hgt.zip',
		'S07E012.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S07E012.hgt.zip',
		'S07E013.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S07E013.hgt.zip',
		'S07E014.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S07E014.hgt.zip',
		'S07E015.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S07E015.hgt.zip',
		'S07E016.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S07E016.hgt.zip',
		'S07E017.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S07E017.hgt.zip',
		'S07E018.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S07E018.hgt.zip',
		'S07E019.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S07E019.hgt.zip',
		'S07E020.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S07E020.hgt.zip',
		'S07E021.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S07E021.hgt.zip',
		'S07E022.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S07E022.hgt.zip',
		'S07E023.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S07E023.hgt.zip',
		'S07E024.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S07E024.hgt.zip',
		'S07E025.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S07E025.hgt.zip',
		'S07E026.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S07E026.hgt.zip',
		'S07E027.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S07E027.hgt.zip',
		'S07E028.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S07E028.hgt.zip',
		'S07E029.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S07E029.hgt.zip',
		'S07E030.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S07E030.hgt.zip',
		'S07E031.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S07E031.hgt.zip',
		'S07E032.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S07E032.hgt.zip',
		'S07E033.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S07E033.hgt.zip',
		'S07E034.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S07E034.hgt.zip',
		'S07E035.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S07E035.hgt.zip',
		'S07E036.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S07E036.hgt.zip',
		'S07E037.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S07E037.hgt.zip',
		'S07E038.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S07E038.hgt.zip',
		'S07E039.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S07E039.hgt.zip',
		'S07E052.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S07E052.hgt.zip',
		'S07E053.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S07E053.hgt.zip',
		'S07E071.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S07E071.hgt.zip',
		'S07E105.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S07E105.hgt.zip',
		'S07E106.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S07E106.hgt.zip',
		'S07E107.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S07E107.hgt.zip',
		'S07E108.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S07E108.hgt.zip',
		'S07E109.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S07E109.hgt.zip',
		'S07E110.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S07E110.hgt.zip',
		'S07E111.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S07E111.hgt.zip',
		'S07E112.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S07E112.hgt.zip',
		'S07E113.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S07E113.hgt.zip',
		'S07E114.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S07E114.hgt.zip',
		'S07E115.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S07E115.hgt.zip',
		'S07E116.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S07E116.hgt.zip',
		'S07E118.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S07E118.hgt.zip',
		'S07E119.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S07E119.hgt.zip',
		'S07E120.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S07E120.hgt.zip',
		'S07E121.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S07E121.hgt.zip',
		'S07E122.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S07E122.hgt.zip',
		'S07E124.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S07E124.hgt.zip',
		'S07E126.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S07E126.hgt.zip',
		'S07E129.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S07E129.hgt.zip',
		'S07E130.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S07E130.hgt.zip',
		'S07E131.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S07E131.hgt.zip',
		'S07E132.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S07E132.hgt.zip',
		'S07E134.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S07E134.hgt.zip',
		'S07E138.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S07E138.hgt.zip',
		'S07E139.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S07E139.hgt.zip',
		'S07E140.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S07E140.hgt.zip',
		'S07E141.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S07E141.hgt.zip',
		'S07E142.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S07E142.hgt.zip',
		'S07E143.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S07E143.hgt.zip',
		'S07E144.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S07E144.hgt.zip',
		'S07E145.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S07E145.hgt.zip',
		'S07E146.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S07E146.hgt.zip',
		'S07E147.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S07E147.hgt.zip',
		'S07E148.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S07E148.hgt.zip',
		'S07E149.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S07E149.hgt.zip',
		'S07E150.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S07E150.hgt.zip',
		'S07E151.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S07E151.hgt.zip',
		'S07E154.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S07E154.hgt.zip',
		'S07E155.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S07E155.hgt.zip',
		'S07E156.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S07E156.hgt.zip',
		'S07E157.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S07E157.hgt.zip',
		'S07E176.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S07E176.hgt.zip',
		'S07E177.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S07E177.hgt.zip',
		'S07W035.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S07W035.hgt.zip',
		'S07W036.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S07W036.hgt.zip',
		'S07W037.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S07W037.hgt.zip',
		'S07W038.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S07W038.hgt.zip',
		'S07W039.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S07W039.hgt.zip',
		'S07W040.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S07W040.hgt.zip',
		'S07W041.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S07W041.hgt.zip',
		'S07W042.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S07W042.hgt.zip',
		'S07W043.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S07W043.hgt.zip',
		'S07W044.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S07W044.hgt.zip',
		'S07W045.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S07W045.hgt.zip',
		'S07W046.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S07W046.hgt.zip',
		'S07W047.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S07W047.hgt.zip',
		'S07W048.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S07W048.hgt.zip',
		'S07W049.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S07W049.hgt.zip',
		'S07W050.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S07W050.hgt.zip',
		'S07W051.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S07W051.hgt.zip',
		'S07W052.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S07W052.hgt.zip',
		'S07W053.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S07W053.hgt.zip',
		'S07W054.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S07W054.hgt.zip',
		'S07W055.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S07W055.hgt.zip',
		'S07W056.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S07W056.hgt.zip',
		'S07W057.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S07W057.hgt.zip',
		'S07W058.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S07W058.hgt.zip',
		'S07W059.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S07W059.hgt.zip',
		'S07W060.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S07W060.hgt.zip',
		'S07W061.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S07W061.hgt.zip',
		'S07W062.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S07W062.hgt.zip',
		'S07W063.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S07W063.hgt.zip',
		'S07W064.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S07W064.hgt.zip',
		'S07W065.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S07W065.hgt.zip',
		'S07W066.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S07W066.hgt.zip',
		'S07W067.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S07W067.hgt.zip',
		'S07W068.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S07W068.hgt.zip',
		'S07W069.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S07W069.hgt.zip',
		'S07W070.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S07W070.hgt.zip',
		'S07W071.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S07W071.hgt.zip',
		'S07W072.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S07W072.hgt.zip',
		'S07W073.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S07W073.hgt.zip',
		'S07W074.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S07W074.hgt.zip',
		'S07W075.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S07W075.hgt.zip',
		'S07W076.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S07W076.hgt.zip',
		'S07W077.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S07W077.hgt.zip',
		'S07W078.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S07W078.hgt.zip',
		'S07W079.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S07W079.hgt.zip',
		'S07W080.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S07W080.hgt.zip',
		'S07W081.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S07W081.hgt.zip',
		'S07W082.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S07W082.hgt.zip',
		'S08E012.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S08E012.hgt.zip',
		'S08E013.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S08E013.hgt.zip',
		'S08E014.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S08E014.hgt.zip',
		'S08E015.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S08E015.hgt.zip',
		'S08E016.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S08E016.hgt.zip',
		'S08E017.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S08E017.hgt.zip',
		'S08E018.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S08E018.hgt.zip',
		'S08E019.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S08E019.hgt.zip',
		'S08E020.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S08E020.hgt.zip',
		'S08E021.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S08E021.hgt.zip',
		'S08E022.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S08E022.hgt.zip',
		'S08E023.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S08E023.hgt.zip',
		'S08E024.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S08E024.hgt.zip',
		'S08E025.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S08E025.hgt.zip',
		'S08E026.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S08E026.hgt.zip',
		'S08E027.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S08E027.hgt.zip',
		'S08E028.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S08E028.hgt.zip',
		'S08E029.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S08E029.hgt.zip',
		'S08E030.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S08E030.hgt.zip',
		'S08E031.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S08E031.hgt.zip',
		'S08E032.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S08E032.hgt.zip',
		'S08E033.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S08E033.hgt.zip',
		'S08E034.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S08E034.hgt.zip',
		'S08E035.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S08E035.hgt.zip',
		'S08E036.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S08E036.hgt.zip',
		'S08E037.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S08E037.hgt.zip',
		'S08E038.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S08E038.hgt.zip',
		'S08E039.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S08E039.hgt.zip',
		'S08E052.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S08E052.hgt.zip',
		'S08E056.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S08E056.hgt.zip',
		'S08E072.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S08E072.hgt.zip',
		'S08E105.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S08E105.hgt.zip',
		'S08E106.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S08E106.hgt.zip',
		'S08E107.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S08E107.hgt.zip',
		'S08E108.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S08E108.hgt.zip',
		'S08E109.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S08E109.hgt.zip',
		'S08E110.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S08E110.hgt.zip',
		'S08E111.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S08E111.hgt.zip',
		'S08E112.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S08E112.hgt.zip',
		'S08E113.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S08E113.hgt.zip',
		'S08E114.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S08E114.hgt.zip',
		'S08E115.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S08E115.hgt.zip',
		'S08E117.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S08E117.hgt.zip',
		'S08E118.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S08E118.hgt.zip',
		'S08E120.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S08E120.hgt.zip',
		'S08E121.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S08E121.hgt.zip',
		'S08E122.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S08E122.hgt.zip',
		'S08E123.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S08E123.hgt.zip',
		'S08E125.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S08E125.hgt.zip',
		'S08E126.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S08E126.hgt.zip',
		'S08E127.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S08E127.hgt.zip',
		'S08E128.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S08E128.hgt.zip',
		'S08E129.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S08E129.hgt.zip',
		'S08E130.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S08E130.hgt.zip',
		'S08E131.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S08E131.hgt.zip',
		'S08E134.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S08E134.hgt.zip',
		'S08E137.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S08E137.hgt.zip',
		'S08E138.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S08E138.hgt.zip',
		'S08E139.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S08E139.hgt.zip',
		'S08E140.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S08E140.hgt.zip',
		'S08E141.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S08E141.hgt.zip',
		'S08E142.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S08E142.hgt.zip',
		'S08E143.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S08E143.hgt.zip',
		'S08E144.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S08E144.hgt.zip',
		'S08E145.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S08E145.hgt.zip',
		'S08E146.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S08E146.hgt.zip',
		'S08E147.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S08E147.hgt.zip',
		'S08E155.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S08E155.hgt.zip',
		'S08E156.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S08E156.hgt.zip',
		'S08E157.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S08E157.hgt.zip',
		'S08E158.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S08E158.hgt.zip',
		'S08E159.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S08E159.hgt.zip',
		'S08E160.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S08E160.hgt.zip',
		'S08E177.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S08E177.hgt.zip',
		'S08E178.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S08E178.hgt.zip',
		'S08W015.hgt': 'https://srtm.kurviger.de/SRTM3//Islands//S08W015.hgt.zip',
		'S08W035.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S08W035.hgt.zip',
		'S08W036.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S08W036.hgt.zip',
		'S08W037.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S08W037.hgt.zip',
		'S08W038.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S08W038.hgt.zip',
		'S08W039.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S08W039.hgt.zip',
		'S08W040.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S08W040.hgt.zip',
		'S08W041.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S08W041.hgt.zip',
		'S08W042.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S08W042.hgt.zip',
		'S08W043.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S08W043.hgt.zip',
		'S08W044.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S08W044.hgt.zip',
		'S08W045.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S08W045.hgt.zip',
		'S08W046.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S08W046.hgt.zip',
		'S08W047.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S08W047.hgt.zip',
		'S08W048.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S08W048.hgt.zip',
		'S08W049.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S08W049.hgt.zip',
		'S08W050.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S08W050.hgt.zip',
		'S08W051.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S08W051.hgt.zip',
		'S08W052.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S08W052.hgt.zip',
		'S08W053.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S08W053.hgt.zip',
		'S08W054.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S08W054.hgt.zip',
		'S08W055.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S08W055.hgt.zip',
		'S08W056.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S08W056.hgt.zip',
		'S08W057.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S08W057.hgt.zip',
		'S08W058.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S08W058.hgt.zip',
		'S08W059.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S08W059.hgt.zip',
		'S08W060.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S08W060.hgt.zip',
		'S08W061.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S08W061.hgt.zip',
		'S08W062.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S08W062.hgt.zip',
		'S08W063.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S08W063.hgt.zip',
		'S08W064.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S08W064.hgt.zip',
		'S08W065.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S08W065.hgt.zip',
		'S08W066.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S08W066.hgt.zip',
		'S08W067.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S08W067.hgt.zip',
		'S08W068.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S08W068.hgt.zip',
		'S08W069.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S08W069.hgt.zip',
		'S08W070.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S08W070.hgt.zip',
		'S08W071.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S08W071.hgt.zip',
		'S08W072.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S08W072.hgt.zip',
		'S08W073.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S08W073.hgt.zip',
		'S08W074.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S08W074.hgt.zip',
		'S08W075.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S08W075.hgt.zip',
		'S08W076.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S08W076.hgt.zip',
		'S08W077.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S08W077.hgt.zip',
		'S08W078.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S08W078.hgt.zip',
		'S08W079.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S08W079.hgt.zip',
		'S08W080.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S08W080.hgt.zip',
		'S08W141.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S08W141.hgt.zip',
		'S09E013.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S09E013.hgt.zip',
		'S09E014.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S09E014.hgt.zip',
		'S09E015.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S09E015.hgt.zip',
		'S09E016.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S09E016.hgt.zip',
		'S09E017.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S09E017.hgt.zip',
		'S09E018.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S09E018.hgt.zip',
		'S09E019.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S09E019.hgt.zip',
		'S09E020.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S09E020.hgt.zip',
		'S09E021.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S09E021.hgt.zip',
		'S09E022.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S09E022.hgt.zip',
		'S09E023.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S09E023.hgt.zip',
		'S09E024.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S09E024.hgt.zip',
		'S09E025.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S09E025.hgt.zip',
		'S09E026.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S09E026.hgt.zip',
		'S09E027.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S09E027.hgt.zip',
		'S09E028.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S09E028.hgt.zip',
		'S09E029.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S09E029.hgt.zip',
		'S09E030.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S09E030.hgt.zip',
		'S09E031.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S09E031.hgt.zip',
		'S09E032.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S09E032.hgt.zip',
		'S09E033.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S09E033.hgt.zip',
		'S09E034.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S09E034.hgt.zip',
		'S09E035.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S09E035.hgt.zip',
		'S09E036.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S09E036.hgt.zip',
		'S09E037.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S09E037.hgt.zip',
		'S09E038.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S09E038.hgt.zip',
		'S09E039.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S09E039.hgt.zip',
		'S09E110.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S09E110.hgt.zip',
		'S09E111.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S09E111.hgt.zip',
		'S09E112.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S09E112.hgt.zip',
		'S09E113.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S09E113.hgt.zip',
		'S09E114.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S09E114.hgt.zip',
		'S09E115.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S09E115.hgt.zip',
		'S09E116.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S09E116.hgt.zip',
		'S09E117.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S09E117.hgt.zip',
		'S09E118.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S09E118.hgt.zip',
		'S09E119.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S09E119.hgt.zip',
		'S09E120.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S09E120.hgt.zip',
		'S09E121.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S09E121.hgt.zip',
		'S09E122.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S09E122.hgt.zip',
		'S09E123.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S09E123.hgt.zip',
		'S09E124.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S09E124.hgt.zip',
		'S09E125.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S09E125.hgt.zip',
		'S09E126.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S09E126.hgt.zip',
		'S09E127.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S09E127.hgt.zip',
		'S09E128.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S09E128.hgt.zip',
		'S09E129.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S09E129.hgt.zip',
		'S09E130.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S09E130.hgt.zip',
		'S09E131.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S09E131.hgt.zip',
		'S09E137.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S09E137.hgt.zip',
		'S09E138.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S09E138.hgt.zip',
		'S09E139.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S09E139.hgt.zip',
		'S09E140.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S09E140.hgt.zip',
		'S09E141.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S09E141.hgt.zip',
		'S09E142.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S09E142.hgt.zip',
		'S09E143.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S09E143.hgt.zip',
		'S09E145.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S09E145.hgt.zip',
		'S09E146.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S09E146.hgt.zip',
		'S09E147.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S09E147.hgt.zip',
		'S09E148.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S09E148.hgt.zip',
		'S09E149.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S09E149.hgt.zip',
		'S09E150.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S09E150.hgt.zip',
		'S09E151.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S09E151.hgt.zip',
		'S09E152.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S09E152.hgt.zip',
		'S09E156.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S09E156.hgt.zip',
		'S09E157.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S09E157.hgt.zip',
		'S09E158.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S09E158.hgt.zip',
		'S09E159.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S09E159.hgt.zip',
		'S09E160.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S09E160.hgt.zip',
		'S09E161.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S09E161.hgt.zip',
		'S09E178.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S09E178.hgt.zip',
		'S09E179.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S09E179.hgt.zip',
		'S09W035.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S09W035.hgt.zip',
		'S09W036.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S09W036.hgt.zip',
		'S09W037.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S09W037.hgt.zip',
		'S09W038.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S09W038.hgt.zip',
		'S09W039.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S09W039.hgt.zip',
		'S09W040.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S09W040.hgt.zip',
		'S09W041.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S09W041.hgt.zip',
		'S09W042.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S09W042.hgt.zip',
		'S09W043.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S09W043.hgt.zip',
		'S09W044.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S09W044.hgt.zip',
		'S09W045.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S09W045.hgt.zip',
		'S09W046.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S09W046.hgt.zip',
		'S09W047.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S09W047.hgt.zip',
		'S09W048.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S09W048.hgt.zip',
		'S09W049.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S09W049.hgt.zip',
		'S09W050.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S09W050.hgt.zip',
		'S09W051.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S09W051.hgt.zip',
		'S09W052.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S09W052.hgt.zip',
		'S09W053.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S09W053.hgt.zip',
		'S09W054.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S09W054.hgt.zip',
		'S09W055.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S09W055.hgt.zip',
		'S09W056.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S09W056.hgt.zip',
		'S09W057.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S09W057.hgt.zip',
		'S09W058.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S09W058.hgt.zip',
		'S09W059.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S09W059.hgt.zip',
		'S09W060.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S09W060.hgt.zip',
		'S09W061.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S09W061.hgt.zip',
		'S09W062.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S09W062.hgt.zip',
		'S09W063.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S09W063.hgt.zip',
		'S09W064.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S09W064.hgt.zip',
		'S09W065.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S09W065.hgt.zip',
		'S09W066.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S09W066.hgt.zip',
		'S09W067.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S09W067.hgt.zip',
		'S09W068.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S09W068.hgt.zip',
		'S09W069.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S09W069.hgt.zip',
		'S09W070.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S09W070.hgt.zip',
		'S09W071.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S09W071.hgt.zip',
		'S09W072.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S09W072.hgt.zip',
		'S09W073.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S09W073.hgt.zip',
		'S09W074.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S09W074.hgt.zip',
		'S09W075.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S09W075.hgt.zip',
		'S09W076.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S09W076.hgt.zip',
		'S09W077.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S09W077.hgt.zip',
		'S09W078.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S09W078.hgt.zip',
		'S09W079.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S09W079.hgt.zip',
		'S09W080.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S09W080.hgt.zip',
		'S09W140.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S09W140.hgt.zip',
		'S09W141.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S09W141.hgt.zip',
		'S09W158.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S09W158.hgt.zip',
		'S09W159.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S09W159.hgt.zip',
		'S09W173.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S09W173.hgt.zip',
		'S10E012.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S10E012.hgt.zip',
		'S10E013.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S10E013.hgt.zip',
		'S10E014.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S10E014.hgt.zip',
		'S10E015.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S10E015.hgt.zip',
		'S10E016.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S10E016.hgt.zip',
		'S10E017.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S10E017.hgt.zip',
		'S10E018.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S10E018.hgt.zip',
		'S10E019.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S10E019.hgt.zip',
		'S10E020.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S10E020.hgt.zip',
		'S10E021.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S10E021.hgt.zip',
		'S10E022.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S10E022.hgt.zip',
		'S10E023.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S10E023.hgt.zip',
		'S10E024.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S10E024.hgt.zip',
		'S10E025.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S10E025.hgt.zip',
		'S10E026.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S10E026.hgt.zip',
		'S10E027.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S10E027.hgt.zip',
		'S10E028.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S10E028.hgt.zip',
		'S10E029.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S10E029.hgt.zip',
		'S10E030.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S10E030.hgt.zip',
		'S10E031.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S10E031.hgt.zip',
		'S10E032.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S10E032.hgt.zip',
		'S10E033.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S10E033.hgt.zip',
		'S10E034.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S10E034.hgt.zip',
		'S10E035.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S10E035.hgt.zip',
		'S10E036.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S10E036.hgt.zip',
		'S10E037.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S10E037.hgt.zip',
		'S10E038.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S10E038.hgt.zip',
		'S10E039.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S10E039.hgt.zip',
		'S10E046.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S10E046.hgt.zip',
		'S10E047.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S10E047.hgt.zip',
		'S10E050.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S10E050.hgt.zip',
		'S10E051.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S10E051.hgt.zip',
		'S10E116.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S10E116.hgt.zip',
		'S10E117.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S10E117.hgt.zip',
		'S10E118.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S10E118.hgt.zip',
		'S10E119.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S10E119.hgt.zip',
		'S10E120.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S10E120.hgt.zip',
		'S10E123.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S10E123.hgt.zip',
		'S10E124.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S10E124.hgt.zip',
		'S10E125.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S10E125.hgt.zip',
		'S10E126.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S10E126.hgt.zip',
		'S10E140.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S10E140.hgt.zip',
		'S10E141.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S10E141.hgt.zip',
		'S10E142.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S10E142.hgt.zip',
		'S10E143.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S10E143.hgt.zip',
		'S10E144.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S10E144.hgt.zip',
		'S10E146.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S10E146.hgt.zip',
		'S10E147.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S10E147.hgt.zip',
		'S10E148.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S10E148.hgt.zip',
		'S10E149.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S10E149.hgt.zip',
		'S10E150.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S10E150.hgt.zip',
		'S10E151.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S10E151.hgt.zip',
		'S10E152.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S10E152.hgt.zip',
		'S10E153.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S10E153.hgt.zip',
		'S10E158.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S10E158.hgt.zip',
		'S10E159.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S10E159.hgt.zip',
		'S10E160.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S10E160.hgt.zip',
		'S10E161.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S10E161.hgt.zip',
		'S10E167.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S10E167.hgt.zip',
		'S10E179.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S10E179.hgt.zip',
		'S10W036.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S10W036.hgt.zip',
		'S10W037.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S10W037.hgt.zip',
		'S10W038.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S10W038.hgt.zip',
		'S10W039.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S10W039.hgt.zip',
		'S10W040.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S10W040.hgt.zip',
		'S10W041.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S10W041.hgt.zip',
		'S10W042.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S10W042.hgt.zip',
		'S10W043.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S10W043.hgt.zip',
		'S10W044.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S10W044.hgt.zip',
		'S10W045.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S10W045.hgt.zip',
		'S10W046.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S10W046.hgt.zip',
		'S10W047.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S10W047.hgt.zip',
		'S10W048.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S10W048.hgt.zip',
		'S10W049.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S10W049.hgt.zip',
		'S10W050.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S10W050.hgt.zip',
		'S10W051.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S10W051.hgt.zip',
		'S10W052.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S10W052.hgt.zip',
		'S10W053.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S10W053.hgt.zip',
		'S10W054.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S10W054.hgt.zip',
		'S10W055.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S10W055.hgt.zip',
		'S10W056.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S10W056.hgt.zip',
		'S10W057.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S10W057.hgt.zip',
		'S10W058.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S10W058.hgt.zip',
		'S10W059.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S10W059.hgt.zip',
		'S10W060.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S10W060.hgt.zip',
		'S10W061.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S10W061.hgt.zip',
		'S10W062.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S10W062.hgt.zip',
		'S10W063.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S10W063.hgt.zip',
		'S10W064.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S10W064.hgt.zip',
		'S10W065.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S10W065.hgt.zip',
		'S10W066.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S10W066.hgt.zip',
		'S10W067.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S10W067.hgt.zip',
		'S10W068.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S10W068.hgt.zip',
		'S10W069.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S10W069.hgt.zip',
		'S10W070.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S10W070.hgt.zip',
		'S10W071.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S10W071.hgt.zip',
		'S10W072.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S10W072.hgt.zip',
		'S10W073.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S10W073.hgt.zip',
		'S10W074.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S10W074.hgt.zip',
		'S10W075.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S10W075.hgt.zip',
		'S10W076.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S10W076.hgt.zip',
		'S10W077.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S10W077.hgt.zip',
		'S10W078.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S10W078.hgt.zip',
		'S10W079.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S10W079.hgt.zip',
		'S10W139.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S10W139.hgt.zip',
		'S10W140.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S10W140.hgt.zip',
		'S10W141.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S10W141.hgt.zip',
		'S10W151.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S10W151.hgt.zip',
		'S10W158.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S10W158.hgt.zip',
		'S10W159.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S10W159.hgt.zip',
		'S10W162.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S10W162.hgt.zip',
		'S10W172.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S10W172.hgt.zip',
		'S11E013.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S11E013.hgt.zip',
		'S11E014.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S11E014.hgt.zip',
		'S11E015.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S11E015.hgt.zip',
		'S11E016.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S11E016.hgt.zip',
		'S11E017.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S11E017.hgt.zip',
		'S11E018.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S11E018.hgt.zip',
		'S11E019.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S11E019.hgt.zip',
		'S11E020.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S11E020.hgt.zip',
		'S11E021.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S11E021.hgt.zip',
		'S11E022.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S11E022.hgt.zip',
		'S11E023.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S11E023.hgt.zip',
		'S11E024.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S11E024.hgt.zip',
		'S11E025.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S11E025.hgt.zip',
		'S11E026.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S11E026.hgt.zip',
		'S11E027.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S11E027.hgt.zip',
		'S11E028.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S11E028.hgt.zip',
		'S11E029.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S11E029.hgt.zip',
		'S11E030.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S11E030.hgt.zip',
		'S11E031.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S11E031.hgt.zip',
		'S11E032.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S11E032.hgt.zip',
		'S11E033.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S11E033.hgt.zip',
		'S11E034.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S11E034.hgt.zip',
		'S11E035.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S11E035.hgt.zip',
		'S11E036.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S11E036.hgt.zip',
		'S11E037.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S11E037.hgt.zip',
		'S11E038.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S11E038.hgt.zip',
		'S11E039.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S11E039.hgt.zip',
		'S11E040.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S11E040.hgt.zip',
		'S11E047.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S11E047.hgt.zip',
		'S11E051.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S11E051.hgt.zip',
		'S11E056.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S11E056.hgt.zip',
		'S11E105.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S11E105.hgt.zip',
		'S11E119.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S11E119.hgt.zip',
		'S11E120.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S11E120.hgt.zip',
		'S11E121.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S11E121.hgt.zip',
		'S11E122.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S11E122.hgt.zip',
		'S11E123.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S11E123.hgt.zip',
		'S11E124.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S11E124.hgt.zip',
		'S11E132.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S11E132.hgt.zip',
		'S11E133.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S11E133.hgt.zip',
		'S11E141.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S11E141.hgt.zip',
		'S11E142.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S11E142.hgt.zip',
		'S11E143.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S11E143.hgt.zip',
		'S11E147.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S11E147.hgt.zip',
		'S11E148.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S11E148.hgt.zip',
		'S11E149.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S11E149.hgt.zip',
		'S11E150.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S11E150.hgt.zip',
		'S11E151.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S11E151.hgt.zip',
		'S11E152.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S11E152.hgt.zip',
		'S11E153.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S11E153.hgt.zip',
		'S11E161.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S11E161.hgt.zip',
		'S11E162.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S11E162.hgt.zip',
		'S11E165.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S11E165.hgt.zip',
		'S11E166.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S11E166.hgt.zip',
		'S11E179.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S11E179.hgt.zip',
		'S11W037.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S11W037.hgt.zip',
		'S11W038.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S11W038.hgt.zip',
		'S11W039.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S11W039.hgt.zip',
		'S11W040.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S11W040.hgt.zip',
		'S11W041.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S11W041.hgt.zip',
		'S11W042.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S11W042.hgt.zip',
		'S11W043.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S11W043.hgt.zip',
		'S11W044.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S11W044.hgt.zip',
		'S11W045.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S11W045.hgt.zip',
		'S11W046.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S11W046.hgt.zip',
		'S11W047.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S11W047.hgt.zip',
		'S11W048.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S11W048.hgt.zip',
		'S11W049.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S11W049.hgt.zip',
		'S11W050.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S11W050.hgt.zip',
		'S11W051.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S11W051.hgt.zip',
		'S11W052.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S11W052.hgt.zip',
		'S11W053.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S11W053.hgt.zip',
		'S11W054.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S11W054.hgt.zip',
		'S11W055.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S11W055.hgt.zip',
		'S11W056.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S11W056.hgt.zip',
		'S11W057.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S11W057.hgt.zip',
		'S11W058.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S11W058.hgt.zip',
		'S11W059.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S11W059.hgt.zip',
		'S11W060.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S11W060.hgt.zip',
		'S11W061.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S11W061.hgt.zip',
		'S11W062.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S11W062.hgt.zip',
		'S11W063.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S11W063.hgt.zip',
		'S11W064.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S11W064.hgt.zip',
		'S11W065.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S11W065.hgt.zip',
		'S11W066.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S11W066.hgt.zip',
		'S11W067.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S11W067.hgt.zip',
		'S11W068.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S11W068.hgt.zip',
		'S11W069.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S11W069.hgt.zip',
		'S11W070.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S11W070.hgt.zip',
		'S11W071.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S11W071.hgt.zip',
		'S11W072.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S11W072.hgt.zip',
		'S11W073.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S11W073.hgt.zip',
		'S11W074.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S11W074.hgt.zip',
		'S11W075.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S11W075.hgt.zip',
		'S11W076.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S11W076.hgt.zip',
		'S11W077.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S11W077.hgt.zip',
		'S11W078.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S11W078.hgt.zip',
		'S11W079.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S11W079.hgt.zip',
		'S11W140.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S11W140.hgt.zip',
		'S11W151.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S11W151.hgt.zip',
		'S11W161.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S11W161.hgt.zip',
		'S11W162.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S11W162.hgt.zip',
		'S11W166.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S11W166.hgt.zip',
		'S12E013.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S12E013.hgt.zip',
		'S12E014.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S12E014.hgt.zip',
		'S12E015.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S12E015.hgt.zip',
		'S12E016.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S12E016.hgt.zip',
		'S12E017.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S12E017.hgt.zip',
		'S12E018.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S12E018.hgt.zip',
		'S12E019.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S12E019.hgt.zip',
		'S12E020.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S12E020.hgt.zip',
		'S12E021.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S12E021.hgt.zip',
		'S12E022.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S12E022.hgt.zip',
		'S12E023.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S12E023.hgt.zip',
		'S12E024.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S12E024.hgt.zip',
		'S12E025.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S12E025.hgt.zip',
		'S12E026.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S12E026.hgt.zip',
		'S12E027.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S12E027.hgt.zip',
		'S12E028.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S12E028.hgt.zip',
		'S12E029.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S12E029.hgt.zip',
		'S12E030.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S12E030.hgt.zip',
		'S12E031.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S12E031.hgt.zip',
		'S12E032.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S12E032.hgt.zip',
		'S12E033.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S12E033.hgt.zip',
		'S12E034.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S12E034.hgt.zip',
		'S12E035.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S12E035.hgt.zip',
		'S12E036.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S12E036.hgt.zip',
		'S12E037.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S12E037.hgt.zip',
		'S12E038.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S12E038.hgt.zip',
		'S12E039.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S12E039.hgt.zip',
		'S12E040.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S12E040.hgt.zip',
		'S12E043.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S12E043.hgt.zip',
		'S12E047.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S12E047.hgt.zip',
		'S12E049.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S12E049.hgt.zip',
		'S12E096.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S12E096.hgt.zip',
		'S12E122.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S12E122.hgt.zip',
		'S12E130.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S12E130.hgt.zip',
		'S12E131.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S12E131.hgt.zip',
		'S12E132.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S12E132.hgt.zip',
		'S12E133.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S12E133.hgt.zip',
		'S12E134.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S12E134.hgt.zip',
		'S12E135.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S12E135.hgt.zip',
		'S12E136.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S12E136.hgt.zip',
		'S12E141.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S12E141.hgt.zip',
		'S12E142.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S12E142.hgt.zip',
		'S12E143.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S12E143.hgt.zip',
		'S12E144.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S12E144.hgt.zip',
		'S12E151.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S12E151.hgt.zip',
		'S12E152.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S12E152.hgt.zip',
		'S12E153.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S12E153.hgt.zip',
		'S12E154.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S12E154.hgt.zip',
		'S12E159.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S12E159.hgt.zip',
		'S12E160.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S12E160.hgt.zip',
		'S12E166.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S12E166.hgt.zip',
		'S12E169.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S12E169.hgt.zip',
		'S12E170.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S12E170.hgt.zip',
		'S12W038.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S12W038.hgt.zip',
		'S12W039.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S12W039.hgt.zip',
		'S12W040.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S12W040.hgt.zip',
		'S12W041.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S12W041.hgt.zip',
		'S12W042.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S12W042.hgt.zip',
		'S12W043.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S12W043.hgt.zip',
		'S12W044.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S12W044.hgt.zip',
		'S12W045.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S12W045.hgt.zip',
		'S12W046.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S12W046.hgt.zip',
		'S12W047.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S12W047.hgt.zip',
		'S12W048.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S12W048.hgt.zip',
		'S12W049.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S12W049.hgt.zip',
		'S12W050.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S12W050.hgt.zip',
		'S12W051.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S12W051.hgt.zip',
		'S12W052.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S12W052.hgt.zip',
		'S12W053.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S12W053.hgt.zip',
		'S12W054.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S12W054.hgt.zip',
		'S12W055.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S12W055.hgt.zip',
		'S12W056.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S12W056.hgt.zip',
		'S12W057.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S12W057.hgt.zip',
		'S12W058.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S12W058.hgt.zip',
		'S12W059.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S12W059.hgt.zip',
		'S12W060.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S12W060.hgt.zip',
		'S12W061.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S12W061.hgt.zip',
		'S12W062.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S12W062.hgt.zip',
		'S12W063.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S12W063.hgt.zip',
		'S12W064.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S12W064.hgt.zip',
		'S12W065.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S12W065.hgt.zip',
		'S12W066.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S12W066.hgt.zip',
		'S12W067.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S12W067.hgt.zip',
		'S12W068.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S12W068.hgt.zip',
		'S12W069.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S12W069.hgt.zip',
		'S12W070.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S12W070.hgt.zip',
		'S12W071.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S12W071.hgt.zip',
		'S12W072.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S12W072.hgt.zip',
		'S12W073.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S12W073.hgt.zip',
		'S12W074.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S12W074.hgt.zip',
		'S12W075.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S12W075.hgt.zip',
		'S12W076.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S12W076.hgt.zip',
		'S12W077.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S12W077.hgt.zip',
		'S12W078.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S12W078.hgt.zip',
		'S12W152.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S12W152.hgt.zip',
		'S12W166.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S12W166.hgt.zip',
		'S12W172.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S12W172.hgt.zip',
		'S13E012.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S13E012.hgt.zip',
		'S13E013.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S13E013.hgt.zip',
		'S13E014.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S13E014.hgt.zip',
		'S13E015.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S13E015.hgt.zip',
		'S13E016.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S13E016.hgt.zip',
		'S13E017.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S13E017.hgt.zip',
		'S13E018.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S13E018.hgt.zip',
		'S13E019.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S13E019.hgt.zip',
		'S13E020.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S13E020.hgt.zip',
		'S13E021.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S13E021.hgt.zip',
		'S13E022.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S13E022.hgt.zip',
		'S13E023.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S13E023.hgt.zip',
		'S13E024.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S13E024.hgt.zip',
		'S13E025.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S13E025.hgt.zip',
		'S13E026.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S13E026.hgt.zip',
		'S13E027.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S13E027.hgt.zip',
		'S13E028.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S13E028.hgt.zip',
		'S13E029.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S13E029.hgt.zip',
		'S13E030.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S13E030.hgt.zip',
		'S13E031.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S13E031.hgt.zip',
		'S13E032.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S13E032.hgt.zip',
		'S13E033.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S13E033.hgt.zip',
		'S13E034.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S13E034.hgt.zip',
		'S13E035.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S13E035.hgt.zip',
		'S13E036.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S13E036.hgt.zip',
		'S13E037.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S13E037.hgt.zip',
		'S13E038.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S13E038.hgt.zip',
		'S13E039.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S13E039.hgt.zip',
		'S13E040.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S13E040.hgt.zip',
		'S13E043.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S13E043.hgt.zip',
		'S13E044.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S13E044.hgt.zip',
		'S13E045.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S13E045.hgt.zip',
		'S13E048.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S13E048.hgt.zip',
		'S13E049.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S13E049.hgt.zip',
		'S13E096.hgt': 'https://srtm.kurviger.de/SRTM3//Eurasia//S13E096.hgt.zip',
		'S13E122.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S13E122.hgt.zip',
		'S13E123.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S13E123.hgt.zip',
		'S13E130.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S13E130.hgt.zip',
		'S13E131.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S13E131.hgt.zip',
		'S13E132.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S13E132.hgt.zip',
		'S13E133.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S13E133.hgt.zip',
		'S13E134.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S13E134.hgt.zip',
		'S13E135.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S13E135.hgt.zip',
		'S13E136.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S13E136.hgt.zip',
		'S13E141.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S13E141.hgt.zip',
		'S13E142.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S13E142.hgt.zip',
		'S13E143.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S13E143.hgt.zip',
		'S13E168.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S13E168.hgt.zip',
		'S13E176.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S13E176.hgt.zip',
		'S13E177.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S13E177.hgt.zip',
		'S13W038.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S13W038.hgt.zip',
		'S13W039.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S13W039.hgt.zip',
		'S13W040.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S13W040.hgt.zip',
		'S13W041.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S13W041.hgt.zip',
		'S13W042.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S13W042.hgt.zip',
		'S13W043.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S13W043.hgt.zip',
		'S13W044.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S13W044.hgt.zip',
		'S13W045.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S13W045.hgt.zip',
		'S13W046.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S13W046.hgt.zip',
		'S13W047.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S13W047.hgt.zip',
		'S13W048.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S13W048.hgt.zip',
		'S13W049.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S13W049.hgt.zip',
		'S13W050.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S13W050.hgt.zip',
		'S13W051.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S13W051.hgt.zip',
		'S13W052.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S13W052.hgt.zip',
		'S13W053.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S13W053.hgt.zip',
		'S13W054.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S13W054.hgt.zip',
		'S13W055.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S13W055.hgt.zip',
		'S13W056.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S13W056.hgt.zip',
		'S13W057.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S13W057.hgt.zip',
		'S13W058.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S13W058.hgt.zip',
		'S13W059.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S13W059.hgt.zip',
		'S13W060.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S13W060.hgt.zip',
		'S13W061.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S13W061.hgt.zip',
		'S13W062.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S13W062.hgt.zip',
		'S13W063.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S13W063.hgt.zip',
		'S13W064.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S13W064.hgt.zip',
		'S13W065.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S13W065.hgt.zip',
		'S13W066.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S13W066.hgt.zip',
		'S13W067.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S13W067.hgt.zip',
		'S13W068.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S13W068.hgt.zip',
		'S13W069.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S13W069.hgt.zip',
		'S13W070.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S13W070.hgt.zip',
		'S13W071.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S13W071.hgt.zip',
		'S13W072.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S13W072.hgt.zip',
		'S13W073.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S13W073.hgt.zip',
		'S13W074.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S13W074.hgt.zip',
		'S13W075.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S13W075.hgt.zip',
		'S13W076.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S13W076.hgt.zip',
		'S13W077.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S13W077.hgt.zip',
		'S13W078.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S13W078.hgt.zip',
		'S14E012.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S14E012.hgt.zip',
		'S14E013.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S14E013.hgt.zip',
		'S14E014.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S14E014.hgt.zip',
		'S14E015.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S14E015.hgt.zip',
		'S14E016.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S14E016.hgt.zip',
		'S14E017.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S14E017.hgt.zip',
		'S14E018.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S14E018.hgt.zip',
		'S14E019.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S14E019.hgt.zip',
		'S14E020.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S14E020.hgt.zip',
		'S14E021.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S14E021.hgt.zip',
		'S14E022.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S14E022.hgt.zip',
		'S14E023.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S14E023.hgt.zip',
		'S14E024.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S14E024.hgt.zip',
		'S14E025.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S14E025.hgt.zip',
		'S14E026.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S14E026.hgt.zip',
		'S14E027.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S14E027.hgt.zip',
		'S14E028.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S14E028.hgt.zip',
		'S14E029.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S14E029.hgt.zip',
		'S14E030.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S14E030.hgt.zip',
		'S14E031.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S14E031.hgt.zip',
		'S14E032.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S14E032.hgt.zip',
		'S14E033.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S14E033.hgt.zip',
		'S14E034.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S14E034.hgt.zip',
		'S14E035.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S14E035.hgt.zip',
		'S14E036.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S14E036.hgt.zip',
		'S14E037.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S14E037.hgt.zip',
		'S14E038.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S14E038.hgt.zip',
		'S14E039.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S14E039.hgt.zip',
		'S14E040.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S14E040.hgt.zip',
		'S14E045.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S14E045.hgt.zip',
		'S14E047.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S14E047.hgt.zip',
		'S14E048.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S14E048.hgt.zip',
		'S14E049.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S14E049.hgt.zip',
		'S14E050.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S14E050.hgt.zip',
		'S14E125.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S14E125.hgt.zip',
		'S14E126.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S14E126.hgt.zip',
		'S14E127.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S14E127.hgt.zip',
		'S14E129.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S14E129.hgt.zip',
		'S14E130.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S14E130.hgt.zip',
		'S14E131.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S14E131.hgt.zip',
		'S14E132.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S14E132.hgt.zip',
		'S14E133.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S14E133.hgt.zip',
		'S14E134.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S14E134.hgt.zip',
		'S14E135.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S14E135.hgt.zip',
		'S14E136.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S14E136.hgt.zip',
		'S14E141.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S14E141.hgt.zip',
		'S14E142.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S14E142.hgt.zip',
		'S14E143.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S14E143.hgt.zip',
		'S14E144.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S14E144.hgt.zip',
		'S14E166.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S14E166.hgt.zip',
		'S14E167.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S14E167.hgt.zip',
		'S14W039.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S14W039.hgt.zip',
		'S14W040.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S14W040.hgt.zip',
		'S14W041.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S14W041.hgt.zip',
		'S14W042.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S14W042.hgt.zip',
		'S14W043.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S14W043.hgt.zip',
		'S14W044.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S14W044.hgt.zip',
		'S14W045.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S14W045.hgt.zip',
		'S14W046.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S14W046.hgt.zip',
		'S14W047.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S14W047.hgt.zip',
		'S14W048.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S14W048.hgt.zip',
		'S14W049.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S14W049.hgt.zip',
		'S14W050.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S14W050.hgt.zip',
		'S14W051.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S14W051.hgt.zip',
		'S14W052.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S14W052.hgt.zip',
		'S14W053.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S14W053.hgt.zip',
		'S14W054.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S14W054.hgt.zip',
		'S14W055.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S14W055.hgt.zip',
		'S14W056.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S14W056.hgt.zip',
		'S14W057.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S14W057.hgt.zip',
		'S14W058.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S14W058.hgt.zip',
		'S14W059.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S14W059.hgt.zip',
		'S14W060.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S14W060.hgt.zip',
		'S14W061.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S14W061.hgt.zip',
		'S14W062.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S14W062.hgt.zip',
		'S14W063.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S14W063.hgt.zip',
		'S14W064.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S14W064.hgt.zip',
		'S14W065.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S14W065.hgt.zip',
		'S14W066.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S14W066.hgt.zip',
		'S14W067.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S14W067.hgt.zip',
		'S14W068.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S14W068.hgt.zip',
		'S14W069.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S14W069.hgt.zip',
		'S14W070.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S14W070.hgt.zip',
		'S14W071.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S14W071.hgt.zip',
		'S14W072.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S14W072.hgt.zip',
		'S14W073.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S14W073.hgt.zip',
		'S14W074.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S14W074.hgt.zip',
		'S14W075.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S14W075.hgt.zip',
		'S14W076.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S14W076.hgt.zip',
		'S14W077.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S14W077.hgt.zip',
		'S14W164.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S14W164.hgt.zip',
		'S14W172.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S14W172.hgt.zip',
		'S14W173.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S14W173.hgt.zip',
		'S14W177.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S14W177.hgt.zip',
		'S15E012.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S15E012.hgt.zip',
		'S15E013.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S15E013.hgt.zip',
		'S15E014.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S15E014.hgt.zip',
		'S15E015.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S15E015.hgt.zip',
		'S15E016.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S15E016.hgt.zip',
		'S15E017.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S15E017.hgt.zip',
		'S15E018.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S15E018.hgt.zip',
		'S15E019.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S15E019.hgt.zip',
		'S15E020.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S15E020.hgt.zip',
		'S15E021.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S15E021.hgt.zip',
		'S15E022.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S15E022.hgt.zip',
		'S15E023.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S15E023.hgt.zip',
		'S15E024.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S15E024.hgt.zip',
		'S15E025.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S15E025.hgt.zip',
		'S15E026.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S15E026.hgt.zip',
		'S15E027.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S15E027.hgt.zip',
		'S15E028.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S15E028.hgt.zip',
		'S15E029.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S15E029.hgt.zip',
		'S15E030.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S15E030.hgt.zip',
		'S15E031.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S15E031.hgt.zip',
		'S15E032.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S15E032.hgt.zip',
		'S15E033.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S15E033.hgt.zip',
		'S15E034.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S15E034.hgt.zip',
		'S15E035.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S15E035.hgt.zip',
		'S15E036.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S15E036.hgt.zip',
		'S15E037.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S15E037.hgt.zip',
		'S15E038.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S15E038.hgt.zip',
		'S15E039.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S15E039.hgt.zip',
		'S15E040.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S15E040.hgt.zip',
		'S15E047.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S15E047.hgt.zip',
		'S15E048.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S15E048.hgt.zip',
		'S15E049.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S15E049.hgt.zip',
		'S15E050.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S15E050.hgt.zip',
		'S15E121.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S15E121.hgt.zip',
		'S15E123.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S15E123.hgt.zip',
		'S15E124.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S15E124.hgt.zip',
		'S15E125.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S15E125.hgt.zip',
		'S15E126.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S15E126.hgt.zip',
		'S15E127.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S15E127.hgt.zip',
		'S15E128.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S15E128.hgt.zip',
		'S15E129.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S15E129.hgt.zip',
		'S15E130.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S15E130.hgt.zip',
		'S15E131.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S15E131.hgt.zip',
		'S15E132.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S15E132.hgt.zip',
		'S15E133.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S15E133.hgt.zip',
		'S15E134.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S15E134.hgt.zip',
		'S15E135.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S15E135.hgt.zip',
		'S15E136.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S15E136.hgt.zip',
		'S15E141.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S15E141.hgt.zip',
		'S15E142.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S15E142.hgt.zip',
		'S15E143.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S15E143.hgt.zip',
		'S15E144.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S15E144.hgt.zip',
		'S15E145.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S15E145.hgt.zip',
		'S15E166.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S15E166.hgt.zip',
		'S15E167.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S15E167.hgt.zip',
		'S15E168.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S15E168.hgt.zip',
		'S15W039.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S15W039.hgt.zip',
		'S15W040.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S15W040.hgt.zip',
		'S15W041.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S15W041.hgt.zip',
		'S15W042.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S15W042.hgt.zip',
		'S15W043.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S15W043.hgt.zip',
		'S15W044.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S15W044.hgt.zip',
		'S15W045.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S15W045.hgt.zip',
		'S15W046.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S15W046.hgt.zip',
		'S15W047.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S15W047.hgt.zip',
		'S15W048.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S15W048.hgt.zip',
		'S15W049.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S15W049.hgt.zip',
		'S15W050.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S15W050.hgt.zip',
		'S15W051.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S15W051.hgt.zip',
		'S15W052.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S15W052.hgt.zip',
		'S15W053.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S15W053.hgt.zip',
		'S15W054.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S15W054.hgt.zip',
		'S15W055.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S15W055.hgt.zip',
		'S15W056.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S15W056.hgt.zip',
		'S15W057.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S15W057.hgt.zip',
		'S15W058.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S15W058.hgt.zip',
		'S15W059.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S15W059.hgt.zip',
		'S15W060.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S15W060.hgt.zip',
		'S15W061.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S15W061.hgt.zip',
		'S15W062.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S15W062.hgt.zip',
		'S15W063.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S15W063.hgt.zip',
		'S15W064.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S15W064.hgt.zip',
		'S15W065.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S15W065.hgt.zip',
		'S15W066.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S15W066.hgt.zip',
		'S15W067.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S15W067.hgt.zip',
		'S15W068.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S15W068.hgt.zip',
		'S15W069.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S15W069.hgt.zip',
		'S15W070.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S15W070.hgt.zip',
		'S15W071.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S15W071.hgt.zip',
		'S15W072.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S15W072.hgt.zip',
		'S15W073.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S15W073.hgt.zip',
		'S15W074.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S15W074.hgt.zip',
		'S15W075.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S15W075.hgt.zip',
		'S15W076.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S15W076.hgt.zip',
		'S15W077.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S15W077.hgt.zip',
		'S15W139.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S15W139.hgt.zip',
		'S15W142.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S15W142.hgt.zip',
		'S15W145.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S15W145.hgt.zip',
		'S15W146.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S15W146.hgt.zip',
		'S15W147.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S15W147.hgt.zip',
		'S15W148.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S15W148.hgt.zip',
		'S15W149.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S15W149.hgt.zip',
		'S15W169.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S15W169.hgt.zip',
		'S15W170.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S15W170.hgt.zip',
		'S15W171.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S15W171.hgt.zip',
		'S15W172.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S15W172.hgt.zip',
		'S15W178.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S15W178.hgt.zip',
		'S15W179.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S15W179.hgt.zip',
		'S16E011.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S16E011.hgt.zip',
		'S16E012.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S16E012.hgt.zip',
		'S16E013.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S16E013.hgt.zip',
		'S16E014.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S16E014.hgt.zip',
		'S16E015.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S16E015.hgt.zip',
		'S16E016.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S16E016.hgt.zip',
		'S16E017.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S16E017.hgt.zip',
		'S16E018.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S16E018.hgt.zip',
		'S16E019.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S16E019.hgt.zip',
		'S16E020.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S16E020.hgt.zip',
		'S16E021.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S16E021.hgt.zip',
		'S16E022.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S16E022.hgt.zip',
		'S16E023.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S16E023.hgt.zip',
		'S16E024.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S16E024.hgt.zip',
		'S16E025.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S16E025.hgt.zip',
		'S16E026.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S16E026.hgt.zip',
		'S16E027.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S16E027.hgt.zip',
		'S16E028.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S16E028.hgt.zip',
		'S16E029.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S16E029.hgt.zip',
		'S16E030.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S16E030.hgt.zip',
		'S16E031.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S16E031.hgt.zip',
		'S16E032.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S16E032.hgt.zip',
		'S16E033.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S16E033.hgt.zip',
		'S16E034.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S16E034.hgt.zip',
		'S16E035.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S16E035.hgt.zip',
		'S16E036.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S16E036.hgt.zip',
		'S16E037.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S16E037.hgt.zip',
		'S16E038.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S16E038.hgt.zip',
		'S16E039.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S16E039.hgt.zip',
		'S16E040.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S16E040.hgt.zip',
		'S16E045.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S16E045.hgt.zip',
		'S16E046.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S16E046.hgt.zip',
		'S16E047.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S16E047.hgt.zip',
		'S16E048.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S16E048.hgt.zip',
		'S16E049.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S16E049.hgt.zip',
		'S16E050.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S16E050.hgt.zip',
		'S16E054.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S16E054.hgt.zip',
		'S16E123.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S16E123.hgt.zip',
		'S16E124.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S16E124.hgt.zip',
		'S16E125.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S16E125.hgt.zip',
		'S16E126.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S16E126.hgt.zip',
		'S16E127.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S16E127.hgt.zip',
		'S16E128.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S16E128.hgt.zip',
		'S16E129.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S16E129.hgt.zip',
		'S16E130.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S16E130.hgt.zip',
		'S16E131.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S16E131.hgt.zip',
		'S16E132.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S16E132.hgt.zip',
		'S16E133.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S16E133.hgt.zip',
		'S16E134.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S16E134.hgt.zip',
		'S16E135.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S16E135.hgt.zip',
		'S16E136.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S16E136.hgt.zip',
		'S16E137.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S16E137.hgt.zip',
		'S16E141.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S16E141.hgt.zip',
		'S16E142.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S16E142.hgt.zip',
		'S16E143.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S16E143.hgt.zip',
		'S16E144.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S16E144.hgt.zip',
		'S16E145.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S16E145.hgt.zip',
		'S16E166.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S16E166.hgt.zip',
		'S16E167.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S16E167.hgt.zip',
		'S16E168.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S16E168.hgt.zip',
		'S16W006.hgt': 'https://srtm.kurviger.de/SRTM3//Islands//S16W006.hgt.zip',
		'S16W039.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S16W039.hgt.zip',
		'S16W040.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S16W040.hgt.zip',
		'S16W041.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S16W041.hgt.zip',
		'S16W042.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S16W042.hgt.zip',
		'S16W043.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S16W043.hgt.zip',
		'S16W044.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S16W044.hgt.zip',
		'S16W045.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S16W045.hgt.zip',
		'S16W046.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S16W046.hgt.zip',
		'S16W047.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S16W047.hgt.zip',
		'S16W048.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S16W048.hgt.zip',
		'S16W049.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S16W049.hgt.zip',
		'S16W050.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S16W050.hgt.zip',
		'S16W051.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S16W051.hgt.zip',
		'S16W052.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S16W052.hgt.zip',
		'S16W053.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S16W053.hgt.zip',
		'S16W054.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S16W054.hgt.zip',
		'S16W055.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S16W055.hgt.zip',
		'S16W056.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S16W056.hgt.zip',
		'S16W057.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S16W057.hgt.zip',
		'S16W058.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S16W058.hgt.zip',
		'S16W059.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S16W059.hgt.zip',
		'S16W060.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S16W060.hgt.zip',
		'S16W061.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S16W061.hgt.zip',
		'S16W062.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S16W062.hgt.zip',
		'S16W063.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S16W063.hgt.zip',
		'S16W064.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S16W064.hgt.zip',
		'S16W065.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S16W065.hgt.zip',
		'S16W066.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S16W066.hgt.zip',
		'S16W067.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S16W067.hgt.zip',
		'S16W068.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S16W068.hgt.zip',
		'S16W069.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S16W069.hgt.zip',
		'S16W070.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S16W070.hgt.zip',
		'S16W071.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S16W071.hgt.zip',
		'S16W072.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S16W072.hgt.zip',
		'S16W073.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S16W073.hgt.zip',
		'S16W074.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S16W074.hgt.zip',
		'S16W075.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S16W075.hgt.zip',
		'S16W076.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S16W076.hgt.zip',
		'S16W141.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S16W141.hgt.zip',
		'S16W143.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S16W143.hgt.zip',
		'S16W145.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S16W145.hgt.zip',
		'S16W146.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S16W146.hgt.zip',
		'S16W147.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S16W147.hgt.zip',
		'S16W148.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S16W148.hgt.zip',
		'S16W149.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S16W149.hgt.zip',
		'S16W155.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S16W155.hgt.zip',
		'S16W174.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S16W174.hgt.zip',
		'S16W176.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S16W176.hgt.zip',
		'S16W180.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S16W180.hgt.zip',
		'S17E011.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S17E011.hgt.zip',
		'S17E012.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S17E012.hgt.zip',
		'S17E013.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S17E013.hgt.zip',
		'S17E014.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S17E014.hgt.zip',
		'S17E015.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S17E015.hgt.zip',
		'S17E016.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S17E016.hgt.zip',
		'S17E017.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S17E017.hgt.zip',
		'S17E018.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S17E018.hgt.zip',
		'S17E019.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S17E019.hgt.zip',
		'S17E020.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S17E020.hgt.zip',
		'S17E021.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S17E021.hgt.zip',
		'S17E022.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S17E022.hgt.zip',
		'S17E023.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S17E023.hgt.zip',
		'S17E024.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S17E024.hgt.zip',
		'S17E025.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S17E025.hgt.zip',
		'S17E026.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S17E026.hgt.zip',
		'S17E027.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S17E027.hgt.zip',
		'S17E028.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S17E028.hgt.zip',
		'S17E029.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S17E029.hgt.zip',
		'S17E030.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S17E030.hgt.zip',
		'S17E031.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S17E031.hgt.zip',
		'S17E032.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S17E032.hgt.zip',
		'S17E033.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S17E033.hgt.zip',
		'S17E034.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S17E034.hgt.zip',
		'S17E035.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S17E035.hgt.zip',
		'S17E036.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S17E036.hgt.zip',
		'S17E037.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S17E037.hgt.zip',
		'S17E038.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S17E038.hgt.zip',
		'S17E039.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S17E039.hgt.zip',
		'S17E040.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S17E040.hgt.zip',
		'S17E043.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S17E043.hgt.zip',
		'S17E044.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S17E044.hgt.zip',
		'S17E045.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S17E045.hgt.zip',
		'S17E046.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S17E046.hgt.zip',
		'S17E047.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S17E047.hgt.zip',
		'S17E048.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S17E048.hgt.zip',
		'S17E049.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S17E049.hgt.zip',
		'S17E050.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S17E050.hgt.zip',
		'S17E059.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S17E059.hgt.zip',
		'S17E122.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S17E122.hgt.zip',
		'S17E123.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S17E123.hgt.zip',
		'S17E124.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S17E124.hgt.zip',
		'S17E125.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S17E125.hgt.zip',
		'S17E126.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S17E126.hgt.zip',
		'S17E127.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S17E127.hgt.zip',
		'S17E128.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S17E128.hgt.zip',
		'S17E129.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S17E129.hgt.zip',
		'S17E130.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S17E130.hgt.zip',
		'S17E131.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S17E131.hgt.zip',
		'S17E132.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S17E132.hgt.zip',
		'S17E133.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S17E133.hgt.zip',
		'S17E134.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S17E134.hgt.zip',
		'S17E135.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S17E135.hgt.zip',
		'S17E136.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S17E136.hgt.zip',
		'S17E137.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S17E137.hgt.zip',
		'S17E138.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S17E138.hgt.zip',
		'S17E139.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S17E139.hgt.zip',
		'S17E140.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S17E140.hgt.zip',
		'S17E141.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S17E141.hgt.zip',
		'S17E142.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S17E142.hgt.zip',
		'S17E143.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S17E143.hgt.zip',
		'S17E144.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S17E144.hgt.zip',
		'S17E145.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S17E145.hgt.zip',
		'S17E146.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S17E146.hgt.zip',
		'S17E149.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S17E149.hgt.zip',
		'S17E150.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S17E150.hgt.zip',
		'S17E167.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S17E167.hgt.zip',
		'S17E168.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S17E168.hgt.zip',
		'S17E177.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S17E177.hgt.zip',
		'S17E178.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S17E178.hgt.zip',
		'S17E179.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S17E179.hgt.zip',
		'S17W006.hgt': 'https://srtm.kurviger.de/SRTM3//Islands//S17W006.hgt.zip',
		'S17W039.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S17W039.hgt.zip',
		'S17W040.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S17W040.hgt.zip',
		'S17W041.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S17W041.hgt.zip',
		'S17W042.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S17W042.hgt.zip',
		'S17W043.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S17W043.hgt.zip',
		'S17W044.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S17W044.hgt.zip',
		'S17W045.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S17W045.hgt.zip',
		'S17W046.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S17W046.hgt.zip',
		'S17W047.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S17W047.hgt.zip',
		'S17W048.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S17W048.hgt.zip',
		'S17W049.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S17W049.hgt.zip',
		'S17W050.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S17W050.hgt.zip',
		'S17W051.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S17W051.hgt.zip',
		'S17W052.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S17W052.hgt.zip',
		'S17W053.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S17W053.hgt.zip',
		'S17W054.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S17W054.hgt.zip',
		'S17W055.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S17W055.hgt.zip',
		'S17W056.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S17W056.hgt.zip',
		'S17W057.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S17W057.hgt.zip',
		'S17W058.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S17W058.hgt.zip',
		'S17W059.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S17W059.hgt.zip',
		'S17W060.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S17W060.hgt.zip',
		'S17W061.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S17W061.hgt.zip',
		'S17W062.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S17W062.hgt.zip',
		'S17W063.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S17W063.hgt.zip',
		'S17W064.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S17W064.hgt.zip',
		'S17W065.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S17W065.hgt.zip',
		'S17W066.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S17W066.hgt.zip',
		'S17W067.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S17W067.hgt.zip',
		'S17W068.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S17W068.hgt.zip',
		'S17W069.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S17W069.hgt.zip',
		'S17W070.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S17W070.hgt.zip',
		'S17W071.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S17W071.hgt.zip',
		'S17W072.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S17W072.hgt.zip',
		'S17W073.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S17W073.hgt.zip',
		'S17W074.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S17W074.hgt.zip',
		'S17W075.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S17W075.hgt.zip',
		'S17W141.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S17W141.hgt.zip',
		'S17W142.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S17W142.hgt.zip',
		'S17W143.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S17W143.hgt.zip',
		'S17W144.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S17W144.hgt.zip',
		'S17W145.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S17W145.hgt.zip',
		'S17W146.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S17W146.hgt.zip',
		'S17W147.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S17W147.hgt.zip',
		'S17W150.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S17W150.hgt.zip',
		'S17W151.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S17W151.hgt.zip',
		'S17W152.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S17W152.hgt.zip',
		'S17W153.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S17W153.hgt.zip',
		'S17W154.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S17W154.hgt.zip',
		'S17W155.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S17W155.hgt.zip',
		'S17W180.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S17W180.hgt.zip',
		'S18E011.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S18E011.hgt.zip',
		'S18E012.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S18E012.hgt.zip',
		'S18E013.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S18E013.hgt.zip',
		'S18E014.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S18E014.hgt.zip',
		'S18E015.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S18E015.hgt.zip',
		'S18E016.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S18E016.hgt.zip',
		'S18E017.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S18E017.hgt.zip',
		'S18E018.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S18E018.hgt.zip',
		'S18E019.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S18E019.hgt.zip',
		'S18E020.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S18E020.hgt.zip',
		'S18E021.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S18E021.hgt.zip',
		'S18E022.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S18E022.hgt.zip',
		'S18E023.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S18E023.hgt.zip',
		'S18E024.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S18E024.hgt.zip',
		'S18E025.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S18E025.hgt.zip',
		'S18E026.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S18E026.hgt.zip',
		'S18E027.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S18E027.hgt.zip',
		'S18E028.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S18E028.hgt.zip',
		'S18E029.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S18E029.hgt.zip',
		'S18E030.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S18E030.hgt.zip',
		'S18E031.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S18E031.hgt.zip',
		'S18E032.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S18E032.hgt.zip',
		'S18E033.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S18E033.hgt.zip',
		'S18E034.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S18E034.hgt.zip',
		'S18E035.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S18E035.hgt.zip',
		'S18E036.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S18E036.hgt.zip',
		'S18E037.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S18E037.hgt.zip',
		'S18E038.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S18E038.hgt.zip',
		'S18E039.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S18E039.hgt.zip',
		'S18E042.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S18E042.hgt.zip',
		'S18E043.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S18E043.hgt.zip',
		'S18E044.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S18E044.hgt.zip',
		'S18E045.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S18E045.hgt.zip',
		'S18E046.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S18E046.hgt.zip',
		'S18E047.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S18E047.hgt.zip',
		'S18E048.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S18E048.hgt.zip',
		'S18E049.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S18E049.hgt.zip',
		'S18E118.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S18E118.hgt.zip',
		'S18E119.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S18E119.hgt.zip',
		'S18E122.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S18E122.hgt.zip',
		'S18E123.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S18E123.hgt.zip',
		'S18E124.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S18E124.hgt.zip',
		'S18E125.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S18E125.hgt.zip',
		'S18E126.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S18E126.hgt.zip',
		'S18E127.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S18E127.hgt.zip',
		'S18E128.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S18E128.hgt.zip',
		'S18E129.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S18E129.hgt.zip',
		'S18E130.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S18E130.hgt.zip',
		'S18E131.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S18E131.hgt.zip',
		'S18E132.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S18E132.hgt.zip',
		'S18E133.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S18E133.hgt.zip',
		'S18E134.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S18E134.hgt.zip',
		'S18E135.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S18E135.hgt.zip',
		'S18E136.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S18E136.hgt.zip',
		'S18E137.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S18E137.hgt.zip',
		'S18E138.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S18E138.hgt.zip',
		'S18E139.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S18E139.hgt.zip',
		'S18E140.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S18E140.hgt.zip',
		'S18E141.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S18E141.hgt.zip',
		'S18E142.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S18E142.hgt.zip',
		'S18E143.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S18E143.hgt.zip',
		'S18E144.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S18E144.hgt.zip',
		'S18E145.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S18E145.hgt.zip',
		'S18E146.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S18E146.hgt.zip',
		'S18E148.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S18E148.hgt.zip',
		'S18E155.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S18E155.hgt.zip',
		'S18E168.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S18E168.hgt.zip',
		'S18E176.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S18E176.hgt.zip',
		'S18E177.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S18E177.hgt.zip',
		'S18E178.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S18E178.hgt.zip',
		'S18E179.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S18E179.hgt.zip',
		'S18W039.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S18W039.hgt.zip',
		'S18W040.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S18W040.hgt.zip',
		'S18W041.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S18W041.hgt.zip',
		'S18W042.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S18W042.hgt.zip',
		'S18W043.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S18W043.hgt.zip',
		'S18W044.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S18W044.hgt.zip',
		'S18W045.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S18W045.hgt.zip',
		'S18W046.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S18W046.hgt.zip',
		'S18W047.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S18W047.hgt.zip',
		'S18W048.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S18W048.hgt.zip',
		'S18W049.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S18W049.hgt.zip',
		'S18W050.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S18W050.hgt.zip',
		'S18W051.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S18W051.hgt.zip',
		'S18W052.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S18W052.hgt.zip',
		'S18W053.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S18W053.hgt.zip',
		'S18W054.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S18W054.hgt.zip',
		'S18W055.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S18W055.hgt.zip',
		'S18W056.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S18W056.hgt.zip',
		'S18W057.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S18W057.hgt.zip',
		'S18W058.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S18W058.hgt.zip',
		'S18W059.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S18W059.hgt.zip',
		'S18W060.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S18W060.hgt.zip',
		'S18W061.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S18W061.hgt.zip',
		'S18W062.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S18W062.hgt.zip',
		'S18W063.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S18W063.hgt.zip',
		'S18W064.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S18W064.hgt.zip',
		'S18W065.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S18W065.hgt.zip',
		'S18W066.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S18W066.hgt.zip',
		'S18W067.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S18W067.hgt.zip',
		'S18W068.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S18W068.hgt.zip',
		'S18W069.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S18W069.hgt.zip',
		'S18W070.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S18W070.hgt.zip',
		'S18W071.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S18W071.hgt.zip',
		'S18W072.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S18W072.hgt.zip',
		'S18W073.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S18W073.hgt.zip',
		'S18W139.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S18W139.hgt.zip',
		'S18W141.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S18W141.hgt.zip',
		'S18W142.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S18W142.hgt.zip',
		'S18W143.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S18W143.hgt.zip',
		'S18W144.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S18W144.hgt.zip',
		'S18W145.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S18W145.hgt.zip',
		'S18W146.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S18W146.hgt.zip',
		'S18W149.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S18W149.hgt.zip',
		'S18W150.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S18W150.hgt.zip',
		'S18W151.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S18W151.hgt.zip',
		'S18W179.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S18W179.hgt.zip',
		'S18W180.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S18W180.hgt.zip',
		'S19E011.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S19E011.hgt.zip',
		'S19E012.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S19E012.hgt.zip',
		'S19E013.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S19E013.hgt.zip',
		'S19E014.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S19E014.hgt.zip',
		'S19E015.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S19E015.hgt.zip',
		'S19E016.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S19E016.hgt.zip',
		'S19E017.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S19E017.hgt.zip',
		'S19E018.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S19E018.hgt.zip',
		'S19E019.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S19E019.hgt.zip',
		'S19E020.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S19E020.hgt.zip',
		'S19E021.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S19E021.hgt.zip',
		'S19E022.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S19E022.hgt.zip',
		'S19E023.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S19E023.hgt.zip',
		'S19E024.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S19E024.hgt.zip',
		'S19E025.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S19E025.hgt.zip',
		'S19E026.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S19E026.hgt.zip',
		'S19E027.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S19E027.hgt.zip',
		'S19E028.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S19E028.hgt.zip',
		'S19E029.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S19E029.hgt.zip',
		'S19E030.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S19E030.hgt.zip',
		'S19E031.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S19E031.hgt.zip',
		'S19E032.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S19E032.hgt.zip',
		'S19E033.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S19E033.hgt.zip',
		'S19E034.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S19E034.hgt.zip',
		'S19E035.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S19E035.hgt.zip',
		'S19E036.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S19E036.hgt.zip',
		'S19E037.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S19E037.hgt.zip',
		'S19E043.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S19E043.hgt.zip',
		'S19E044.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S19E044.hgt.zip',
		'S19E045.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S19E045.hgt.zip',
		'S19E046.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S19E046.hgt.zip',
		'S19E047.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S19E047.hgt.zip',
		'S19E048.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S19E048.hgt.zip',
		'S19E049.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S19E049.hgt.zip',
		'S19E121.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S19E121.hgt.zip',
		'S19E122.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S19E122.hgt.zip',
		'S19E123.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S19E123.hgt.zip',
		'S19E124.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S19E124.hgt.zip',
		'S19E125.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S19E125.hgt.zip',
		'S19E126.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S19E126.hgt.zip',
		'S19E127.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S19E127.hgt.zip',
		'S19E128.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S19E128.hgt.zip',
		'S19E129.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S19E129.hgt.zip',
		'S19E130.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S19E130.hgt.zip',
		'S19E131.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S19E131.hgt.zip',
		'S19E132.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S19E132.hgt.zip',
		'S19E133.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S19E133.hgt.zip',
		'S19E134.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S19E134.hgt.zip',
		'S19E135.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S19E135.hgt.zip',
		'S19E136.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S19E136.hgt.zip',
		'S19E137.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S19E137.hgt.zip',
		'S19E138.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S19E138.hgt.zip',
		'S19E139.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S19E139.hgt.zip',
		'S19E140.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S19E140.hgt.zip',
		'S19E141.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S19E141.hgt.zip',
		'S19E142.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S19E142.hgt.zip',
		'S19E143.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S19E143.hgt.zip',
		'S19E144.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S19E144.hgt.zip',
		'S19E145.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S19E145.hgt.zip',
		'S19E146.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S19E146.hgt.zip',
		'S19E162.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S19E162.hgt.zip',
		'S19E163.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S19E163.hgt.zip',
		'S19E168.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S19E168.hgt.zip',
		'S19E169.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S19E169.hgt.zip',
		'S19E177.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S19E177.hgt.zip',
		'S19E178.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S19E178.hgt.zip',
		'S19E179.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S19E179.hgt.zip',
		'S19W040.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S19W040.hgt.zip',
		'S19W041.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S19W041.hgt.zip',
		'S19W042.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S19W042.hgt.zip',
		'S19W043.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S19W043.hgt.zip',
		'S19W044.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S19W044.hgt.zip',
		'S19W045.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S19W045.hgt.zip',
		'S19W046.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S19W046.hgt.zip',
		'S19W047.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S19W047.hgt.zip',
		'S19W048.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S19W048.hgt.zip',
		'S19W049.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S19W049.hgt.zip',
		'S19W050.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S19W050.hgt.zip',
		'S19W051.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S19W051.hgt.zip',
		'S19W052.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S19W052.hgt.zip',
		'S19W053.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S19W053.hgt.zip',
		'S19W054.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S19W054.hgt.zip',
		'S19W055.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S19W055.hgt.zip',
		'S19W056.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S19W056.hgt.zip',
		'S19W057.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S19W057.hgt.zip',
		'S19W058.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S19W058.hgt.zip',
		'S19W059.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S19W059.hgt.zip',
		'S19W060.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S19W060.hgt.zip',
		'S19W061.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S19W061.hgt.zip',
		'S19W062.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S19W062.hgt.zip',
		'S19W063.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S19W063.hgt.zip',
		'S19W064.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S19W064.hgt.zip',
		'S19W065.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S19W065.hgt.zip',
		'S19W066.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S19W066.hgt.zip',
		'S19W067.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S19W067.hgt.zip',
		'S19W068.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S19W068.hgt.zip',
		'S19W069.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S19W069.hgt.zip',
		'S19W070.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S19W070.hgt.zip',
		'S19W071.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S19W071.hgt.zip',
		'S19W137.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S19W137.hgt.zip',
		'S19W138.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S19W138.hgt.zip',
		'S19W139.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S19W139.hgt.zip',
		'S19W140.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S19W140.hgt.zip',
		'S19W141.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S19W141.hgt.zip',
		'S19W142.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S19W142.hgt.zip',
		'S19W143.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S19W143.hgt.zip',
		'S19W160.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S19W160.hgt.zip',
		'S19W164.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S19W164.hgt.zip',
		'S19W170.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S19W170.hgt.zip',
		'S19W174.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S19W174.hgt.zip',
		'S19W175.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S19W175.hgt.zip',
		'S19W179.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S19W179.hgt.zip',
		'S19W180.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S19W180.hgt.zip',
		'S20E012.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S20E012.hgt.zip',
		'S20E013.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S20E013.hgt.zip',
		'S20E014.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S20E014.hgt.zip',
		'S20E015.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S20E015.hgt.zip',
		'S20E016.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S20E016.hgt.zip',
		'S20E017.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S20E017.hgt.zip',
		'S20E018.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S20E018.hgt.zip',
		'S20E019.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S20E019.hgt.zip',
		'S20E020.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S20E020.hgt.zip',
		'S20E021.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S20E021.hgt.zip',
		'S20E022.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S20E022.hgt.zip',
		'S20E023.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S20E023.hgt.zip',
		'S20E024.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S20E024.hgt.zip',
		'S20E025.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S20E025.hgt.zip',
		'S20E026.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S20E026.hgt.zip',
		'S20E027.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S20E027.hgt.zip',
		'S20E028.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S20E028.hgt.zip',
		'S20E029.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S20E029.hgt.zip',
		'S20E030.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S20E030.hgt.zip',
		'S20E031.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S20E031.hgt.zip',
		'S20E032.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S20E032.hgt.zip',
		'S20E033.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S20E033.hgt.zip',
		'S20E034.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S20E034.hgt.zip',
		'S20E035.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S20E035.hgt.zip',
		'S20E044.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S20E044.hgt.zip',
		'S20E045.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S20E045.hgt.zip',
		'S20E046.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S20E046.hgt.zip',
		'S20E047.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S20E047.hgt.zip',
		'S20E048.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S20E048.hgt.zip',
		'S20E049.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S20E049.hgt.zip',
		'S20E057.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S20E057.hgt.zip',
		'S20E063.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S20E063.hgt.zip',
		'S20E118.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S20E118.hgt.zip',
		'S20E119.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S20E119.hgt.zip',
		'S20E120.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S20E120.hgt.zip',
		'S20E121.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S20E121.hgt.zip',
		'S20E122.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S20E122.hgt.zip',
		'S20E123.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S20E123.hgt.zip',
		'S20E124.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S20E124.hgt.zip',
		'S20E125.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S20E125.hgt.zip',
		'S20E126.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S20E126.hgt.zip',
		'S20E127.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S20E127.hgt.zip',
		'S20E128.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S20E128.hgt.zip',
		'S20E129.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S20E129.hgt.zip',
		'S20E130.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S20E130.hgt.zip',
		'S20E131.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S20E131.hgt.zip',
		'S20E132.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S20E132.hgt.zip',
		'S20E133.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S20E133.hgt.zip',
		'S20E134.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S20E134.hgt.zip',
		'S20E135.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S20E135.hgt.zip',
		'S20E136.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S20E136.hgt.zip',
		'S20E137.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S20E137.hgt.zip',
		'S20E138.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S20E138.hgt.zip',
		'S20E139.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S20E139.hgt.zip',
		'S20E140.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S20E140.hgt.zip',
		'S20E141.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S20E141.hgt.zip',
		'S20E142.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S20E142.hgt.zip',
		'S20E143.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S20E143.hgt.zip',
		'S20E144.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S20E144.hgt.zip',
		'S20E145.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S20E145.hgt.zip',
		'S20E146.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S20E146.hgt.zip',
		'S20E147.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S20E147.hgt.zip',
		'S20E148.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S20E148.hgt.zip',
		'S20E158.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S20E158.hgt.zip',
		'S20E163.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S20E163.hgt.zip',
		'S20E169.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S20E169.hgt.zip',
		'S20E170.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S20E170.hgt.zip',
		'S20E177.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S20E177.hgt.zip',
		'S20E178.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S20E178.hgt.zip',
		'S20E179.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S20E179.hgt.zip',
		'S20W040.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S20W040.hgt.zip',
		'S20W041.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S20W041.hgt.zip',
		'S20W042.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S20W042.hgt.zip',
		'S20W043.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S20W043.hgt.zip',
		'S20W044.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S20W044.hgt.zip',
		'S20W045.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S20W045.hgt.zip',
		'S20W046.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S20W046.hgt.zip',
		'S20W047.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S20W047.hgt.zip',
		'S20W048.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S20W048.hgt.zip',
		'S20W049.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S20W049.hgt.zip',
		'S20W050.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S20W050.hgt.zip',
		'S20W051.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S20W051.hgt.zip',
		'S20W052.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S20W052.hgt.zip',
		'S20W053.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S20W053.hgt.zip',
		'S20W054.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S20W054.hgt.zip',
		'S20W055.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S20W055.hgt.zip',
		'S20W056.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S20W056.hgt.zip',
		'S20W057.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S20W057.hgt.zip',
		'S20W058.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S20W058.hgt.zip',
		'S20W059.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S20W059.hgt.zip',
		'S20W060.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S20W060.hgt.zip',
		'S20W061.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S20W061.hgt.zip',
		'S20W062.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S20W062.hgt.zip',
		'S20W063.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S20W063.hgt.zip',
		'S20W064.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S20W064.hgt.zip',
		'S20W065.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S20W065.hgt.zip',
		'S20W066.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S20W066.hgt.zip',
		'S20W067.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S20W067.hgt.zip',
		'S20W068.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S20W068.hgt.zip',
		'S20W069.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S20W069.hgt.zip',
		'S20W070.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S20W070.hgt.zip',
		'S20W071.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S20W071.hgt.zip',
		'S20W139.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S20W139.hgt.zip',
		'S20W140.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S20W140.hgt.zip',
		'S20W141.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S20W141.hgt.zip',
		'S20W142.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S20W142.hgt.zip',
		'S20W145.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S20W145.hgt.zip',
		'S20W146.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S20W146.hgt.zip',
		'S20W158.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S20W158.hgt.zip',
		'S20W159.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S20W159.hgt.zip',
		'S20W170.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S20W170.hgt.zip',
		'S20W175.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S20W175.hgt.zip',
		'S20W176.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S20W176.hgt.zip',
		'S20W179.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S20W179.hgt.zip',
		'S20W180.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S20W180.hgt.zip',
		'S21E013.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S21E013.hgt.zip',
		'S21E014.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S21E014.hgt.zip',
		'S21E015.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S21E015.hgt.zip',
		'S21E016.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S21E016.hgt.zip',
		'S21E017.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S21E017.hgt.zip',
		'S21E018.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S21E018.hgt.zip',
		'S21E019.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S21E019.hgt.zip',
		'S21E020.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S21E020.hgt.zip',
		'S21E021.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S21E021.hgt.zip',
		'S21E022.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S21E022.hgt.zip',
		'S21E023.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S21E023.hgt.zip',
		'S21E024.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S21E024.hgt.zip',
		'S21E025.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S21E025.hgt.zip',
		'S21E026.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S21E026.hgt.zip',
		'S21E027.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S21E027.hgt.zip',
		'S21E028.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S21E028.hgt.zip',
		'S21E029.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S21E029.hgt.zip',
		'S21E030.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S21E030.hgt.zip',
		'S21E031.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S21E031.hgt.zip',
		'S21E032.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S21E032.hgt.zip',
		'S21E033.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S21E033.hgt.zip',
		'S21E034.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S21E034.hgt.zip',
		'S21E035.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S21E035.hgt.zip',
		'S21E043.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S21E043.hgt.zip',
		'S21E044.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S21E044.hgt.zip',
		'S21E045.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S21E045.hgt.zip',
		'S21E046.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S21E046.hgt.zip',
		'S21E047.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S21E047.hgt.zip',
		'S21E048.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S21E048.hgt.zip',
		'S21E055.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S21E055.hgt.zip',
		'S21E057.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S21E057.hgt.zip',
		'S21E115.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S21E115.hgt.zip',
		'S21E116.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S21E116.hgt.zip',
		'S21E117.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S21E117.hgt.zip',
		'S21E118.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S21E118.hgt.zip',
		'S21E119.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S21E119.hgt.zip',
		'S21E120.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S21E120.hgt.zip',
		'S21E121.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S21E121.hgt.zip',
		'S21E122.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S21E122.hgt.zip',
		'S21E123.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S21E123.hgt.zip',
		'S21E124.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S21E124.hgt.zip',
		'S21E125.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S21E125.hgt.zip',
		'S21E126.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S21E126.hgt.zip',
		'S21E127.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S21E127.hgt.zip',
		'S21E128.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S21E128.hgt.zip',
		'S21E129.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S21E129.hgt.zip',
		'S21E130.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S21E130.hgt.zip',
		'S21E131.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S21E131.hgt.zip',
		'S21E132.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S21E132.hgt.zip',
		'S21E133.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S21E133.hgt.zip',
		'S21E134.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S21E134.hgt.zip',
		'S21E135.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S21E135.hgt.zip',
		'S21E136.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S21E136.hgt.zip',
		'S21E137.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S21E137.hgt.zip',
		'S21E138.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S21E138.hgt.zip',
		'S21E139.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S21E139.hgt.zip',
		'S21E140.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S21E140.hgt.zip',
		'S21E141.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S21E141.hgt.zip',
		'S21E142.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S21E142.hgt.zip',
		'S21E143.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S21E143.hgt.zip',
		'S21E144.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S21E144.hgt.zip',
		'S21E145.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S21E145.hgt.zip',
		'S21E146.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S21E146.hgt.zip',
		'S21E147.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S21E147.hgt.zip',
		'S21E148.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S21E148.hgt.zip',
		'S21E149.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S21E149.hgt.zip',
		'S21E150.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S21E150.hgt.zip',
		'S21E154.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S21E154.hgt.zip',
		'S21E163.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S21E163.hgt.zip',
		'S21E164.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S21E164.hgt.zip',
		'S21E165.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S21E165.hgt.zip',
		'S21E166.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S21E166.hgt.zip',
		'S21E167.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S21E167.hgt.zip',
		'S21E169.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S21E169.hgt.zip',
		'S21W029.hgt': 'https://srtm.kurviger.de/SRTM3//Islands//S21W029.hgt.zip',
		'S21W030.hgt': 'https://srtm.kurviger.de/SRTM3//Islands//S21W030.hgt.zip',
		'S21W041.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S21W041.hgt.zip',
		'S21W042.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S21W042.hgt.zip',
		'S21W043.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S21W043.hgt.zip',
		'S21W044.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S21W044.hgt.zip',
		'S21W045.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S21W045.hgt.zip',
		'S21W046.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S21W046.hgt.zip',
		'S21W047.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S21W047.hgt.zip',
		'S21W048.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S21W048.hgt.zip',
		'S21W049.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S21W049.hgt.zip',
		'S21W050.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S21W050.hgt.zip',
		'S21W051.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S21W051.hgt.zip',
		'S21W052.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S21W052.hgt.zip',
		'S21W053.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S21W053.hgt.zip',
		'S21W054.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S21W054.hgt.zip',
		'S21W055.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S21W055.hgt.zip',
		'S21W056.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S21W056.hgt.zip',
		'S21W057.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S21W057.hgt.zip',
		'S21W058.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S21W058.hgt.zip',
		'S21W059.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S21W059.hgt.zip',
		'S21W060.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S21W060.hgt.zip',
		'S21W061.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S21W061.hgt.zip',
		'S21W062.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S21W062.hgt.zip',
		'S21W063.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S21W063.hgt.zip',
		'S21W064.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S21W064.hgt.zip',
		'S21W065.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S21W065.hgt.zip',
		'S21W066.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S21W066.hgt.zip',
		'S21W067.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S21W067.hgt.zip',
		'S21W068.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S21W068.hgt.zip',
		'S21W069.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S21W069.hgt.zip',
		'S21W070.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S21W070.hgt.zip',
		'S21W071.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S21W071.hgt.zip',
		'S21W139.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S21W139.hgt.zip',
		'S21W140.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S21W140.hgt.zip',
		'S21W144.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S21W144.hgt.zip',
		'S21W158.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S21W158.hgt.zip',
		'S21W159.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S21W159.hgt.zip',
		'S21W175.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S21W175.hgt.zip',
		'S21W176.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S21W176.hgt.zip',
		'S21W179.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S21W179.hgt.zip',
		'S22E013.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S22E013.hgt.zip',
		'S22E014.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S22E014.hgt.zip',
		'S22E015.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S22E015.hgt.zip',
		'S22E016.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S22E016.hgt.zip',
		'S22E017.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S22E017.hgt.zip',
		'S22E018.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S22E018.hgt.zip',
		'S22E019.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S22E019.hgt.zip',
		'S22E020.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S22E020.hgt.zip',
		'S22E021.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S22E021.hgt.zip',
		'S22E022.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S22E022.hgt.zip',
		'S22E023.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S22E023.hgt.zip',
		'S22E024.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S22E024.hgt.zip',
		'S22E025.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S22E025.hgt.zip',
		'S22E026.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S22E026.hgt.zip',
		'S22E027.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S22E027.hgt.zip',
		'S22E028.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S22E028.hgt.zip',
		'S22E029.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S22E029.hgt.zip',
		'S22E030.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S22E030.hgt.zip',
		'S22E031.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S22E031.hgt.zip',
		'S22E032.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S22E032.hgt.zip',
		'S22E033.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S22E033.hgt.zip',
		'S22E034.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S22E034.hgt.zip',
		'S22E035.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S22E035.hgt.zip',
		'S22E043.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S22E043.hgt.zip',
		'S22E044.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S22E044.hgt.zip',
		'S22E045.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S22E045.hgt.zip',
		'S22E046.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S22E046.hgt.zip',
		'S22E047.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S22E047.hgt.zip',
		'S22E048.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S22E048.hgt.zip',
		'S22E055.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S22E055.hgt.zip',
		'S22E113.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S22E113.hgt.zip',
		'S22E114.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S22E114.hgt.zip',
		'S22E115.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S22E115.hgt.zip',
		'S22E116.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S22E116.hgt.zip',
		'S22E117.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S22E117.hgt.zip',
		'S22E118.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S22E118.hgt.zip',
		'S22E119.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S22E119.hgt.zip',
		'S22E120.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S22E120.hgt.zip',
		'S22E121.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S22E121.hgt.zip',
		'S22E122.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S22E122.hgt.zip',
		'S22E123.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S22E123.hgt.zip',
		'S22E124.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S22E124.hgt.zip',
		'S22E125.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S22E125.hgt.zip',
		'S22E126.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S22E126.hgt.zip',
		'S22E127.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S22E127.hgt.zip',
		'S22E128.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S22E128.hgt.zip',
		'S22E129.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S22E129.hgt.zip',
		'S22E130.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S22E130.hgt.zip',
		'S22E131.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S22E131.hgt.zip',
		'S22E132.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S22E132.hgt.zip',
		'S22E133.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S22E133.hgt.zip',
		'S22E134.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S22E134.hgt.zip',
		'S22E135.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S22E135.hgt.zip',
		'S22E136.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S22E136.hgt.zip',
		'S22E137.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S22E137.hgt.zip',
		'S22E138.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S22E138.hgt.zip',
		'S22E139.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S22E139.hgt.zip',
		'S22E140.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S22E140.hgt.zip',
		'S22E141.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S22E141.hgt.zip',
		'S22E142.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S22E142.hgt.zip',
		'S22E143.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S22E143.hgt.zip',
		'S22E144.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S22E144.hgt.zip',
		'S22E145.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S22E145.hgt.zip',
		'S22E146.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S22E146.hgt.zip',
		'S22E147.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S22E147.hgt.zip',
		'S22E148.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S22E148.hgt.zip',
		'S22E149.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S22E149.hgt.zip',
		'S22E150.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S22E150.hgt.zip',
		'S22E151.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S22E151.hgt.zip',
		'S22E152.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S22E152.hgt.zip',
		'S22E153.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S22E153.hgt.zip',
		'S22E154.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S22E154.hgt.zip',
		'S22E155.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S22E155.hgt.zip',
		'S22E158.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S22E158.hgt.zip',
		'S22E164.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S22E164.hgt.zip',
		'S22E165.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S22E165.hgt.zip',
		'S22E166.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S22E166.hgt.zip',
		'S22E167.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S22E167.hgt.zip',
		'S22E168.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S22E168.hgt.zip',
		'S22W041.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S22W041.hgt.zip',
		'S22W042.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S22W042.hgt.zip',
		'S22W043.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S22W043.hgt.zip',
		'S22W044.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S22W044.hgt.zip',
		'S22W045.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S22W045.hgt.zip',
		'S22W046.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S22W046.hgt.zip',
		'S22W047.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S22W047.hgt.zip',
		'S22W048.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S22W048.hgt.zip',
		'S22W049.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S22W049.hgt.zip',
		'S22W050.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S22W050.hgt.zip',
		'S22W051.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S22W051.hgt.zip',
		'S22W052.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S22W052.hgt.zip',
		'S22W053.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S22W053.hgt.zip',
		'S22W054.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S22W054.hgt.zip',
		'S22W055.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S22W055.hgt.zip',
		'S22W056.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S22W056.hgt.zip',
		'S22W057.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S22W057.hgt.zip',
		'S22W058.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S22W058.hgt.zip',
		'S22W059.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S22W059.hgt.zip',
		'S22W060.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S22W060.hgt.zip',
		'S22W061.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S22W061.hgt.zip',
		'S22W062.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S22W062.hgt.zip',
		'S22W063.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S22W063.hgt.zip',
		'S22W064.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S22W064.hgt.zip',
		'S22W065.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S22W065.hgt.zip',
		'S22W066.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S22W066.hgt.zip',
		'S22W067.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S22W067.hgt.zip',
		'S22W068.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S22W068.hgt.zip',
		'S22W069.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S22W069.hgt.zip',
		'S22W070.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S22W070.hgt.zip',
		'S22W071.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S22W071.hgt.zip',
		'S22W136.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S22W136.hgt.zip',
		'S22W137.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S22W137.hgt.zip',
		'S22W139.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S22W139.hgt.zip',
		'S22W140.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S22W140.hgt.zip',
		'S22W141.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S22W141.hgt.zip',
		'S22W155.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S22W155.hgt.zip',
		'S22W158.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S22W158.hgt.zip',
		'S22W160.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S22W160.hgt.zip',
		'S22W175.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S22W175.hgt.zip',
		'S22W176.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S22W176.hgt.zip',
		'S22W179.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S22W179.hgt.zip',
		'S23E014.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S23E014.hgt.zip',
		'S23E015.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S23E015.hgt.zip',
		'S23E016.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S23E016.hgt.zip',
		'S23E017.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S23E017.hgt.zip',
		'S23E018.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S23E018.hgt.zip',
		'S23E019.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S23E019.hgt.zip',
		'S23E020.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S23E020.hgt.zip',
		'S23E021.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S23E021.hgt.zip',
		'S23E022.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S23E022.hgt.zip',
		'S23E023.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S23E023.hgt.zip',
		'S23E024.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S23E024.hgt.zip',
		'S23E025.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S23E025.hgt.zip',
		'S23E026.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S23E026.hgt.zip',
		'S23E027.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S23E027.hgt.zip',
		'S23E028.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S23E028.hgt.zip',
		'S23E029.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S23E029.hgt.zip',
		'S23E030.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S23E030.hgt.zip',
		'S23E031.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S23E031.hgt.zip',
		'S23E032.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S23E032.hgt.zip',
		'S23E033.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S23E033.hgt.zip',
		'S23E034.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S23E034.hgt.zip',
		'S23E035.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S23E035.hgt.zip',
		'S23E040.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S23E040.hgt.zip',
		'S23E043.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S23E043.hgt.zip',
		'S23E044.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S23E044.hgt.zip',
		'S23E045.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S23E045.hgt.zip',
		'S23E046.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S23E046.hgt.zip',
		'S23E047.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S23E047.hgt.zip',
		'S23E048.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S23E048.hgt.zip',
		'S23E113.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S23E113.hgt.zip',
		'S23E114.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S23E114.hgt.zip',
		'S23E115.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S23E115.hgt.zip',
		'S23E116.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S23E116.hgt.zip',
		'S23E117.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S23E117.hgt.zip',
		'S23E118.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S23E118.hgt.zip',
		'S23E119.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S23E119.hgt.zip',
		'S23E120.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S23E120.hgt.zip',
		'S23E121.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S23E121.hgt.zip',
		'S23E122.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S23E122.hgt.zip',
		'S23E123.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S23E123.hgt.zip',
		'S23E124.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S23E124.hgt.zip',
		'S23E125.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S23E125.hgt.zip',
		'S23E126.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S23E126.hgt.zip',
		'S23E127.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S23E127.hgt.zip',
		'S23E128.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S23E128.hgt.zip',
		'S23E129.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S23E129.hgt.zip',
		'S23E130.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S23E130.hgt.zip',
		'S23E131.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S23E131.hgt.zip',
		'S23E132.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S23E132.hgt.zip',
		'S23E133.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S23E133.hgt.zip',
		'S23E134.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S23E134.hgt.zip',
		'S23E135.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S23E135.hgt.zip',
		'S23E136.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S23E136.hgt.zip',
		'S23E137.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S23E137.hgt.zip',
		'S23E138.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S23E138.hgt.zip',
		'S23E139.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S23E139.hgt.zip',
		'S23E140.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S23E140.hgt.zip',
		'S23E141.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S23E141.hgt.zip',
		'S23E142.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S23E142.hgt.zip',
		'S23E143.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S23E143.hgt.zip',
		'S23E144.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S23E144.hgt.zip',
		'S23E145.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S23E145.hgt.zip',
		'S23E146.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S23E146.hgt.zip',
		'S23E147.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S23E147.hgt.zip',
		'S23E148.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S23E148.hgt.zip',
		'S23E149.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S23E149.hgt.zip',
		'S23E150.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S23E150.hgt.zip',
		'S23E152.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S23E152.hgt.zip',
		'S23E155.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S23E155.hgt.zip',
		'S23E165.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S23E165.hgt.zip',
		'S23E166.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S23E166.hgt.zip',
		'S23E167.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S23E167.hgt.zip',
		'S23E168.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S23E168.hgt.zip',
		'S23E171.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S23E171.hgt.zip',
		'S23E172.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S23E172.hgt.zip',
		'S23W041.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S23W041.hgt.zip',
		'S23W042.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S23W042.hgt.zip',
		'S23W043.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S23W043.hgt.zip',
		'S23W044.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S23W044.hgt.zip',
		'S23W045.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S23W045.hgt.zip',
		'S23W046.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S23W046.hgt.zip',
		'S23W047.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S23W047.hgt.zip',
		'S23W048.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S23W048.hgt.zip',
		'S23W049.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S23W049.hgt.zip',
		'S23W050.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S23W050.hgt.zip',
		'S23W051.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S23W051.hgt.zip',
		'S23W052.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S23W052.hgt.zip',
		'S23W053.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S23W053.hgt.zip',
		'S23W054.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S23W054.hgt.zip',
		'S23W055.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S23W055.hgt.zip',
		'S23W056.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S23W056.hgt.zip',
		'S23W057.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S23W057.hgt.zip',
		'S23W058.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S23W058.hgt.zip',
		'S23W059.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S23W059.hgt.zip',
		'S23W060.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S23W060.hgt.zip',
		'S23W061.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S23W061.hgt.zip',
		'S23W062.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S23W062.hgt.zip',
		'S23W063.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S23W063.hgt.zip',
		'S23W064.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S23W064.hgt.zip',
		'S23W065.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S23W065.hgt.zip',
		'S23W066.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S23W066.hgt.zip',
		'S23W067.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S23W067.hgt.zip',
		'S23W068.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S23W068.hgt.zip',
		'S23W069.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S23W069.hgt.zip',
		'S23W070.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S23W070.hgt.zip',
		'S23W071.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S23W071.hgt.zip',
		'S23W135.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S23W135.hgt.zip',
		'S23W137.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S23W137.hgt.zip',
		'S23W139.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S23W139.hgt.zip',
		'S23W152.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S23W152.hgt.zip',
		'S23W153.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S23W153.hgt.zip',
		'S23W177.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S23W177.hgt.zip',
		'S24E014.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S24E014.hgt.zip',
		'S24E015.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S24E015.hgt.zip',
		'S24E016.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S24E016.hgt.zip',
		'S24E017.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S24E017.hgt.zip',
		'S24E018.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S24E018.hgt.zip',
		'S24E019.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S24E019.hgt.zip',
		'S24E020.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S24E020.hgt.zip',
		'S24E021.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S24E021.hgt.zip',
		'S24E022.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S24E022.hgt.zip',
		'S24E023.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S24E023.hgt.zip',
		'S24E024.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S24E024.hgt.zip',
		'S24E025.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S24E025.hgt.zip',
		'S24E026.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S24E026.hgt.zip',
		'S24E027.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S24E027.hgt.zip',
		'S24E028.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S24E028.hgt.zip',
		'S24E029.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S24E029.hgt.zip',
		'S24E030.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S24E030.hgt.zip',
		'S24E031.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S24E031.hgt.zip',
		'S24E032.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S24E032.hgt.zip',
		'S24E033.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S24E033.hgt.zip',
		'S24E034.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S24E034.hgt.zip',
		'S24E035.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S24E035.hgt.zip',
		'S24E043.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S24E043.hgt.zip',
		'S24E044.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S24E044.hgt.zip',
		'S24E045.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S24E045.hgt.zip',
		'S24E046.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S24E046.hgt.zip',
		'S24E047.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S24E047.hgt.zip',
		'S24E113.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S24E113.hgt.zip',
		'S24E114.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S24E114.hgt.zip',
		'S24E115.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S24E115.hgt.zip',
		'S24E116.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S24E116.hgt.zip',
		'S24E117.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S24E117.hgt.zip',
		'S24E118.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S24E118.hgt.zip',
		'S24E119.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S24E119.hgt.zip',
		'S24E120.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S24E120.hgt.zip',
		'S24E121.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S24E121.hgt.zip',
		'S24E122.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S24E122.hgt.zip',
		'S24E123.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S24E123.hgt.zip',
		'S24E124.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S24E124.hgt.zip',
		'S24E125.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S24E125.hgt.zip',
		'S24E126.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S24E126.hgt.zip',
		'S24E127.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S24E127.hgt.zip',
		'S24E128.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S24E128.hgt.zip',
		'S24E129.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S24E129.hgt.zip',
		'S24E130.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S24E130.hgt.zip',
		'S24E131.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S24E131.hgt.zip',
		'S24E132.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S24E132.hgt.zip',
		'S24E133.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S24E133.hgt.zip',
		'S24E134.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S24E134.hgt.zip',
		'S24E135.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S24E135.hgt.zip',
		'S24E136.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S24E136.hgt.zip',
		'S24E137.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S24E137.hgt.zip',
		'S24E138.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S24E138.hgt.zip',
		'S24E139.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S24E139.hgt.zip',
		'S24E140.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S24E140.hgt.zip',
		'S24E141.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S24E141.hgt.zip',
		'S24E142.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S24E142.hgt.zip',
		'S24E143.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S24E143.hgt.zip',
		'S24E144.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S24E144.hgt.zip',
		'S24E145.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S24E145.hgt.zip',
		'S24E146.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S24E146.hgt.zip',
		'S24E147.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S24E147.hgt.zip',
		'S24E148.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S24E148.hgt.zip',
		'S24E149.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S24E149.hgt.zip',
		'S24E150.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S24E150.hgt.zip',
		'S24E151.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S24E151.hgt.zip',
		'S24E152.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S24E152.hgt.zip',
		'S24E155.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S24E155.hgt.zip',
		'S24W042.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S24W042.hgt.zip',
		'S24W043.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S24W043.hgt.zip',
		'S24W044.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S24W044.hgt.zip',
		'S24W045.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S24W045.hgt.zip',
		'S24W046.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S24W046.hgt.zip',
		'S24W047.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S24W047.hgt.zip',
		'S24W048.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S24W048.hgt.zip',
		'S24W049.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S24W049.hgt.zip',
		'S24W050.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S24W050.hgt.zip',
		'S24W051.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S24W051.hgt.zip',
		'S24W052.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S24W052.hgt.zip',
		'S24W053.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S24W053.hgt.zip',
		'S24W054.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S24W054.hgt.zip',
		'S24W055.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S24W055.hgt.zip',
		'S24W056.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S24W056.hgt.zip',
		'S24W057.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S24W057.hgt.zip',
		'S24W058.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S24W058.hgt.zip',
		'S24W059.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S24W059.hgt.zip',
		'S24W060.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S24W060.hgt.zip',
		'S24W061.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S24W061.hgt.zip',
		'S24W062.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S24W062.hgt.zip',
		'S24W063.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S24W063.hgt.zip',
		'S24W064.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S24W064.hgt.zip',
		'S24W065.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S24W065.hgt.zip',
		'S24W066.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S24W066.hgt.zip',
		'S24W067.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S24W067.hgt.zip',
		'S24W068.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S24W068.hgt.zip',
		'S24W069.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S24W069.hgt.zip',
		'S24W070.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S24W070.hgt.zip',
		'S24W071.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S24W071.hgt.zip',
		'S24W131.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S24W131.hgt.zip',
		'S24W135.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S24W135.hgt.zip',
		'S24W136.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S24W136.hgt.zip',
		'S24W138.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S24W138.hgt.zip',
		'S24W148.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S24W148.hgt.zip',
		'S24W150.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S24W150.hgt.zip',
		'S25E014.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S25E014.hgt.zip',
		'S25E015.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S25E015.hgt.zip',
		'S25E016.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S25E016.hgt.zip',
		'S25E017.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S25E017.hgt.zip',
		'S25E018.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S25E018.hgt.zip',
		'S25E019.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S25E019.hgt.zip',
		'S25E020.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S25E020.hgt.zip',
		'S25E021.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S25E021.hgt.zip',
		'S25E022.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S25E022.hgt.zip',
		'S25E023.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S25E023.hgt.zip',
		'S25E024.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S25E024.hgt.zip',
		'S25E025.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S25E025.hgt.zip',
		'S25E026.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S25E026.hgt.zip',
		'S25E027.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S25E027.hgt.zip',
		'S25E028.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S25E028.hgt.zip',
		'S25E029.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S25E029.hgt.zip',
		'S25E030.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S25E030.hgt.zip',
		'S25E031.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S25E031.hgt.zip',
		'S25E032.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S25E032.hgt.zip',
		'S25E033.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S25E033.hgt.zip',
		'S25E034.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S25E034.hgt.zip',
		'S25E035.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S25E035.hgt.zip',
		'S25E043.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S25E043.hgt.zip',
		'S25E044.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S25E044.hgt.zip',
		'S25E045.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S25E045.hgt.zip',
		'S25E046.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S25E046.hgt.zip',
		'S25E047.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S25E047.hgt.zip',
		'S25E113.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S25E113.hgt.zip',
		'S25E114.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S25E114.hgt.zip',
		'S25E115.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S25E115.hgt.zip',
		'S25E116.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S25E116.hgt.zip',
		'S25E117.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S25E117.hgt.zip',
		'S25E118.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S25E118.hgt.zip',
		'S25E119.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S25E119.hgt.zip',
		'S25E120.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S25E120.hgt.zip',
		'S25E121.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S25E121.hgt.zip',
		'S25E122.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S25E122.hgt.zip',
		'S25E123.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S25E123.hgt.zip',
		'S25E124.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S25E124.hgt.zip',
		'S25E125.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S25E125.hgt.zip',
		'S25E126.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S25E126.hgt.zip',
		'S25E127.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S25E127.hgt.zip',
		'S25E128.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S25E128.hgt.zip',
		'S25E129.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S25E129.hgt.zip',
		'S25E130.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S25E130.hgt.zip',
		'S25E131.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S25E131.hgt.zip',
		'S25E132.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S25E132.hgt.zip',
		'S25E133.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S25E133.hgt.zip',
		'S25E134.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S25E134.hgt.zip',
		'S25E135.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S25E135.hgt.zip',
		'S25E136.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S25E136.hgt.zip',
		'S25E137.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S25E137.hgt.zip',
		'S25E138.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S25E138.hgt.zip',
		'S25E139.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S25E139.hgt.zip',
		'S25E140.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S25E140.hgt.zip',
		'S25E141.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S25E141.hgt.zip',
		'S25E142.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S25E142.hgt.zip',
		'S25E143.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S25E143.hgt.zip',
		'S25E144.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S25E144.hgt.zip',
		'S25E145.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S25E145.hgt.zip',
		'S25E146.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S25E146.hgt.zip',
		'S25E147.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S25E147.hgt.zip',
		'S25E148.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S25E148.hgt.zip',
		'S25E149.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S25E149.hgt.zip',
		'S25E150.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S25E150.hgt.zip',
		'S25E151.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S25E151.hgt.zip',
		'S25E152.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S25E152.hgt.zip',
		'S25E153.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S25E153.hgt.zip',
		'S25W046.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S25W046.hgt.zip',
		'S25W047.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S25W047.hgt.zip',
		'S25W048.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S25W048.hgt.zip',
		'S25W049.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S25W049.hgt.zip',
		'S25W050.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S25W050.hgt.zip',
		'S25W051.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S25W051.hgt.zip',
		'S25W052.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S25W052.hgt.zip',
		'S25W053.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S25W053.hgt.zip',
		'S25W054.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S25W054.hgt.zip',
		'S25W055.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S25W055.hgt.zip',
		'S25W056.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S25W056.hgt.zip',
		'S25W057.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S25W057.hgt.zip',
		'S25W058.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S25W058.hgt.zip',
		'S25W059.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S25W059.hgt.zip',
		'S25W060.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S25W060.hgt.zip',
		'S25W061.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S25W061.hgt.zip',
		'S25W062.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S25W062.hgt.zip',
		'S25W063.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S25W063.hgt.zip',
		'S25W064.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S25W064.hgt.zip',
		'S25W065.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S25W065.hgt.zip',
		'S25W066.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S25W066.hgt.zip',
		'S25W067.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S25W067.hgt.zip',
		'S25W068.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S25W068.hgt.zip',
		'S25W069.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S25W069.hgt.zip',
		'S25W070.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S25W070.hgt.zip',
		'S25W071.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S25W071.hgt.zip',
		'S25W125.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S25W125.hgt.zip',
		'S25W129.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S25W129.hgt.zip',
		'S26E014.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S26E014.hgt.zip',
		'S26E015.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S26E015.hgt.zip',
		'S26E016.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S26E016.hgt.zip',
		'S26E017.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S26E017.hgt.zip',
		'S26E018.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S26E018.hgt.zip',
		'S26E019.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S26E019.hgt.zip',
		'S26E020.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S26E020.hgt.zip',
		'S26E021.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S26E021.hgt.zip',
		'S26E022.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S26E022.hgt.zip',
		'S26E023.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S26E023.hgt.zip',
		'S26E024.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S26E024.hgt.zip',
		'S26E025.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S26E025.hgt.zip',
		'S26E026.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S26E026.hgt.zip',
		'S26E027.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S26E027.hgt.zip',
		'S26E028.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S26E028.hgt.zip',
		'S26E029.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S26E029.hgt.zip',
		'S26E030.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S26E030.hgt.zip',
		'S26E031.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S26E031.hgt.zip',
		'S26E032.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S26E032.hgt.zip',
		'S26E033.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S26E033.hgt.zip',
		'S26E034.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S26E034.hgt.zip',
		'S26E044.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S26E044.hgt.zip',
		'S26E045.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S26E045.hgt.zip',
		'S26E046.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S26E046.hgt.zip',
		'S26E047.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S26E047.hgt.zip',
		'S26E112.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S26E112.hgt.zip',
		'S26E113.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S26E113.hgt.zip',
		'S26E114.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S26E114.hgt.zip',
		'S26E115.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S26E115.hgt.zip',
		'S26E116.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S26E116.hgt.zip',
		'S26E117.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S26E117.hgt.zip',
		'S26E118.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S26E118.hgt.zip',
		'S26E119.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S26E119.hgt.zip',
		'S26E120.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S26E120.hgt.zip',
		'S26E121.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S26E121.hgt.zip',
		'S26E122.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S26E122.hgt.zip',
		'S26E123.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S26E123.hgt.zip',
		'S26E124.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S26E124.hgt.zip',
		'S26E125.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S26E125.hgt.zip',
		'S26E126.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S26E126.hgt.zip',
		'S26E127.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S26E127.hgt.zip',
		'S26E128.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S26E128.hgt.zip',
		'S26E129.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S26E129.hgt.zip',
		'S26E130.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S26E130.hgt.zip',
		'S26E131.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S26E131.hgt.zip',
		'S26E132.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S26E132.hgt.zip',
		'S26E133.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S26E133.hgt.zip',
		'S26E134.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S26E134.hgt.zip',
		'S26E135.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S26E135.hgt.zip',
		'S26E136.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S26E136.hgt.zip',
		'S26E137.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S26E137.hgt.zip',
		'S26E138.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S26E138.hgt.zip',
		'S26E139.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S26E139.hgt.zip',
		'S26E140.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S26E140.hgt.zip',
		'S26E141.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S26E141.hgt.zip',
		'S26E142.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S26E142.hgt.zip',
		'S26E143.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S26E143.hgt.zip',
		'S26E144.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S26E144.hgt.zip',
		'S26E145.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S26E145.hgt.zip',
		'S26E146.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S26E146.hgt.zip',
		'S26E147.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S26E147.hgt.zip',
		'S26E148.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S26E148.hgt.zip',
		'S26E149.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S26E149.hgt.zip',
		'S26E150.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S26E150.hgt.zip',
		'S26E151.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S26E151.hgt.zip',
		'S26E152.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S26E152.hgt.zip',
		'S26E153.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S26E153.hgt.zip',
		'S26W048.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S26W048.hgt.zip',
		'S26W049.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S26W049.hgt.zip',
		'S26W050.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S26W050.hgt.zip',
		'S26W051.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S26W051.hgt.zip',
		'S26W052.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S26W052.hgt.zip',
		'S26W053.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S26W053.hgt.zip',
		'S26W054.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S26W054.hgt.zip',
		'S26W055.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S26W055.hgt.zip',
		'S26W056.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S26W056.hgt.zip',
		'S26W057.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S26W057.hgt.zip',
		'S26W058.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S26W058.hgt.zip',
		'S26W059.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S26W059.hgt.zip',
		'S26W060.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S26W060.hgt.zip',
		'S26W061.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S26W061.hgt.zip',
		'S26W062.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S26W062.hgt.zip',
		'S26W063.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S26W063.hgt.zip',
		'S26W064.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S26W064.hgt.zip',
		'S26W065.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S26W065.hgt.zip',
		'S26W066.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S26W066.hgt.zip',
		'S26W067.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S26W067.hgt.zip',
		'S26W068.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S26W068.hgt.zip',
		'S26W069.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S26W069.hgt.zip',
		'S26W070.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S26W070.hgt.zip',
		'S26W071.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S26W071.hgt.zip',
		'S26W131.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S26W131.hgt.zip',
		'S27E014.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S27E014.hgt.zip',
		'S27E015.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S27E015.hgt.zip',
		'S27E016.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S27E016.hgt.zip',
		'S27E017.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S27E017.hgt.zip',
		'S27E018.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S27E018.hgt.zip',
		'S27E019.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S27E019.hgt.zip',
		'S27E020.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S27E020.hgt.zip',
		'S27E021.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S27E021.hgt.zip',
		'S27E022.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S27E022.hgt.zip',
		'S27E023.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S27E023.hgt.zip',
		'S27E024.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S27E024.hgt.zip',
		'S27E025.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S27E025.hgt.zip',
		'S27E026.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S27E026.hgt.zip',
		'S27E027.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S27E027.hgt.zip',
		'S27E028.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S27E028.hgt.zip',
		'S27E029.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S27E029.hgt.zip',
		'S27E030.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S27E030.hgt.zip',
		'S27E031.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S27E031.hgt.zip',
		'S27E032.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S27E032.hgt.zip',
		'S27E113.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S27E113.hgt.zip',
		'S27E114.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S27E114.hgt.zip',
		'S27E115.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S27E115.hgt.zip',
		'S27E116.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S27E116.hgt.zip',
		'S27E117.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S27E117.hgt.zip',
		'S27E118.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S27E118.hgt.zip',
		'S27E119.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S27E119.hgt.zip',
		'S27E120.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S27E120.hgt.zip',
		'S27E121.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S27E121.hgt.zip',
		'S27E122.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S27E122.hgt.zip',
		'S27E123.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S27E123.hgt.zip',
		'S27E124.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S27E124.hgt.zip',
		'S27E125.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S27E125.hgt.zip',
		'S27E126.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S27E126.hgt.zip',
		'S27E127.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S27E127.hgt.zip',
		'S27E128.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S27E128.hgt.zip',
		'S27E129.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S27E129.hgt.zip',
		'S27E130.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S27E130.hgt.zip',
		'S27E131.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S27E131.hgt.zip',
		'S27E132.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S27E132.hgt.zip',
		'S27E133.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S27E133.hgt.zip',
		'S27E134.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S27E134.hgt.zip',
		'S27E135.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S27E135.hgt.zip',
		'S27E136.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S27E136.hgt.zip',
		'S27E137.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S27E137.hgt.zip',
		'S27E138.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S27E138.hgt.zip',
		'S27E139.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S27E139.hgt.zip',
		'S27E140.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S27E140.hgt.zip',
		'S27E141.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S27E141.hgt.zip',
		'S27E142.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S27E142.hgt.zip',
		'S27E143.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S27E143.hgt.zip',
		'S27E144.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S27E144.hgt.zip',
		'S27E145.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S27E145.hgt.zip',
		'S27E146.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S27E146.hgt.zip',
		'S27E147.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S27E147.hgt.zip',
		'S27E148.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S27E148.hgt.zip',
		'S27E149.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S27E149.hgt.zip',
		'S27E150.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S27E150.hgt.zip',
		'S27E151.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S27E151.hgt.zip',
		'S27E152.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S27E152.hgt.zip',
		'S27E153.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S27E153.hgt.zip',
		'S27W049.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S27W049.hgt.zip',
		'S27W050.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S27W050.hgt.zip',
		'S27W051.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S27W051.hgt.zip',
		'S27W052.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S27W052.hgt.zip',
		'S27W053.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S27W053.hgt.zip',
		'S27W054.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S27W054.hgt.zip',
		'S27W055.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S27W055.hgt.zip',
		'S27W056.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S27W056.hgt.zip',
		'S27W057.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S27W057.hgt.zip',
		'S27W058.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S27W058.hgt.zip',
		'S27W059.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S27W059.hgt.zip',
		'S27W060.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S27W060.hgt.zip',
		'S27W061.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S27W061.hgt.zip',
		'S27W062.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S27W062.hgt.zip',
		'S27W063.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S27W063.hgt.zip',
		'S27W064.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S27W064.hgt.zip',
		'S27W065.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S27W065.hgt.zip',
		'S27W066.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S27W066.hgt.zip',
		'S27W067.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S27W067.hgt.zip',
		'S27W068.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S27W068.hgt.zip',
		'S27W069.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S27W069.hgt.zip',
		'S27W070.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S27W070.hgt.zip',
		'S27W071.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S27W071.hgt.zip',
		'S27W080.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S27W080.hgt.zip',
		'S27W081.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S27W081.hgt.zip',
		'S27W106.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S27W106.hgt.zip',
		'S28E015.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S28E015.hgt.zip',
		'S28E016.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S28E016.hgt.zip',
		'S28E017.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S28E017.hgt.zip',
		'S28E018.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S28E018.hgt.zip',
		'S28E019.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S28E019.hgt.zip',
		'S28E020.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S28E020.hgt.zip',
		'S28E021.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S28E021.hgt.zip',
		'S28E022.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S28E022.hgt.zip',
		'S28E023.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S28E023.hgt.zip',
		'S28E024.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S28E024.hgt.zip',
		'S28E025.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S28E025.hgt.zip',
		'S28E026.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S28E026.hgt.zip',
		'S28E027.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S28E027.hgt.zip',
		'S28E028.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S28E028.hgt.zip',
		'S28E029.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S28E029.hgt.zip',
		'S28E030.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S28E030.hgt.zip',
		'S28E031.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S28E031.hgt.zip',
		'S28E032.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S28E032.hgt.zip',
		'S28E113.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S28E113.hgt.zip',
		'S28E114.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S28E114.hgt.zip',
		'S28E115.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S28E115.hgt.zip',
		'S28E116.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S28E116.hgt.zip',
		'S28E117.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S28E117.hgt.zip',
		'S28E118.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S28E118.hgt.zip',
		'S28E119.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S28E119.hgt.zip',
		'S28E120.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S28E120.hgt.zip',
		'S28E121.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S28E121.hgt.zip',
		'S28E122.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S28E122.hgt.zip',
		'S28E123.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S28E123.hgt.zip',
		'S28E124.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S28E124.hgt.zip',
		'S28E125.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S28E125.hgt.zip',
		'S28E126.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S28E126.hgt.zip',
		'S28E127.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S28E127.hgt.zip',
		'S28E128.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S28E128.hgt.zip',
		'S28E129.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S28E129.hgt.zip',
		'S28E130.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S28E130.hgt.zip',
		'S28E131.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S28E131.hgt.zip',
		'S28E132.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S28E132.hgt.zip',
		'S28E133.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S28E133.hgt.zip',
		'S28E134.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S28E134.hgt.zip',
		'S28E135.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S28E135.hgt.zip',
		'S28E136.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S28E136.hgt.zip',
		'S28E137.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S28E137.hgt.zip',
		'S28E138.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S28E138.hgt.zip',
		'S28E139.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S28E139.hgt.zip',
		'S28E140.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S28E140.hgt.zip',
		'S28E141.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S28E141.hgt.zip',
		'S28E142.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S28E142.hgt.zip',
		'S28E143.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S28E143.hgt.zip',
		'S28E144.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S28E144.hgt.zip',
		'S28E145.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S28E145.hgt.zip',
		'S28E146.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S28E146.hgt.zip',
		'S28E147.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S28E147.hgt.zip',
		'S28E148.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S28E148.hgt.zip',
		'S28E149.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S28E149.hgt.zip',
		'S28E150.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S28E150.hgt.zip',
		'S28E151.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S28E151.hgt.zip',
		'S28E152.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S28E152.hgt.zip',
		'S28E153.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S28E153.hgt.zip',
		'S28W049.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S28W049.hgt.zip',
		'S28W050.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S28W050.hgt.zip',
		'S28W051.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S28W051.hgt.zip',
		'S28W052.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S28W052.hgt.zip',
		'S28W053.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S28W053.hgt.zip',
		'S28W054.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S28W054.hgt.zip',
		'S28W055.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S28W055.hgt.zip',
		'S28W056.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S28W056.hgt.zip',
		'S28W057.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S28W057.hgt.zip',
		'S28W058.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S28W058.hgt.zip',
		'S28W059.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S28W059.hgt.zip',
		'S28W060.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S28W060.hgt.zip',
		'S28W061.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S28W061.hgt.zip',
		'S28W062.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S28W062.hgt.zip',
		'S28W063.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S28W063.hgt.zip',
		'S28W064.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S28W064.hgt.zip',
		'S28W065.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S28W065.hgt.zip',
		'S28W066.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S28W066.hgt.zip',
		'S28W067.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S28W067.hgt.zip',
		'S28W068.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S28W068.hgt.zip',
		'S28W069.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S28W069.hgt.zip',
		'S28W070.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S28W070.hgt.zip',
		'S28W071.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S28W071.hgt.zip',
		'S28W072.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S28W072.hgt.zip',
		'S28W110.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S28W110.hgt.zip',
		'S28W144.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S28W144.hgt.zip',
		'S28W145.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S28W145.hgt.zip',
		'S29E015.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S29E015.hgt.zip',
		'S29E016.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S29E016.hgt.zip',
		'S29E017.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S29E017.hgt.zip',
		'S29E018.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S29E018.hgt.zip',
		'S29E019.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S29E019.hgt.zip',
		'S29E020.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S29E020.hgt.zip',
		'S29E021.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S29E021.hgt.zip',
		'S29E022.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S29E022.hgt.zip',
		'S29E023.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S29E023.hgt.zip',
		'S29E024.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S29E024.hgt.zip',
		'S29E025.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S29E025.hgt.zip',
		'S29E026.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S29E026.hgt.zip',
		'S29E027.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S29E027.hgt.zip',
		'S29E028.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S29E028.hgt.zip',
		'S29E029.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S29E029.hgt.zip',
		'S29E030.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S29E030.hgt.zip',
		'S29E031.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S29E031.hgt.zip',
		'S29E032.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S29E032.hgt.zip',
		'S29E113.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S29E113.hgt.zip',
		'S29E114.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S29E114.hgt.zip',
		'S29E115.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S29E115.hgt.zip',
		'S29E116.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S29E116.hgt.zip',
		'S29E117.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S29E117.hgt.zip',
		'S29E118.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S29E118.hgt.zip',
		'S29E119.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S29E119.hgt.zip',
		'S29E120.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S29E120.hgt.zip',
		'S29E121.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S29E121.hgt.zip',
		'S29E122.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S29E122.hgt.zip',
		'S29E123.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S29E123.hgt.zip',
		'S29E124.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S29E124.hgt.zip',
		'S29E125.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S29E125.hgt.zip',
		'S29E126.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S29E126.hgt.zip',
		'S29E127.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S29E127.hgt.zip',
		'S29E128.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S29E128.hgt.zip',
		'S29E129.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S29E129.hgt.zip',
		'S29E130.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S29E130.hgt.zip',
		'S29E131.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S29E131.hgt.zip',
		'S29E132.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S29E132.hgt.zip',
		'S29E133.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S29E133.hgt.zip',
		'S29E134.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S29E134.hgt.zip',
		'S29E135.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S29E135.hgt.zip',
		'S29E136.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S29E136.hgt.zip',
		'S29E137.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S29E137.hgt.zip',
		'S29E138.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S29E138.hgt.zip',
		'S29E139.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S29E139.hgt.zip',
		'S29E140.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S29E140.hgt.zip',
		'S29E141.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S29E141.hgt.zip',
		'S29E142.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S29E142.hgt.zip',
		'S29E143.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S29E143.hgt.zip',
		'S29E144.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S29E144.hgt.zip',
		'S29E145.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S29E145.hgt.zip',
		'S29E146.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S29E146.hgt.zip',
		'S29E147.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S29E147.hgt.zip',
		'S29E148.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S29E148.hgt.zip',
		'S29E149.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S29E149.hgt.zip',
		'S29E150.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S29E150.hgt.zip',
		'S29E151.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S29E151.hgt.zip',
		'S29E152.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S29E152.hgt.zip',
		'S29E153.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S29E153.hgt.zip',
		'S29E167.hgt': 'https://srtm.kurviger.de/SRTM3//Islands//S29E167.hgt.zip',
		'S29W049.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S29W049.hgt.zip',
		'S29W050.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S29W050.hgt.zip',
		'S29W051.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S29W051.hgt.zip',
		'S29W052.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S29W052.hgt.zip',
		'S29W053.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S29W053.hgt.zip',
		'S29W054.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S29W054.hgt.zip',
		'S29W055.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S29W055.hgt.zip',
		'S29W056.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S29W056.hgt.zip',
		'S29W057.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S29W057.hgt.zip',
		'S29W058.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S29W058.hgt.zip',
		'S29W059.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S29W059.hgt.zip',
		'S29W060.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S29W060.hgt.zip',
		'S29W061.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S29W061.hgt.zip',
		'S29W062.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S29W062.hgt.zip',
		'S29W063.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S29W063.hgt.zip',
		'S29W064.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S29W064.hgt.zip',
		'S29W065.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S29W065.hgt.zip',
		'S29W066.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S29W066.hgt.zip',
		'S29W067.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S29W067.hgt.zip',
		'S29W068.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S29W068.hgt.zip',
		'S29W069.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S29W069.hgt.zip',
		'S29W070.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S29W070.hgt.zip',
		'S29W071.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S29W071.hgt.zip',
		'S29W072.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S29W072.hgt.zip',
		'S30E016.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S30E016.hgt.zip',
		'S30E017.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S30E017.hgt.zip',
		'S30E018.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S30E018.hgt.zip',
		'S30E019.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S30E019.hgt.zip',
		'S30E020.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S30E020.hgt.zip',
		'S30E021.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S30E021.hgt.zip',
		'S30E022.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S30E022.hgt.zip',
		'S30E023.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S30E023.hgt.zip',
		'S30E024.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S30E024.hgt.zip',
		'S30E025.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S30E025.hgt.zip',
		'S30E026.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S30E026.hgt.zip',
		'S30E027.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S30E027.hgt.zip',
		'S30E028.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S30E028.hgt.zip',
		'S30E029.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S30E029.hgt.zip',
		'S30E030.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S30E030.hgt.zip',
		'S30E031.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S30E031.hgt.zip',
		'S30E114.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S30E114.hgt.zip',
		'S30E115.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S30E115.hgt.zip',
		'S30E116.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S30E116.hgt.zip',
		'S30E117.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S30E117.hgt.zip',
		'S30E118.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S30E118.hgt.zip',
		'S30E119.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S30E119.hgt.zip',
		'S30E120.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S30E120.hgt.zip',
		'S30E121.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S30E121.hgt.zip',
		'S30E122.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S30E122.hgt.zip',
		'S30E123.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S30E123.hgt.zip',
		'S30E124.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S30E124.hgt.zip',
		'S30E125.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S30E125.hgt.zip',
		'S30E126.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S30E126.hgt.zip',
		'S30E127.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S30E127.hgt.zip',
		'S30E128.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S30E128.hgt.zip',
		'S30E129.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S30E129.hgt.zip',
		'S30E130.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S30E130.hgt.zip',
		'S30E131.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S30E131.hgt.zip',
		'S30E132.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S30E132.hgt.zip',
		'S30E133.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S30E133.hgt.zip',
		'S30E134.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S30E134.hgt.zip',
		'S30E135.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S30E135.hgt.zip',
		'S30E136.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S30E136.hgt.zip',
		'S30E137.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S30E137.hgt.zip',
		'S30E138.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S30E138.hgt.zip',
		'S30E139.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S30E139.hgt.zip',
		'S30E140.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S30E140.hgt.zip',
		'S30E141.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S30E141.hgt.zip',
		'S30E142.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S30E142.hgt.zip',
		'S30E143.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S30E143.hgt.zip',
		'S30E144.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S30E144.hgt.zip',
		'S30E145.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S30E145.hgt.zip',
		'S30E146.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S30E146.hgt.zip',
		'S30E147.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S30E147.hgt.zip',
		'S30E148.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S30E148.hgt.zip',
		'S30E149.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S30E149.hgt.zip',
		'S30E150.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S30E150.hgt.zip',
		'S30E151.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S30E151.hgt.zip',
		'S30E152.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S30E152.hgt.zip',
		'S30E153.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S30E153.hgt.zip',
		'S30E167.hgt': 'https://srtm.kurviger.de/SRTM3//Islands//S30E167.hgt.zip',
		'S30W050.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S30W050.hgt.zip',
		'S30W051.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S30W051.hgt.zip',
		'S30W052.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S30W052.hgt.zip',
		'S30W053.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S30W053.hgt.zip',
		'S30W054.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S30W054.hgt.zip',
		'S30W055.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S30W055.hgt.zip',
		'S30W056.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S30W056.hgt.zip',
		'S30W057.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S30W057.hgt.zip',
		'S30W058.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S30W058.hgt.zip',
		'S30W059.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S30W059.hgt.zip',
		'S30W060.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S30W060.hgt.zip',
		'S30W061.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S30W061.hgt.zip',
		'S30W062.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S30W062.hgt.zip',
		'S30W063.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S30W063.hgt.zip',
		'S30W064.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S30W064.hgt.zip',
		'S30W065.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S30W065.hgt.zip',
		'S30W066.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S30W066.hgt.zip',
		'S30W067.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S30W067.hgt.zip',
		'S30W068.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S30W068.hgt.zip',
		'S30W069.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S30W069.hgt.zip',
		'S30W070.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S30W070.hgt.zip',
		'S30W071.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S30W071.hgt.zip',
		'S30W072.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S30W072.hgt.zip',
		'S30W178.hgt': 'https://srtm.kurviger.de/SRTM3//Islands//S30W178.hgt.zip',
		'S31E017.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S31E017.hgt.zip',
		'S31E018.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S31E018.hgt.zip',
		'S31E019.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S31E019.hgt.zip',
		'S31E020.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S31E020.hgt.zip',
		'S31E021.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S31E021.hgt.zip',
		'S31E022.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S31E022.hgt.zip',
		'S31E023.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S31E023.hgt.zip',
		'S31E024.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S31E024.hgt.zip',
		'S31E025.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S31E025.hgt.zip',
		'S31E026.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S31E026.hgt.zip',
		'S31E027.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S31E027.hgt.zip',
		'S31E028.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S31E028.hgt.zip',
		'S31E029.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S31E029.hgt.zip',
		'S31E030.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S31E030.hgt.zip',
		'S31E114.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S31E114.hgt.zip',
		'S31E115.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S31E115.hgt.zip',
		'S31E116.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S31E116.hgt.zip',
		'S31E117.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S31E117.hgt.zip',
		'S31E118.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S31E118.hgt.zip',
		'S31E119.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S31E119.hgt.zip',
		'S31E120.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S31E120.hgt.zip',
		'S31E121.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S31E121.hgt.zip',
		'S31E122.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S31E122.hgt.zip',
		'S31E123.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S31E123.hgt.zip',
		'S31E124.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S31E124.hgt.zip',
		'S31E125.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S31E125.hgt.zip',
		'S31E126.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S31E126.hgt.zip',
		'S31E127.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S31E127.hgt.zip',
		'S31E128.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S31E128.hgt.zip',
		'S31E129.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S31E129.hgt.zip',
		'S31E130.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S31E130.hgt.zip',
		'S31E131.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S31E131.hgt.zip',
		'S31E132.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S31E132.hgt.zip',
		'S31E133.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S31E133.hgt.zip',
		'S31E134.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S31E134.hgt.zip',
		'S31E135.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S31E135.hgt.zip',
		'S31E136.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S31E136.hgt.zip',
		'S31E137.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S31E137.hgt.zip',
		'S31E138.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S31E138.hgt.zip',
		'S31E139.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S31E139.hgt.zip',
		'S31E140.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S31E140.hgt.zip',
		'S31E141.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S31E141.hgt.zip',
		'S31E142.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S31E142.hgt.zip',
		'S31E143.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S31E143.hgt.zip',
		'S31E144.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S31E144.hgt.zip',
		'S31E145.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S31E145.hgt.zip',
		'S31E146.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S31E146.hgt.zip',
		'S31E147.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S31E147.hgt.zip',
		'S31E148.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S31E148.hgt.zip',
		'S31E149.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S31E149.hgt.zip',
		'S31E150.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S31E150.hgt.zip',
		'S31E151.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S31E151.hgt.zip',
		'S31E152.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S31E152.hgt.zip',
		'S31E153.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S31E153.hgt.zip',
		'S31W051.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S31W051.hgt.zip',
		'S31W052.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S31W052.hgt.zip',
		'S31W053.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S31W053.hgt.zip',
		'S31W054.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S31W054.hgt.zip',
		'S31W055.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S31W055.hgt.zip',
		'S31W056.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S31W056.hgt.zip',
		'S31W057.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S31W057.hgt.zip',
		'S31W058.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S31W058.hgt.zip',
		'S31W059.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S31W059.hgt.zip',
		'S31W060.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S31W060.hgt.zip',
		'S31W061.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S31W061.hgt.zip',
		'S31W062.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S31W062.hgt.zip',
		'S31W063.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S31W063.hgt.zip',
		'S31W064.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S31W064.hgt.zip',
		'S31W065.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S31W065.hgt.zip',
		'S31W066.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S31W066.hgt.zip',
		'S31W067.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S31W067.hgt.zip',
		'S31W068.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S31W068.hgt.zip',
		'S31W069.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S31W069.hgt.zip',
		'S31W070.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S31W070.hgt.zip',
		'S31W071.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S31W071.hgt.zip',
		'S31W072.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S31W072.hgt.zip',
		'S31W179.hgt': 'https://srtm.kurviger.de/SRTM3//Islands//S31W179.hgt.zip',
		'S32E017.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S32E017.hgt.zip',
		'S32E018.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S32E018.hgt.zip',
		'S32E019.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S32E019.hgt.zip',
		'S32E020.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S32E020.hgt.zip',
		'S32E021.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S32E021.hgt.zip',
		'S32E022.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S32E022.hgt.zip',
		'S32E023.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S32E023.hgt.zip',
		'S32E024.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S32E024.hgt.zip',
		'S32E025.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S32E025.hgt.zip',
		'S32E026.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S32E026.hgt.zip',
		'S32E027.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S32E027.hgt.zip',
		'S32E028.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S32E028.hgt.zip',
		'S32E029.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S32E029.hgt.zip',
		'S32E030.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S32E030.hgt.zip',
		'S32E115.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S32E115.hgt.zip',
		'S32E116.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S32E116.hgt.zip',
		'S32E117.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S32E117.hgt.zip',
		'S32E118.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S32E118.hgt.zip',
		'S32E119.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S32E119.hgt.zip',
		'S32E120.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S32E120.hgt.zip',
		'S32E121.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S32E121.hgt.zip',
		'S32E122.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S32E122.hgt.zip',
		'S32E123.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S32E123.hgt.zip',
		'S32E124.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S32E124.hgt.zip',
		'S32E125.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S32E125.hgt.zip',
		'S32E126.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S32E126.hgt.zip',
		'S32E127.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S32E127.hgt.zip',
		'S32E128.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S32E128.hgt.zip',
		'S32E129.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S32E129.hgt.zip',
		'S32E130.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S32E130.hgt.zip',
		'S32E131.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S32E131.hgt.zip',
		'S32E132.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S32E132.hgt.zip',
		'S32E133.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S32E133.hgt.zip',
		'S32E134.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S32E134.hgt.zip',
		'S32E135.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S32E135.hgt.zip',
		'S32E136.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S32E136.hgt.zip',
		'S32E137.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S32E137.hgt.zip',
		'S32E138.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S32E138.hgt.zip',
		'S32E139.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S32E139.hgt.zip',
		'S32E140.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S32E140.hgt.zip',
		'S32E141.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S32E141.hgt.zip',
		'S32E142.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S32E142.hgt.zip',
		'S32E143.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S32E143.hgt.zip',
		'S32E144.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S32E144.hgt.zip',
		'S32E145.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S32E145.hgt.zip',
		'S32E146.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S32E146.hgt.zip',
		'S32E147.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S32E147.hgt.zip',
		'S32E148.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S32E148.hgt.zip',
		'S32E149.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S32E149.hgt.zip',
		'S32E150.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S32E150.hgt.zip',
		'S32E151.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S32E151.hgt.zip',
		'S32E152.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S32E152.hgt.zip',
		'S32E153.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S32E153.hgt.zip',
		'S32E159.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S32E159.hgt.zip',
		'S32W051.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S32W051.hgt.zip',
		'S32W052.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S32W052.hgt.zip',
		'S32W053.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S32W053.hgt.zip',
		'S32W054.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S32W054.hgt.zip',
		'S32W055.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S32W055.hgt.zip',
		'S32W056.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S32W056.hgt.zip',
		'S32W057.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S32W057.hgt.zip',
		'S32W058.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S32W058.hgt.zip',
		'S32W059.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S32W059.hgt.zip',
		'S32W060.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S32W060.hgt.zip',
		'S32W061.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S32W061.hgt.zip',
		'S32W062.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S32W062.hgt.zip',
		'S32W063.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S32W063.hgt.zip',
		'S32W064.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S32W064.hgt.zip',
		'S32W065.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S32W065.hgt.zip',
		'S32W066.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S32W066.hgt.zip',
		'S32W067.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S32W067.hgt.zip',
		'S32W068.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S32W068.hgt.zip',
		'S32W069.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S32W069.hgt.zip',
		'S32W070.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S32W070.hgt.zip',
		'S32W071.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S32W071.hgt.zip',
		'S32W072.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S32W072.hgt.zip',
		'S32W179.hgt': 'https://srtm.kurviger.de/SRTM3//Islands//S32W179.hgt.zip',
		'S33E017.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S33E017.hgt.zip',
		'S33E018.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S33E018.hgt.zip',
		'S33E019.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S33E019.hgt.zip',
		'S33E020.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S33E020.hgt.zip',
		'S33E021.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S33E021.hgt.zip',
		'S33E022.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S33E022.hgt.zip',
		'S33E023.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S33E023.hgt.zip',
		'S33E024.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S33E024.hgt.zip',
		'S33E025.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S33E025.hgt.zip',
		'S33E026.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S33E026.hgt.zip',
		'S33E027.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S33E027.hgt.zip',
		'S33E028.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S33E028.hgt.zip',
		'S33E029.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S33E029.hgt.zip',
		'S33E115.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S33E115.hgt.zip',
		'S33E116.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S33E116.hgt.zip',
		'S33E117.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S33E117.hgt.zip',
		'S33E118.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S33E118.hgt.zip',
		'S33E119.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S33E119.hgt.zip',
		'S33E120.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S33E120.hgt.zip',
		'S33E121.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S33E121.hgt.zip',
		'S33E122.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S33E122.hgt.zip',
		'S33E123.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S33E123.hgt.zip',
		'S33E124.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S33E124.hgt.zip',
		'S33E125.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S33E125.hgt.zip',
		'S33E126.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S33E126.hgt.zip',
		'S33E127.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S33E127.hgt.zip',
		'S33E128.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S33E128.hgt.zip',
		'S33E132.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S33E132.hgt.zip',
		'S33E133.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S33E133.hgt.zip',
		'S33E134.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S33E134.hgt.zip',
		'S33E135.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S33E135.hgt.zip',
		'S33E136.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S33E136.hgt.zip',
		'S33E137.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S33E137.hgt.zip',
		'S33E138.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S33E138.hgt.zip',
		'S33E139.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S33E139.hgt.zip',
		'S33E140.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S33E140.hgt.zip',
		'S33E141.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S33E141.hgt.zip',
		'S33E142.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S33E142.hgt.zip',
		'S33E143.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S33E143.hgt.zip',
		'S33E144.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S33E144.hgt.zip',
		'S33E145.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S33E145.hgt.zip',
		'S33E146.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S33E146.hgt.zip',
		'S33E147.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S33E147.hgt.zip',
		'S33E148.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S33E148.hgt.zip',
		'S33E149.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S33E149.hgt.zip',
		'S33E150.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S33E150.hgt.zip',
		'S33E151.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S33E151.hgt.zip',
		'S33E152.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S33E152.hgt.zip',
		'S33W052.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S33W052.hgt.zip',
		'S33W053.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S33W053.hgt.zip',
		'S33W054.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S33W054.hgt.zip',
		'S33W055.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S33W055.hgt.zip',
		'S33W056.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S33W056.hgt.zip',
		'S33W057.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S33W057.hgt.zip',
		'S33W058.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S33W058.hgt.zip',
		'S33W059.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S33W059.hgt.zip',
		'S33W060.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S33W060.hgt.zip',
		'S33W061.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S33W061.hgt.zip',
		'S33W062.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S33W062.hgt.zip',
		'S33W063.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S33W063.hgt.zip',
		'S33W064.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S33W064.hgt.zip',
		'S33W065.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S33W065.hgt.zip',
		'S33W066.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S33W066.hgt.zip',
		'S33W067.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S33W067.hgt.zip',
		'S33W068.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S33W068.hgt.zip',
		'S33W069.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S33W069.hgt.zip',
		'S33W070.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S33W070.hgt.zip',
		'S33W071.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S33W071.hgt.zip',
		'S33W072.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S33W072.hgt.zip',
		'S34E017.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S34E017.hgt.zip',
		'S34E018.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S34E018.hgt.zip',
		'S34E019.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S34E019.hgt.zip',
		'S34E020.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S34E020.hgt.zip',
		'S34E021.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S34E021.hgt.zip',
		'S34E022.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S34E022.hgt.zip',
		'S34E023.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S34E023.hgt.zip',
		'S34E024.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S34E024.hgt.zip',
		'S34E025.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S34E025.hgt.zip',
		'S34E026.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S34E026.hgt.zip',
		'S34E027.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S34E027.hgt.zip',
		'S34E114.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S34E114.hgt.zip',
		'S34E115.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S34E115.hgt.zip',
		'S34E116.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S34E116.hgt.zip',
		'S34E117.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S34E117.hgt.zip',
		'S34E118.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S34E118.hgt.zip',
		'S34E119.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S34E119.hgt.zip',
		'S34E120.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S34E120.hgt.zip',
		'S34E121.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S34E121.hgt.zip',
		'S34E122.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S34E122.hgt.zip',
		'S34E123.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S34E123.hgt.zip',
		'S34E124.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S34E124.hgt.zip',
		'S34E134.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S34E134.hgt.zip',
		'S34E135.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S34E135.hgt.zip',
		'S34E136.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S34E136.hgt.zip',
		'S34E137.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S34E137.hgt.zip',
		'S34E138.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S34E138.hgt.zip',
		'S34E139.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S34E139.hgt.zip',
		'S34E140.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S34E140.hgt.zip',
		'S34E141.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S34E141.hgt.zip',
		'S34E142.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S34E142.hgt.zip',
		'S34E143.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S34E143.hgt.zip',
		'S34E144.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S34E144.hgt.zip',
		'S34E145.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S34E145.hgt.zip',
		'S34E146.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S34E146.hgt.zip',
		'S34E147.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S34E147.hgt.zip',
		'S34E148.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S34E148.hgt.zip',
		'S34E149.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S34E149.hgt.zip',
		'S34E150.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S34E150.hgt.zip',
		'S34E151.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S34E151.hgt.zip',
		'S34W053.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S34W053.hgt.zip',
		'S34W054.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S34W054.hgt.zip',
		'S34W055.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S34W055.hgt.zip',
		'S34W056.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S34W056.hgt.zip',
		'S34W057.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S34W057.hgt.zip',
		'S34W058.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S34W058.hgt.zip',
		'S34W059.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S34W059.hgt.zip',
		'S34W060.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S34W060.hgt.zip',
		'S34W061.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S34W061.hgt.zip',
		'S34W062.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S34W062.hgt.zip',
		'S34W063.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S34W063.hgt.zip',
		'S34W064.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S34W064.hgt.zip',
		'S34W065.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S34W065.hgt.zip',
		'S34W066.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S34W066.hgt.zip',
		'S34W067.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S34W067.hgt.zip',
		'S34W068.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S34W068.hgt.zip',
		'S34W069.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S34W069.hgt.zip',
		'S34W070.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S34W070.hgt.zip',
		'S34W071.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S34W071.hgt.zip',
		'S34W072.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S34W072.hgt.zip',
		'S34W079.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S34W079.hgt.zip',
		'S34W081.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S34W081.hgt.zip',
		'S35E018.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S35E018.hgt.zip',
		'S35E019.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S35E019.hgt.zip',
		'S35E020.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S35E020.hgt.zip',
		'S35E021.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S35E021.hgt.zip',
		'S35E022.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S35E022.hgt.zip',
		'S35E023.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S35E023.hgt.zip',
		'S35E024.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S35E024.hgt.zip',
		'S35E025.hgt': 'https://srtm.kurviger.de/SRTM3//Africa//S35E025.hgt.zip',
		'S35E114.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S35E114.hgt.zip',
		'S35E115.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S35E115.hgt.zip',
		'S35E116.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S35E116.hgt.zip',
		'S35E117.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S35E117.hgt.zip',
		'S35E118.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S35E118.hgt.zip',
		'S35E119.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S35E119.hgt.zip',
		'S35E120.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S35E120.hgt.zip',
		'S35E121.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S35E121.hgt.zip',
		'S35E122.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S35E122.hgt.zip',
		'S35E123.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S35E123.hgt.zip',
		'S35E134.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S35E134.hgt.zip',
		'S35E135.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S35E135.hgt.zip',
		'S35E136.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S35E136.hgt.zip',
		'S35E137.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S35E137.hgt.zip',
		'S35E138.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S35E138.hgt.zip',
		'S35E139.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S35E139.hgt.zip',
		'S35E140.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S35E140.hgt.zip',
		'S35E141.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S35E141.hgt.zip',
		'S35E142.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S35E142.hgt.zip',
		'S35E143.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S35E143.hgt.zip',
		'S35E144.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S35E144.hgt.zip',
		'S35E145.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S35E145.hgt.zip',
		'S35E146.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S35E146.hgt.zip',
		'S35E147.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S35E147.hgt.zip',
		'S35E148.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S35E148.hgt.zip',
		'S35E149.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S35E149.hgt.zip',
		'S35E150.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S35E150.hgt.zip',
		'S35E151.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S35E151.hgt.zip',
		'S35E172.hgt': 'https://srtm.kurviger.de/SRTM3//Islands//S35E172.hgt.zip',
		'S35E173.hgt': 'https://srtm.kurviger.de/SRTM3//Islands//S35E173.hgt.zip',
		'S35W054.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S35W054.hgt.zip',
		'S35W055.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S35W055.hgt.zip',
		'S35W056.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S35W056.hgt.zip',
		'S35W057.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S35W057.hgt.zip',
		'S35W058.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S35W058.hgt.zip',
		'S35W059.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S35W059.hgt.zip',
		'S35W060.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S35W060.hgt.zip',
		'S35W061.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S35W061.hgt.zip',
		'S35W062.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S35W062.hgt.zip',
		'S35W063.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S35W063.hgt.zip',
		'S35W064.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S35W064.hgt.zip',
		'S35W065.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S35W065.hgt.zip',
		'S35W066.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S35W066.hgt.zip',
		'S35W067.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S35W067.hgt.zip',
		'S35W068.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S35W068.hgt.zip',
		'S35W069.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S35W069.hgt.zip',
		'S35W070.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S35W070.hgt.zip',
		'S35W071.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S35W071.hgt.zip',
		'S35W072.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S35W072.hgt.zip',
		'S35W073.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S35W073.hgt.zip',
		'S36E116.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S36E116.hgt.zip',
		'S36E117.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S36E117.hgt.zip',
		'S36E118.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S36E118.hgt.zip',
		'S36E135.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S36E135.hgt.zip',
		'S36E136.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S36E136.hgt.zip',
		'S36E137.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S36E137.hgt.zip',
		'S36E138.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S36E138.hgt.zip',
		'S36E139.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S36E139.hgt.zip',
		'S36E140.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S36E140.hgt.zip',
		'S36E141.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S36E141.hgt.zip',
		'S36E142.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S36E142.hgt.zip',
		'S36E143.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S36E143.hgt.zip',
		'S36E144.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S36E144.hgt.zip',
		'S36E145.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S36E145.hgt.zip',
		'S36E146.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S36E146.hgt.zip',
		'S36E147.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S36E147.hgt.zip',
		'S36E148.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S36E148.hgt.zip',
		'S36E149.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S36E149.hgt.zip',
		'S36E150.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S36E150.hgt.zip',
		'S36E173.hgt': 'https://srtm.kurviger.de/SRTM3//Islands//S36E173.hgt.zip',
		'S36E174.hgt': 'https://srtm.kurviger.de/SRTM3//Islands//S36E174.hgt.zip',
		'S36E175.hgt': 'https://srtm.kurviger.de/SRTM3//Islands//S36E175.hgt.zip',
		'S36W058.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S36W058.hgt.zip',
		'S36W059.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S36W059.hgt.zip',
		'S36W060.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S36W060.hgt.zip',
		'S36W061.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S36W061.hgt.zip',
		'S36W062.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S36W062.hgt.zip',
		'S36W063.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S36W063.hgt.zip',
		'S36W064.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S36W064.hgt.zip',
		'S36W065.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S36W065.hgt.zip',
		'S36W066.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S36W066.hgt.zip',
		'S36W067.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S36W067.hgt.zip',
		'S36W068.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S36W068.hgt.zip',
		'S36W069.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S36W069.hgt.zip',
		'S36W070.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S36W070.hgt.zip',
		'S36W071.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S36W071.hgt.zip',
		'S36W072.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S36W072.hgt.zip',
		'S36W073.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S36W073.hgt.zip',
		'S37E136.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S37E136.hgt.zip',
		'S37E137.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S37E137.hgt.zip',
		'S37E139.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S37E139.hgt.zip',
		'S37E140.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S37E140.hgt.zip',
		'S37E141.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S37E141.hgt.zip',
		'S37E142.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S37E142.hgt.zip',
		'S37E143.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S37E143.hgt.zip',
		'S37E144.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S37E144.hgt.zip',
		'S37E145.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S37E145.hgt.zip',
		'S37E146.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S37E146.hgt.zip',
		'S37E147.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S37E147.hgt.zip',
		'S37E148.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S37E148.hgt.zip',
		'S37E149.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S37E149.hgt.zip',
		'S37E150.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S37E150.hgt.zip',
		'S37E173.hgt': 'https://srtm.kurviger.de/SRTM3//Islands//S37E173.hgt.zip',
		'S37E174.hgt': 'https://srtm.kurviger.de/SRTM3//Islands//S37E174.hgt.zip',
		'S37E175.hgt': 'https://srtm.kurviger.de/SRTM3//Islands//S37E175.hgt.zip',
		'S37E176.hgt': 'https://srtm.kurviger.de/SRTM3//Islands//S37E176.hgt.zip',
		'S37W057.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S37W057.hgt.zip',
		'S37W058.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S37W058.hgt.zip',
		'S37W059.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S37W059.hgt.zip',
		'S37W060.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S37W060.hgt.zip',
		'S37W061.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S37W061.hgt.zip',
		'S37W062.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S37W062.hgt.zip',
		'S37W063.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S37W063.hgt.zip',
		'S37W064.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S37W064.hgt.zip',
		'S37W065.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S37W065.hgt.zip',
		'S37W066.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S37W066.hgt.zip',
		'S37W067.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S37W067.hgt.zip',
		'S37W068.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S37W068.hgt.zip',
		'S37W069.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S37W069.hgt.zip',
		'S37W070.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S37W070.hgt.zip',
		'S37W071.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S37W071.hgt.zip',
		'S37W072.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S37W072.hgt.zip',
		'S37W073.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S37W073.hgt.zip',
		'S37W074.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S37W074.hgt.zip',
		'S38E077.hgt': 'https://srtm.kurviger.de/SRTM3//Islands//S38E077.hgt.zip',
		'S38E139.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S38E139.hgt.zip',
		'S38E140.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S38E140.hgt.zip',
		'S38E141.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S38E141.hgt.zip',
		'S38E142.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S38E142.hgt.zip',
		'S38E143.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S38E143.hgt.zip',
		'S38E144.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S38E144.hgt.zip',
		'S38E145.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S38E145.hgt.zip',
		'S38E146.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S38E146.hgt.zip',
		'S38E147.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S38E147.hgt.zip',
		'S38E148.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S38E148.hgt.zip',
		'S38E149.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S38E149.hgt.zip',
		'S38E150.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S38E150.hgt.zip',
		'S38E174.hgt': 'https://srtm.kurviger.de/SRTM3//Islands//S38E174.hgt.zip',
		'S38E175.hgt': 'https://srtm.kurviger.de/SRTM3//Islands//S38E175.hgt.zip',
		'S38E176.hgt': 'https://srtm.kurviger.de/SRTM3//Islands//S38E176.hgt.zip',
		'S38E177.hgt': 'https://srtm.kurviger.de/SRTM3//Islands//S38E177.hgt.zip',
		'S38E178.hgt': 'https://srtm.kurviger.de/SRTM3//Islands//S38E178.hgt.zip',
		'S38W013.hgt': 'https://srtm.kurviger.de/SRTM3//Islands//S38W013.hgt.zip',
		'S38W057.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S38W057.hgt.zip',
		'S38W058.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S38W058.hgt.zip',
		'S38W059.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S38W059.hgt.zip',
		'S38W060.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S38W060.hgt.zip',
		'S38W061.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S38W061.hgt.zip',
		'S38W062.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S38W062.hgt.zip',
		'S38W063.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S38W063.hgt.zip',
		'S38W064.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S38W064.hgt.zip',
		'S38W065.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S38W065.hgt.zip',
		'S38W066.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S38W066.hgt.zip',
		'S38W067.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S38W067.hgt.zip',
		'S38W068.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S38W068.hgt.zip',
		'S38W069.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S38W069.hgt.zip',
		'S38W070.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S38W070.hgt.zip',
		'S38W071.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S38W071.hgt.zip',
		'S38W072.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S38W072.hgt.zip',
		'S38W073.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S38W073.hgt.zip',
		'S38W074.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S38W074.hgt.zip',
		'S39E077.hgt': 'https://srtm.kurviger.de/SRTM3//Islands//S39E077.hgt.zip',
		'S39E140.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S39E140.hgt.zip',
		'S39E141.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S39E141.hgt.zip',
		'S39E142.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S39E142.hgt.zip',
		'S39E143.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S39E143.hgt.zip',
		'S39E144.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S39E144.hgt.zip',
		'S39E145.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S39E145.hgt.zip',
		'S39E146.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S39E146.hgt.zip',
		'S39E147.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S39E147.hgt.zip',
		'S39E174.hgt': 'https://srtm.kurviger.de/SRTM3//Islands//S39E174.hgt.zip',
		'S39E175.hgt': 'https://srtm.kurviger.de/SRTM3//Islands//S39E175.hgt.zip',
		'S39E176.hgt': 'https://srtm.kurviger.de/SRTM3//Islands//S39E176.hgt.zip',
		'S39E177.hgt': 'https://srtm.kurviger.de/SRTM3//Islands//S39E177.hgt.zip',
		'S39E178.hgt': 'https://srtm.kurviger.de/SRTM3//Islands//S39E178.hgt.zip',
		'S39W058.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S39W058.hgt.zip',
		'S39W059.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S39W059.hgt.zip',
		'S39W060.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S39W060.hgt.zip',
		'S39W061.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S39W061.hgt.zip',
		'S39W062.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S39W062.hgt.zip',
		'S39W063.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S39W063.hgt.zip',
		'S39W064.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S39W064.hgt.zip',
		'S39W065.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S39W065.hgt.zip',
		'S39W066.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S39W066.hgt.zip',
		'S39W067.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S39W067.hgt.zip',
		'S39W068.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S39W068.hgt.zip',
		'S39W069.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S39W069.hgt.zip',
		'S39W070.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S39W070.hgt.zip',
		'S39W071.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S39W071.hgt.zip',
		'S39W072.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S39W072.hgt.zip',
		'S39W073.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S39W073.hgt.zip',
		'S39W074.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S39W074.hgt.zip',
		'S40E143.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S40E143.hgt.zip',
		'S40E144.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S40E144.hgt.zip',
		'S40E146.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S40E146.hgt.zip',
		'S40E147.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S40E147.hgt.zip',
		'S40E148.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S40E148.hgt.zip',
		'S40E173.hgt': 'https://srtm.kurviger.de/SRTM3//Islands//S40E173.hgt.zip',
		'S40E174.hgt': 'https://srtm.kurviger.de/SRTM3//Islands//S40E174.hgt.zip',
		'S40E175.hgt': 'https://srtm.kurviger.de/SRTM3//Islands//S40E175.hgt.zip',
		'S40E176.hgt': 'https://srtm.kurviger.de/SRTM3//Islands//S40E176.hgt.zip',
		'S40E177.hgt': 'https://srtm.kurviger.de/SRTM3//Islands//S40E177.hgt.zip',
		'S40E178.hgt': 'https://srtm.kurviger.de/SRTM3//Islands//S40E178.hgt.zip',
		'S40W062.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S40W062.hgt.zip',
		'S40W063.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S40W063.hgt.zip',
		'S40W064.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S40W064.hgt.zip',
		'S40W065.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S40W065.hgt.zip',
		'S40W066.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S40W066.hgt.zip',
		'S40W067.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S40W067.hgt.zip',
		'S40W068.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S40W068.hgt.zip',
		'S40W069.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S40W069.hgt.zip',
		'S40W070.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S40W070.hgt.zip',
		'S40W071.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S40W071.hgt.zip',
		'S40W072.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S40W072.hgt.zip',
		'S40W073.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S40W073.hgt.zip',
		'S40W074.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S40W074.hgt.zip',
		'S41E143.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S41E143.hgt.zip',
		'S41E144.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S41E144.hgt.zip',
		'S41E145.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S41E145.hgt.zip',
		'S41E146.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S41E146.hgt.zip',
		'S41E147.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S41E147.hgt.zip',
		'S41E148.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S41E148.hgt.zip',
		'S41E172.hgt': 'https://srtm.kurviger.de/SRTM3//Islands//S41E172.hgt.zip',
		'S41E173.hgt': 'https://srtm.kurviger.de/SRTM3//Islands//S41E173.hgt.zip',
		'S41E174.hgt': 'https://srtm.kurviger.de/SRTM3//Islands//S41E174.hgt.zip',
		'S41E175.hgt': 'https://srtm.kurviger.de/SRTM3//Islands//S41E175.hgt.zip',
		'S41E176.hgt': 'https://srtm.kurviger.de/SRTM3//Islands//S41E176.hgt.zip',
		'S41W010.hgt': 'https://srtm.kurviger.de/SRTM3//Islands//S41W010.hgt.zip',
		'S41W011.hgt': 'https://srtm.kurviger.de/SRTM3//Islands//S41W011.hgt.zip',
		'S41W063.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S41W063.hgt.zip',
		'S41W064.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S41W064.hgt.zip',
		'S41W065.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S41W065.hgt.zip',
		'S41W066.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S41W066.hgt.zip',
		'S41W067.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S41W067.hgt.zip',
		'S41W068.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S41W068.hgt.zip',
		'S41W069.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S41W069.hgt.zip',
		'S41W070.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S41W070.hgt.zip',
		'S41W071.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S41W071.hgt.zip',
		'S41W072.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S41W072.hgt.zip',
		'S41W073.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S41W073.hgt.zip',
		'S41W074.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S41W074.hgt.zip',
		'S42E144.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S42E144.hgt.zip',
		'S42E145.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S42E145.hgt.zip',
		'S42E146.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S42E146.hgt.zip',
		'S42E147.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S42E147.hgt.zip',
		'S42E148.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S42E148.hgt.zip',
		'S42E171.hgt': 'https://srtm.kurviger.de/SRTM3//Islands//S42E171.hgt.zip',
		'S42E172.hgt': 'https://srtm.kurviger.de/SRTM3//Islands//S42E172.hgt.zip',
		'S42E173.hgt': 'https://srtm.kurviger.de/SRTM3//Islands//S42E173.hgt.zip',
		'S42E174.hgt': 'https://srtm.kurviger.de/SRTM3//Islands//S42E174.hgt.zip',
		'S42E175.hgt': 'https://srtm.kurviger.de/SRTM3//Islands//S42E175.hgt.zip',
		'S42E176.hgt': 'https://srtm.kurviger.de/SRTM3//Islands//S42E176.hgt.zip',
		'S42W063.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S42W063.hgt.zip',
		'S42W064.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S42W064.hgt.zip',
		'S42W065.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S42W065.hgt.zip',
		'S42W066.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S42W066.hgt.zip',
		'S42W067.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S42W067.hgt.zip',
		'S42W068.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S42W068.hgt.zip',
		'S42W069.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S42W069.hgt.zip',
		'S42W070.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S42W070.hgt.zip',
		'S42W071.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S42W071.hgt.zip',
		'S42W072.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S42W072.hgt.zip',
		'S42W073.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S42W073.hgt.zip',
		'S42W074.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S42W074.hgt.zip',
		'S42W075.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S42W075.hgt.zip',
		'S43E145.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S43E145.hgt.zip',
		'S43E146.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S43E146.hgt.zip',
		'S43E147.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S43E147.hgt.zip',
		'S43E148.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S43E148.hgt.zip',
		'S43E170.hgt': 'https://srtm.kurviger.de/SRTM3//Islands//S43E170.hgt.zip',
		'S43E171.hgt': 'https://srtm.kurviger.de/SRTM3//Islands//S43E171.hgt.zip',
		'S43E172.hgt': 'https://srtm.kurviger.de/SRTM3//Islands//S43E172.hgt.zip',
		'S43E173.hgt': 'https://srtm.kurviger.de/SRTM3//Islands//S43E173.hgt.zip',
		'S43E174.hgt': 'https://srtm.kurviger.de/SRTM3//Islands//S43E174.hgt.zip',
		'S43W064.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S43W064.hgt.zip',
		'S43W065.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S43W065.hgt.zip',
		'S43W066.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S43W066.hgt.zip',
		'S43W067.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S43W067.hgt.zip',
		'S43W068.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S43W068.hgt.zip',
		'S43W069.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S43W069.hgt.zip',
		'S43W070.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S43W070.hgt.zip',
		'S43W071.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S43W071.hgt.zip',
		'S43W072.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S43W072.hgt.zip',
		'S43W073.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S43W073.hgt.zip',
		'S43W074.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S43W074.hgt.zip',
		'S43W075.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S43W075.hgt.zip',
		'S44E145.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S44E145.hgt.zip',
		'S44E146.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S44E146.hgt.zip',
		'S44E147.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S44E147.hgt.zip',
		'S44E148.hgt': 'https://srtm.kurviger.de/SRTM3//Australia//S44E148.hgt.zip',
		'S44E168.hgt': 'https://srtm.kurviger.de/SRTM3//Islands//S44E168.hgt.zip',
		'S44E169.hgt': 'https://srtm.kurviger.de/SRTM3//Islands//S44E169.hgt.zip',
		'S44E170.hgt': 'https://srtm.kurviger.de/SRTM3//Islands//S44E170.hgt.zip',
		'S44E171.hgt': 'https://srtm.kurviger.de/SRTM3//Islands//S44E171.hgt.zip',
		'S44E172.hgt': 'https://srtm.kurviger.de/SRTM3//Islands//S44E172.hgt.zip',
		'S44E173.hgt': 'https://srtm.kurviger.de/SRTM3//Islands//S44E173.hgt.zip',
		'S44W065.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S44W065.hgt.zip',
		'S44W066.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S44W066.hgt.zip',
		'S44W067.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S44W067.hgt.zip',
		'S44W068.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S44W068.hgt.zip',
		'S44W069.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S44W069.hgt.zip',
		'S44W070.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S44W070.hgt.zip',
		'S44W071.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S44W071.hgt.zip',
		'S44W072.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S44W072.hgt.zip',
		'S44W073.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S44W073.hgt.zip',
		'S44W074.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S44W074.hgt.zip',
		'S44W075.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S44W075.hgt.zip',
		'S44W176.hgt': 'https://srtm.kurviger.de/SRTM3//Islands//S44W176.hgt.zip',
		'S44W177.hgt': 'https://srtm.kurviger.de/SRTM3//Islands//S44W177.hgt.zip',
		'S45E167.hgt': 'https://srtm.kurviger.de/SRTM3//Islands//S45E167.hgt.zip',
		'S45E168.hgt': 'https://srtm.kurviger.de/SRTM3//Islands//S45E168.hgt.zip',
		'S45E169.hgt': 'https://srtm.kurviger.de/SRTM3//Islands//S45E169.hgt.zip',
		'S45E170.hgt': 'https://srtm.kurviger.de/SRTM3//Islands//S45E170.hgt.zip',
		'S45E171.hgt': 'https://srtm.kurviger.de/SRTM3//Islands//S45E171.hgt.zip',
		'S45W066.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S45W066.hgt.zip',
		'S45W067.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S45W067.hgt.zip',
		'S45W068.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S45W068.hgt.zip',
		'S45W069.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S45W069.hgt.zip',
		'S45W070.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S45W070.hgt.zip',
		'S45W071.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S45W071.hgt.zip',
		'S45W072.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S45W072.hgt.zip',
		'S45W073.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S45W073.hgt.zip',
		'S45W074.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S45W074.hgt.zip',
		'S45W075.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S45W075.hgt.zip',
		'S45W076.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S45W076.hgt.zip',
		'S45W176.hgt': 'https://srtm.kurviger.de/SRTM3//Islands//S45W176.hgt.zip',
		'S45W177.hgt': 'https://srtm.kurviger.de/SRTM3//Islands//S45W177.hgt.zip',
		'S46E050.hgt': 'https://srtm.kurviger.de/SRTM3//Islands//S46E050.hgt.zip',
		'S46E166.hgt': 'https://srtm.kurviger.de/SRTM3//Islands//S46E166.hgt.zip',
		'S46E167.hgt': 'https://srtm.kurviger.de/SRTM3//Islands//S46E167.hgt.zip',
		'S46E168.hgt': 'https://srtm.kurviger.de/SRTM3//Islands//S46E168.hgt.zip',
		'S46E169.hgt': 'https://srtm.kurviger.de/SRTM3//Islands//S46E169.hgt.zip',
		'S46E170.hgt': 'https://srtm.kurviger.de/SRTM3//Islands//S46E170.hgt.zip',
		'S46E171.hgt': 'https://srtm.kurviger.de/SRTM3//Islands//S46E171.hgt.zip',
		'S46W066.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S46W066.hgt.zip',
		'S46W067.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S46W067.hgt.zip',
		'S46W068.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S46W068.hgt.zip',
		'S46W069.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S46W069.hgt.zip',
		'S46W070.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S46W070.hgt.zip',
		'S46W071.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S46W071.hgt.zip',
		'S46W072.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S46W072.hgt.zip',
		'S46W073.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S46W073.hgt.zip',
		'S46W074.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S46W074.hgt.zip',
		'S46W075.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S46W075.hgt.zip',
		'S46W076.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S46W076.hgt.zip',
		'S47E037.hgt': 'https://srtm.kurviger.de/SRTM3//Islands//S47E037.hgt.zip',
		'S47E038.hgt': 'https://srtm.kurviger.de/SRTM3//Islands//S47E038.hgt.zip',
		'S47E050.hgt': 'https://srtm.kurviger.de/SRTM3//Islands//S47E050.hgt.zip',
		'S47E051.hgt': 'https://srtm.kurviger.de/SRTM3//Islands//S47E051.hgt.zip',
		'S47E052.hgt': 'https://srtm.kurviger.de/SRTM3//Islands//S47E052.hgt.zip',
		'S47E166.hgt': 'https://srtm.kurviger.de/SRTM3//Islands//S47E166.hgt.zip',
		'S47E167.hgt': 'https://srtm.kurviger.de/SRTM3//Islands//S47E167.hgt.zip',
		'S47E168.hgt': 'https://srtm.kurviger.de/SRTM3//Islands//S47E168.hgt.zip',
		'S47E169.hgt': 'https://srtm.kurviger.de/SRTM3//Islands//S47E169.hgt.zip',
		'S47E170.hgt': 'https://srtm.kurviger.de/SRTM3//Islands//S47E170.hgt.zip',
		'S47W067.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S47W067.hgt.zip',
		'S47W068.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S47W068.hgt.zip',
		'S47W069.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S47W069.hgt.zip',
		'S47W070.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S47W070.hgt.zip',
		'S47W071.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S47W071.hgt.zip',
		'S47W072.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S47W072.hgt.zip',
		'S47W073.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S47W073.hgt.zip',
		'S47W074.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S47W074.hgt.zip',
		'S47W075.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S47W075.hgt.zip',
		'S47W076.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S47W076.hgt.zip',
		'S48E167.hgt': 'https://srtm.kurviger.de/SRTM3//Islands//S48E167.hgt.zip',
		'S48E168.hgt': 'https://srtm.kurviger.de/SRTM3//Islands//S48E168.hgt.zip',
		'S48E179.hgt': 'https://srtm.kurviger.de/SRTM3//Islands//S48E179.hgt.zip',
		'S48W066.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S48W066.hgt.zip',
		'S48W067.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S48W067.hgt.zip',
		'S48W068.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S48W068.hgt.zip',
		'S48W069.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S48W069.hgt.zip',
		'S48W070.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S48W070.hgt.zip',
		'S48W071.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S48W071.hgt.zip',
		'S48W072.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S48W072.hgt.zip',
		'S48W073.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S48W073.hgt.zip',
		'S48W074.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S48W074.hgt.zip',
		'S48W075.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S48W075.hgt.zip',
		'S48W076.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S48W076.hgt.zip',
		'S49E068.hgt': 'https://srtm.kurviger.de/SRTM3//Islands//S49E068.hgt.zip',
		'S49E069.hgt': 'https://srtm.kurviger.de/SRTM3//Islands//S49E069.hgt.zip',
		'S49E166.hgt': 'https://srtm.kurviger.de/SRTM3//Islands//S49E166.hgt.zip',
		'S49W066.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S49W066.hgt.zip',
		'S49W067.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S49W067.hgt.zip',
		'S49W068.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S49W068.hgt.zip',
		'S49W069.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S49W069.hgt.zip',
		'S49W070.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S49W070.hgt.zip',
		'S49W071.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S49W071.hgt.zip',
		'S49W072.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S49W072.hgt.zip',
		'S49W073.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S49W073.hgt.zip',
		'S49W074.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S49W074.hgt.zip',
		'S49W075.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S49W075.hgt.zip',
		'S49W076.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S49W076.hgt.zip',
		'S50E068.hgt': 'https://srtm.kurviger.de/SRTM3//Islands//S50E068.hgt.zip',
		'S50E069.hgt': 'https://srtm.kurviger.de/SRTM3//Islands//S50E069.hgt.zip',
		'S50E070.hgt': 'https://srtm.kurviger.de/SRTM3//Islands//S50E070.hgt.zip',
		'S50E178.hgt': 'https://srtm.kurviger.de/SRTM3//Islands//S50E178.hgt.zip',
		'S50W068.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S50W068.hgt.zip',
		'S50W069.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S50W069.hgt.zip',
		'S50W070.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S50W070.hgt.zip',
		'S50W071.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S50W071.hgt.zip',
		'S50W072.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S50W072.hgt.zip',
		'S50W073.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S50W073.hgt.zip',
		'S50W074.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S50W074.hgt.zip',
		'S50W075.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S50W075.hgt.zip',
		'S50W076.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S50W076.hgt.zip',
		'S51E068.hgt': 'https://srtm.kurviger.de/SRTM3//Islands//S51E068.hgt.zip',
		'S51E165.hgt': 'https://srtm.kurviger.de/SRTM3//Islands//S51E165.hgt.zip',
		'S51E166.hgt': 'https://srtm.kurviger.de/SRTM3//Islands//S51E166.hgt.zip',
		'S51W062.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S51W062.hgt.zip',
		'S51W068.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S51W068.hgt.zip',
		'S51W069.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S51W069.hgt.zip',
		'S51W070.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S51W070.hgt.zip',
		'S51W071.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S51W071.hgt.zip',
		'S51W072.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S51W072.hgt.zip',
		'S51W073.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S51W073.hgt.zip',
		'S51W074.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S51W074.hgt.zip',
		'S51W075.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S51W075.hgt.zip',
		'S51W076.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S51W076.hgt.zip',
		'S52W058.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S52W058.hgt.zip',
		'S52W059.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S52W059.hgt.zip',
		'S52W060.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S52W060.hgt.zip',
		'S52W061.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S52W061.hgt.zip',
		'S52W062.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S52W062.hgt.zip',
		'S52W069.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S52W069.hgt.zip',
		'S52W070.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S52W070.hgt.zip',
		'S52W071.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S52W071.hgt.zip',
		'S52W072.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S52W072.hgt.zip',
		'S52W073.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S52W073.hgt.zip',
		'S52W074.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S52W074.hgt.zip',
		'S52W075.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S52W075.hgt.zip',
		'S52W076.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S52W076.hgt.zip',
		'S53E073.hgt': 'https://srtm.kurviger.de/SRTM3//Islands//S53E073.hgt.zip',
		'S53E168.hgt': 'https://srtm.kurviger.de/SRTM3//Islands//S53E168.hgt.zip',
		'S53E169.hgt': 'https://srtm.kurviger.de/SRTM3//Islands//S53E169.hgt.zip',
		'S53W059.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S53W059.hgt.zip',
		'S53W060.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S53W060.hgt.zip',
		'S53W061.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S53W061.hgt.zip',
		'S53W062.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S53W062.hgt.zip',
		'S53W069.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S53W069.hgt.zip',
		'S53W070.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S53W070.hgt.zip',
		'S53W071.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S53W071.hgt.zip',
		'S53W072.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S53W072.hgt.zip',
		'S53W073.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S53W073.hgt.zip',
		'S53W074.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S53W074.hgt.zip',
		'S53W075.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S53W075.hgt.zip',
		'S53W076.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S53W076.hgt.zip',
		'S54E072.hgt': 'https://srtm.kurviger.de/SRTM3//Islands//S54E072.hgt.zip',
		'S54E073.hgt': 'https://srtm.kurviger.de/SRTM3//Islands//S54E073.hgt.zip',
		'S54W038.hgt': 'https://srtm.kurviger.de/SRTM3//Islands//S54W038.hgt.zip',
		'S54W039.hgt': 'https://srtm.kurviger.de/SRTM3//Islands//S54W039.hgt.zip',
		'S54W068.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S54W068.hgt.zip',
		'S54W069.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S54W069.hgt.zip',
		'S54W070.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S54W070.hgt.zip',
		'S54W071.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S54W071.hgt.zip',
		'S54W072.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S54W072.hgt.zip',
		'S54W073.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S54W073.hgt.zip',
		'S54W074.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S54W074.hgt.zip',
		'S54W075.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S54W075.hgt.zip',
		'S55E003.hgt': 'https://srtm.kurviger.de/SRTM3//Islands//S55E003.hgt.zip',
		'S55E158.hgt': 'https://srtm.kurviger.de/SRTM3//Islands//S55E158.hgt.zip',
		'S55W036.hgt': 'https://srtm.kurviger.de/SRTM3//Islands//S55W036.hgt.zip',
		'S55W037.hgt': 'https://srtm.kurviger.de/SRTM3//Islands//S55W037.hgt.zip',
		'S55W038.hgt': 'https://srtm.kurviger.de/SRTM3//Islands//S55W038.hgt.zip',
		'S55W039.hgt': 'https://srtm.kurviger.de/SRTM3//Islands//S55W039.hgt.zip',
		'S55W064.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S55W064.hgt.zip',
		'S55W065.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S55W065.hgt.zip',
		'S55W066.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S55W066.hgt.zip',
		'S55W067.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S55W067.hgt.zip',
		'S55W068.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S55W068.hgt.zip',
		'S55W069.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S55W069.hgt.zip',
		'S55W070.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S55W070.hgt.zip',
		'S55W071.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S55W071.hgt.zip',
		'S55W072.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S55W072.hgt.zip',
		'S55W073.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S55W073.hgt.zip',
		'S55W074.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S55W074.hgt.zip',
		'S56E158.hgt': 'https://srtm.kurviger.de/SRTM3//Islands//S56E158.hgt.zip',
		'S56W035.hgt': 'https://srtm.kurviger.de/SRTM3//Islands//S56W035.hgt.zip',
		'S56W067.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S56W067.hgt.zip',
		'S56W068.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S56W068.hgt.zip',
		'S56W069.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S56W069.hgt.zip',
		'S56W070.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S56W070.hgt.zip',
		'S56W071.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S56W071.hgt.zip',
		'S56W072.hgt': 'https://srtm.kurviger.de/SRTM3//South_America//S56W072.hgt.zip',
	},
}
